Package: bayesRecon
Type: Package
Date: 2026-04-16
Title: Probabilistic Reconciliation via Conditioning
Version: 1.0.1
Authors@R: c(person(given = "Dario",
             family = "Azzimonti",
             role = c("aut","cre"),
             email = "dario.azzimonti@gmail.com",
             comment = c(ORCID = "0000-0001-5080-3061")),
       person(given = "Lorenzo",
             family = "Zambon",
             role = c("aut"),
             email = "lorenzo.zambon@idsia.ch",
             comment = c(ORCID = "0000-0002-8939-993X")),
       person(given = "Stefano",
             family = "Damato",
             role = c("aut"),
             email = "stefano.damato@idsia.ch",
             comment = c(ORCID = "0009-0001-0225-3736")),
      person(given = "Nicolò",
             family = "Rubattu",
             role = c("aut"),
             email = "nicolo.rubattu@idsia.ch",
             comment = c(ORCID = "0000-0002-2703-1005")),
      person(given = "Giorgio",
             family = "Corani",
             role = c("aut"),
             email = "giorgio.corani@idsia.ch",
             comment = c(ORCID = "0000-0002-1541-8384")))
Maintainer: Dario Azzimonti <dario.azzimonti@gmail.com>
Description: Provides methods for probabilistic reconciliation of hierarchical forecasts of time series. 
             The available methods include analytical Gaussian reconciliation (Corani et al., 2021) 
             <doi:10.1007/978-3-030-67664-3_13>, 
             MCMC reconciliation of count time series (Corani et al., 2024) 
             <doi:10.1016/j.ijforecast.2023.04.003>, 
             Bottom-Up Importance Sampling (Zambon et al., 2024) 
             <doi:10.1007/s11222-023-10343-y>,
             methods for the reconciliation of mixed hierarchies (Mix-Cond and TD-cond) (Zambon et al., 2024) 
             <https://proceedings.mlr.press/v244/zambon24a.html>,
             analytical reconciliation with Bayesian treatment of the covariance matrix (Carrara et al., 2025)
             <doi: 10.48550/arXiv.2506.19554>.
License: LGPL (>= 3)
Depends: R (>= 4.1.0)
Imports: stats, utils, lpSolve (>= 5.6.18), nloptr
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
Suggests: knitr, rmarkdown, forecast, glarma, scoringRules, ggplot2,
        testthat (>= 3.0.0)
Config/testthat/edition: 3
VignetteBuilder: knitr
URL: https://github.com/IDSIA/bayesRecon,
        https://idsia.github.io/bayesRecon/
BugReports: https://github.com/IDSIA/bayesRecon/issues
NeedsCompilation: no
Packaged: 2026-04-16 09:54:44 UTC; dario.azzimonti
Author: Dario Azzimonti [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-5080-3061>),
  Lorenzo Zambon [aut] (ORCID: <https://orcid.org/0000-0002-8939-993X>),
  Stefano Damato [aut] (ORCID: <https://orcid.org/0009-0001-0225-3736>),
  Nicolò Rubattu [aut] (ORCID: <https://orcid.org/0000-0002-2703-1005>),
  Giorgio Corani [aut] (ORCID: <https://orcid.org/0000-0002-1541-8384>)
Repository: CRAN
Date/Publication: 2026-04-16 10:32:05 UTC
Built: R 4.6.0; ; 2026-04-26 12:40:57 UTC; unix

# Generated by roxygen2: do not edit by hand

export(.check_hierarchical)
export(.compute_naive_cov)
export(.core_reconc_BUIS)
export(.core_reconc_MixCond)
export(.core_reconc_TDcond)
export(.core_reconc_t)
export(.get_Au)
export(.get_HG)
export(.lowest_lev)
export(PMF_get_mean)
export(PMF_get_quantile)
export(PMF_get_var)
export(PMF_sample)
export(PMF_summary)
export(get_reconc_matrices)
export(multi_log_score_optimization)
export(reconc_BUIS)
export(reconc_MCMC)
export(reconc_MixCond)
export(reconc_TDcond)
export(reconc_gaussian)
export(reconc_t)
export(schaferStrimmer_cov)
export(temporal_aggregation)
import(nloptr)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# bayesRecon: BAyesian reCONciliation of hierarchical forecasts <a href="https://idsia.github.io/bayesRecon/"><img src="man/figures/logo.png" align="right" height="150" alt="bayesRecon website" /></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/IDSIA/bayesRecon/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/IDSIA/bayesRecon/actions/workflows/R-CMD-check.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/bayesRecon)](https://CRAN.R-project.org/package=bayesRecon)
[![](http://cranlogs.r-pkg.org/badges/grand-total/bayesRecon)](https://cran.r-project.org/package=bayesRecon)
[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
[![License: LGPL (\>=
3)](https://img.shields.io/badge/license-LGPL%20(%3E=%203)-yellow.svg)](https://www.gnu.org/licences/lgpl-3.0)
[![coverage](https://coveralls.io/repos/github/IDSIA/bayesRecon/badge.svg)](https://coveralls.io/github/IDSIA/bayesRecon)
<!-- badges: end -->

The package `bayesRecon` implements several methods for probabilistic
reconciliation of hierarchical time series forecasts.

The reconciliation functions are:

- `reconc_gaussian`: reconciliation via conditioning of multivariate
  Gaussian base forecasts; this is done analytically;
- `reconc_t`: reconciliation via conditioning of Gaussian forecasts with
  uncertain covariance matrix; the reconciled forecasts are multivariate
  Student-t; this is done analytically;
- `reconc_BUIS`: reconciliation via conditioning of any probabilistic
  forecast via bottom-up importance sampling; an alternative method for
  discrete forecasts is implemented in `reconc_MCMC`, but we recommend
  using `reconc_BUIS`;
- `reconc_MixCond` and `reconc_TDcond`: reconciliation of mixed
  hierarchies, where the upper forecasts are multivariate Gaussian and
  the bottom forecasts are discrete distributions; `reconc_MixCond`
  implements conditioning via importance sampling, while `reconc_TDcond`
  implements top-down conditioning.

## News

:boom: \[2026-03-05\] bayesRecon v1.0: major update to the API, added
`reconc_t`.

:boom: \[2024-05-29\] Added `reconc_MixCond` and `reconc_TDcond` and the
vignette “Reconciliation of M5 hierarchy with mixed-type forecasts”.

:boom: \[2023-12-19\] Added the vignette “Properties of the reconciled
distribution via conditioning”.

:boom: \[2023-08-23\] Added the vignette “Probabilistic Reconciliation
via Conditioning with bayesRecon”. Added the `schaferStrimmer_cov`
function.

:boom: \[2023-05-26\] bayesRecon v0.1.0 is released!

## Installation

You can install the **stable** version on [R
CRAN](https://cran.r-project.org/package=bayesRecon)

``` r
install.packages("bayesRecon", dependencies = TRUE)
```

You can also install the **development** version from
[Github](https://github.com/IDSIA/bayesRecon)

``` r
# install.packages("devtools")
devtools::install_github("IDSIA/bayesRecon", build_vignettes = TRUE, dependencies = TRUE)
```

## Usage

The package `bayesRecon` implements functions for probabilistic forecast
reconciliation. In the following examples, we show how to use the main
reconciliation functions of the package for different types of base
forecasts.

In each example, we

1.  generate simulated hierarchical time series;
2.  compute the base forecasts for each series;
3.  reconcile the forecasts using the functions from the `bayesRecon`
    package.

### Example 1: Gaussian forecast distributions

<img src="./man/figures/hier_small_README.png" width="50%" style="display: block; margin: auto;" />

<br />

Let us consider a hierarchy with 4 bottom time series and 3 upper time
series, as shown in the figure above. The hierarchy is specified by the
*aggregation matrix* **A**:

``` r
A <- matrix(c(
  1, 1, 1, 1,
  1, 1, 0, 0,
  0, 0, 1, 1
), nrow = 3, byrow = TRUE)
```

In this example, we assume that the base forecasts are multivariate
Gaussian, which is a common choice for real-valued time series.

To generate the hierarchical time series, we first randomly simulate the
bottom series using an AR(1) process and then aggregate them using **A**
to obtain the upper series.

``` r
set.seed(1234)

# Simulate bottom series from AR(1) processes
n_obs <- 12  # length of the time series
B_ts <- matrix(nrow = 4, ncol = n_obs)
for (j in 1:4) {
  B_ts[j, ] <- arima.sim(model = list(ar = 0.8), n = n_obs, sd = 0.5)
}
# Aggregate to obtain upper series 
U_ts <- A %*% B_ts
# Convert to mts
B_ts <- ts(t(B_ts))
U_ts <- ts(t(U_ts))
Y_ts <- cbind(U_ts, B_ts)
```

We compute the base forecasts using an ETS model with Gaussian
predictive distribution, implemented in the
[`forecast`](https://cran.r-project.org/package=forecast) package. For
simplicity, we only compute one-step-ahead forecasts, but the same
procedure can be applied to multi-step-ahead forecasts.

We also save the in-sample residuals of the fitted models, as we later
need them to estimate the covariance matrix of the base forecasts.

``` r
library(forecast)

base_fc_mean <- c()
residuals <- matrix(nrow = n_obs, ncol = ncol(Y_ts))
for (j in 1:ncol(Y_ts)) {
  fit <- forecast::ets(Y_ts[,j], additive.only = TRUE)  # fit ets on each time series
  base_fc_mean[j] <- as.numeric(forecast::forecast(fit, h = 1)$mean)
  residuals[,j] <- fit$residuals
}
```

We then analytically compute the reconciled forecasts via conditioning
using the `reconc_gaussian` and the `reconc_t` functions. The reconciled
forecasts produced by `reconc_gaussian` are multivariate Gaussian, and
they are equivalent to MinT reconciliation ([Zambon et
al. 2024](https://doi.org/10.1016/j.ijforecast.2023.12.004)). The
`reconc_t` method adopts a Bayesian approach to account for the
uncertainty of the covariance matrix of the base forecasts; the
reconciled forecasts, which are multivariate Student-t, are typically
better calibrated (see [Carrara et
al. 2025](https://arxiv.org/abs/2506.19554) for details).

``` r
library(bayesRecon)

# Reconcile with both methods
rec_g <- reconc_gaussian(
  A,
  base_fc_mean,
  residuals = residuals,
  return_upper = TRUE
)
rec_t <- reconc_t(
  A,
  base_fc_mean,
  y_train = Y_ts,
  residuals = residuals,
  return_upper = TRUE
)

# Compare means of the base and reconciled forecasts
means <- rbind(c(base_fc_mean),
               c(rec_g$upper_rec_mean, rec_g$bottom_rec_mean),
               c(rec_t$upper_rec_mean, rec_t$bottom_rec_mean))
rownames(means) <- c("base", "reconc_gaussian", "reconc_t")
colnames(means) <- c("T", "A", "B", "AA", "AB", "BA", "BB")
print(round(means, 2))
#>                     T     A     B    AA    AB    BA    BB
#> base            -2.64 -2.24 -0.35 -2.06 -0.18 -0.19 -0.47
#> reconc_gaussian -2.73 -2.25 -0.49 -2.05 -0.20 -0.14 -0.35
#> reconc_t        -2.87 -2.37 -0.50 -2.02 -0.35 -0.11 -0.39
```

Finally, we compare the reconciled forecast distributions for the top
series T obtained with the two methods by plotting their marginal
densities.

<img src="man/figures/README-unnamed-chunk-7-1.png" width="75%" style="display: block; margin: auto;" />

### Example 2: discrete forecast distributions

We consider the same hierarchy of Example 1; however, we assume that the
base forecasts are Poisson distributions, which is a common choice for
count time series.

We randomly generate the bottom series using Poisson distributions with
time-varying rates that include a monthly seasonal pattern; we then
aggregate them using **A** to obtain the upper series.

``` r
set.seed(123)
n_obs <- 60

# Bottom time series are obtained by drawing from Poisson distributions with time-varying rates
lambda_bls <- c(3, 4, 5, 6)  # baseline Poisson rates for bottom series
seas <- 1.5*sin(2*pi*(1:n_obs)/12)  # specify monthly seasonality (period = 12) 
lambdas <- outer(lambda_bls, seas, FUN = "+")  # adds seasonality to each baseline (4 x n_obs matrix)
lambdas <- lambdas + matrix(rnorm(4 * n_obs, sd = 0.1), nrow = 4) # add small Gaussian noise to rates

# Simulate bottom-level count time series
B_ts <- matrix(rpois(4 * n_obs, lambdas), nrow = 4)
# Aggregate to obtain upper series
U_ts <- A %*% B_ts
# Convert to mts
B_ts <- ts(t(B_ts), frequency = 12)
U_ts <- ts(t(U_ts), frequency = 12)
Y_ts <- cbind(U_ts, B_ts)
```

We compute the one-step-ahead *base forecasts* using the package
[`glarma`](https://cran.r-project.org/package=glarma), which is specific
for count time series. We forecast using a `glarma` model with Poisson
predictive distribution.

``` r
library(glarma)

base_fc <- list()
for (j in 1:ncol(Y_ts)) {
  yj <- Y_ts[,j]  # time series j
  X <- matrix(c(rep(1, length(yj)), seas), ncol = 2)  # matrix of exogenous regressors
  X_new <- matrix(c(1, 1.5*sin(2*pi*(n_obs + 1)/12)), ncol = 2)  # regressors for the forecast period
  fit <- glarma::glarma(yj, X, type = "Poi")  # fit the model
  # Compute base forecasts, which are Poisson distributions specified by the rate parameter 
  fc_lambda <- glarma::forecast(fit, newdata = X_new, n.ahead = 1)$mu  
  # Save the base forecast parameters in a list of lists, one for each series
  base_fc[[j]] <- list(lambda = fc_lambda)  
}
```

We then compute the reconciled forecasts using the bottom-up importance
sampling (BUIS) algorithm (see [Zambon et
al. 2024](https://doi.org/10.1007/s11222-023-10343-y) for details). The
output of `reconc_BUIS` is a joint sample from the reconciled
distribution, which can be used to compute any desired summary
(e.g. mean, quantiles, etc.).

``` r
rec_buis <- reconc_BUIS(
  A,
  base_fc,
  in_type = "params",
  distr = "poisson",
  num_samples = 20000
)
samples_buis <- rbind(rec_buis$upper_rec_samples,
                      rec_buis$bottom_rec_samples)
rownames(samples_buis) <- c("T", "A", "B", "AA", "AB", "BA", "BB")

# Compute reconciled means
print(round(rowMeans(samples_buis), 2))
#>     T     A     B    AA    AB    BA    BB 
#> 20.69  8.36 12.33  4.05  4.31  5.83  6.50

# Compute upper quantiles of the reconciled forecast distributions:
print(apply(samples_buis, 1, quantile, probs = c(0.80, 0.95)))
#>      T  A  B AA AB BA BB
#> 80% 23 10 14  5  6  8  8
#> 95% 26 12 16  7  7  9 10
```

Finally, we compare the base and reconciled forecasts for the top series
T by plotting the base and reconciled forecast distributions.

<img src="man/figures/README-unnamed-chunk-11-1.png" width="75%" style="display: block; margin: auto;" />

Similar results can be obtained with `reconc_MCMC`, which is a
bare-bones implementation of the Metropolis-Hastings algorithm. However,
we recommend using `reconc_BUIS` rather than `reconc_MCMC` for
reconciling discrete forecasts.

### Example 3: mixed-type forecast distributions

In many large hierarchies the bottom series are low-count integers
(e.g., item-level sales), while the upper series can be considered as
real-valued due to the smoothing effect of aggregation (e.g., total
sales). These hierarchies are often referred to as *mixed*, since
forecasts for the bottom series are discrete distributions, while
forecasts for the upper series are continuous distributions. The
functions `reconc_MixCond` and `reconc_TDcond` handle this mixed case:
they take a list of discrete distributions for the bottom level and a
multivariate Gaussian for the upper levels. These functions implement
different methods for reconciling mixed hierarchies; we recommend using
`reconc_MixCond` for moderately sized hierarchies and `reconc_TDcond`
for large hierarchies (see [Zambon et
al. 2024](https://proceedings.mlr.press/v244/zambon24a.html) for
details).

Let us consider a hierarchy with 3 upper series and 52 bottom series
arranged in 2 groups of 26:

<img src="./man/figures/hier_large_README.png" width="50%" style="display: block; margin: auto;" />

<br />

We randomly generate the bottom count time series as in Ex. 2; we then
aggregate them using **A** to obtain the upper series.

``` r
set.seed(12)
n_b   <- 52   # number of bottom series 
n_u   <- 3    # number of upper series
n_obs <- 60   # series length

# Build aggregation matrix A for the hierarchy in the figure above
A <- rbind(rep(1, n_b),
           c(rep(1, 26), rep(0, 26)),
           c(rep(0, 26), rep(1, 26)))

# Assume Poisson data generating process + monthly seasonality
lambda_levels <- runif(n_b, min = 0.1, max = 2)
seas <- 1 + .5*sin(2*pi*(1:n_obs)/12)
lambdas <- outer(lambda_levels, seas, FUN = "*")

# Generate bottom series
B_ts <- matrix(rpois(n_obs * n_b, lambdas), nrow = n_b)
# Aggregate to obtain upper series
U_ts <- A %*% B_ts
# Convert to mts
B_ts <- ts(t(B_ts), frequency = 12)
U_ts <- ts(t(U_ts), frequency = 12)
```

We show a comparison of upper and bottom time series. Even though the
bottom series are made of low counts, the upper series can be considered
as real-valued due to the smoothing effect of aggregation.

<img src="man/figures/README-unnamed-chunk-14-1.png" width="75%" style="display: block; margin: auto;" />

We compute the one-step-ahead base forecasts for each upper series with
an additive ETS model, implemented in the
[`forecast`](https://cran.r-project.org/package=forecast) package. We
use the covariance matrix of the in-sample residuals, estimated via
shrinkage using the `schaferStrimmer_cov` function, as the joint
forecast covariance of the upper series.

``` r
mu_u <- numeric(n_u)
residuals_u <- matrix(nrow = n_obs, ncol = n_u)
for (j in seq_len(n_u)) {
  fit <- forecast::ets(ts(U_ts[, j]), additive.only = TRUE)
  mu_u[j] <- as.numeric(forecast::forecast(fit, h = 1)$mean)
  residuals_u[, j] <- fit$residuals
}
# Estimate the covariance matrix via Schafer-Strimmer shrinkage from in-sample residuals
Sigma_u <- bayesRecon::schaferStrimmer_cov(residuals_u)$shrink_cov 
# Save upper base forecasts as a list with mean and covariance
base_fc_upper <- list(mean = mu_u, cov = Sigma_u)  
```

We compute the one-step-ahead *base forecasts* for the bottom series
using the package [`glarma`](https://cran.r-project.org/package=glarma).
The base forecasts are Poisson distributions.

``` r
base_fc_bottom <- list()
for (j in seq_len(n_b)) {
  bj <- B_ts[,j]
  X <- matrix(c(rep(1, length(bj)), seas), ncol = 2)  # matrix of exogenous regressors
  X_new <- matrix(c(1, 1 + .5*sin(2*pi*(n_obs + 1)/12)), ncol = 2)  # regressors for the forecast period
  fit <- glarma(bj, X, type = "Poi")
  # Bottom base forecasts are Poisson distributions, specified by the rate parameter 
  fc_lambda <- glarma::forecast(fit, newdata = X_new, n.ahead = 1)$mu
  # Save the parameters of bottom base forecasts in a list of lists, one for each series
  base_fc_bottom[[j]] <- list(lambda = fc_lambda)
}
```

We reconcile using both `reconc_MixCond` (importance-sampling based
conditioning) and `reconc_TDcond` (top-down conditioning). These
functions implement different methods for reconciling mixed hierarchies,
but they share the same input arguments and output structure.

``` r
res_mc <- reconc_MixCond(
  A, base_fc_bottom, base_fc_upper,
  bottom_in_type = "params", distr = "poisson",
  num_samples = 2e4, 
  return_type = "pmf"
)

res_td <- reconc_TDcond(
  A, base_fc_bottom, base_fc_upper,
  bottom_in_type = "params", distr = "poisson",
  num_samples = 2e4, 
  return_type = "pmf"
)
```

The joint forecast distribution can be obtained by specifying
`return_type = "samples"`. In this case, since we set
`return_type = "pmf"`, the functions return the reconciled marginal
forecast distributions as probability mass functions (PMFs). From these
PMFs, we can compute any desired summary (e.g. mean, quantiles, etc.)
using the `PMF` functions.

``` r
# Compare the upper means of the base and reconciled forecasts
upper_means <- rbind(
  base    = mu_u,
  MixCond = sapply(res_mc$upper_rec_pmf, PMF_get_mean),
  TDcond  = sapply(res_td$upper_rec_pmf, PMF_get_mean)
)
colnames(upper_means) <- c("T", "A", "B")
print(round(upper_means, 2))
#>             T     A     B
#> base    56.98 19.84 34.31
#> MixCond 59.24 23.55 35.68
#> TDcond  55.99 20.42 35.57

# Compare the 95% upper quantiles of the base and reconciled forecast distributions
upper_q <- rbind(
  base    = sapply(mu_u, function(m) qnorm(0.95, mean = m, sd = sqrt(Sigma_u[1,1]))),
  MixCond = sapply(res_mc$upper_rec_pmf, PMF_get_quantile, p = 0.95),
  TDcond  = sapply(res_td$upper_rec_pmf, PMF_get_quantile, p = 0.95)
)
colnames(upper_q) <- c("T", "A", "B")
print(round(upper_q, 2))
#>             T    A     B
#> base    81.93 44.8 59.26
#> MixCond 71.00 30.0 44.00
#> TDcond  81.00 34.0 51.00
```

Finally, we compare the base forecast and the two reconciled forecast
distributions for the top series T. The base distribution is a Gaussian
density (line); the reconciled distributions are discrete PMFs (bars).
The black triangle indicates the actual value of T. We refer to [Zambon
et al. 2024](https://proceedings.mlr.press/v244/zambon24a.html) for a
detailed comparison of the two methods for reconciling mixed hierarchies
of different sizes.

<img src="man/figures/README-unnamed-chunk-19-1.png" width="75%" style="display: block; margin: auto;" />

## References

Carrara, C., Corani, G., Azzimonti, D., Zambon, L. (2025). *Modeling the
uncertainty on the covariance matrix for probabilistic forecast
reconciliation*. arXiv preprint arXiv:2506.19554. [Available
here](https://arxiv.org/abs/2506.19554)

Corani, G., Azzimonti, D., Augusto, J.P.S.C., Zaffalon, M. (2021).
*Probabilistic Reconciliation of Hierarchical Forecast via Bayes’ Rule*.
ECML PKDD 2020. Lecture Notes in Computer Science, vol 12459.
[DOI](https://doi.org/10.1007/978-3-030-67664-3_13)

Corani, G., Azzimonti, D., Rubattu, N. (2024). *Probabilistic
reconciliation of count time series*. International Journal of
Forecasting 40 (2), 457-469.
[DOI](https://doi.org/10.1016/j.ijforecast.2023.04.003)

Zambon, L., Azzimonti, D. & Corani, G. (2024). *Efficient probabilistic
reconciliation of forecasts for real-valued and count time series*.
Statistics and Computing 34 (1), 21.
[DOI](https://doi.org/10.1007/s11222-023-10343-y)

Zambon, L., Agosto, A., Giudici, P., Corani, G. (2024). *Properties of
the reconciled distributions for Gaussian and count forecasts*.
International Journal of Forecasting 40 (4), 1438-1448.
[DOI](https://doi.org/10.1016/j.ijforecast.2023.12.004)

Zambon, L., Azzimonti, D., Rubattu, N., Corani, G. (2024).
*Probabilistic reconciliation of mixed-type hierarchical time series*.
Proceedings of the Fortieth Conference on Uncertainty in Artificial
Intelligence, in Proceedings of Machine Learning Research 244:4078-4095.
[Available here](https://proceedings.mlr.press/v244/zambon24a.html).

## Contributors

<!-- prettier-ignore-start -->

<!-- markdownlint-disable -->

<table>

<tbody>

<tr>

<td align="center" valign="top" width="14.28%">

<a href="https://sites.google.com/view/darioazzimonti/home">
<img src="https://github.com/dazzimonti.png" width="100px;" alt="Dario Azzimonti" style="border-radius:50%;border:1px solid #646464;"/><br />
<sub><b>Dario Azzimonti</b></sub></a><br />
<sub>(Maintainer)</sub><br />
<a href="mailto:dario.azzimonti@gmail.com?subject=[bayesRecon package]">Email</a>
</td>

<td align="center" valign="top" width="14.28%">

<a href="#">
<img src="https://github.com/LorenzoZambon.png" width="100px;" alt="Lorenzo Zambon" style="border-radius:50%;border:1px solid #646464;"/><br />
<sub><b>Lorenzo Zambon</b></sub></a><br /> <sub> </sub><br />
<a href="mailto:lorenzo.zambon@idsia.ch?subject=[bayesRecon package]">Email</a>
</td>

<td align="center" valign="top" width="14.28%">

<a href="#">
<img src="https://github.com/StefanoDamato.png" width="100px;" alt="Stefano Damato" style="border-radius:50%;border:1px solid #646464;"/><br />
<sub><b>Stefano Damato</b></sub></a><br /> <sub> </sub><br />
<a href="mailto:stefano.damato@idsia.ch?subject=[bayesRecon package]">Email</a>
</td>

<td align="center" valign="top" width="14.28%">

<a href="#">
<img src="https://github.com/nicorbtt.png" width="100px;" alt="Nicolò Rubattu" style="border-radius:50%;border:1px solid #646464;"/><br />
<sub><b>Nicolò Rubattu</b></sub></a><br /> <sub> </sub><br />
<a href="mailto:nico.rubattu@gmail.com?subject=[bayesRecon package]">Email</a>
</td>

<td align="center" valign="top" width="14.28%">

<a href="https://sites.google.com/site/awerbhjkl678214/home">
<img src="https://github.com/gcorani.png" width="100px;" alt="Giorgio Corani" style="border-radius:50%;border:1px solid #646464;"/><br />
<sub><b>Giorgio Corani</b></sub></a><br /> <sub> </sub><br />
<a href="mailto:giorgio.corani@idsia.ch?subject=[bayesRecon package]">Email</a>
</td>

</tr>

</tbody>

</table>

<!-- markdownlint-restore -->

<!-- prettier-ignore-end -->

## Getting help

If you encounter a clear bug, please file a minimal reproducible example
on [GitHub](https://github.com/IDSIA/bayesRecon/issues).

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval=TRUE ### !!!! set to FALSE here to render only the text !!!!
)
set.seed(42)

## ----klippy, echo=FALSE, include=TRUE, eval=FALSE-----------------------------
#  klippy::klippy(position = c('top', 'right'), tooltip_message = 'Copy', tooltip_success = 'Done', color="black")

## ----install, eval=FALSE------------------------------------------------------
#  install.packages('bayesRecon', dependencies = TRUE)

## ----load---------------------------------------------------------------------
library(bayesRecon)

## ----carpart-plot, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 1**: Carpart - monthly car part sales.", fig.dim = c(6, 3)----
layout(mat = matrix(c(1, 2), nrow = 1, ncol = 2), widths = c(2, 1))
plot(carparts_example, xlab = "Time", ylab = "Car part sales", main = NULL)
hist(carparts_example, xlab = "Car part sales", main = NULL)

## ----train-test---------------------------------------------------------------
train <- window(carparts_example, end = c(2001, 3))
test  <- window(carparts_example, start = c(2001, 4))

## ----temp-agg-----------------------------------------------------------------
train.agg <- bayesRecon::temporal_aggregation(train, agg_levels = c(2, 3, 4, 6, 12))
levels <- c("Annual", "Biannual", "4-Monthly", "Quarterly", "2-Monthly", "Monthly")
names(train.agg) <- levels

## ----temp-agg-plot, dpi=300, fig.show="hold", out.width="100%", out.heigth="100%", fig.align='center', fig.cap="**Figure 2**: The aggregated time series of the temporal hierarchy.", fig.dim=c(6,3.5)----
par(mfrow = c(2, 3), mai = c(0.6, 0.6, 0.5, 0.5))
for (l in levels) {
  plot(train.agg[[l]], xlab = "Time", ylab = "Car part sales", main = l)
}

## ----hier-fore, cache=TRUE----------------------------------------------------
# install.packages("glarma", dependencies = TRUE)
#library(glarma)

fc.samples <- list()
D <- 20000
fc.count <- 1

# iterating over the temporal aggregation levels
for (l in seq_along(train.agg)) {
  f.level <- frequency(train.agg[[l]])
  print(paste("Forecasting at ", levels[l], "...", sep = ""))
  # fit an independent model for each aggregation level
  model <- glarma::glarma(
    train.agg[[l]],
    phiLags = if (f.level == 1) 1 else 1:(min(6, f.level - 1)),
    thetaLags = if (f.level == 1) NULL else f.level,
    X = cbind(intercept = rep(1, length(train.agg[[l]]))),
    offset = cbind(intercept = rep(0, length(train.agg[[l]]))),
    type = "Poi"
  )
  # forecast 1 year ahead
  h <- f.level
  tmp <- matrix(data = NA, nrow = h, ncol = D)
  for (s in (1:D)) {
    # each call to 'forecast.glarma' returns a simulation path
    tmp[, s] <- glarma::forecast(
      model,
      n.ahead = h,
      newdata = cbind(intercept = rep(1, h)),
      newoffset = rep(0, h)
    )$Y
  }
  # collect the forecasted samples
  for (i in 1:h) {
    fc.samples[[fc.count]] <- tmp[i, ]
    fc.count <- fc.count + 1
  }
}


## ----aggregationMatrix--------------------------------------------------------
recon.matrices <- bayesRecon::get_reconc_matrices(agg_levels = c(2, 3, 4, 6, 12), h = 12)
# Aggregation matrix
A <- recon.matrices$A 

## ----reconc-------------------------------------------------------------------
recon.res <- bayesRecon::reconc_BUIS(
  A,
  base_forecasts = fc.samples,
  in_type = "samples",
  distr = "discrete",
  seed = 42
)

## ----res----------------------------------------------------------------------
reconciled_samples <- recon.res$reconciled_samples
dim(reconciled_samples)


## ----metrics------------------------------------------------------------------
# install.packages("scoringRules", dependencies = TRUE)
library(scoringRules)

ae.fc <- list()
ae.reconc <- list()
crps.fc <- list()
crps.reconc <- list()
for (h in 1:length(test)) {
  y.hat_ <- median(fc.samples[[nrow(A) + h]])
  y.reconc_ <- median(recon.res$bottom_reconciled_samples[, h])
  # Compute Absolute Errors
  ae.fc[[h]] <- abs(test[h] - y.hat_)
  ae.reconc[[h]] <- abs(test[h] - y.reconc_)
  # Compute Continuous Ranked Probability Score (CRPS)
  crps.fc[[h]] <-
    scoringRules::crps_sample(y = test[h], dat = fc.samples[[nrow(A) + h]])
  crps.reconc[[h]] <-
    scoringRules::crps_sample(y = test[h], dat = recon.res$bottom_reconciled_samples[, h])
}

mae.fc <- mean(unlist(ae.fc))
mae.reconc <- mean(unlist(ae.reconc))
crps.fc <- mean(unlist(crps.fc))
crps.reconc <- mean(unlist(crps.reconc))
metrics <- data.frame(
  row.names = c("MAE", "CRPS"),
  base.forecasts = c(mae.fc, crps.fc),
  reconciled.forecasts = c(mae.reconc, crps.reconc)
)
metrics


## ----m3-plot, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 3**: M3 - N1485 time series.", fig.dim = c(6, 3)----
plot(M3_example$train, xlab = "Time", ylab = "y", main = "N1485")

## ----m3-agg-------------------------------------------------------------------
train.agg <- bayesRecon::temporal_aggregation(M3_example$train)
levels <- c("Annual", "Biannual", "4-Monthly", "Quarterly", "2-Monthly", "Monthly")
names(train.agg) <- levels

## ----m3-fore------------------------------------------------------------------
# install.packages("forecast", dependencies = TRUE)
library(forecast)

H <- length(M3_example$test)
H

fc <- list()
level.idx <- 1
fc.idx <- 1
for (level in train.agg) {
  level.name <- names(train.agg)[level.idx]
  # fit an ETS model for each temporal level
  model <- ets(level)
  # generate forecasts for each level within 18 months
  h <- floor(H / (12 / frequency(level)))
  print(paste("Forecasting at ", level.name, ", h=", h, "...", sep = ""))
  level.fc <- forecast(model, h = h)
  # save mean and sd of the gaussian predictive distribution
  for (i in 1:h) {
    fc[[fc.idx]] <- list(mean = level.fc$mean[[i]],
                         sd = (level.fc$upper[, "95%"][[i]] - level.fc$mean[[i]]) / qnorm(0.975))
    fc.idx <- fc.idx + 1
  }
  level.idx <- level.idx + 1
}

## ----m3-rmat, dpi=300, out.width = '70%', fig.align='center', fig.cap="**Figure 4**: M3 - The aggregation matrix A (red=1, yellow=0).", fig.dim = c(8, 8)----
rmat <- get_reconc_matrices(agg_levels = c(2, 3, 4, 6, 12), h = 18)

par(mai = c(1,1,0.5,0.5))
image(1:ncol(rmat$A), 1:nrow(rmat$A), 
      t(apply(t(rmat$A),1,rev)), 
      xaxt='n', yaxt='n', ylab = "", xlab = levels[6])
axis(1, at=1:ncol(rmat$A), label=1:ncol(rmat$A), las=1)
axis(2, at=c(23,22,19,15,9), label=levels[1:5], las=2)

## ----m3-reco------------------------------------------------------------------
recon.gauss <- bayesRecon::reconc_gaussian(
  A = rmat$A,
  base_forecasts.mu = sapply(fc, "[[", 1),
  base_forecasts.Sigma = diag(sapply(fc, "[[", 2)) ^ 2
)

reconc.buis <- bayesRecon::reconc_BUIS(
  A = rmat$A,
  base_forecasts = fc,
  in_type = "params",
  distr = "gaussian",
  num_samples = 20000,
  seed = 42
)

# check that the algorithms return consistent results
round(rbind(
  c(rmat$S %*% recon.gauss$bottom_reconciled_mean),
  rowMeans(reconc.buis$reconciled_samples)
))

## ----m3-plotfore, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 5**: M3 - Base and reconciled forecasts. The black line shows the actual data (dashed in the test). The orange line is the  mean of the base forecasts, the blu line is the reconciled mean. We also show the 95% prediction intervals.", fig.dim = c(6, 4)----
yhat.mu <- tail(sapply(fc, "[[", 1), 18)
yhat.sigma <- tail(sapply(fc, "[[", 2), 18)
yhat.hi95 <- qnorm(0.975, mean = yhat.mu, sd = yhat.sigma)
yhat.lo95 <- qnorm(0.025, mean = yhat.mu, sd = yhat.sigma)
yreconc.mu <- rowMeans(reconc.buis$bottom_reconciled_samples)
yreconc.hi95 <- apply(reconc.buis$bottom_reconciled_samples, 1, 
                      function(x) quantile(x, 0.975))
yreconc.lo95 <- apply(reconc.buis$bottom_reconciled_samples, 1, 
                      function(x) quantile(x, 0.025))

ylim_ <- c(min(M3_example$train, M3_example$test, yhat.lo95, yreconc.lo95) - 1, 
           max(M3_example$train, M3_example$test, yhat.hi95, yreconc.hi95) + 1)

plot(M3_example$train, xlim = c(1990, 1995.6), ylim = ylim_, 
     ylab = "y", main = "N1485 Forecasts")
lines(M3_example$test, lty = "dashed")
lines(ts(yhat.mu, start = start(M3_example$test), frequency = 12), 
      col = "coral", lwd = 2)
lines(ts(yreconc.mu, start = start(M3_example$test), frequency = 12), 
      col = "blue2", lwd = 2)
xtest <- time(M3_example$test)
polygon(c(xtest, rev(xtest)), c(yhat.mu, rev(yhat.hi95)), 
        col = "#FF7F5066", border = "#FF7F5066")
polygon(c(xtest, rev(xtest)), c(yhat.mu, rev(yhat.lo95)), 
        col = "#FF7F5066", border = "#FF7F5066")
polygon(c(xtest, rev(xtest)), c(yreconc.mu, rev(yreconc.hi95)), 
        col = "#0000EE4D", border = "#0000EE4D")
polygon(c(xtest, rev(xtest)), c(yreconc.mu, rev(yreconc.lo95)), 
        col = "#0000EE4D", border = "#0000EE4D")

## ----infants-forecasts--------------------------------------------------------
# install.packages("forecast", dependencies = TRUE)
library(forecast)

fc <- list()
residuals <- matrix(NA,
                    nrow = length(infantMortality$total),
                    ncol = length(infantMortality))
fc.idx <- 1
for (s in infantMortality) {
  s.name <- names(infantMortality)[fc.idx]
  print(paste("Forecasting at ", s.name, "...", sep = ""))
  # fit an auto.arima model and forecast with h=1
  model <- auto.arima(s)
  s.fc <- forecast(model, h = 1)
  # save mean and sd of the gaussian predictive distribution
  fc[[s.name]] <- c(s.fc$mean,
                    (s.fc$upper[, "95%"][[1]] - s.fc$mean) / qnorm(0.975))
  residuals[, fc.idx] <- s.fc$residuals
  fc.idx <- fc.idx + 1
}

## ----infants-s, dpi=300, out.width = '70%', fig.align='center', fig.cap="**Figure 6**: Infants mortality - The aggregation matrix A (red=1, yellow=0).", fig.dim = c(8, 8)----
# we have 16 bottom time series, and 11 upper time series
A <- matrix(data = c(1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
                     1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                     0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,
                     0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,
                     0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,
                     0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,
                     0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,
                     0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,
                     0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,
                     1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,
                     0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1), byrow=TRUE, ncol = 16)

# plot of A
par(mai = c(1.5,1,0.5,0.5))
image(1:ncol(A), 1:nrow(A), 
      t(apply(t(A),1,rev)), 
      xaxt='n', yaxt='n', ann=FALSE)
axis(1, at=1:ncol(A), label=names(infantMortality)[12:27], las=2)
axis(2, at=c(1:11), label=rev(names(infantMortality)[1:11]), las=2)

## ----infants reconc-----------------------------------------------------------
# means
mu <- sapply(fc, "[[", 1)
# Shrinkage covariance
shrink.res <- bayesRecon::schaferStrimmer_cov(residuals)
print(paste("The estimated shrinkage intensity is", round(shrink.res$lambda_star, 3)))
Sigma <- shrink.res$shrink_cov

## ----infants-recon------------------------------------------------------------
recon.gauss <- bayesRecon::reconc_gaussian(A,
                                           base_forecasts.mu = mu,
                                           base_forecasts.Sigma = Sigma)

bottom_mu_reconc <- recon.gauss$bottom_reconciled_mean
bottom_Sigma_reconc <- recon.gauss$bottom_reconciled_covariance

# Obtain reconciled mu and Sigma for the upper variable
upper_mu_reconc <- A %*% bottom_mu_reconc
upper_Sigma_reconc <- A %*% bottom_Sigma_reconc %*% t(A)

upper_mu_reconc

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(bayesRecon)

## ----M5hier, fig.cap="**Figure 1**: graph of the M5 hierarchy.", out.width = '100%', echo = FALSE----
knitr::include_graphics("img/M5store_hier.png")

## ----InitializeHierarchy------------------------------------------------------
# Hierarchy composed by 3060 time series: 3049 bottom and 11 upper
n_b <- 3049
n_u <- 11
n <- n_b + n_u

# Load matrix A
A <- M5_CA1_basefc$A

# Load base forecasts:
base_fc_upper <- M5_CA1_basefc$upper
base_fc_bottom <- M5_CA1_basefc$bottom

# We will save all the results in the list rec_fc
rec_fc <- list(
  Gauss      = list(),
  Mixed_cond = list(),
  TD_cond    = list()
)

## ----readBaseForecasts--------------------------------------------------------
# Parameters of the upper base forecast distributions
mu_u <- unlist(lapply(base_fc_upper, "[[", "mu")) # upper means
# Compute the (shrinked) covariance matrix of the residuals
residuals.upper <- lapply(base_fc_upper, "[[", "residuals")
residuals.upper <- t(do.call("rbind", residuals.upper))
Sigma_u <- schaferStrimmer_cov(residuals.upper)$shrink_cov

# Parameters of the bottom base forecast distributions
mu_b <- c()
sd_b <- c()
for (fc_b in base_fc_bottom) {
  pmf <- fc_b$pmf
  mu_b <- c(mu_b, PMF_get_mean(pmf))
  sd_b <- c(sd_b, PMF_get_var(pmf)**0.5)
}
Sigma_b <- diag(sd_b**2)

# Mean and covariance matrix of the base forecasts
base_fc_mean <- c(mu_u, mu_b)
base_fc_cov <- matrix(0, nrow = n, ncol = n)
base_fc_cov[1:n_u, 1:n_u] <- Sigma_u
base_fc_cov[(n_u + 1):n, (n_u + 1):n] <- Sigma_b

## ----GaussianReconciliation---------------------------------------------------
# Gaussian reconciliation
start <- Sys.time()
gauss <- reconc_gaussian(A, base_fc_mean, base_fc_cov)
stop <- Sys.time()

rec_fc$Gauss <- list(
  mu_b = gauss$bottom_rec_mean,
  Sigma_b = gauss$bottom_rec_cov,
  mu_u = A %*% gauss$bottom_rec_mean,
  Sigma_u = A %*% gauss$bottom_rec_cov %*% t(A)
)

Gauss_time <- as.double(round(difftime(stop, start, units = "secs"), 2))
cat("Time taken by Gaussian reconciliation: ", Gauss_time, "s")

## ----MixedCondReconciliation--------------------------------------------------
seed <- 1
N_samples_IS <- 5e4

# Base forecasts
fc_upper_4rec <- list(mean = mu_u, cov = Sigma_u)
fc_bottom_4rec <- lapply(base_fc_bottom, "[[", "pmf") # list of PMFs

# MixCond reconciliation
start <- Sys.time()
mix_cond <- reconc_MixCond(A, fc_bottom_4rec, fc_upper_4rec,
  bottom_in_type = "pmf",
  num_samples = N_samples_IS, return_type = "pmf", seed = seed
)
stop <- Sys.time()

rec_fc$Mixed_cond <- list(
  bottom = mix_cond$bottom_rec_pmf,
  upper  = mix_cond$upper_rec_pmf,
  ESS    = mix_cond$ESS
)

MixCond_time <- as.double(round(difftime(stop, start, units = "secs"), 2))
cat("Computational time for Mix-cond reconciliation: ", MixCond_time, "s")

## ----TDcondReconciliation-----------------------------------------------------
N_samples_TD <- 1e4

# TDcond reconciliation
start <- Sys.time()
td <- reconc_TDcond(A, fc_bottom_4rec, fc_upper_4rec,
  bottom_in_type = "pmf", num_samples = N_samples_TD,
  return_type = "pmf", seed = seed
)
stop <- Sys.time()

## ----print computational time TD cond-----------------------------------------
rec_fc$TD_cond <- list(
  bottom = td$bottom_rec_pmf,
  upper  = td$upper_rec_pmf
)

TDCond_time <- as.double(round(difftime(stop, start, units = "secs"), 2))
cat("Computational time for TD-cond reconciliation: ", TDCond_time, "s")

## ----InitializeMetrics--------------------------------------------------------
# Parameters for computing the scores
alpha <- 0.1 # MIS uses 90% coverage intervals
jitt <- 1e-9 # jitter for numerical stability

# Save actual values
actuals_u <- unlist(lapply(base_fc_upper, "[[", "actual"))
actuals_b <- unlist(lapply(base_fc_bottom, "[[", "actual"))
actuals <- c(actuals_u, actuals_b)

# Scaling factor for computing MASE
Q_u <- M5_CA1_basefc$Q_u
Q_b <- M5_CA1_basefc$Q_b
Q <- c(Q_u, Q_b)

# Initialize lists to save the results
mase <- list()
mis <- list()
rps <- list()

## ----metricFunctions,include=FALSE--------------------------------------------
# Functions for computing the scores of a PMF
AE_pmf <- function(pmf, actual) {
  return(abs(PMF_get_quantile(pmf, p = 0.5) - actual))
}
MIS_pmf <- function(pmf, actual, alpha) {
  u <- PMF_get_quantile(pmf, p = 1 - (alpha / 2))
  l <- PMF_get_quantile(pmf, p = alpha / 2)
  return(u - l + (2 / alpha) * (l - actual) * (actual < l) +
    (2 / alpha) * (actual - u) * (actual > u))
}
RPS_pmf <- function(pmf, actual) {
  cdf <- cumsum(pmf) / sum(pmf)
  M <- length(cdf)
  # if actual is outside the supp of pmf, add ones to the end of the cdf:
  if (actual >= M) {
    cdf <- c(cdf, rep(1, (actual - M + 1)))
    M <- length(cdf)
  }
  cdf_act <- (0:(M - 1)) >= actual # unit step function in actual
  crps_ <- sum((cdf - cdf_act)**2)
  return(crps_)
}

# Function for computing the MIS of (possibly truncated) Gaussian forecasts
MIS_gauss <- function(mus, sds, actuals, alpha, trunc = FALSE) {
  z <- qnorm(1 - (alpha / 2))
  u <- mus + z * sds
  l <- mus - z * sds
  # If it is truncated, set negative quantiles to zero
  if (trunc) {
    l[l < 0] <- 0
    u[u < 0] <- 0
  }
  return(u - l + (2 / alpha) * (l - actuals) * (actuals < l) +
    (2 / alpha) * (actuals - u) * (actuals > u))
}

## ----computeScores------------------------------------------------------------
# Compute scores for the base forecasts
# Upper
mu_u <- unlist(lapply(base_fc_upper, "[[", "mu"))
sd_u <- unlist(lapply(base_fc_upper, "[[", "sigma"))
mase$base[1:n_u] <- abs(mu_u - actuals_u) / Q_u
mis$base[1:n_u] <- MIS_gauss(mu_u, sd_u, actuals_u, alpha)
rps$base[1:n_u] <- scoringRules::crps(actuals_u, "norm", mean = mu_u, sd = sd_u)
# Bottom
pmfs <- lapply(base_fc_bottom, "[[", "pmf")
mase$base[(n_u + 1):n] <- mapply(AE_pmf, pmfs, actuals_b) / Q_b
mis$base[(n_u + 1):n] <- mapply(MIS_pmf, pmfs, actuals_b, MoreArgs = list(alpha = alpha))
rps$base[(n_u + 1):n] <- mapply(RPS_pmf, pmfs, actuals_b)

# Compute scores for Gauss reconciliation
mu <- c(rec_fc$Gauss$mu_u, rec_fc$Gauss$mu_b)
sd <- c(diag(rec_fc$Gauss$Sigma_u), diag(rec_fc$Gauss$Sigma_b))**0.5
sd <- sd + jitt
mase$Gauss <- abs(mu - actuals) / Q
mis$Gauss <- MIS_gauss(mu, sd, actuals, alpha)
rps$Gauss <- scoringRules::crps(actuals, "norm", mean = mu, sd = sd)

# Compute scores for Mix-cond reconciliation
pmfs <- c(rec_fc$Mixed_cond$upper, rec_fc$Mixed_cond$bottom)
mase$MixCond <- mapply(AE_pmf, pmfs, actuals) / Q
mis$MixCond <- mapply(MIS_pmf, pmfs, actuals, MoreArgs = list(alpha = alpha))
rps$MixCond <- mapply(RPS_pmf, pmfs, actuals)

# Compute scores for TD-cond reconciliation
pmfs <- c(rec_fc$TD_cond$upper, rec_fc$TD_cond$bottom)
mase$TDcond <- mapply(AE_pmf, pmfs, actuals) / Q
mis$TDcond <- mapply(MIS_pmf, pmfs, actuals, MoreArgs = list(alpha = alpha))
rps$TDcond <- mapply(RPS_pmf, pmfs, actuals)

## ----skillScoreFunction,include=FALSE-----------------------------------------
# Function for computing the skill score
skill.score <- function(ref, met) {
  s <- (2 * (ref - met) / (ref + met)) * 100
  s[is.na(s)] <- 0 # if both numerator and denominator are 0, set skill score to 0
  return(s)
}

## ----ComputeScores------------------------------------------------------------
scores <- list(
  mase = mase,
  mis = mis,
  rps = rps
)
scores_ <- names(scores)

ref_met <- "base"
methods_ <- c("Gauss", "MixCond", "TDcond")

# For each score and method we compute the skill score with respect to the base forecasts
skill_scores <- list()
for (s in scores_) {
  skill_scores[[s]] <- list()
  for (met in methods_) {
    skill_scores[[s]][["upper"]][[met]] <- skill.score(
      scores[[s]][[ref_met]][1:n_u],
      scores[[s]][[met]][1:n_u]
    )
    skill_scores[[s]][["bottom"]][[met]] <- skill.score(
      scores[[s]][[ref_met]][(n_u + 1):n],
      scores[[s]][[met]][(n_u + 1):n]
    )
  }
}

## ----AverageScores------------------------------------------------------------
mean_skill_scores <- list()

for (s in scores_) {
  mean_skill_scores[[s]] <- rbind(
    data.frame(lapply(skill_scores[[s]][["upper"]], mean)),
    data.frame(lapply(skill_scores[[s]][["bottom"]], mean))
  )
  rownames(mean_skill_scores[[s]]) <- c("upper", "bottom")
}

## ----printMASETable-----------------------------------------------------------
knitr::kable(mean_skill_scores$mase, digits = 2, caption = "Mean skill score on MASE.", align = "lccc")

## ----PrintMIStable------------------------------------------------------------
knitr::kable(mean_skill_scores$mis, digits = 2, caption = "Mean skill score on MIS.")

## ----printRPStable------------------------------------------------------------
knitr::kable(mean_skill_scores$rps, digits = 2, caption = "Mean skill score on RPS.")

## ----MASEboxplots, fig.cap="**Figure 2**: boxplot of MASE skill scores for upper and bottom time series.", fig.width=7,fig.height=8----
custom_colors <- c(
  "#a8a8e4",
  "#a9c7e4",
  "#aae4df"
)

# Boxplots of MASE skill scores
par(mfrow = c(2, 1))
boxplot(skill_scores$mase$upper,
  main = "MASE upper time series",
  col = custom_colors, ylim = c(-80, 80)
)
abline(h = 0, lty = 3)
boxplot(skill_scores$mase$bottom,
  main = "MASE bottom time series",
  col = custom_colors, ylim = c(-200, 200)
)
abline(h = 0, lty = 3)

## ----setupParams,eval=TRUE,include=FALSE--------------------------------------
par(mfrow = c(1, 1))

## ----MISboxplots, fig.cap="**Figure 3**: boxplot of MIS skill scores for upper and bottom time series.", fig.width=7,fig.height=8----
# Boxplots of MIS skill scores
par(mfrow = c(2, 1))
boxplot(skill_scores$mis$upper,
  main = "MIS upper time series",
  col = custom_colors, ylim = c(-150, 150)
)
abline(h = 0, lty = 3)
boxplot(skill_scores$mis$bottom,
  main = "MIS bottom time series",
  col = custom_colors, ylim = c(-200, 200)
)
abline(h = 0, lty = 3)

## ----eval=TRUE,include=FALSE--------------------------------------------------
par(mfrow = c(1, 1))

## ----RPSboxplots,fig.cap="**Figure 4**: boxplot of RPS skill scores for upper and bottom time series.", fig.width=7,fig.height=8----
# Boxplots of RPS skill scores
par(mfrow = c(2, 1))
boxplot(skill_scores$rps$upper,
  main = "RPS upper time series",
  col = custom_colors, ylim = c(-80, 80)
)
abline(h = 0, lty = 3)
boxplot(skill_scores$rps$bottom,
  main = "RPS bottom time series",
  col = custom_colors, ylim = c(-200, 200)
)
abline(h = 0, lty = 3)

## ----eval=TRUE,include=FALSE--------------------------------------------------
par(mfrow = c(1, 1))

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(bayesRecon)

## ----out.width = '50%', echo = FALSE------------------------------------------
knitr::include_graphics("img/finTS_hier.jpg")

## -----------------------------------------------------------------------------
# Hierarchy composed by 6 time series: 5 bottom and 1 upper
n_b <- 5
n_u <- 1
n <- n_b + n_u   

A <- matrix(1, ncol = n_b, nrow = n_u)  # aggregation matrix

# Actual values:
actuals <- data.frame(extr_mkt_events)  # convert to data frame
# Base forecasts:
base_fc <- extr_mkt_events_basefc

N <- nrow(actuals)  # number of days (3508)

# # If you want to run only N reconciliations (instead of 3508):
# N <- 200
# actuals <- actuals[1:N,]
# base_fc$mu <- base_fc$mu[1:N,]

## -----------------------------------------------------------------------------
# We need to save the mean and median of the reconciled distribution
# in order to compute the squared error and the absolute error:
rec_means   <- matrix(NA, ncol = n, nrow = N)
rec_medians <- matrix(NA, ncol = n, nrow = N)

# We need to save the lower and upper quantiles of the reconciled distribution
# in order to compute the interval score:
rec_L <- matrix(NA, ncol = n, nrow = N)
rec_U <- matrix(NA, ncol = n, nrow = N)
int_cov = 0.9   # use 90% interval         
q1 <- (1 - int_cov) / 2
q2 <- (1 + int_cov) / 2

# Set the number of samples to draw from the reconciled distribution:
N_samples <- 1e4

start <- Sys.time()
for (j in 1:N) {
  # Base forecasts:
  base_fc_j <- c()   
  for (i in 1:n) {
    base_fc_j[[i]] <- list(size = base_fc$size[[i]], mu = base_fc$mu[[j,i]])
  }
  
  # Reconcile via importance sampling:
  buis <- reconc_BUIS(A, base_fc_j, "params", "nbinom", 
                      num_samples = N_samples, seed = 42)
  samples_y <- buis$reconciled_samples
  
  # Save mean, median, and lower and upper quantiles:
  rec_means[j,]   <- rowMeans(samples_y)
  rec_medians[j,] <- apply(samples_y, 1, median)
  rec_L[j,]       <- apply(samples_y, 1, quantile, q1)
  rec_U[j,]       <- apply(samples_y, 1, quantile, q2)
}
stop <- Sys.time()
cat("Computational time for ", N, " reconciliations: ",     
    round(difftime(stop, start, units = "secs"), 2), "s")

## -----------------------------------------------------------------------------
base_means   <- base_fc$mu
base_medians <- matrix(NA, ncol = n, nrow = N)
base_L       <- matrix(NA, ncol = n, nrow = N)
base_U       <- matrix(NA, ncol = n, nrow = N)

for (i in 1:n) {
  base_medians[,i] <- sapply(base_means[,i], 
                       function(mu) qnbinom(p=0.5, size=base_fc$size[[i]], mu=mu))
  base_L[,i]       <- sapply(base_means[,i], 
                       function(mu) qnbinom(p=q1,  size=base_fc$size[[i]], mu=mu))
  base_U[,i]       <- sapply(base_means[,i], 
                       function(mu) qnbinom(p=q2,  size=base_fc$size[[i]], mu=mu))
}

## -----------------------------------------------------------------------------
# Compute the squared errors
SE_base <- (base_means - actuals)^2
SE_rec  <- (rec_means - actuals)^2

# Compute the absolute errors
AE_base <- abs(base_medians - actuals)
AE_rec  <- abs(rec_medians - actuals)

# Define the function for the interval score
int_score <- function(l, u, actual, int_cov = 0.9) {
  is <- (u - l) +
  2 / (1-int_cov) * (actual - u) * (actual>u) +
  2 / (1-int_cov) * (l - actual) * (l>actual) 
  return(is)
}

# Compute the interval scores
IS_base <- mapply(int_score, base_L, base_U, data.matrix(actuals))
IS_base <- matrix(IS_base, nrow = N)
IS_rec <- mapply(int_score, rec_L, rec_U, data.matrix(actuals))
IS_rec <- matrix(IS_rec, nrow = N)

## -----------------------------------------------------------------------------
SS_AE <- (AE_base - AE_rec) / (AE_base + AE_rec) * 2
SS_SE <- (SE_base - SE_rec) / (SE_base + SE_rec) * 2
SS_IS <- (IS_base - IS_rec) / (IS_base + IS_rec) * 2
SS_AE[is.na(SS_AE)] <- 0
SS_SE[is.na(SS_SE)] <- 0
SS_IS[is.na(SS_IS)] <- 0

mean_skill_scores <- c(round(colMeans(SS_IS), 2),
                       round(colMeans(SS_SE), 2),
                       round(colMeans(SS_AE), 2))
mean_skill_scores <- data.frame(t(matrix(mean_skill_scores, nrow = n)))
colnames(mean_skill_scores) <- names(actuals)
rownames(mean_skill_scores) <- c("Interval score", "Squared error", "Absolute error")
knitr::kable(mean_skill_scores)

## -----------------------------------------------------------------------------
# Now we draw a larger number of samples:
N_samples <- 1e5

### Example of concordant-shift effect
j <- 124
base_fc_j <- c()   
for (i in 1:n) base_fc_j[[i]] <- list(size = extr_mkt_events_basefc$size[[i]],
                                      mu = extr_mkt_events_basefc$mu[[j,i]])
# Reconcile
buis <- reconc_BUIS(A, base_fc_j, "params", "nbinom", 
                    num_samples = N_samples, seed = 42)
samples_y <- buis$reconciled_samples

# The reconciled mean is lower than both the base and bottom-up means:
means <- c(round(extr_mkt_events_basefc$mu[[j,1]], 2),
           round(sum(extr_mkt_events_basefc$mu[j,2:n]), 2),
           round(mean(samples_y[1,]), 2))
col_names <- c("Base upper mean", "Bottom-up upper mean", "Reconciled upper mean")
knitr::kable(matrix(means, nrow=1), col.names = col_names)

### Example of combination effect 
j <- 1700
base_fc_j <- c()   
for (i in 1:n) base_fc_j[[i]] <- list(size = extr_mkt_events_basefc$size[[i]],
                                      mu = extr_mkt_events_basefc$mu[[j,i]])
# Reconcile 
buis <- reconc_BUIS(A, base_fc_j, "params", "nbinom", 
                    num_samples = N_samples, seed = 42)
samples_y <- buis$reconciled_samples

# The reconciled mean is between the base and the bottom-up mean:
means <- c(round(extr_mkt_events_basefc$mu[[j,1]], 2),
           round(sum(extr_mkt_events_basefc$mu[j,2:n]), 2),
           round(mean(samples_y[1,]), 2))
col_names <- c("Base upper mean", "Bottom-up upper mean", "Reconciled upper mean")
knitr::kable(matrix(means, nrow=1), col.names = col_names)

## -----------------------------------------------------------------------------
j <- 2308
base_fc_j <- c()   
for (i in 1:n) base_fc_j[[i]] <- list(size = extr_mkt_events_basefc$size[[i]],
                                      mu = extr_mkt_events_basefc$mu[[j,i]])
# Reconcile
buis <- reconc_BUIS(A, base_fc_j, "params", "nbinom", num_samples = N_samples, seed = 42)
samples_y <- buis$reconciled_samples

# Compute the variance of the base and reconciled bottom forecasts
base_bottom_var <- mapply(function(mu, size) var(rnbinom(n = 1e5, size = size, mu = mu)),
                          extr_mkt_events_basefc$mu[j,2:n], 
                          extr_mkt_events_basefc$size[2:n])
rec_bottom_var <- apply(samples_y[2:n,], MARGIN = 1, var) 

# The reconciled variance is greater than the base variance:
bottom_var <- rbind(base_bottom_var, rec_bottom_var)
rownames(bottom_var) <- c("var base", "var reconc")
knitr::kable(round(bottom_var, 2))

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----load, message=FALSE, warning=FALSE---------------------------------------
# load the packages
library(bayesRecon)
library(forecast) # base forecasts
library(ggplot2)  # plots

## ----swiss-tourism-plot, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 1**: Swiss tourism - monthly Swiss overnight stays.", fig.dim = c(8, 4)----
# save all time series 
Y = swiss_tourism$ts

# plot the first (top) time series
autoplot(Y[,1], ylab = "Overnight stays in Switzerland",linewidth=0.9)+
  scale_y_continuous(labels = function(x) paste0(formatC(x / 1e6, format = "g"), "M"))

## ----Swiss tourism------------------------------------------------------------
# Save aggregation matrix
A = swiss_tourism$agg_mat

# Number of bottom and upper time series
n_b = ncol(A)
n_u = nrow(A)
n = n_b + n_u

# Frequency is monthly:
print(frequency(Y))

# Select the length of the training set and the forecast horizon
L = 60    
h = 1    

# Select the training set and the actuals for the forecast horizon
train = window(Y, end = time(Y)[L])
actuals = window(Y, start= time(Y)[L + 1], end = time(Y)[L + h])

## ----Base-forecasts and residuals computation---------------------------------
# Compute base forecasts and residuals for each time series
base_fc = rep(NA, n)
res = matrix(NA, ncol = n, nrow = L)
for (i in 1:n){
    models = forecast::ets(train[,i], model = "AZZ")
    f = forecast(models, h = h)
    base_fc[i] = f$mean
    res[, i] = models$residuals
}

## ----t-reconciliation---------------------------------------------------------
t_rec_results = reconc_t(A, base_fc_mean = base_fc, 
                         y_train = train, 
                         residuals = res,
                         return_parameters = TRUE,
                         return_upper = TRUE)

## ----Base---------------------------------------------------------------------
# Base forecasts
Base_mean = base_fc
Base_cov_mat = crossprod(res)/nrow(res) # covariance of the residuals

## ----Gaussian-----------------------------------------------------------------
# Gaussian/MinT: compute reconciliation with bayesRecon
gauss_results = reconc_gaussian(A, base_fc, residuals = res, return_upper = TRUE)
# Reconciled mean for the whole hierarchy:
MinT_reconciled_mean = c(gauss_results$upper_rec_mean, 
                         gauss_results$bottom_rec_mean)

## ----compute uppers, echo=FALSE, eval=TRUE------------------------------------
t_Rec_reconciled_mean <- c(t_rec_results$upper_rec_mean,t_rec_results$bottom_rec_mean)
t_Rec_corr_with_upper = A %*% t_rec_results$bottom_rec_scale_matrix
t_Rec_scale_par <- rbind(cbind(t_rec_results$upper_rec_scale_matrix, t_Rec_corr_with_upper),
                         cbind(t(t_Rec_corr_with_upper), t_rec_results$bottom_rec_scale_matrix))

# Index of the upper variable to plot
i_upper <- 1  # change this if a different index is needed

# Extract distribution's parameters for each method

# MinT
# Build the full reconciled covariance matrix
MinT_corr_with_upper = A %*% gauss_results$bottom_rec_cov
MinT_cov_mat = rbind(cbind(gauss_results$upper_rec_cov, MinT_corr_with_upper),
                        cbind(t(MinT_corr_with_upper),gauss_results$bottom_rec_cov))
mu_MinT <- as.numeric(MinT_reconciled_mean[i_upper])
sd_MinT <- sqrt(MinT_cov_mat[i_upper, i_upper])

# t-Rec
mu_tRec <- as.numeric(t_Rec_reconciled_mean[i_upper])
scale_tRec <- sqrt(t_Rec_scale_par[i_upper, i_upper])
df_tRec <- t_rec_results$bottom_rec_df

# Base
mu_Base <- as.numeric(Base_mean[i_upper])
sd_Base <- sqrt(Base_cov_mat[i_upper, i_upper])

# Create a grid of x values
x_vals <- seq(min(mu_MinT, mu_tRec, mu_Base) - 4*max(sd_MinT, scale_tRec, sd_Base),
              max(mu_MinT, mu_tRec, mu_Base) + 4*max(sd_MinT, scale_tRec, sd_Base),
              length.out = 1000)

# Compute densities
dens_MinT <- dnorm(x_vals, mean = mu_MinT, sd = sd_MinT)
dens_tRec <- dt((x_vals - mu_tRec) / scale_tRec, df = df_tRec) /scale_tRec
dens_Base <- dnorm(x_vals, mean = mu_Base, sd = sd_Base)


## ----comparison-plot, echo=FALSE, eval=TRUE, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 2**: Predictive densities of the upper time series obtained with MinT (purple), t-Rec (green) and Base (blue). The black triangle indicates the actual value.", fig.dim = c(8, 4), warning=FALSE----

dens_df <- data.frame(
  x       = rep(x_vals, 3),
  Method  = rep(c("MinT", "t-Rec", "Base"), each = length(x_vals)),
  Density = c(dens_MinT, dens_tRec, dens_Base)
)

method_colors <- c(
  "MinT" = "#440154",
  "t-Rec" = "#29AF7F",
  "Base" = "#3B528B"
)

ggplot(dens_df, aes(x = x, y = Density, color = Method)) +
  geom_area(
    data = dens_df[dens_df$Method %in% c("MinT", "t-Rec"),],
    aes(fill = Method),
    position = "identity",
    alpha = 0.3,
    color = NA
  ) +
  geom_line(linewidth = 1.5) +
  scale_color_manual(values = method_colors) +
  scale_fill_manual(values = method_colors) +
  geom_point(aes(x = actuals[i_upper], y = 0, shape = "Actual value"), color = "black", size = 3) +
  scale_shape_manual(values = c("Actual value" = 17)) +
  scale_x_continuous(labels = function(x) paste0(formatC(x / 1e3, format = "g"), "k")) +
  guides(fill = "none") +
  labs(
    title = "Predictive densities of upper time series",
    x = "",
    y = ""
  ) +
  theme_minimal(base_size = 12) +
  theme(
    legend.position = "top",
    legend.title = element_blank(),
    plot.title       = element_text(size = 18)
  )

## ----compute density----------------------------------------------------------
# Select which series to plot
i = 1     # CH
j = 19    # GR

# density of the inverse gamma
dinvgamma <- function(x, shape, rate) {
  dgamma(1/x, shape = shape, rate = rate) / x^2
}

# density of the standard deviation (square root transform of variance)
d_std_dev <- function(x, shape,rate){
  dinvgamma(x^2, shape = shape, rate = rate)*2*x
}

# compute the density of the variance parameters for CH
mean_ch <- t_rec_results$posterior_Psi[i, i]/(t_rec_results$posterior_nu - n - 1)
x_ch <- sqrt(seq(mean_ch*0.5, mean_ch*1.7, length.out = 1000))
shape_ch <- (t_rec_results$posterior_nu - n + 1) /2
rate_ch  <- t_rec_results$posterior_Psi[i, i] / 2
dens_ch <- d_std_dev(x_ch, shape = shape_ch, rate = rate_ch)

# compute the density of the variance parameters for GR
mean_gr <- t_rec_results$posterior_Psi[j, j]/(t_rec_results$posterior_nu - n - 1)
x_gr <- sqrt(seq(mean_gr*0.5, mean_gr*1.7, length.out = 1000))
shape_gr <- (t_rec_results$posterior_nu - n + 1) /2
rate_gr  <- t_rec_results$posterior_Psi[j, j] / 2
dens_gr <- d_std_dev(x_gr, shape = shape_gr, rate = rate_gr)

## ----density plot, echo=FALSE, eval=TRUE, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 3**: Density of the posterior standard deviation of the forecasts.", fig.dim = c(8, 4), warning=FALSE----

df_dens <- data.frame(
  x = c(x_ch, x_gr), 
  y = c(dens_ch, dens_gr), 
  panel = rep(c("CH", "GR"), each = length(dens_ch))
)

# Plot the density of the standard deviation parameters for CH and GR
ggplot(df_dens, aes(x = x, y = y)) +
  geom_line(linewidth = 1.5, color="darkolivegreen", alpha=0.7) +
  scale_x_continuous(labels = function(x) paste0(formatC(x / 1e3, format = "g"), "k")) +
  guides(fill = "none") +
  labs(
    title = "Posterior standard deviation of the forecasts",
    x = "",
    y = ""
  ) +
  facet_wrap(~ panel, nrow = 1, scales = "free") +
  theme_minimal(base_size = 12) +
  theme(
    legend.position = "top",
    legend.title = element_blank(),
    plot.title       = element_text(size = 18)
  )

## ----generate IW samples------------------------------------------------------
# generate k samples from an IW(Psi, nu) distribution
rinvwishart <- function(k, nu, Psi, seed=42) {
  p <- nrow(Psi)
  Sigma <- solve(Psi)
  
  set.seed(seed)
  all_W <- rWishart(k, df = nu, Sigma = Sigma)
  
  W <- array(NA, dim = c(p, p, k))
  for (i in 1:k) {
    W[,,i] <- solve(all_W[,,i])
  }
  return(W)
}

IW_post_samples <- rinvwishart(k=1000, nu = t_rec_results$posterior_nu,
                            Psi = t_rec_results$posterior_Psi)


## ----compute correlations, echo=FALSE, eval=TRUE------------------------------
corr_mint <- cov2cor(MinT_cov_mat)
corr_base <- cov2cor(Base_cov_mat)
corr_tRec <- cov2cor(t_rec_results$posterior_Psi/(t_rec_results$posterior_nu - n - 1))

## ----plot densities, echo=FALSE, eval=TRUE, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 4**: Density of the posterior correlation between CH and GR obtained with t-Rec.", fig.dim = c(8, 4), warning=FALSE----
# Compute correlation samples for all sample matrices generated above
corr_post <- array(apply(IW_post_samples, FUN = function(M) cov2cor(M), MARGIN = c(3)),
                   dim = dim(IW_post_samples))

df_corr <- data.frame(x = corr_post[i, j, ])

ggplot(df_corr, aes(x = x)) +
  geom_density(fill = "#00BA38", alpha = 0.4) +
  geom_vline(xintercept = corr_mint[i, j], linetype = "dashed", color = "black") +
  annotate("text", x = corr_mint[i, j], y = max(density(corr_post[i, j, ])$y) * 0.95, label = "MinT", 
           hjust = -0.2, vjust = 0, size = 4.5) +
  labs(x = "", y = "Density",
       title = expression(paste("Posterior correlation ", rho["CH,GR"]))) +
  theme_minimal() +
  theme(
    legend.position  = "none",
    plot.title       = element_text(size = 18)
  )

## ----echo=FALSE,eval=TRUE-----------------------------------------------------
# Print table with values for Std and correlation
knitr::kable(data.frame(
  Method = c("Base", "MinT", "t-Rec"),
  sigma2_CH   = sqrt(c(Base_cov_mat[i, i], MinT_cov_mat[i, i], t_rec_results$posterior_Psi[i, i]/(t_rec_results$posterior_nu - n - 1))),
  sigma2_GR   = sqrt(c(Base_cov_mat[j, j], MinT_cov_mat[j, j], t_rec_results$posterior_Psi[j, j]/(t_rec_results$posterior_nu - n - 1))),
  rho_CH_GR   = c(corr_base[i, j], corr_mint[i, j], corr_tRec[i, j])
), digits = c(0, 0, 0, 2),
   format = "html",
   col.names = c("Method",
                 "$\\hat{\\sigma}_{\\text{CH}}$",
                 "$\\hat{\\sigma}_{\\text{GR}}$",
                 "$\\hat{\\rho}_{\\text{CH,GR}}$"),
   escape = FALSE,
   table.attr = "style='width:auto;'",
   format.args = list(big.mark = ",", scientific = FALSE),
   caption = "**Table 1**: Standard deviation and correlation estimates for CH and GR.")

#-------------------------------------------------------------------------------
#' @title MCMC for Probabilistic Reconciliation of forecasts via conditioning
#'
#' @description
#'
#' Uses Markov Chain Monte Carlo algorithm to draw samples from the reconciled
#' forecast distribution, which is obtained via conditioning.
#'
#' This is a bare-bones implementation of the Metropolis-Hastings algorithm,
#' we suggest the usage of tools to check the convergence.
#' The function only works with Poisson or Negative Binomial base forecasts.
#'
#' The function [reconc_BUIS()] is generally faster on most hierarchies.
#'
#'
#'
#' @param A aggregation matrix (n_upper x n_bottom).
#' @param base_fc list of the parameters of the base forecast distributions, see details.
#' @param distr a string describing the type of predictive distribution.
#' @param num_samples number of samples to draw using MCMC.
#' @param tuning_int number of iterations between scale updates of the proposal.
#' @param init_scale initial scale of the proposal.
#' @param burn_in number of initial samples to be discarded.
#' @param return_upper Logical. If \code{TRUE} (default), also returns the reconciled samples for the
#'        upper time series. Default is \code{TRUE}.
#' @param seed seed for reproducibility.
#'
#' @details
#'
#' The parameter `base_fc` is a list containing n = n_upper + n_bottom elements.
#' Each element is a list containing the estimated:
#'
#' * mean and sd for the Gaussian base forecast, see \link[stats]{Normal}, if `distr`='gaussian';
#' * lambda for the Poisson base forecast, see \link[stats]{Poisson}, if `distr`='poisson';
#' * size and prob (or mu) for the negative binomial base forecast, see \link[stats]{NegBinomial}, if `distr`='nbinom'.
#'
#' The first n_upper elements of the list are the upper base forecasts, in the order given by the rows of A.
#' The elements from n_upper+1 until the end of the list are the bottom base forecasts, in the order given by the columns of A.
#'
#' @return A list containing the reconciled forecasts. The list has the following named elements:
#'
#' * `bottom_rec_samples`: a matrix (n_bottom x `num_samples`) containing reconciled samples for the bottom time series;
#' * `upper_rec_samples`: a matrix (n_upper x `num_samples`) containing reconciled samples for the upper time series (only if `return_upper = TRUE`).
#'
#' @examples
#'
#' library(bayesRecon)
#'
#' # Create a minimal hierarchy with 2 bottom and 1 upper variable
#' rec_mat <- get_reconc_matrices(agg_levels = c(1, 2), h = 2)
#' A <- rec_mat$A
#'
#' # Set the parameters of the Poisson base forecast distributions
#' lambda1 <- 2
#' lambda2 <- 4
#' lambdaY <- 9
#' lambdas <- c(lambdaY, lambda1, lambda2)
#'
#' base_fc <- list()
#' for (i in 1:length(lambdas)) {
#'   base_fc[[i]] <- list(lambda = lambdas[i])
#' }
#'
#' # Sample from the reconciled forecast distribution using MCMC
#' mcmc <- reconc_MCMC(A, base_fc,
#'   distr = "poisson",
#'   num_samples = 30000, seed = 42
#' )
#' samples_mcmc <- rbind(mcmc$upper_rec_samples, mcmc$bottom_rec_samples)
#'
#' # Compare the reconciled means with those obtained via BUIS
#' buis <- reconc_BUIS(A, base_fc,
#'   in_type = "params",
#'   distr = "poisson", num_samples = 100000, seed = 42
#' )
#' samples_buis <- rbind(buis$upper_rec_samples, buis$bottom_rec_samples)
#'
#' print(rowMeans(samples_mcmc))
#' print(rowMeans(samples_buis))
#'
#' @references
#' Corani, G., Azzimonti, D., Rubattu, N. (2024).
#' *Probabilistic reconciliation of count time series*.
#' International Journal of Forecasting 40 (2), 457-469.
#' \doi{10.1016/j.ijforecast.2023.04.003}.
#'
#' @seealso
#' [reconc_BUIS()]
#'
#' @export
reconc_MCMC <- function(A,
                        base_fc,
                        distr,
                        num_samples = 10000,
                        tuning_int = 100,
                        init_scale = 1,
                        burn_in = 1000,
                        return_upper = TRUE,
                        seed = NULL) {
  if (!is.null(seed)) set.seed(seed)

  # Ensure that data inputs are valid
  if (distr == "gaussian") {
    stop("MCMC for Gaussian distributions is not implemented")
  }

  n_bottom <- ncol(A)
  n_ts <- nrow(A) + ncol(A)

  # Transform distr into list
  if (!is.list(distr)) {
    distr <- rep(list(distr), n_ts)
  }
  # Check input
  .check_input_BUIS(A, base_fc, in_type = as.list(rep("params", n_ts)), distr = distr)

  # the first burn_in samples will be removed
  num_samples <- num_samples + burn_in

  # Set the covariance matrix of the proposal (identity matrix)
  cov_mat_prop <- diag(rep(1, n_bottom))

  # Set the counter for tuning
  c_tuning <- tuning_int

  # Initialize acceptance counter
  accept_count <- 0

  # Create empty matrix for the samples from MCMC
  b <- matrix(nrow = num_samples, ncol = n_bottom)

  # Get matrix A and bottom base forecasts
  split_hierarchy_res <- list(
    A = A,
    upper = base_fc[1:nrow(A)],
    bottom = base_fc[(nrow(A) + 1):n_ts],
    upper_idxs = 1:nrow(A),
    bottom_idxs = (nrow(A) + 1):n_ts
  )
  bottom_base_fc <- split_hierarchy_res$bottom
  bottom_distr <- distr[split_hierarchy_res$bottom_idxs]

  # Initialize first sample (draw from base distribution)
  b[1, ] <- .initialize_b(bottom_base_fc, bottom_distr)

  # Initialize prop list
  old_prop <- list(
    "b" = b[1, ],
    "scale" = init_scale
  )

  # Run the chain
  for (i in 2:num_samples) {
    if (c_tuning == 0) {
      old_prop$acc_rate <- accept_count / tuning_int # set acc_rate
      accept_count <- 0 # reset acceptance counter
      c_tuning <- tuning_int # reset tuning counter
    }

    prop <- .proposal(old_prop, cov_mat_prop)
    b_prop <- prop$b
    alpha <- .accept_prob(b_prop, b[i - 1, ], A, distr, base_fc)

    if (stats::runif(1) < alpha) {
      b[i, ] <- b_prop
      accept_count <- accept_count + 1
    } else {
      b[i, ] <- b[i - 1, ]
    }

    old_prop <- list(
      "b" = b[i, ],
      "scale" = prop$scale
    )

    c_tuning <- c_tuning - 1
  }

  b_samples <- t(b[(burn_in + 1):num_samples, ]) # output shape: n_bottom x num_samples
  u_samples <- A %*% b_samples

  out <- list(bottom_rec_samples = b_samples)
  if (return_upper) {
    out$upper_rec_samples <- u_samples
  }

  return(out)
}


#-------------------------------------------------------------------------------
##################################
.initialize_b <- function(bottom_base_fc, bottom_distr) {
  b <- c()
  for (i in 1:length(bottom_distr)) {
    b[i] <- .distr_sample(bottom_base_fc[[i]], bottom_distr[[i]], 1)
  }

  return(b)
}


#-------------------------------------------------------------------------------
##################################
# @title Compute acceptance probability
# @param b proposal state
# @param b0 current state
# @param A aggregation matrix
# @param distr list of strings specifying the distribution of each variable
# @param params list of the parameters of the distributions
# @return the acceptance probability alpha
.accept_prob <- function(b, b0, A, distr, params) {
  alpha <- .target_pmf(b, A, distr, params) / .target_pmf(b0, A, distr, params)

  return(min(1, alpha))
}


##################################
.target_pmf <- function(b, A, distr, params) {
  n_ts <- nrow(A) + ncol(A)

  y <- rbind(A, diag(nrow = ncol(A), ncol = ncol(A))) %*% b

  pmf <- 1
  for (j in 1:n_ts) {
    pmf <- pmf * .distr_pmf(y[[j]], params[[j]], distr[[j]])
  }

  return(pmf)
}


#-------------------------------------------------------------------------------
##################################
# @title Generate a proposal for MCMC step
# @param prev_prop a list containing b, scale, acc_rate
# @param cov_mat_prop the covariance matrix (ncol(cov_mat_prop)=length(b)) for the normal proposal
# @return a list containing the new b, scale, acc_rate
.proposal <- function(prev_prop, cov_mat_prop) {
  # extract previous proposal
  b0 <- prev_prop$b
  old_scale <- prev_prop$scale
  acc_rate <- prev_prop$acc_rate

  if (!is.null(acc_rate)) {
    # Compute the scaling parameter
    scale <- .tune(old_scale, acc_rate)
  } else {
    scale <- old_scale
  }


  n_x <- length(b0)

  if (n_x != ncol(cov_mat_prop)) {
    stop(sprintf("Error in .proposal: previous state dim (%s) and
                 covariance dim (%s) do not match", n_x, ncol(cov_mat_prop)))
  }

  # We always do an independent Gaussian proposal
  dd <- stats::rnorm(n_x) * diag(cov_mat_prop) * scale

  # CHANGE HERE for mixed case
  dd <- round(dd, 0)
  b <- b0 + dd

  prop <- list(b = b, acc_rate = acc_rate, scale = scale)
  return(prop)
}


# scaling parameter
# we use the same variance adaptation used in pymc3
# Rate    Variance adaptation
# ----    -------------------
# <0.001        x 0.1
# <0.05         x 0.5
# <0.2          x 0.9
# >0.5          x 1.1
# >0.75         x 2
# >0.95         x 10
.tune <- function(scale, acc_rate) {
  if (acc_rate < 0.001) {
    return(scale * 0.1)
  } else if (acc_rate < 0.05) {
    return(scale * 0.5)
  } else if (acc_rate < 0.2) {
    return(scale * 0.9)
  } else if (acc_rate > 0.5) {
    return(scale * 1.1)
  } else if (acc_rate > 0.75) {
    return(scale * 2)
  } else if (acc_rate > 0.95) {
    return(scale * 10)
  }

  return(scale)
}

test_that("MCMC Monthly, in_type=='params', distr='poisson'", {
  A <- read.csv(file = "dataForTests/Monthly-Poisson_A.csv", header = FALSE)
  A <- as.matrix(A)
  base_forecasts_in <- read.csv(file = "dataForTests/Monthly-Poisson_basef.csv", header = FALSE)
  base_forecasts <- list()
  for (i in 1:nrow(base_forecasts_in)) {
    base_forecasts[[i]] <- list(lambda = base_forecasts_in[i, 1])
  }
  res.buis <- reconc_BUIS(A, base_forecasts,
    in_type = "params", distr = "poisson", num_samples = 100000, seed = 42
  )

  res.mcmc <- reconc_MCMC(A, base_fc = base_forecasts, distr = "poisson", num_samples = 100000, seed = 42)

  m <- (rowMeans(res.buis$bottom_rec_samples) - rowMeans(res.mcmc$bottom_rec_samples)) / rowMeans(res.buis$bottom_rec_samples)
  expect_equal(max(abs(m)) < 0.02, TRUE)
})

#' @section Learn more:
#'
#' To learn more about `bayesRecon`, start with the vignettes: `browseVignettes(package = "bayesRecon")`
#'
#' @section Reconciliation functions:
#'
#' The package implements reconciliation via conditioning for probabilistic forecasts
#' of hierarchical time series. The reconciliation functions are:
#'
#' * [reconc_gaussian()]: reconciliation via conditioning assuming multivariate Gaussian
#'    base forecasts; this is done analytically;
#' * [reconc_t()]: reconciliation via conditioning assuming multivariate Gaussian
#'    base forecasts with uncertain covariance matrix; 
#'    the reconciled forecasts are multivariate Student-t; this is done analytically;
#' * [reconc_BUIS()]: reconciliation via conditioning of any probabilistic forecast
#'    via bottom-up importance sampling; an alternative method for discrete forecasts 
#'    is implemented in [reconc_MCMC()], but we recommend using `reconc_BUIS`;
#' * [reconc_MixCond()] and [reconc_TDcond()]: reconciliation of mixed hierarchies, where
#'    the upper forecasts are multivariate Gaussian and the bottom forecasts are discrete distributions;
#'    `reconc_MixCond` implements conditioning via importance sampling, 
#'    while `reconc_TDcond` implements top-down conditioning.
#'
#' @section Utility functions:
#'
#' * [temporal_aggregation()]: temporal aggregation of a given time series object of class \link[stats]{ts};
#' * [get_reconc_matrices()]: aggregation and summing matrices for a temporal hierarchy
#'    of time series from user-selected list of aggregation levels;
#' * [schaferStrimmer_cov()]: computes the Schäfer-Strimmer shrinkage estimator for the covariance matrix;
#' * [multi_log_score_optimization()]: estimates the optimal degrees of freedom for [reconc_t()] 
#'    by maximizing the leave-one-out (LOO) multivariate log density;
#' * [PMF]: functions for handling PMF objects (sampling, computing statistics like mean, variance, quantiles, and summaries).
#'
#' @references
#' 
#' Carrara, C., Corani, G., Azzimonti, D., & Zambon, L. (2025).
#' *Modeling the uncertainty on the covariance matrix for probabilistic forecast reconciliation*.
#' arXiv preprint arXiv:2506.19554. <https://arxiv.org/abs/2506.19554>.
#' 
#' Corani, G., Azzimonti, D., Augusto, J.P.S.C., Zaffalon, M. (2021).
#' *Probabilistic Reconciliation of Hierarchical Forecast via Bayes' Rule*.
#' ECML PKDD 2020. Lecture Notes in Computer Science, vol 12459.
#' \doi{10.1007/978-3-030-67664-3_13}.
#'
#' Corani, G., Azzimonti, D., Rubattu, N. (2024).
#' *Probabilistic reconciliation of count time series*.
#' International Journal of Forecasting 40 (2), 457-469.
#' \doi{10.1016/j.ijforecast.2023.04.003}.
#'
#' Zambon, L., Azzimonti, D. & Corani, G. (2024).
#' *Efficient probabilistic reconciliation of forecasts for real-valued and count time series*.
#' Statistics and Computing 34 (1), 21.
#' \doi{10.1007/s11222-023-10343-y}.
#'
#' Zambon, L., Agosto, A., Giudici, P., Corani, G. (2024).
#' *Properties of the reconciled distributions for Gaussian and count forecasts*.
#' International Journal of Forecasting (in press).
#' \doi{10.1016/j.ijforecast.2023.12.004}.
#'
#' Zambon, L., Azzimonti, D., Rubattu, N., Corani, G. (2024).
#' *Probabilistic reconciliation of mixed-type hierarchical time series*.
#' Proceedings of the Fortieth Conference on Uncertainty in Artificial Intelligence,
#' PMLR 244:4078-4095. <https://proceedings.mlr.press/v244/zambon24a.html>.
#'
#'
#' @keywords internal
"_PACKAGE"

## usethis namespace: start
## usethis namespace: end
NULL

###############################################################################
# Reconciliation with Bottom-Up Importance Sampling (BUIS)
###############################################################################

# Checks that there is no bottom continuous variable child of a
# discrete upper variable
.check_hierfamily_rel <- function(sh.res, distr, debug = FALSE) {
  for (bi in seq_along(distr[sh.res$bottom_idxs])) {
    distr_bottom <- distr[sh.res$bottom_idxs][[bi]]
    rel_upper_i <- sh.res$A[, bi]
    rel_distr_upper <- unlist(distr[sh.res$upper_idxs])[rel_upper_i == 1]
    err_message <- "A continuous bottom distribution cannot be child of a discrete one."
    if (distr_bottom == "continuous") {
      if (sum(rel_distr_upper == "discrete") | sum(rel_distr_upper %in% .DISCR_DISTR)) {
        if (debug) {
          return(-1)
        } else {
          stop(err_message)
        }
      }
    }
    if (distr_bottom %in% .CONT_DISTR) {
      if (sum(rel_distr_upper == "discrete") | sum(rel_distr_upper %in% .DISCR_DISTR)) {
        if (debug) {
          return(-1)
        } else {
          stop(err_message)
        }
      }
    }
  }
  if (debug) {
    return(0)
  }
}

.emp_pmf <- function(l, density_samples) {
  empirical_pmf <- PMF_from_samples(density_samples)
  w <- sapply(l, function(i) empirical_pmf[i + 1])
  return(w)
}

.compute_weights <- function(b, u, in_type_, distr_) {
  if (in_type_ == "samples") {
    if (distr_ == "discrete") {
      # Discrete samples
      w <- .emp_pmf(b, u)
    } else if (distr_ == "continuous") {
      # KDE
      d <- stats::density(u, bw = "SJ", n = 2**16)
      df <- stats::approxfun(d)
      w <- df(b)
    }
    # be sure no NA are returned, if NA, we want 0:
    # for the discrete branch:   if b_i !in u        --> NA
    # for the continuous branch: if b_i !in range(u) --> NA
    w[is.na(w)] <- 0
  } else if (in_type_ == "params") {
    w <- .distr_pmf(b, u, distr_) # this never returns NA
  }
  # be sure not to return all 0 weights, return ones instead
  # if (sum(w) == 0) { w = w + 1 }
  return(w)
}

#' @title BUIS for Probabilistic Reconciliation of forecasts via conditioning
#'
#' @description
#'
#' Uses the Bottom-Up Importance Sampling algorithm to draw samples from the reconciled
#' forecast distribution, obtained via conditioning.
#'
#' @details
#'
#' The parameter `base_fc` is a list containing n = n_upper + n_bottom elements.
#' The first n_upper elements of the list are the upper base forecasts, in the order given by the rows of A.
#' The elements from n_upper+1 until the end of the list are the bottom base forecasts, in the order given by the columns of A.
#'
#' The i-th element depends on the values of `in_type[[i]]` and `distr[[i]]`.
#'
#' If `in_type[[i]]`='samples', then `base_fc[[i]]` is a vector containing samples from the base forecast distribution.
#'
#' If `in_type[[i]]`='params', then `base_fc[[i]]` is a list containing the estimated:
#'
#' * mean and sd for the Gaussian base forecast if `distr[[i]]`='gaussian', see \link[stats]{Normal};
#' * lambda for the Poisson base forecast if `distr[[i]]`='poisson', see \link[stats]{Poisson};
#' * size and prob (or mu) for the negative binomial base forecast if `distr[[i]]`='nbinom', see \link[stats]{NegBinomial}.
#'
#' See the description of the parameters `in_type` and `distr` for more details.
#'
#' Warnings are triggered from the Importance Sampling step if:
#'
#' * weights are all zeros, then the upper is ignored during reconciliation;
#' * the effective sample size is < 200;
#' * the effective sample size is < 1% of the sample size (`num_samples` if `in_type` is 'params' or the size of the base forecast if if `in_type` is 'samples').
#'
#' Note that warnings are an indication that the base forecasts might have issues.
#' Please check the base forecasts in case of warnings.
#'
#' @param A aggregation matrix (n_upper x n_bottom).
#' @param base_fc A list containing the base_forecasts, see details.
#' @param in_type A string or a list of length n_upper + n_bottom. If it is a list the i-th element is a string with two possible values:
#'
#' * 'samples' if the i-th base forecasts are in the form of samples;
#' * 'params'  if the i-th base forecasts are in the form of estimated parameters.
#'
#' If it `in_type` is a string it is assumed that all base forecasts are of the same type.
#'
#' @param distr A string or a list of length n_upper + n_bottom describing the type of base forecasts.
#' If it is a list the i-th element is a string with two possible values:
#'
#' * 'continuous' or 'discrete' if `in_type[[i]]`='samples';
#' * 'gaussian', 'poisson' or 'nbinom' if `in_type[[i]]`='params'.
#'
#' If `distr` is a string it is assumed that all distributions are of the same type.
#'
#' @param num_samples Number of samples drawn from the reconciled distribution.
#'        This is ignored if `bottom_in_type='samples'`; in this case, the number of reconciled samples is equal to
#'        the number of samples of the base forecasts.
#'
#' @param suppress_warnings Logical. If \code{TRUE}, no warnings about effective sample size
#'        are triggered. If \code{FALSE}, warnings are generated. Default is \code{FALSE}. See Details.
#' @param return_upper Logical, whether to return the reconciled parameters for the upper variables (default is TRUE).
#' @param seed Seed for reproducibility.
#'
#' @return A list containing the reconciled forecasts. The list has the following named elements:
#'
#' * `bottom_rec_samples`: a matrix (n_bottom x `num_samples`) containing the reconciled samples for the bottom time series;
#' * `upper_rec_samples`: (only if `return_upper = TRUE`) a matrix (n_upper x `num_samples`) containing the reconciled samples for the upper time series.
#'
#' @examples
#'
#' library(bayesRecon)
#'
#' # Create a minimal hierarchy with 2 bottom and 1 upper variable
#' rec_mat <- get_reconc_matrices(agg_levels = c(1, 2), h = 2)
#' A <- rec_mat$A
#' S <- rec_mat$S
#'
#'
#' # 1) Gaussian base forecasts
#'
#' # Set the parameters of the Gaussian base forecast distributions
#' mu1 <- 2
#' mu2 <- 4
#' muY <- 9
#' mus <- c(muY, mu1, mu2)
#'
#' sigma1 <- 2
#' sigma2 <- 2
#' sigmaY <- 3
#' sigmas <- c(sigmaY, sigma1, sigma2)
#'
#' base_fc <- list()
#' for (i in 1:length(mus)) {
#'   base_fc[[i]] <- list(mean = mus[[i]], sd = sigmas[[i]])
#' }
#'
#'
#' # Sample from the reconciled forecast distribution using the BUIS algorithm
#' buis <- reconc_BUIS(A, base_fc,
#'   in_type = "params",
#'   distr = "gaussian", num_samples = 100000, seed = 42
#' )
#'
#' samples_buis <- rbind(buis$upper_rec_samples, buis$bottom_rec_samples)
#'
#' # In the Gaussian case, the reconciled distribution is still Gaussian and can be
#' # computed in closed form
#' Sigma <- diag(sigmas^2) # transform into covariance matrix
#' analytic_rec <- reconc_gaussian(A,
#'   base_fc_mean = mus,
#'   base_fc_cov = Sigma
#' )
#'
#' # Compare the reconciled means obtained analytically and via BUIS
#' print(c(S %*% analytic_rec$bottom_rec_mean))
#' print(rowMeans(samples_buis))
#'
#'
#' # 2) Poisson base forecasts
#'
#' # Set the parameters of the Poisson base forecast distributions
#' lambda1 <- 2
#' lambda2 <- 4
#' lambdaY <- 9
#' lambdas <- c(lambdaY, lambda1, lambda2)
#'
#' base_fc <- list()
#' for (i in 1:length(lambdas)) {
#'   base_fc[[i]] <- list(lambda = lambdas[i])
#' }
#'
#' # Sample from the reconciled forecast distribution using the BUIS algorithm
#' buis <- reconc_BUIS(A, base_fc,
#'   in_type = "params",
#'   distr = "poisson", num_samples = 100000, seed = 42
#' )
#' samples_buis <- rbind(buis$upper_rec_samples, buis$bottom_rec_samples)
#'
#' # Print the reconciled means
#' print(rowMeans(samples_buis))
#'
#' @references
#' Zambon, L., Azzimonti, D. & Corani, G. (2024).
#' *Efficient probabilistic reconciliation of forecasts for real-valued and count time series*.
#' Statistics and Computing 34 (1), 21.
#' \doi{10.1007/s11222-023-10343-y}.
#'
#'
#' @seealso
#' [reconc_gaussian()]
#'
#' @export
reconc_BUIS <- function(A,
                        base_fc,
                        in_type,
                        distr,
                        num_samples = 2e4,
                        suppress_warnings = FALSE,
                        return_upper = TRUE,
                        seed = NULL) {
  if (!is.null(seed)) set.seed(seed)
  n_upper <- nrow(A)
  n_bottom <- ncol(A)
  n_tot <- length(base_fc)

  # Transform distr and in_type into lists
  if (!is.list(distr)) {
    distr <- rep(list(distr), n_tot)
  }
  if (!is.list(in_type)) {
    in_type <- rep(list(in_type), n_tot)
  }

  # Ensure that data inputs are valid
  .check_input_BUIS(A, base_fc, in_type, distr)

  # Split bottoms, uppers
  # the first nrow(A) elements of base_fc are upper
  # the second ncol(A) elements of base_fc are lower

  split_hierarchy_res <- list(
    A = A,
    upper = base_fc[1:nrow(A)],
    bottom = base_fc[(nrow(A) + 1):n_tot],
    upper_idxs = 1:nrow(A),
    bottom_idxs = (nrow(A) + 1):n_tot
  )
  upper_base_fc <- split_hierarchy_res$upper
  bottom_base_fc <- split_hierarchy_res$bottom

  # Check on continuous/discrete in relationship to the hierarchy
  .check_hierfamily_rel(split_hierarchy_res, distr)

  # H, G
  is_hier <- .check_hierarchical(A)
  # If A is hierarchical we do not solve the integer linear programming problem
  if (is_hier) {
    H <- A
    G <- NULL
    upper_base_fc_H <- upper_base_fc
    upper_base_fc_G <- NULL
    in_typeH <- in_type[split_hierarchy_res$upper_idxs]
    distr_H <- distr[split_hierarchy_res$upper_idxs]
    in_typeG <- NULL
    distr_G <- NULL
  } else {
    get_HG_res <- .get_HG(A, upper_base_fc, distr[split_hierarchy_res$upper_idxs], in_type[split_hierarchy_res$upper_idxs])
    H <- get_HG_res$H
    upper_base_fc_H <- get_HG_res$Hv
    G <- get_HG_res$G
    upper_base_fc_G <- get_HG_res$Gv
    in_typeH <- get_HG_res$Hin_type
    distr_H <- get_HG_res$Hdistr
    in_typeG <- get_HG_res$Gin_type
    distr_G <- get_HG_res$Gdistr
  }

  # Reconciliation using BUIS

  # 1. Bottom samples
  B <- list()
  in_type_bottom <- in_type[split_hierarchy_res$bottom_idxs]
  for (bi in 1:n_bottom) {
    if (in_type_bottom[[bi]] == "samples") {
      B[[bi]] <- unlist(bottom_base_fc[[bi]])
    } else if (in_type_bottom[[bi]] == "params") {
      B[[bi]] <- .distr_sample(
        bottom_base_fc[[bi]],
        distr[split_hierarchy_res$bottom_idxs][[bi]],
        num_samples
      )
    }
  }
  B <- do.call("cbind", B) # B is a matrix (num_samples x n_bottom)

  out <- .core_reconc_BUIS(
    A = A, H = H, G = G, B = B,
    upper_base_fc_H = upper_base_fc_H,
    in_typeH = in_typeH, distr_H = distr_H,
    upper_base_fc_G = upper_base_fc_G,
    in_typeG = in_typeG, distr_G = distr_G,
    .comp_w = .compute_weights,
    suppress_warnings = suppress_warnings,
    return_upper = return_upper
  )

  # # Bottom-Up IS on the hierarchical part
  # for (hi in 1:nrow(H)) {
  #   c = H[hi, ]
  #   b_mask = (c != 0)
  #   weights = .compute_weights(
  #     b = (B %*% c),
  #     # (num_samples x 1)
  #     u = upper_base_fc_H[[hi]],
  #     in_type_ = in_typeH[[hi]],
  #     distr_ = distr_H[[hi]]
  #   )
  #   check_weights_res = .check_weights(weights)
  #   if (check_weights_res$warning & !suppress_warnings) {
  #     warning_msg = check_weights_res$warning_msg
  #     # add information to the warning message
  #     upper_fromA_i = which(lapply(seq_len(nrow(A)), function(i) sum(abs(A[i,] - c))) == 0)
  #     for (wmsg in warning_msg) {
  #       wmsg = paste(wmsg, paste0("Check the upper forecast at index: ", upper_fromA_i,"."))
  #       warning(wmsg)
  #     }
  #   }
  #   if(check_weights_res$warning & (1 %in% check_weights_res$warning_code)){
  #     next
  #   }
  #   B[, b_mask] = .resample(B[, b_mask], weights)
  # }

  # if (!is.null(G)) {
  #   # Plain IS on the additional constraints
  #   weights = matrix(1, nrow = nrow(B))
  #   for (gi in 1:nrow(G)) {
  #     c = G[gi, ]
  #     weights = weights * .compute_weights(
  #       b = (B %*% c),
  #       u = upper_base_fc_G[[gi]],
  #       in_type_ = in_typeG[[gi]],
  #       distr_ = distr_G[[gi]]
  #     )
  #   }
  #   check_weights_res = .check_weights(weights)
  #   if (check_weights_res$warning & !suppress_warnings) {
  #     warning_msg = check_weights_res$warning_msg
  #     # add information to the warning message
  #     upper_fromA_i = c()
  #     for (gi in 1:nrow(G)) {
  #       c = G[gi, ]
  #       upper_fromA_i = c(upper_fromA_i,
  #                         which(lapply(seq_len(nrow(A)), function(i) sum(abs(A[i,] - c))) == 0))
  #     }
  #     for (wmsg in warning_msg) {
  #       wmsg = paste(wmsg, paste0("Check the upper forecasts at index: ", paste0("{",paste(upper_fromA_i, collapse = ","), "}.")))
  #       warning(wmsg)
  #     }
  #   }
  #   if(!(check_weights_res$warning & (1 %in% check_weights_res$warning_code))){
  #     B = .resample(B, weights)
  #   }

  # }

  return(out)
}

#' Core Reconciliation via Bayesian Universality Information Sharing
#'
#' Internal function that performs the core reconciliation logic for the BUIS method, which
#' reconciles forecasts using importance sampling based on both hierarchical and equality
#' constraints through a Bayesian framework.
#'
#' @param A Matrix defining the overall hierarchy.
#' @param H Matrix defining hierarchical constraints.
#' @param G Matrix defining general linear constraints.
#' @param B Matrix of bottom level base forecast samples.
#' @param upper_base_fc_H List of upper base forecasts for hierarchical constraints.
#' @param in_typeH Character string specifying input type for H forecasts ('pmf', 'samples', or 'params').
#' @param distr_H Character string specifying distribution type for H forecasts ('poisson' or 'nbinom').
#' @param upper_base_fc_G List of upper base forecasts for general constraints.
#' @param in_typeG Character string specifying input type for G forecasts ('pmf', 'samples', or 'params').
#' @param distr_G Character string specifying distribution type for G forecasts ('poisson' or 'nbinom').
#' @param .comp_w Function to compute weights for importance sampling. Default is `.compute_weights`.
#' @param suppress_warnings Logical. If TRUE, suppresses warnings about sample quality. Default is FALSE.
#'
#' @return A list containing:
#'   \itemize{
#'     \item `bottom_rec`: List with reconciled bottom forecasts (pmf and/or samples).
#'     \item `upper_rec_H`: List with reconciled upper forecasts for H constraints.
#'     \item `upper_rec_G`: List with reconciled upper forecasts for G constraints.
#'   }
#'
#' @keywords internal
#' @export
.core_reconc_BUIS <- function(A,
                              H, G,
                              B,
                              upper_base_fc_H,
                              in_typeH,
                              distr_H,
                              upper_base_fc_G,
                              in_typeG,
                              distr_G,
                              .comp_w = .compute_weights,
                              suppress_warnings = FALSE,
                              return_upper = TRUE) {
  # Hierarchical part
  for (hi in 1:nrow(H)) {
    c <- H[hi, ]
    b_mask <- (c != 0)
    weights <- .comp_w(
      b = (B %*% c),
      # (num_samples x 1)
      u = upper_base_fc_H[[hi]],
      in_type_ = in_typeH[[hi]],
      distr_ = distr_H[[hi]]
    )
    check_weights_res <- .check_weights(weights)
    if (check_weights_res$warning & !suppress_warnings) {
      warning_msg <- check_weights_res$warning_msg
      # add information to the warning message
      upper_fromA_i <- which(lapply(seq_len(nrow(A)), function(i) sum(abs(A[i, ] - c))) == 0)
      for (wmsg in warning_msg) {
        wmsg <- paste(wmsg, paste0("Check the upper forecast at index: ", upper_fromA_i, "."))
        warning(wmsg)
      }
    }
    if (check_weights_res$warning & (1 %in% check_weights_res$warning_code)) {
      next
    }
    B[, b_mask] <- .resample(B[, b_mask], weights)
  }

  # Non-hierarchical part
  if (!is.null(G)) {
    # Plain IS on the additional constraints
    weights <- matrix(1, nrow = nrow(B))
    for (gi in 1:nrow(G)) {
      c <- G[gi, ]
      weights <- weights * .comp_w(
        b = (B %*% c),
        u = upper_base_fc_G[[gi]],
        in_type_ = in_typeG[[gi]],
        distr_ = distr_G[[gi]]
      )
    }
    check_weights_res <- .check_weights(weights)
    if (check_weights_res$warning & !suppress_warnings) {
      warning_msg <- check_weights_res$warning_msg
      # add information to the warning message
      upper_fromA_i <- c()
      for (gi in 1:nrow(G)) {
        c <- G[gi, ]
        upper_fromA_i <- c(
          upper_fromA_i,
          which(lapply(seq_len(nrow(A)), function(i) sum(abs(A[i, ] - c))) == 0)
        )
      }
      for (wmsg in warning_msg) {
        wmsg <- paste(wmsg, paste0("Check the upper forecasts at index: ", paste0("{", paste(upper_fromA_i, collapse = ","), "}.")))
        warning(wmsg)
      }
    }
    if (!(check_weights_res$warning & (1 %in% check_weights_res$warning_code))) {
      B <- .resample(B, weights)
    }
  }
  B <- t(B)
  U <- A %*% B
  out <- list(bottom_rec_samples = B)
  if (return_upper) {
    out$upper_rec_samples <- U
  }
  return(out)
}

test_that("Test hierfamily", {
  S <- matrix(
    data = c(
      1, 1, 1, 1, 1, 1,
      1, 1, 0, 0, 0, 0,
      0, 0, 1, 1, 0, 0,
      0, 0, 0, 0, 1, 1,
      1, 0, 0, 0, 0, 0,
      0, 1, 0, 0, 0, 0,
      0, 0, 1, 0, 0, 0,
      0, 0, 0, 1, 0, 0,
      0, 0, 0, 0, 1, 0,
      0, 0, 0, 0, 0, 1
    ), ncol = 6, byrow = TRUE
  )
  Y <- c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
  split_hierarchy_res <- .split_hierarchy(S, Y)

  distr <- list(
    "continuous",
    "continuous", "continuous", "continuous",
    "continuous", "continuous", "continuous", "continuous", "continuous", "continuous"
  )
  expect_equal(.check_hierfamily_rel(split_hierarchy_res, distr, debug = TRUE), 0)

  distr <- list(
    "continuous",
    "discrete", "continuous", "continuous",
    "continuous", "continuous", "continuous", "continuous", "continuous", "continuous"
  )
  expect_equal(.check_hierfamily_rel(split_hierarchy_res, distr, debug = TRUE), -1)

  distr <- list(
    "discrete",
    "continuous", "continuous", "continuous",
    "continuous", "continuous", "continuous", "discrete", "continuous", "continuous"
  )
  expect_equal(.check_hierfamily_rel(split_hierarchy_res, distr, debug = TRUE), -1)

  distr <- list(
    "discrete",
    "continuous", "continuous", "continuous",
    "discrete", "discrete", "discrete", "discrete", "discrete", "discrete"
  )
  expect_equal(.check_hierfamily_rel(split_hierarchy_res, distr, debug = TRUE), 0)

  distr <- list(
    "continuous",
    "continuous", "continuous", "continuous",
    "discrete", "discrete", "discrete", "discrete", "discrete", "discrete"
  )
  expect_equal(.check_hierfamily_rel(split_hierarchy_res, distr, debug = TRUE), 0)

  distr <- list(
    "gaussian",
    "continuous", "continuous", "continuous",
    "discrete", "discrete", "discrete", "discrete", "discrete", "discrete"
  )
  expect_equal(.check_hierfamily_rel(split_hierarchy_res, distr, debug = TRUE), 0)

  distr <- list(
    "gaussian",
    "gaussian", "gaussian", "gaussian",
    "nbinom", "poisson", "poisson", "nbinom", "nbinom", "nbinom"
  )
  expect_equal(.check_hierfamily_rel(split_hierarchy_res, distr, debug = TRUE), 0)

  distr <- list(
    "gaussian",
    "gaussian", "gaussian", "nbinom",
    "nbinom", "poisson", "poisson", "nbinom", "nbinom", "nbinom"
  )
  expect_equal(.check_hierfamily_rel(split_hierarchy_res, distr, debug = TRUE), 0)

  distr <- list(
    "gaussian",
    "gaussian", "gaussian", "nbinom",
    "nbinom", "poisson", "poisson", "nbinom", "gaussian", "nbinom"
  )
  expect_equal(.check_hierfamily_rel(split_hierarchy_res, distr, debug = TRUE), -1)

  distr <- list(
    "gaussian",
    "gaussian", "gaussian", "nbinom",
    "nbinom", "poisson", "poisson", "nbinom", "continuous", "nbinom"
  )
  expect_equal(.check_hierfamily_rel(split_hierarchy_res, distr, debug = TRUE), -1)
})

test_that("Test effective sample size", {
  S <- matrix(data = c(1, 1, 1, 0, 0, 1), nrow = 3, byrow = TRUE)

  # -----------
  n <- 200
  b1 <- stats::rpois(n = n, lambda = 3)
  b2 <- stats::rpois(n = n, lambda = 4)
  u <- stats::rnorm(n = n, mean = 30, sd = 1)
  B <- cbind(b1, b2)
  c <- matrix(S[1, ])
  b <- (B %*% c)

  w <- .compute_weights(b, u, "samples", "continuous")

  check_w <- .check_weights(w, n_eff_min = 200)
  expect_equal(check_w$warning, TRUE)
  expect_equal(check_w$warning_code, 1)
  expect_equal(check_w$n_eff, n)

  # Try the warning message
  # base_forecast = list(u,b1,b2)
  # A = matrix(S[1,],nrow=1)
  # a = reconc_BUIS(A, base_forecast, in_type = "samples", distr = list("continuous","discrete","discrete"), seed=42)

  # -----------
  n <- 199
  b1 <- stats::rpois(n = n, lambda = 3)
  b2 <- stats::rpois(n = n, lambda = 4)
  u <- stats::rnorm(n = n, mean = 30, sd = 1)
  B <- cbind(b1, b2)
  c <- matrix(S[1, ])
  b <- (B %*% c)

  w <- .compute_weights(b, u, "samples", "continuous")

  check_w <- .check_weights(w, n_eff_min = 200)
  expect_equal(check_w$warning, TRUE)
  expect_equal(check_w$warning_code, 1)
  expect_equal(check_w$n_eff, n)

  # Try the warning message
  # base_forecast = list(u,b1,b2)
  # A = matrix(S[1,],nrow=1)
  # a = bayesRecon::reconc_BUIS(A, base_forecast, in_type = "samples", distr = list("continuous","discrete","discrete"), seed=42)

  # -----------
  n <- 2000
  b1 <- stats::rpois(n = n, lambda = 3)
  b2 <- stats::rpois(n = n, lambda = 4)
  u <- stats::rnorm(n = n, mean = 18, sd = 1)
  B <- cbind(b1, b2)
  c <- matrix(S[1, ])
  b <- (B %*% c)

  w <- .compute_weights(b, u, "samples", "continuous")

  check_w <- .check_weights(w, n_eff_min = 200, p_n_eff = 0.01)
  expect_equal(check_w$warning, TRUE)
  expect_equal(check_w$warning_code, c(2, 3))
  expect_equal(check_w$n_eff < 200, TRUE)

  # Try the warning message
  # base_forecast = list(u,b1,b2)
  # A = matrix(S[1,],nrow=1)
  # a = bayesRecon::reconc_BUIS(A, base_forecast, in_type = "samples", distr = list("continuous","discrete","discrete"), seed=42)
})

# A pmf is represented as normalized numeric vector v:
# for each j = 0, ..., M, the probability of j is the value v[[j+1]]

###

# Fit a Negative Binomial distribution on a given vector of samples.
# If data are underdispersed, fit a Poisson.
# If min_supp is specified, the returned pmf must have minimum length of min_supp+1
# Use Rtoll instead of 1e-6?
.fit_static_negbin <- function(v_, toll = .NEGBIN_TOLL) {
  v <- v_[!is.na(v_)] # remove NA
  Mu <- mean(v)
  Var <- stats::var(v)
  if (Var <= Mu) { # if data are underdispersed, fit Poisson
    M <- stats::qpois(1 - toll, Mu)
    pmf <- stats::dpois(0:M, Mu)
  } else { # else, fit Negative Binomial
    size <- Mu^2 / (Var - Mu)
    M <- stats::qnbinom(1 - toll, size = size, mu = Mu)
    pmf <- stats::dnbinom(0:M, size = size, mu = Mu)
  }
  return(pmf / sum(pmf))
}

# Estimate the pmf from a vector of non-negative discrete samples.
# If there are NA, they are removed before computing the pmf.
# Several estimates are possible: naive empirical pmf, parametric, KDE (...)
PMF_from_samples <- function(v_,
                             estim_type = "naive",
                             weights_ = NULL,
                             estim_params = NULL,
                             min_supp = NULL,
                             al_smooth = .ALPHA_SMOOTHING,
                             check_in = TRUE) {
  # First, remove NA
  v <- v_[!is.na(v_)]

  if (check_in) {
    # Check that samples are discrete and non-negative
    .check_discrete_samples(v)
    if (any(v < 0)) stop("Input error: the are negative samples")
  }

  if (estim_type == "naive") {
    if (!is.null(estim_params)) {
      warning("Not yet implemented. Do not specify estim_params if estim_type = 'naive'")
    }

    # TODO: add possibility of doing a smoothing (e.g. Laplace)
    #       maybe default = TRUE only if there are holes?
    # TODO: implement min_supp

    if (is.null(weights_)) {
      pmf <- tabulate(v + 1) / length(v) # the support starts from 0
    } else {
      weights <- weights_[!is.na(v_)]
      weights <- weights / sum(weights)
      pmf <- rep(0, max(v) + 1)
      for (vv in unique(v)) {
        pmf[[vv + 1]] <- sum(weights[v == vv])
      }
    }
  } else if (estim_type == "parametric") {
    if (!is.null(estim_params)) {
      stop("Not yet implemented. Do not specify estim_params if estim_type = 'parametric'")
    }
    # TODO: add more flexibility in the parametric estim (add other distr, e.g. for underdispersed data)

    pmf <- .fit_static_negbin(v)
  } else if (estim_type == "kde") {
    stop("Kernel density estimation not yet implemented")
    # TODO
  } else {
    stop("The choice of estim_type is not valid")
  }

  # pad with zeros to reach the specified length
  if (!is.null(min_supp) && length(pmf) <= min_supp) {
    pmf <- c(pmf, rep(0, min_supp - length(pmf) + 1))
  }

  if (!is.null(al_smooth)) pmf <- PMF_smoothing(pmf, al_smooth, laplace = T)

  return(pmf)
}


# Compute the pmf from a parametric distribution
PMF_from_params <- function(params, distr, Rtoll = .RTOLL) {
  # Check that the distribution is implemented, and that the params are ok
  if (!(distr %in% .DISCR_DISTR)) {
    stop(paste0(
      "Input error: distr must be one of {",
      paste(.DISCR_DISTR, collapse = ", "), "}"
    ))
  }
  .check_distr_params(distr, params)
  # Compute the pmf
  switch(distr,
    "poisson" = {
      lambda <- params$lambda
      M <- stats::qpois(1 - Rtoll, lambda)
      pmf <- stats::dpois(0:M, lambda)
    },
    "nbinom" = {
      size <- params$size
      prob <- params$prob
      mu <- params$mu
      if (!is.null(prob)) {
        M <- stats::qnbinom(1 - Rtoll, size = size, prob = prob)
        pmf <- stats::dnbinom(0:M, size = size, prob = prob)
      } else if (!is.null(mu)) {
        M <- stats::qnbinom(1 - Rtoll, size = size, mu = mu)
        pmf <- stats::dnbinom(0:M, size = size, mu = mu)
      }
    },
  )
  pmf <- pmf / sum(pmf)
  return(pmf)
}

#' @title PMF operations and utilities
#'
#' @description
#'
#' A set of functions for working with Probability Mass Functions (PMFs) in bayesRecon.
#' A PMF is represented as a normalized numeric vector where element `v[j+1]` represents
#' the probability of value `j` (support starts from 0).
#'
#' These functions provide utilities for:
#' * Drawing samples from PMFs
#' * Computing summary statistics (mean, variance, quantiles)
#' * Summarizing PMF distributions
#'
#' @usage
#' PMF_sample(pmf, N_samples)
#' PMF_get_mean(pmf)
#' PMF_get_var(pmf)
#' PMF_get_quantile(pmf, p)
#' PMF_summary(pmf, Ltoll = .TOLL, Rtoll = .RTOLL)
#'
#' @param pmf A PMF object (numeric vector where element j+1 is the probability of value j).
#' @param N_samples Number of samples to draw from the PMF.
#' @param p Probability level for quantile computation. Either a number or a vector with all elements between 0 and 1.
#' @param Ltoll Lower tolerance for computing the minimum of the PMF (default: 1e-15).
#' @param Rtoll Upper tolerance for computing the maximum of the PMF (default: 1e-9).
#'
#' @examples
#' library(bayesRecon)
#'
#' # Let's build the pmf of a Binomial distribution with parameters n and p
#' n <- 10
#' p <- 0.6
#' pmf_binomial <- apply(matrix(seq(0, n)), MARGIN = 1, FUN = \(x) dbinom(x, size = n, prob = p))
#'
#' # Draw samples from the PMF object
#' set.seed(1)
#' samples <- PMF_sample(pmf = pmf_binomial, N_samples = 1e4)
#'
#' # Compute statistics
#' PMF_get_mean(pmf_binomial) # Mean: should be close to n*p = 6
#' PMF_get_var(pmf_binomial) # Variance: should be close to n*p*(1-p) = 2.4
#' PMF_get_quantile(pmf_binomial, 0.5) # Median
#' PMF_summary(pmf_binomial) # Full summary
#'
#' @section Functions:
#'
#' \describe{
#'   \item{\code{PMF_sample(pmf, N_samples)}}{
#'     Draws random samples from the probability distribution specified by the PMF.
#'     Uses sampling with replacement from the discrete support values, weighted by
#'     their probabilities. This is useful for generating synthetic data or for
#'     Monte Carlo simulations.
#'   }
#'
#'   \item{\code{PMF_get_mean(pmf)}}{
#'     Computes the expected value (mean) of the distribution represented by the PMF.
#'     The mean is calculated as the sum of each value in the support weighted by
#'     its probability: \eqn{\sum_{x} x \cdot P(X=x)}.
#'   }
#'
#'   \item{\code{PMF_get_var(pmf)}}{
#'     Computes the variance of the distribution represented by the PMF.
#'     The variance measures the spread of the distribution and is calculated as
#'     \eqn{E[X^2] - (E[X])^2}, where \eqn{E[X]} is the mean.
#'   }
#'
#'   \item{\code{PMF_get_quantile(pmf, p)}}{
#'     Computes the quantile of the distribution at probability level \code{p}.
#'     Returns the smallest value \code{x} such that the cumulative probability up to \code{x}
#'     is greater than or equal to \code{p}. For example, \code{p=0.5} gives the median.
#'     The parameter \code{p} can be a single number or a vector of probabilities.
#'   }
#'
#'   \item{\code{PMF_summary(pmf, Ltoll, Rtoll)}}{
#'     Provides a comprehensive summary of the distribution including minimum, maximum,
#'     quartiles, median, and mean. The minimum and maximum are determined based on
#'     probability thresholds to handle the potentially infinite tails of discrete distributions.
#'   }
#' }
#'
#' @name PMF
#' @export
PMF_sample <- function(pmf, N_samples) {
  s <- sample(0:(length(pmf) - 1), prob = pmf, replace = TRUE, size = N_samples)
  return(s)
}

#' @rdname PMF
#' @export
PMF_get_mean <- function(pmf) {
  supp <- 0:(length(pmf) - 1)
  m <- pmf %*% supp
  return(m)
}

#' @rdname PMF
#' @export
PMF_get_var <- function(pmf) {
  supp <- 0:(length(pmf) - 1)
  v <- pmf %*% supp^2 - PMF_get_mean(pmf)^2
  return(v)
}


#' @rdname PMF
#' @export
PMF_get_quantile <- function(pmf, p) {
  if (min(p) <= 0 | max(p) >= 1) {
    stop("Input error: probability p must be between 0 and 1")
  }
  cdf <- cumsum(pmf)
  x <- pmin(findInterval(p, cdf) + 1, length(cdf))
  return(x - 1)
}


#' @rdname PMF
#' @export
PMF_summary <- function(pmf, Ltoll = .TOLL, Rtoll = .RTOLL) {
  min_pmf <- min(which(pmf > Ltoll)) - 1
  max_pmf <- max(which(pmf > Rtoll)) - 1
  all_summaries <- data.frame(
    "Min." = min_pmf,
    `1st Qu.` = PMF_get_quantile(pmf, 0.25),
    "Median" = PMF_get_quantile(pmf, 0.5),
    "Mean" = PMF_get_mean(pmf),
    `3rd Qu.` = PMF_get_quantile(pmf, 0.75),
    "Max" = max_pmf,
    check.names = FALSE
  )
  return(all_summaries)
}

# Apply smoothing to a pmf.
# If the smoothing parameter alpha is not specified, it is set to the min of pmf.
# If laplace is set to TRUE, add alpha to all the points.
# Otherwise, add alpha only to points with zero mass.
PMF_smoothing <- function(pmf, alpha = .ALPHA_SMOOTHING, laplace = FALSE) {
  if (is.null(alpha)) alpha <- min(pmf[pmf != 0])

  if (laplace) {
    pmf <- pmf + rep(alpha, length(pmf))
  } else {
    pmf[pmf == 0] <- alpha
  }


  return(pmf / sum(pmf))
}


# Compute convolution between 2 pmfs. Then, for numerical reasons:
# -removes small values at the end of the vector (< Rtoll)
# -set to zero all the values to the left of the support
# -set to zero small values (< toll)
PMF_conv <- function(pmf1, pmf2, toll = .TOLL, Rtoll = .RTOLL) {
  pmf <- stats::convolve(pmf1, rev(pmf2), type = "open")
  # Look for last value > Rtoll and remove all the elements after it:
  last_pos <- max(which(pmf > Rtoll))
  pmf <- pmf[1:last_pos]
  # Set to zero values smaller than toll:
  pmf[pmf < toll] <- 0
  # Set to zero elements at the left of m1 + m2, which are the minima of the supports
  # of pmf1 and pmf2: guarantees that supp(v) is not "larger" than supp(v1) + supp(v2)
  m1 <- min(which(pmf1 > 0))
  m2 <- min(which(pmf2 > 0))
  m <- m1 + m2 - 1
  if (m > 1) pmf[1:(m - 1)] <- 0
  pmf <- pmf / sum(pmf) # re-normalize
  return(pmf)
}

# Computes the pmf of the bottom-up distribution analytically
# l_pmf: list of bottom pmfs
# toll and Rtoll: used during convolution (see PMF_conv)
# smoothing: whether to apply smoothing to the bottom pmfs to "cover the holes"
# al_smooth, lap_smooth: smoothing parameters (see PMF_smoothing)
# Returns:
# -the bottom-up pmf, if return_all=FALSE
# -otherwise, a list of lists of pmfs for all the steps of the algorithm;
#  they correspond to the variables of the "auxiliary binary tree"
PMF_bottom_up <- function(l_pmf, toll = .TOLL, Rtoll = .RTOLL, return_all = FALSE,
                          smoothing = TRUE, al_smooth = .ALPHA_SMOOTHING, lap_smooth = .LAP_SMOOTHING) {
  # Smoothing to "cover the holes" in the supports of the bottom pmfs
  if (smoothing) {
    l_pmf <- lapply(l_pmf, PMF_smoothing,
      alpha = al_smooth, laplace = lap_smooth
    )
  }

  # In case we have an upper which is a duplicate of a bottom,
  # the bottom up is simply that bottom.
  if (length(l_pmf) == 1) {
    if (return_all) {
      return(list(l_pmf))
    } else {
      return(l_pmf[[1]])
    }
  }

  # Doesn't do convolutions sequentially
  # Instead, for each iteration (while) it creates a new list of vectors
  # by doing convolution between 1 and 2, 3 and 4, ...
  # Then, the new list has length halved (if L is odd, just copy the last element)
  # Ends when the list has length 1: contains just 1 vector that is the convolution
  # of all the vectors of the list
  old_v <- l_pmf
  l_l_v <- list(old_v) # list with all the step-by-step lists of pmf
  L <- length(old_v)
  while (L > 1) {
    new_v <- c()
    for (j in 1:(L %/% 2)) {
      new_v <- c(new_v, list(PMF_conv(old_v[[2 * j - 1]], old_v[[2 * j]],
        toll = toll, Rtoll = Rtoll
      )))
    }
    if (L %% 2 == 1) new_v <- c(new_v, list(old_v[[L]]))
    old_v <- new_v
    l_l_v <- c(l_l_v, list(old_v))
    L <- length(old_v)
  }

  if (return_all) {
    return(l_l_v)
  } else {
    return(new_v[[1]])
  }
}

# Given a vector v_u and a list of bottom pmf l_pmf,
# checks if the elements of v_u are contained in the support of the bottom-up distr
# Returns a vector with the same length of v_u with TRUE if it is contained and FALSE otherwise
PMF_check_support <- function(v_u, l_pmf, toll = .TOLL, Rtoll = .RTOLL,
                              smoothing = TRUE, al_smooth = .ALPHA_SMOOTHING, lap_smooth = .LAP_SMOOTHING) {
  pmf_u <- PMF_bottom_up(l_pmf,
    toll = toll, Rtoll = Rtoll, return_all = FALSE,
    smoothing = smoothing, al_smooth = al_smooth, lap_smooth = lap_smooth
  )
  # The elements of v_u must be in the support of pmf_u
  supp_u <- which(pmf_u > 0) - 1
  mask <- v_u %in% supp_u
  return(mask)
}

# Compute the tempered pmf
# The pmf is raised to the power of 1/temp, and then normalized
# temp must be a positive number
PMF_tempering <- function(pmf, temp) {
  if (temp <= 0) stop("temp must be positive")
  if (temp == 1) {
    return(pmf)
  }

  temp_pmf <- pmf**(1 / temp)
  return(temp_pmf / sum(temp_pmf))
}

test_that("PMF_conv", {
  # Generate a size and a prob each binomial
  s1 <- 20
  p1 <- 0.6
  s2 <- 30
  p2 <- 0.7

  # Compute the pmf for the two binomials
  pmf1 <- dbinom(seq(0, s1), size = s1, prob = p1)
  pmf2 <- dbinom(seq(0, s2), size = s2, prob = p2)

  # True mean of the convolution
  expected_conv_mean <- s1 * p1 + s2 * p2
  expected_conv_var <- s1 * p1 * (1 - p1) + s2 * p2 * (1 - p2)

  # Compute the PMF of the convolution
  pmf_conv <- PMF_conv(pmf1 = pmf1, pmf2 = pmf2)

  abs(PMF_get_mean(pmf_conv) - expected_conv_mean)
  abs(PMF_get_var(pmf_conv) - expected_conv_var)

  # Check if the convolution mean is equal to the truth
  expect_lt(abs(PMF_get_mean(pmf_conv) - expected_conv_mean), 1e-6)

  # Check if the convolution var is equal to the truth
  expect_lt(abs(PMF_get_var(pmf_conv) - expected_conv_var), 8e-5)
})

test_that("PMF_bottom_up", {
  # Test with 10 bottom
  n_bottom <- 10

  # Create sizes and probs for nbinom bottom distributions
  sizes <- c(seq(11, 15), seq(19, 15))
  probs <- c(rep(0.4, 5), rep(0.7, 5))
  distr <- "nbinom"

  # Compute true BU params (mean/var) and bottom pmfs
  true_bu_mean <- 0
  true_bu_var <- 0
  bott_pmfs <- list()
  for (i in seq(n_bottom)) {
    true_bu_mean <- true_bu_mean + sizes[i] * (1 - probs[i]) / probs[i]
    true_bu_var <- true_bu_var + sizes[i] * (1 - probs[i]) / probs[i]^2
    params <- list(size = sizes[i], prob = probs[i])
    bott_pmfs[[i]] <- PMF_from_params(params = params, distr = distr)
  }
  # Run PMF_bottom_up to get the bottom up pmf
  bottom_up_pmf <- PMF_bottom_up(l_pmf = bott_pmfs)

  # Check if true mean is close enough to bottom up pmf mean
  expect_lt(abs(PMF_get_mean(bottom_up_pmf) - true_bu_mean) / true_bu_mean, 2e-6)

  # Check if true var is close enough to bottom up pmf var
  expect_lt(abs(PMF_get_var(bottom_up_pmf) - true_bu_var) / true_bu_var, 6e-5)
})

test_that("PMF_quantile", {
  n_samples <- 1e5
  size <- 10
  prob <- 0.6
  p <- 0.01

  x <- rnbinom(n_samples, size = size, prob = prob)
  pmf <- PMF_from_samples(x)
  q <- PMF_get_quantile(pmf, p)
  qq <- qnbinom(p, size = size, prob = prob)

  expect_equal(q, qq)
})

test_that("PMF_quantile multivariate", {
  n_samples <- 1e5
  size <- 50
  prob <- 0.5
  p <- seq(from=0.1,to=0.9,length.out=5)
  
  set.seed(42)
  x <- rnbinom(n_samples, size = size, prob = prob)
  pmf <- PMF_from_samples(x)
  q <- PMF_get_quantile(pmf, p)
  qq <- qnbinom(p, size = size, prob = prob)
  
  expect_equal(q, qq)
})

################################################################################
# IMPLEMENTED DISTRIBUTIONS

.DISTR_TYPES <- c("continuous", "discrete")
.DISCR_DISTR <- c("poisson", "nbinom")
.CONT_DISTR <- c("gaussian")

################################################################################
# PARAMETERS FOR PMF CONVOLUTION AND SMOOTHING

.TOLL <- 1e-15
.RTOLL <- 1e-9
.ALPHA_SMOOTHING <- 1e-9
.LAP_SMOOTHING <- FALSE

################################################################################
# OTHER PARAMETERS

.NEGBIN_TOLL <- 1e-6 
# used when fitting a Negative Binomial distribution
.L_SHRINK_RECONC_T <- 1e-4  
# used for shrinking the empirical covariance matrix in reconc_t
.MIN_FRACTION_SAMPLES_OK <- 0.5  
# used in reconc_TDcond: if the fraction of reconciled upper samples that lie 
#in the support of the bottom-up distribution is less than this threshold, the function stops

################################################################################
# CHECK INPUT

# Function to check values allowed in S.
.check_S <- function(S) {
  if (!identical(sort(unique(as.vector(S))), c(0, 1))) {
    stop("Input error in S: S must be a matrix containing only 0s and 1s.")
  }

  if (!all(colSums(S) > 1)) {
    stop("Input error in S: all bottom level forecasts must aggregate into an upper.")
  }

  if (nrow(unique(S)) != nrow(S)) {
    warning("S has some repeated rows.")
  }

  # Check that each bottom has a corresponding row with with one 1 and the rest 0s.
  if (nrow(unique(S[rowSums(S) == 1, ])) < ncol(S)) {
    stop("Input error in S: there is at least one bottom that does not have a row with one 1 and the rest 0s.")
  }
}

# Function to check aggregation matrix A
.check_A <- function(A) {
  if (!all(A %in% c(0, 1))) {
    stop("Input error in A: A must be a matrix containing only 0s and 1s.")
  }

  if (any(colSums(A) == 0)) {
    stop("Input error in A: some columns do not have any 1.
          All bottom level forecasts must aggregate into an upper.")
  }

  if (nrow(unique(A)) != nrow(A)) {
    warning("A has some repeated rows.")
  }
}

# Check if it is a covariance matrix (i.e. symmetric p.d.)
.check_cov <- function(cov_matrix, Sigma_str, pd_check = FALSE, symm_check = FALSE) {
  # Check if the matrix is square
  if (!is.matrix(cov_matrix) || nrow(cov_matrix) != ncol(cov_matrix)) {
    stop(paste0(Sigma_str, " is not square"))
  }

  # Check if the matrix is positive semi-definite
  if (pd_check) {
    eigen_values <- eigen(cov_matrix, symmetric = TRUE)$values
    if (any(eigen_values <= 0)) {
      stop(paste0(Sigma_str, " is not positive semi-definite"))
    }
  }
  if (symm_check) {
    # Check if the matrix is symmetric
    if (!isSymmetric(cov_matrix)) {
      stop(paste0(Sigma_str, " is not symmetric"))
    }
  }
  # Check if the diagonal elements are non-negative
  if (any(diag(cov_matrix) < 0)) {
    stop(paste0(Sigma_str, ": some elements on the diagonal are negative"))
  }
  # If all checks pass, return TRUE
  return(TRUE)
}

# Checks if the input is a real number
.check_real_number <- function(x) {
  return(length(x) == 1 & is.numeric(x))
}

# Checks if the input is a positive number
.check_positive_number <- function(x) {
  return(length(x) == 1 && is.numeric(x) && x > 0)
}

# Check that the distr is implemented
.check_implemented_distr <- function(distr) {
  if (!(distr %in% c(.DISCR_DISTR, .CONT_DISTR))) {
    stop(paste(
      "Input error: the distribution must be one of {",
      paste(c(.DISCR_DISTR, .CONT_DISTR), collapse = ", "), "}"
    ))
  }
}

# Check the parameters of distr
.check_distr_params <- function(distr, params) {
  .check_implemented_distr(distr)
  if (!is.list(params)) {
    stop("Input error: the parameters of the distribution must be given as a list.")
  }
  switch(distr,
    "gaussian" = {
      mean <- params$mean
      sd <- params$sd
      if (!.check_real_number(mean)) {
        stop("Input error: mean of Gaussian must be a real number")
      }
      if (!.check_positive_number(sd)) {
        stop("Input error: sd of Gaussian must be a positive number")
      }
    },
    "poisson" = {
      lambda <- params$lambda
      if (!.check_positive_number(lambda)) {
        stop("Input error: lambda of Poisson must be a positive number")
      }
    },
    "nbinom" = {
      size <- params$size
      prob <- params$prob
      mu <- params$mu
      # Check that size is specified, and that is a positive number
      if (is.null(size)) {
        stop("Input error: size parameter for the nbinom distribution must be specified")
      }
      if (!.check_positive_number(size)) {
        stop("Input error: size of nbinom must be a positive number")
      }
      # Check that exactly one of prob, mu is specified
      if (!is.null(prob) & !is.null(mu)) {
        stop("Input error: prob and mu for the nbinom distribution are both specified ")
      } else if (is.null(prob) & is.null(mu)) {
        stop("Input error: either prob or mu must be specified")
      } else {
        if (!is.null(prob)) {
          if (!.check_positive_number(prob) | prob > 1) {
            stop("Input error: prob of nbinom must be positive and <= 1")
          }
        } else if (!is.null(mu)) {
          if (!.check_positive_number(mu)) {
            stop("Input error: mu of nbinom must be positive")
          }
        }
      }
    },
  )
}

# Check that the samples are discrete
.check_discrete_samples <- function(samples) {
  if (!isTRUE(as.vector(all.equal(as.vector(samples), as.integer(samples))))) {
    stop("Input error: samples are not all discrete")
  }
}

# Check input for BUIS (and for MH)
# base_fc, in_type, and distr must be list
.check_input_BUIS <- function(A, base_fc, in_type, distr) {
  .check_A(A)

  n_tot_A <- ncol(A) + nrow(A)

  # Check in_type
  if (!is.list(in_type)) {
    stop("Input error: in_type must be a list")
  }
  if (!(n_tot_A == length(in_type))) {
    stop("Input error: ncol(A)+nrow(A) != length(in_type)")
  }
  for (i in 1:n_tot_A) {
    if (!(in_type[[i]] %in% c("params", "samples"))) {
      stop("Input error: in_type[[", i, "]] must be either 'samples' or 'params'")
    }
  }

  # Check distr and base forecasts
  if (!is.list(distr)) {
    stop("Input error: distr must be a list")
  }
  if (!(n_tot_A == length(distr))) {
    stop("Input error: ncol(A)+nrow(A) != length(distr)")
  }
  if (!is.list(base_fc)) {
    stop("Input error: base_fc must be a list")
  }
  if (!(n_tot_A == length(base_fc))) {
    stop("Input error: ncol(A)+nrow(A) != length(base_fc)")
  }
  for (i in 1:n_tot_A) {
    if (in_type[[i]] == "params") {
      .check_distr_params(distr[[i]], base_fc[[i]])
    } else if (in_type[[i]] == "samples") {
      if (!(distr[[i]] %in% .DISTR_TYPES)) {
        stop(paste(
          "Input error: the distribution must be one of {",
          paste(.DISTR_TYPES, collapse = ", "), "}"
        ))
      }
      if (distr[[i]] == "discrete") {
        .check_discrete_samples(base_fc[[i]])
      }
      # TODO: check sample size?
    } else {
      stop("Input error: in_type[[", i, "]] must be either 'samples' or 'params'")
    }
  }
}

# Check input for TDcond
.check_input_TD <- function(A, base_fc_bottom, base_fc_upper,
                            bottom_in_type, distr,
                            return_type) {
  .check_A(A)

  n_b <- ncol(A) # number of bottom TS
  n_u <- nrow(A) # number of upper TS

  if (!(bottom_in_type %in% c("pmf", "samples", "params"))) {
    stop("Input error: bottom_in_type must be either 'pmf', 'samples', or 'params'")
  }
  if (!(return_type %in% c("pmf", "samples", "all"))) {
    stop("Input error: return_type must be either 'pmf', 'samples', or 'all'")
  }
  if (length(base_fc_bottom) != n_b) {
    stop("Input error: length of base_fc_bottom does not match with A")
  }
  # If cov is a number, transform into a matrix
  if (length(base_fc_upper$cov) == 1) {
    base_fc_upper$cov <- as.matrix(base_fc_upper$cov)
  }
  # Check the dimensions of mean and cov
  if (length(base_fc_upper$mean) != n_u | any(dim(base_fc_upper$cov) != c(n_u, n_u))) {
    stop("Input error: the dimensions of the upper parameters do not match with A")
  }
  # Check that cov is a covariance matrix (symmetric positive semi-definite)
  .check_cov(base_fc_upper$cov, "Upper covariance matrix", symm_check = TRUE)

  # If bottom_in_type is not "params" but distr is specified, throw a warning
  if (bottom_in_type %in% c("pmf", "samples") & !is.null(distr)) {
    warning(paste0("Since bottom_in_type = '", bottom_in_type, "', the input distr is ignored"))
  }
  # If bottom_in_type is params, distr must be one of the implemented discrete distr.
  # Also, check the parameters
  if (bottom_in_type == "params") {
    if (is.null(distr)) {
      stop("Input error: if bottom_in_type = 'params', distr must be specified")
    }
    if (!(distr %in% .DISCR_DISTR)) {
      stop(paste0(
        "Input error: distr must be one of {",
        paste(.DISCR_DISTR, collapse = ", "), "}"
      ))
    }
    for (i in 1:n_b) {
      .check_distr_params(distr, base_fc_bottom[[i]])
    }
  }
}

.check_input_t <- function(A, base_fc_mean, y_train, residuals, ...) {
  .check_A(A)

  n_b <- ncol(A) # number of bottom TS
  n_u <- nrow(A) # number of upper TS

  if (!is.vector(base_fc_mean)) {
    stop("Input error: base_fc_mean must be a vector")
  }
  n <- length(base_fc_mean)
  if (n_u + n_b != n) {
    stop("Input error: the length of base_fc_mean must be equal to nrow(A) + ncol(A)")
  }
  
  add_args <- list(...)
  prior <- add_args$prior
  posterior <- add_args$posterior
  freq <- add_args$freq
  criterion <- add_args$criterion

  ##############################################################################
  ### CASE 1 ###
  # If posterior is provided, check if is a list with entries nu and Psi and extract values
  if (!is.null(posterior)) {
    if (is.list(posterior)) {
      nu_post <- posterior$nu
      Psi_post <- posterior$Psi
      if (is.null(nu_post) | is.null(Psi_post)) {
        stop("Input error: posterior must be a list with entries nu and Psi")
      } else if (!is.numeric(nu_post) | length(nu_post) != 1 | nu_post <= n-1) {
        stop("Input error: nu in posterior must be a number greater than n. of series - 1")
      } else if (!is.matrix(Psi_post) | any(dim(Psi_post) != c(n, n))) {
        stop("Input error: Psi in posterior must be a matrix with dimensions compatible with base_fc_mean")
      }
      # If posterior is provided, then y_train, residuals, and prior are ignored:
      # if they are provided, throw a warning
      if (!is.null(y_train)) {
        warning("Input warning: posterior is provided, ignoring y_train")
      }
      if (!is.null(residuals)) {
        warning("Input warning: posterior is provided, ignoring residuals")
      }
      if (!is.null(prior)) {
        warning("Input warning: posterior is provided, ignoring prior")
      }
    } else {
      stop("Input error: posterior must be a list with entries nu and Psi")
    }

    ### CASE 2 ###
    # If posterior not provided, first check that residuals are provided
  } else {
    if (is.null(residuals)) {
      stop("Input error: either posterior or residuals must be provided")
    }
    if (!is.matrix(residuals)) {
      stop("Input error: residuals must be a matrix")
    }
    if (ncol(residuals) != n) {
      stop("Input error: number of columns of residuals must be equal to length of base_fc_mean")
    }

    L <- nrow(residuals) # number of residual samples (i.e., training length)
    if (L < 10) {
      warning("Warning: number of rows of residuals is less than 10, covariance estimation may be inaccurate")
    }
    # TODO: implement fallback

    ### CASE 2a ###
    # If prior is provided, check if is a list with entries nu and Psi and extract values
    if (!is.null(prior)) {
      if (is.list(prior)) {
        nu_prior <- prior$nu
        Psi_prior <- prior$Psi
        if (is.null(nu_prior) | is.null(Psi_prior)) {
          stop("Input error: prior must be a list with entries nu and Psi")
        } else if (!is.numeric(nu_prior) | length(nu_prior) != 1 | nu_prior <= n-1) {
          stop("Input error: nu in prior must be a number greater than n. of series - 1")
        } else if (!is.matrix(Psi_prior) | any(dim(Psi_prior) != c(n, n))) {
          stop("Input error: Psi in prior must be a matrix with dimensions compatible with base_fc_mean")
        }
        # If prior is provided, then y_train is ignored: if it is provided, throw a warning
        if (!is.null(y_train)) {
          warning("Input warning: prior is provided, ignoring y_train")
        }
      } else {
        stop("Input error: prior must be a list with entries nu and Psi")
      }
    }

    ### CASE 2b ###
    # If prior not provided:
    # - compute Psi using the (shrinked) covariance matrix of the residuals of the naive
    #   or seasonal naive forecasts
    # - set nu using LOOCV
    else {
      if (is.null(y_train)) {
        stop("Input error: y_train must be provided when neither prior nor posterior are given")
      }
      if (!is.matrix(y_train)) {  
        stop("Input error: y_train must be a matrix")
      }
      # this works also if y_train is a mts; TODO: allow for other types, e.g. data.frame
      if (ncol(y_train) != n) {
        stop("Input error: number of columns of y_train must be equal to length of base_fc_mean")
      }
      if (nrow(y_train) != L) {
        warning("Numbers of rows of y_train and of residuals are different!")
      }
      
      if (!is.null(freq)) {
        if (!is.numeric(freq) | length(freq) != 1 | freq < 1 | (freq %% 1) != 0) {
          stop("Input error: freq must be a positive integer")
        }
      }
      
      # Current logic: 
      # * if y_train is a mts object: check if freq is provided and is equal to the frequency of y_train; 
      #                               if they are different, throw a warning and ignore the frequency of y_train (see compute_naive_cov function)
      # * if y_train is not a mts object: check if freq is provided and is a positive integer; 
      #                                   if not, throw a warning  
      if (stats::is.mts(y_train)) {
        if (!is.null(freq) && freq != stats::frequency(y_train)) {
          warning("Input warning: the provided freq is different from the frequency of y_train.
                   The frequency of y_train will be ignored.")
        } 
      } else {
        if (is.null(freq)) {
          warning("Input warning: y_train is not a mts object and freq is not provided. 
                   The prior will be set assuming no seasonality.")
        }
      }
      
      if (!is.null(criterion)) {
        if (!(criterion %in% c("RSS", "seas-test"))) {
          stop("Input error: criterion must be either 'RSS' or 'seas-test'")
        }
      }
    }
  }
}

# Check importance sampling weights
.check_weights <- function(w, n_eff_min = 200, p_n_eff = 0.01) {
  warning <- FALSE
  warning_code <- c()
  warning_msg <- c()

  n <- length(w)
  n_eff <- n

  # 1. w==0
  if (all(w == 0)) {
    warning <- TRUE
    warning_code <- c(warning_code, 1)
    warning_msg <- c(
      warning_msg,
      "Importance Sampling: all the weights are zeros. This is probably caused by a strong incoherence between bottom and upper base forecasts."
    )
  } else {
    # Effective sample size
    w <- w / sum(w)
    n_eff <- 1 / sum(w^2)

    # 2. n_eff < threshold
    if (n_eff < n_eff_min) {
      warning <- TRUE
      warning_code <- c(warning_code, 2)
      warning_msg <- c(
        warning_msg,
        paste0("Importance Sampling: effective_sample_size= ", round(n_eff, 2), " (< ", n_eff_min, ").")
      )
    }

    # 3. n_eff < p*n, e.g. p = 0.05
    if (n_eff < p_n_eff * n) {
      warning <- TRUE
      warning_code <- c(warning_code, 3)
      warning_msg <- c(
        warning_msg,
        paste0("Importance Sampling: effective_sample_size= ", round(n_eff, 2), " (< ", round(p_n_eff * 100, 2), "%).")
      )
    }
  }
  res <- list(
    warning = warning,
    warning_code = warning_code,
    warning_msg = warning_msg,
    n_eff = n_eff
  )

  return(res)
}

################################################################################
# SAMPLE

# Sample from one of the implemented distributions
.distr_sample <- function(params, distr, n) {
  .check_distr_params(distr, params)
  switch(distr,
    "gaussian" = {
      mean <- params$mean
      sd <- params$sd
      samples <- stats::rnorm(n = n, mean = mean, sd = sd)
    },
    "poisson" = {
      lambda <- params$lambda
      samples <- stats::rpois(n = n, lambda = lambda)
    },
    "nbinom" = {
      size <- params$size
      prob <- params$prob
      mu <- params$mu
      if (!is.null(prob)) {
        samples <- stats::rnbinom(n = n, size = size, prob = prob)
      } else if (!is.null(mu)) {
        samples <- stats::rnbinom(n = n, size = size, mu = mu)
      }
    },
  )
  return(samples)
}

# Sample from a multivariate Gaussian distribution with specified mean and cov. matrix
.MVN_sample <- function(n_samples, mu, Sigma) {
  n <- length(mu)
  if (any(dim(Sigma) != c(n, n))) {
    stop("Dimensions of mean and covariance matrix are not compatible!")
  }
  .check_cov(Sigma, "Covariance matrix", pd_check = FALSE, symm_check = FALSE)

  Z <- matrix(stats::rnorm(n * n_samples), ncol = n)

  Ch <- tryCatch(base::chol(Sigma),
    error = function(e) stop(paste0(e, "check the covariance in .MVN_sample, the Cholesky fails."))
  )

  samples <- Z %*% Ch + matrix(mu, nrow = n_samples, ncol = n, byrow = TRUE)
  return(samples)
}

# Compute the MVN density
.MVN_density <- function(x, mu, Sigma, max_size_x = 5e3, suppress_warnings = TRUE) {
  # save dimension of mu
  n <- length(mu)

  # Check Sigma
  if (any(dim(Sigma) != c(n, n))) {
    stop("Dimension of mu and Sigma are not compatible!")
  }
  .check_cov(Sigma, "Sigma", pd_check = FALSE, symm_check = FALSE)

  # x must be a matrix with ncol = n (nrow is the number of points to evaluate)
  # or a vector with length n (in which case it is transformed into a matrix)
  if (is.vector(x)) {
    if (length(x) != n) stop("Length of x must be the same of mu")
    x <- matrix(x, ncol = length(x))
  } else if (is.matrix(x)) {
    if (ncol(x) != n) stop("The number of columns of x must be equal to the length of mu")
  } else {
    stop("x must be either a vector or a matrix")
  }

  # Compute Cholesky of Sigma
  chol_S <- tryCatch(base::chol(Sigma),
    error = function(e) stop(paste0(e, "check the covariance in .MVN_density, the Cholesky fails."))
  )

  # Constant of the loglikelihood (computed here because it is always the same)
  const <- -sum(log(diag(chol_S))) - 0.5 * n * log(2 * pi)

  # This part breaks down the evaluation of the density eval into batches, for memory
  rows_x <- nrow(x)

  if (rows_x > max_size_x) {
    logval <- rep(0, rows_x)

    # Compute how many batches we need
    num_backsolves <- rows_x %/% max_size_x

    if (!suppress_warnings) {
      warning_msg <- paste0("x has ", rows_x, " rows, the density evaluation is broken down into ", num_backsolves, " pieces for memory preservation.")
      warning(warning_msg)
    }

    for (j in seq(num_backsolves)) {
      idx_to_select <- (1 + (j - 1) * max_size_x):((j) * max_size_x)
      # Do one backsolve for each batch
      tmp <- backsolve(chol_S, t(x[idx_to_select, ]) - mu, transpose = TRUE)
      rss <- colSums(tmp^2)

      # Update the logval for those indices
      logval[idx_to_select] <- const - 0.5 * rss
    }

    # Last indices: if the number of rows of x is not exactly divided by the size of the batches
    remainder <- rows_x %% max_size_x
    if (remainder != 0) {
      idx_to_select <- (1 + (num_backsolves) * max_size_x):(remainder + (num_backsolves) * max_size_x)
      # Do backsolve on the remaining indices
      tmp <- backsolve(chol_S, t(x[idx_to_select, ]) - mu, transpose = TRUE)
      rss <- colSums(tmp^2)

      logval[idx_to_select] <- const - 0.5 * rss
    }
  } else {
    tmp <- backsolve(chol_S, t(x) - mu, transpose = TRUE)
    rss <- colSums(tmp^2)

    logval <- const - 0.5 * rss
  }

  return(exp(logval))
}

# Resample from weighted sample
.resample <- function(S_, weights, num_samples = NA) {
  if (is.na(num_samples)) {
    num_samples <- length(weights)
  }

  if (nrow(S_) != length(weights)) {
    stop("Error in .resample: nrow(S_) != length(weights)")
  }

  tmp_idx <- sample(x = 1:nrow(S_), num_samples, replace = TRUE, prob = weights)
  return(S_[tmp_idx, ])
}

################################################################################
# Miscellaneous

# Compute the pmf of the distribution specified by distr and params at the points x
.distr_pmf <- function(x, params, distr) {
  .check_distr_params(distr, params)
  switch(distr,
    "gaussian" = {
      mean <- params$mean
      sd <- params$sd
      pmf <- stats::dnorm(x = x, mean = mean, sd = sd)
    },
    "poisson" = {
      lambda <- params$lambda
      pmf <- stats::dpois(x = x, lambda = lambda)
    },
    "nbinom" = {
      size <- params$size
      prob <- params$prob
      mu <- params$mu
      if (!is.null(prob)) {
        pmf <- stats::dnbinom(x = x, size = size, prob = prob)
      } else if (!is.null(mu)) {
        pmf <- stats::dnbinom(x = x, size = size, mu = mu)
      }
    },
  )
  return(pmf)
}

.shape <- function(m) {
  print(paste0("(", nrow(m), ",", ncol(m), ")"))
}

################################################################################
# Functions for tests

.gen_gaussian <- function(params_file, seed = NULL) {
  if (!is.null(seed)) set.seed(seed)
  params <- utils::read.csv(file = params_file, header = FALSE)
  out <- list()
  for (i in 1:nrow(params)) {
    out[[i]] <- stats::rnorm(n = 1e6, mean = params[[1]][i], sd = params[[2]][i])
  }
  return(out)
}

.gen_poisson <- function(params_file, seed = NULL) {
  if (!is.null(seed)) set.seed(seed)
  params <- utils::read.csv(file = params_file, header = FALSE)
  out <- list()
  for (i in 1:nrow(params)) {
    out[[i]] <- stats::rpois(n = 1e6, lambda = params[[1]][i])
  }
  return(out)
}

test_that("sampling from univariate normal", {
  # Generate 1e4 samples from univariate Gaussian
  params <- list(mean = 42, sd = 1)
  distr <- "gaussian"
  n <- 1e4
  samples <- .distr_sample(params, distr, n)

  # Compute empirical mean and sd
  sam_mean <- mean(samples)
  sam_sd <- sd(samples)

  # Check how close empirical values are to the truth
  m <- abs(sam_mean - 42) / 42
  s <- abs(sam_sd - 1)

  expect_equal(m < 2e-3, TRUE)
  expect_equal(s < 4e-2, TRUE)
})

test_that("sampling from univariate nbinom", {
  # Generate 1e4 samples from negative binomial (size, prob)
  params <- list(size = 12, prob = 0.8)
  distr <- "nbinom"
  n <- 1e4
  samples <- .distr_sample(params, distr, n)

  # Compute empirical mean
  sam_mean <- mean(samples)
  true_mean <- params$size * (1 - params$prob) / params$prob

  # Check how close empirical values are to the truth
  m <- abs(sam_mean - true_mean) / true_mean

  expect_equal(m < 3e-2, TRUE)

  # Generate 1e4 samples from negative binomial (size, mu)
  params <- list(size = 12, mu = true_mean)
  distr <- "nbinom"
  n <- 1e4
  samples <- .distr_sample(params, distr, n)

  # Compute empirical mean
  sam_mean <- mean(samples)

  # Check how close empirical values are to the truth
  m <- abs(sam_mean - params$mu) / params$mu

  expect_equal(m < 3e-2, TRUE)

  # Check if it returns an error when all 3 parameters are specified
  params <- list(size = 12, mu = true_mean, prob = 0.8)
  distr <- "nbinom"
  n <- 1e4
  expect_error(.distr_sample(params, distr, n))

  # Check if it returns an error when size is not specified
  params <- list(mu = true_mean, prob = 0.8)
  distr <- "nbinom"
  n <- 1e4
  expect_error(.distr_sample(params, distr, n))
})

test_that("sampling from univariate poisson", {
  # Generate 1e4 samples from poisson
  params <- list(lambda = 10)
  distr <- "poisson"
  n <- 1e4
  samples <- .distr_sample(params, distr, n)

  # Compute empirical mean
  sam_mean <- mean(samples)

  # Check how close empirical values are to the truth
  m <- abs(sam_mean - 10) / 10

  expect_equal(m < 3e-2, TRUE)
})

test_that("sampling from multivariate normal", {
  # Generate 1e4 samples from bivariate normal
  mu <- c(10, 10)
  Sigma <- matrix(c(1, 0.7, 0.7, 1), nrow = 2)
  n <- 1e4
  samples <- .MVN_sample(n, mu, Sigma)

  # Compute empirical mean
  sam_mean <- colMeans(samples)

  # Check how close empirical values are to the truth
  m <- abs(sam_mean - 10) / 10

  expect_equal(all(m < 8e-3), TRUE)
})

test_that("MVN density works", {
  # Create 3x3 covariance matrix
  L <- matrix(0, nrow = 3, ncol = 3)
  L[lower.tri(L, diag = TRUE)] <- c(0.9, 0.8, 0.5, 0.9, 0.2, 0.6)
  Sigma <- L %*% t(L)

  # create mean vector
  mu <- c(0, 1, -1)

  # matrix where to evaluate the MVN
  xx <- matrix(c(
    0, 2, 1,
    2, 3, 4,
    0.5, 0.5, 0.5,
    0, 1, -1
  ), ncol = 3, byrow = TRUE)

  res <- .MVN_density(x = xx, mu = mu, Sigma = Sigma)

  true_val <- c(8.742644e-04, 1.375497e-11, 3.739985e-03, 1.306453e-01)
  expect_equal(res, true_val, tolerance = 1e-6)

  # Check if block-evaluation works
  xx <- matrix(runif(3 * 1e4), ncol = 3, byrow = TRUE)

  res_chuncks <- .MVN_density(x = xx, mu = mu, Sigma = Sigma)
  res_all <- .MVN_density(x = xx, mu = mu, Sigma = Sigma, max_size_x = 1e4)

  expect_equal(res_chuncks, res_all)
})

test_that("Monthly, in_type=='params', distr='gaussian'", {
  # Run IS Reconc
  A <- read.csv(file = "dataForTests/Monthly-Gaussian_A.csv", header = FALSE)
  A <- as.matrix(A)
  base_forecasts_in <- read.csv(file = "dataForTests/Monthly-Gaussian_basef.csv", header = FALSE)
  base_forecasts <- list()
  for (i in 1:nrow(base_forecasts_in)) {
    base_forecasts[[i]] <- list(
      mean = base_forecasts_in[i, 1],
      sd = base_forecasts_in[i, 2]
    )
  }
  res.buis <- reconc_BUIS(A, base_forecasts,
    in_type = "params", distr = "gaussian", num_samples = 100000, seed = 42
  )
  # Run Gauss Reconc
  res.gauss <- reconc_gaussian(A, base_forecasts_in[[1]], diag(base_forecasts_in[[2]]^2))
  # Test on bottom
  n_upper <- nrow(A)
  n_bottom <- ncol(A)
  m <- mean(rowMeans(res.buis$bottom_rec_samples) - as.numeric(res.gauss$bottom_rec_mean))
  expect_equal(abs(m) < 8e-3, TRUE)
})

test_that("Weekly, in_type=='params', distr='gaussian'", {
  # Run IS Reconc
  A <- read.csv(file = "dataForTests/Weekly-Gaussian_A.csv", header = FALSE)
  A <- as.matrix(A)
  mode(A) <- "numeric"
  base_forecasts_in <- read.csv(file = "dataForTests/Weekly-Gaussian_basef.csv", header = FALSE)
  base_forecasts <- list()
  for (i in 1:nrow(base_forecasts_in)) {
    base_forecasts[[i]] <- list(
      mean = base_forecasts_in[i, 1],
      sd = base_forecasts_in[i, 2]
    )
  }
  res.buis <- reconc_BUIS(A, base_forecasts,
    in_type = "params", distr = "gaussian", num_samples = 100000, seed = 42
  )
  # Run Gauss Reconc
  res.gauss <- reconc_gaussian(A, base_forecasts_in[[1]], diag(base_forecasts_in[[2]]^2))
  # Test on bottom
  n_upper <- nrow(A)
  n_bottom <- ncol(A)
  m <- mean(rowMeans(res.buis$bottom_rec_samples) - as.numeric(res.gauss$bottom_rec_mean))
  expect_equal(abs(m) < 2e-2, TRUE)
})

test_that("Monthly, in_type=='params', distr='poisson'", {
  A <- read.csv(file = "dataForTests/Monthly-Poisson_A.csv", header = FALSE)
  A <- as.matrix(A)
  base_forecasts_in <- read.csv(file = "dataForTests/Monthly-Poisson_basef.csv", header = FALSE)
  base_forecasts <- list()
  for (i in 1:nrow(base_forecasts_in)) {
    base_forecasts[[i]] <- list(lambda = base_forecasts_in[i, 1])
  }
  res.buis <- reconc_BUIS(A, base_forecasts,
    in_type = "params", distr = "poisson", num_samples = 100000, seed = 42
  )
  expect_no_error(res.buis)
})


test_that("Monthly, in_type=='params', distr='nbinom'", {
  A <- read.csv(file = "dataForTests/Monthly-NegBin_A.csv", header = FALSE)
  A <- as.matrix(A)
  base_forecasts_in <- read.csv(file = "dataForTests/Monthly-NegBin_basef.csv", header = FALSE)
  base_forecasts <- list()
  for (i in 1:nrow(base_forecasts_in)) {
    base_forecasts[[i]] <- list(
      size = base_forecasts_in[i, 2],
      mu = base_forecasts_in[i, 1]
    )
  }
  res.buis <- reconc_BUIS(A, base_forecasts,
    in_type = "params", distr = "nbinom", num_samples = 10000, seed = 42
  )
  expect_no_error(res.buis)
})


test_that("Monthly, in_type=='samples', distr='continuous'", {
  # Run IS Reconc from samples
  A <- read.csv(file = "dataForTests/Monthly-Gaussian_A.csv", header = FALSE)
  A <- as.matrix(A)
  base_forecasts <- .gen_gaussian("dataForTests/Monthly-Gaussian_basef.csv", seed = 42)
  res.buis_samples <- reconc_BUIS(A, base_forecasts, in_type = "samples", distr = "continuous", seed = 42)
  # Run IS Reconc
  base_forecasts_in <- read.csv(file = "dataForTests/Monthly-Gaussian_basef.csv", header = FALSE)
  base_forecasts <- list()
  for (i in 1:nrow(base_forecasts_in)) {
    base_forecasts[[i]] <- list(
      mean = base_forecasts_in[i, 1],
      sd = base_forecasts_in[i, 2]
    )
  }
  res.buis <- reconc_BUIS(A, base_forecasts, in_type = "params", distr = "gaussian", num_samples = 100000, seed = 42)
  m <- mean(rowMeans(res.buis$bottom_rec_samples) - rowMeans(res.buis_samples$bottom_rec_samples))
  expect_equal(abs(m) < 1e-2, TRUE)
})

test_that("Monthly, in_type=='samples', distr='discrete'", {
  # Run IS Reconc from samples
  A <- read.csv(file = "dataForTests/Monthly-Poisson_A.csv", header = FALSE)
  A <- as.matrix(A)
  base_forecasts <- .gen_poisson("dataForTests/Monthly-Poisson_basef.csv", seed = 42)
  res.buis_samples <- reconc_BUIS(A, base_forecasts, in_type = "samples", distr = "discrete", seed = 42)
  # Run IS Reconc
  base_forecasts_in <- read.csv(file = "dataForTests/Monthly-Poisson_basef.csv", header = FALSE)
  base_forecasts <- list()
  for (i in 1:nrow(base_forecasts_in)) {
    base_forecasts[[i]] <- list(lambda = base_forecasts_in[i, 1])
  }
  res.buis <- reconc_BUIS(A, base_forecasts, in_type = "params", distr = "poisson", num_samples = 100000, seed = 42)
  m <- mean(rowMeans(res.buis$bottom_rec_samples) - rowMeans(res.buis_samples$bottom_rec_samples))
  expect_equal(abs(m) < 1.5e-2, TRUE)
})

test_that("Monthly simple, in_type=='params', distr='nbinom'", {
  # Read samples from dataForTests (reproducibility)
  vals <- read.csv(file = "dataForTests/Monthly-Count_ts.csv", header = FALSE)

  # Create a count time series with monthly observations for 10 years
  y <- ts(data = vals, frequency = 12)

  # Create the aggregated yearly time series
  y_agg <- temporal_aggregation(y, agg_levels = c(1, 12))

  # We use a marginal forecast that computes for each month
  # the empirical mean and variance
  # the forecast is a negative binomial with those params
  fc_bottom <- list()
  for (i in seq(12)) {
    mm <- mean(y_agg$`f=12`[seq(i, 120, 12)])
    vv <- max(var(y_agg$`f=12`[seq(i, 120, 12)]), mm + 0.5)
    # cat("i: ",i, "mean: ",mm, "var: ",vv, "size: ",mm^2/(vv-mm), "prob: ",mm/vv, "\n")

    fc_bottom[[i]] <- list(size = mm^2 / (vv - mm), mu = mm)
  }

  # We compute the empirical mean and variance of the yearly ts
  # we forecast with a negative binomial with those parameters
  mm <- mean(y_agg$`f=1`)
  vv <- var(y_agg$`f=1`)
  fc_upper <- list(size = mm^2 / (vv - mm), prob = mm / vv)

  # Obtain the aggregation matrix for this hierarchy
  rec_mat <- get_reconc_matrices(c(1, 12), 12)

  base_forecasts <- append(list(fc_upper), fc_bottom)
  res.buis_params <- reconc_BUIS(rec_mat$A, base_forecasts, in_type = "params", distr = "nbinom", seed = 42)


  fc_upper_gauss <- list(mean = mm, cov = matrix(vv))
  res.mixCond <- reconc_MixCond(rec_mat$A, fc_bottom, fc_upper_gauss, bottom_in_type = "params", distr = "nbinom")
  upp_pmf <- PMF_from_samples(as.integer(res.buis_params$upper_rec_samples))

  expect_equal(res.mixCond$upper_rec_pmf[[1]], upp_pmf, tolerance = 0.1)
})


# reconc_gaussian with residuals -----------------------------------------------

# Shared fixture
set.seed(42)
.n_gauss   <- 3L
.L_gauss   <- 60L
.A_gauss   <- matrix(c(1, 1), nrow = 1)
.res_gauss <- matrix(rnorm(.L_gauss * .n_gauss), nrow = .L_gauss, ncol = .n_gauss)
.mu_gauss  <- c(11, 4, 6)  # incoherent: 11 != 4 + 6 = 10

test_that("reconc_gaussian runs without error when residuals are provided", {
  expect_no_error(
    reconc_gaussian(.A_gauss, base_fc_mean = .mu_gauss, residuals = .res_gauss)
  )
})

test_that("reconc_gaussian output structure is correct with residuals", {
  result <- reconc_gaussian(.A_gauss, base_fc_mean = .mu_gauss, residuals = .res_gauss)

  expect_named(result, c("bottom_rec_mean", "bottom_rec_cov"), ignore.order = TRUE)
  expect_length(result$bottom_rec_mean, ncol(.A_gauss))
  expect_equal(dim(result$bottom_rec_cov), c(ncol(.A_gauss), ncol(.A_gauss)))
})

test_that("reconc_gaussian reconciled covariance is symmetric and positive definite with residuals", {
  result <- reconc_gaussian(.A_gauss, base_fc_mean = .mu_gauss, residuals = .res_gauss)

  S <- result$bottom_rec_cov
  expect_equal(S, t(S), tolerance = 1e-12)
  eigs <- eigen(S, symmetric = TRUE, only.values = TRUE)$values
  expect_true(all(eigs > 0))
})

test_that("reconc_gaussian satisfies hierarchical constraint with residuals", {
  result <- reconc_gaussian(.A_gauss, base_fc_mean = .mu_gauss, residuals = .res_gauss,
    return_upper = TRUE
  )

  expect_equal(
    as.vector(.A_gauss %*% result$bottom_rec_mean),
    as.vector(result$upper_rec_mean),
    tolerance = 1e-10
  )
  expect_equal(
    .A_gauss %*% result$bottom_rec_cov %*% t(.A_gauss),
    result$upper_rec_cov,
    tolerance = 1e-10
  )
})

test_that("return_upper=TRUE returns upper fields with residuals", {
  result <- reconc_gaussian(.A_gauss, base_fc_mean = .mu_gauss, residuals = .res_gauss,
    return_upper = TRUE
  )
  expect_true(all(c("upper_rec_mean", "upper_rec_cov") %in% names(result)))
})

# Regression test: hard-coded reference values
# Reference computed with:
#   set.seed(42)
#   n_series <- 3L; L_train <- 60L
#   A <- matrix(c(1,1), nrow=1)
#   res_mat <- matrix(rnorm(L_train * n_series), nrow=L_train, ncol=n_series)
#   mu <- c(11, 4, 6)
#   reconc_gaussian(A, base_fc_mean=mu, residuals=res_mat, return_upper=TRUE)

test_that("reconc_gaussian with residuals matches hard-coded reference values", {
  set.seed(42)
  n_series <- 3L
  L_train  <- 60L
  A        <- matrix(c(1, 1), nrow = 1)
  res_mat  <- matrix(rnorm(L_train * n_series), nrow = L_train, ncol = n_series)
  mu       <- c(11, 4, 6)

  result <- reconc_gaussian(A, base_fc_mean = mu, residuals = res_mat, return_upper = TRUE)

  # bottom_rec_mean
  expect_equal(result$bottom_rec_mean, c(4.284368, 6.268177), tolerance = 1e-5)

  # bottom_rec_cov
  expect_equal(
    result$bottom_rec_cov,
    matrix(c(0.5930640, -0.2222453, -0.2222453, 0.5719506), nrow = 2),
    tolerance = 1e-5
  )

  # upper_rec_mean
  expect_equal(result$upper_rec_mean, 10.55254, tolerance = 1e-4)

  # upper_rec_cov
  expect_equal(result$upper_rec_cov, matrix(0.7205241), tolerance = 1e-5)
})

##############################################################################

test_that("reconc_TDcond simple example", {
  # Simple example with
  # - 12 bottom
  # - 10 upper: year, 6 bi-monthly, 3 quarterly
  A <- matrix(
    data = c(
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
      1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
      1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1
    ),
    nrow = 10, byrow = TRUE
  )

  # Define means and vars for the forecasts
  means <- c(90, 31, 32, 31, 33, 31, 32, 62, 63, 64, rep(15, 12))
  vars <- c(20, 4, 4, 4, 4, 4, 4, 8, 8, 8, rep(2, 12))^2

  # create the lists for reconciliation
  ## upper
  fc_upper <- list(
    mean = means[1:10],
    cov = diag(vars[1:10])
  )

  ## bottom
  fc_bottom <- list()
  for (i in seq(ncol(A))) {
    fc_bottom[[i]] <- as.integer(.distr_sample(list(mean = means[i + 10], sd = vars[i + 10]), "gaussian", 2e4))
    fc_bottom[[i]][which(fc_bottom[[i]] < 0)] <- 0 # set-negative-to-zero
  }


  # reconciliation with TDcond
  res.TDcond <- reconc_TDcond(A, fc_bottom, fc_upper,
    bottom_in_type = "samples",
    num_samples = 2e4, return_type = "pmf",
    seed = 42
  )

  res.TDcond2 <- reconc_TDcond(A, fc_bottom, fc_upper,
    bottom_in_type = "samples",
    num_samples = 2e4, return_type = "samples",
    seed = 42
  )

  res.TDcond3 <- reconc_TDcond(A, fc_bottom, fc_upper,
    bottom_in_type = "samples",
    num_samples = 2e4, return_type = "all",
    seed = 42
  )

  # Check if all return_type return identical results
  expect_identical(res.TDcond$bottom_rec_pmf, res.TDcond3$bottom_rec_pmf)
  expect_identical(res.TDcond2$bottom_rec_samples, res.TDcond3$bottom_rec_samples)

  # Compute the reconciliation analytically (everything Gaussian)
  ## Save bottom forecast parameters
  fc_bott_gauss <- list(
    mean = means[11:22],
    cov = diag(vars[11:22])
  )

  # Compute the reconciled precision
  inv_B <- diag(1 / diag(fc_bott_gauss$cov))
  inv_U <- diag(1 / diag(fc_upper$cov))
  At_inv_U_A <- crossprod(A, inv_U) %*% A
  # Here we use the reduced A with only the lowest level
  Au <- A[.lowest_lev(A), ]
  inv_A_B_At <- solve(Au %*% tcrossprod(fc_bott_gauss$cov, Au))

  # formulas for the reconciled precision, covariance and mean
  bott_reconc_Prec <- inv_B + At_inv_U_A - crossprod(Au, inv_A_B_At) %*% Au
  bott_reconc_cov <- solve(bott_reconc_Prec)
  bott_reconc_mean <- fc_bott_gauss$mean + tcrossprod(bott_reconc_cov, A) %*% inv_U %*% (fc_upper$mean - A %*% fc_bott_gauss$mean)

  # compute the difference between empirical and analytical
  m_diff <- unlist(lapply(res.TDcond$bottom_rec_pmf, PMF_get_mean)) - bott_reconc_mean

  expect_true(all(abs(m_diff / bott_reconc_mean) < 8e-3))


  # The variances are different
  # unlist(lapply(res.TDcond$pmf$bottom,PMF_get_var))
  # diag(bott_reconc_cov)
})

# Sample from the distribution p(B_1, B_2 | B_1 + B_2 = u),
# where B_1 and B_2 are distributed as pmf1 and pmf2.
# u is a vector
.cond_biv_sampling <- function(u, pmf1, pmf2) {
  # In this way then we iterate over the one with shorter support:
  sw <- FALSE
  if (length(pmf1) > length(pmf2)) {
    pmf_ <- pmf1
    pmf1 <- pmf2
    pmf2 <- pmf_
    sw <- TRUE
  }

  b1 <- rep(NA, length(u)) # initialize empty vector

  for (u_uniq in unique(u)) { # loop over different values of u

    len_supp1 <- length(pmf1)
    supp1 <- 0:(len_supp1 - 1)
    p1 <- pmf1

    supp2 <- u_uniq - supp1
    supp2[supp2 < 0] <- Inf # trick to get NA when we access pmf2 outside the support
    p2 <- pmf2[supp2 + 1] # +1 because support starts from 0, but vector indexing from 1
    p2[is.na(p2)] <- 0 # set NA to zero

    p <- p1 * p2
    p <- p / sum(p)

    u_posit <- (u == u_uniq)
    b1[u_posit] <- sample(supp1, size = sum(u_posit), replace = TRUE, prob = p)
  }

  if (sw) b1 <- u - b1 # if we have switched, switch back

  return(list(b1, u - b1))
}

# Given a vector u of the upper values and a list of the bottom distr pmfs,
# returns samples (dim: n_bottom x length(u)) from the conditional distr
# of the bottom given the upper values
.TD_sampling <- function(u, bott_pmf,
                         toll = .TOLL, Rtoll = .RTOLL, smoothing = TRUE,
                         al_smooth = .ALPHA_SMOOTHING, lap_smooth = .LAP_SMOOTHING) {
  # If the bottom pmf list contains only 1 element,
  # then the TD samples are simply a copy of the upper samples.
  if (length(bott_pmf) == 1) {
    return(matrix(u, nrow = 1))
  }

  l_l_pmf <- rev(PMF_bottom_up(bott_pmf,
    toll = toll, Rtoll = Rtoll, return_all = TRUE,
    smoothing = smoothing, al_smooth = al_smooth, lap_smooth = lap_smooth
  ))

  b_old <- matrix(u, nrow = 1)
  for (l_pmf in l_l_pmf[2:length(l_l_pmf)]) {
    L <- length(l_pmf)
    b_new <- matrix(ncol = length(u), nrow = L)
    for (j in 1:(L %/% 2)) {
      b <- .cond_biv_sampling(b_old[j, ], l_pmf[[2 * j - 1]], l_pmf[[2 * j]])
      b_new[2 * j - 1, ] <- b[[1]]
      b_new[2 * j, ] <- b[[2]]
    }
    if (L %% 2 == 1) b_new[L, ] <- b_old[L %/% 2 + 1, ]
    b_old <- b_new
  }

  return(b_new)
}

#' @rdname Mixed_reconciliation
#' @examples
#'
#' library(bayesRecon)
#'
#' # Consider a simple hierarchy with two bottom and one upper
#' A <- matrix(c(1, 1), nrow = 1)
#' # The bottom forecasts are Poisson with lambda=15
#' lambda <- 15
#' n_tot <- 60
#' base_fc_bottom <- list()
#' base_fc_bottom[[1]] <- apply(matrix(seq(0, n_tot)), MARGIN = 1,
#'                              FUN = \(x) dpois(x, lambda = lambda))
#' base_fc_bottom[[2]] <- apply(matrix(seq(0, n_tot)), MARGIN = 1,
#'                              FUN = \(x) dpois(x, lambda = lambda))
#'
#' # The upper forecast is a Normal with mean 40 and std 5
#' base_fc_upper <- list(mean = 40, cov = matrix(c(5^2)))
#'
#' # Reconcile with reconc_TDcond
#' res.TDcond <- reconc_TDcond(A, base_fc_bottom, base_fc_upper)
#'
#' # Note that the bottom distributions are shifted to the right
#' PMF_summary(res.TDcond$bottom_rec_pmf[[1]])
#' PMF_summary(base_fc_bottom[[1]])
#'
#' PMF_summary(res.TDcond$bottom_rec_pmf[[2]])
#' PMF_summary(base_fc_bottom[[2]])
#'
#' # The upper distribution remains similar
#' PMF_summary(res.TDcond$upper_rec_pmf[[1]])
#' PMF_get_var(res.TDcond$upper_rec_pmf[[1]])
#'
#' ## Example 2: reconciliation with unbalanced hierarchy
#' # We consider the example in Fig. 9 of Zambon et al. (2024).
#'
#' # The hierarchy has 5 bottoms and 3 uppers
#' A <- matrix(c(
#'   1, 1, 1, 1, 1,
#'   1, 1, 0, 0, 0,
#'   0, 0, 1, 1, 0
#' ), nrow = 3, byrow = TRUE)
#' # Note that the 5th bottom only appears in the highest level, this is an unbalanced hierarchy.
#' n_upper <- nrow(A)
#' n_bottom <- ncol(A)
#'
#' # The bottom forecasts are Poisson with lambda=15
#' lambda <- 15
#' n_tot <- 60
#' base_fc_bottom <- list()
#' for (i in seq(n_bottom)) {
#'   base_fc_bottom[[i]] <- apply(matrix(seq(0, n_tot)), MARGIN = 1,
#'                                FUN = \(x) dpois(x, lambda = lambda))
#' }
#'
#' # The upper forecasts are a multivariate Gaussian
#' mean <- c(75, 30, 30)
#' cov <- matrix(c(
#'   5^2, 5, 5,
#'   5, 10, 0,
#'   5, 0, 10
#' ), nrow = 3, byrow = TRUE)
#'
#' base_fc_upper <- list(mean = mean, cov = cov)
#' \dontrun{
#' # If we reconcile with reconc_TDcond it won't work (unbalanced hierarchy)
#' res.TDcond <- reconc_TDcond(A, base_fc_bottom, base_fc_upper)
#' }
#'
#' # We can balance the hierarchy by duplicating the node b5:
#' # i) consider the time series observations for b5 as the upper u4,
#' # ii) fit the multivariate ts model for u1, u2, u3, u4.
#'
#' # In this example we simply assume that the forecast for u1-u4 is
#' # Gaussian with the mean and variance of u4 given by the parameters in b5.
#' mean_b5 <- lambda
#' var_b5 <- lambda
#' mean <- c(75, 30, 30, mean_b5)
#' cov <- matrix(c(
#'   5^2, 5, 5, 5,
#'   5, 10, 0, 0,
#'   5, 0, 10, 0,
#'   5, 0, 0, var_b5
#' ), nrow = 4, byrow = TRUE)
#' base_fc_upper <- list(mean = mean, cov = cov)
#'
#' # We also need to update the aggregation matrix
#' A <- matrix(c(
#'   1, 1, 1, 1, 1,
#'   1, 1, 0, 0, 0,
#'   0, 0, 1, 1, 0,
#'   0, 0, 0, 0, 1
#' ), nrow = 4, byrow = TRUE)
#'
#' # We can now reconcile with TDcond
#' res.TDcond <- reconc_TDcond(A, base_fc_bottom, base_fc_upper)
#'
#' # Note that the reconciled distribution of b5 and u4 are identical,
#' # keep this in mind when using the results of your reconciliation!
#' max(abs(res.TDcond$bottom_rec_pmf[[5]] - res.TDcond$upper_rec_pmf[[4]]))
#'
#' @export
reconc_TDcond <- function(A, base_fc_bottom, base_fc_upper,
                          bottom_in_type = "pmf", distr = NULL,
                          num_samples = 2e4, return_type = "pmf",
                          return_upper = TRUE,
                          suppress_warnings = TRUE, seed = NULL) {
  if (!is.null(seed)) set.seed(seed)

  # Check inputs
  .check_input_TD(
    A, base_fc_bottom, base_fc_upper,
    bottom_in_type, distr,
    return_type
  )

  # Get mean and covariance matrix of the MVN upper base forecasts
  mean_upper <- base_fc_upper$mean
  cov_upper <- as.matrix(base_fc_upper$cov)

  # Prepare list of bottom pmf
  if (bottom_in_type == "pmf") {
    L_pmf <- base_fc_bottom
  } else if (bottom_in_type == "samples") {
    L_pmf <- lapply(base_fc_bottom, PMF_from_samples)
  } else if (bottom_in_type == "params") {
    L_pmf <- lapply(base_fc_bottom, PMF_from_params, distr = distr)
  }

  out <- .core_reconc_TDcond(
    A, mean_upper, cov_upper, L_pmf, num_samples,
    return_type = return_type, 
    return_upper = return_upper,
    suppress_warnings = suppress_warnings
  )

  return(out)
}

#' Core Reconciliation via Top-Down Conditioning for Mixed Hierarchies
#'
#' Internal function that performs the core reconciliation logic using top-down conditioning
#' (TD) for mixed hierarchies. First, upper forecasts are reconciled analytically via conditioning
#' (if necessary), then bottom distributions are updated through probabilistic top-down procedure
#' by conditioning on the reconciled upper values.
#'
#' @param A Matrix (n_upper x n_bottom) defining the hierarchy where upper = A %*% bottom.
#' @param mean_upper Vector of mean forecasts for upper level.
#' @param cov_upper Covariance matrix of upper level forecasts.
#' @param L_pmf List of PMF objects representing the bottom level base forecasts.
#' @param num_samples Number of samples to draw from the reconciled distribution.
#' @param return_type Character string specifying return format: 'pmf', 'samples', or 'all'.
#' @param return_upper Logical, whether to return reconciled upper forecasts (default TRUE).
#' @param suppress_warnings Logical, whether to suppress warnings about samples outside support (default TRUE).
#' @param min_fraction_samples_ok Numeric between 0 and 1, minimum fraction of reconciled 
#' upper samples that must lie in the support of the bottom-up distribution (default 0.5). 
#' If the fraction is below this threshold, the function returns an error.
#'
#' @return A list containing:
#'   \itemize{
#'     \item `bottom_rec_pmf`: list of PMF objects for each bottom series (only if `return_type` is 'pmf' or 'all').
#'     \item `bottom_rec_samples`: matrix (n_bottom x num_samples) of reconciled bottom samples (only if `return_type` is 'samples' or 'all').
#'     \item `upper_rec_pmf`: list of PMF objects for each upper series (only if `return_type` is 'pmf' or 'all', and `return_upper = TRUE`).
#'     \item `upper_rec_samples`: matrix (n_upper x num_samples) of reconciled upper samples (only if `return_type` is 'samples' or 'all', and `return_upper = TRUE`).
#'   }
#'
#' @details
#' The function internally:
#' \enumerate{
#'   \item Identifies the "lowest upper" nodes in the hierarchy.
#'   \item If all uppers are lowest-uppers, samples directly from the upper MVN. Otherwise,
#'         analytically reconciles the upper hierarchy and samples from the lowest level.
#'   \item Reconciles bottom distributions by conditioning on the sampled/reconciled upper values
#'         using the probabilistic top-down algorithm.
#'   \item Discards samples that fall outside the support of the bottom-up distribution.
#' }
#'
#' @keywords internal
#' @export
.core_reconc_TDcond <- function(A, mean_upper, cov_upper, L_pmf, num_samples,
                                return_type, return_upper = TRUE,
                                suppress_warnings = TRUE,
                                min_fraction_samples_ok = .MIN_FRACTION_SAMPLES_OK) {
  # Find the "lowest upper"
  n_u <- nrow(A)
  n_b <- ncol(A)
  lowest_rows <- .lowest_lev(A)
  n_u_low <- length(lowest_rows) # number of lowest upper
  n_u_upp <- n_u - n_u_low # number of "upper upper"

  ### Get upper samples
  if (n_u == n_u_low) {
    # If all the upper are lowest-upper, just sample from the base distribution
    U <- .MVN_sample(num_samples, mean_upper, cov_upper) # (dim: num_samples x n_u_low)
    U <- round(U) # round to integer
    mode(U) <- "integer" # convert to integer
    U_js <- asplit(U, MARGIN = 2) # split into list of column vectors
  } else {
    # Else, analytically reconcile the upper and then sample from the lowest-uppers

    # Get the aggregation matrix A_u for the upper sub-hierarchy
    A_u <- .get_Au(A, lowest_rows)

    # Analytically reconcile the upper
    # The entries of mean_upper must be in the correct order, i.e. rows of A_u (upper), columns of A_u (bottom)
    mean_upper_ord <- c(mean_upper[-lowest_rows], mean_upper[lowest_rows])
    # Same for cov_upper
    cov_upper_ord <- matrix(nrow = n_u, ncol = n_u)
    cov_upper_ord[1:n_u_upp, 1:n_u_upp] <- cov_upper[-lowest_rows, -lowest_rows]
    cov_upper_ord[1:n_u_upp, (n_u_upp + 1):n_u] <- cov_upper[-lowest_rows, lowest_rows]
    cov_upper_ord[(n_u_upp + 1):n_u, 1:n_u_upp] <- cov_upper[lowest_rows, -lowest_rows]
    cov_upper_ord[(n_u_upp + 1):n_u, (n_u_upp + 1):n_u] <- cov_upper[lowest_rows, lowest_rows]
    rec_gauss_u <- reconc_gaussian(A_u, mean_upper_ord, cov_upper_ord)

    # Sample from reconciled MVN on the lowest level of the upper (dim: num_samples x n_u_low)
    U <- .MVN_sample(
      n_samples = num_samples,
      mu = rec_gauss_u$bottom_rec_mean,
      Sigma = rec_gauss_u$bottom_rec_cov
    )
    U <- round(U) # round
    mode(U) <- "integer" # convert to integer
    U_js <- asplit(U, MARGIN = 2) # split into list of column vectors
  }

  # Prepare list of lists of bottom pmf relative to each lowest upper
  L_pmf_js <- list()
  for (j in lowest_rows) {
    Aj <- A[j, ]
    L_pmf_js <- c(L_pmf_js, list(L_pmf[as.logical(Aj)]))
  }

  # Check that each multiv. sample of U is contained in the supp of the bottom-up distr
  samp_ok <- mapply(PMF_check_support, U_js, L_pmf_js)
  samp_ok <- rowSums(samp_ok) == n_u_low

  # Check if the fraction of samples that are in the support of the bottom-up distribution 
  # is above the threshold; if not, stop 
  if (mean(samp_ok) < min_fraction_samples_ok) {
    stop(paste0("The fraction of reconciled upper samples that lie in the support 
                 of the bottom-up distribution is",
                 round(mean(samp_ok * 100), 1), 
                "which is below the minimum threshold (",
                 round(min_fraction_samples_ok * 100, 1),  "). ",
                "Consider increasing the variance of the base forecasts."))
  }

  # Resample the upper samples that are in the support of the bottom-up distribution, if necessary.
  # The number of output samples is equal to the number of input samples, but some of the output samples 
  # are duplicates of the "good" input samples.
  if (sum(samp_ok) < num_samples) {
    res_idxs <- sample(which(samp_ok), size = num_samples, replace = TRUE)
    U_js <- lapply(U_js, "[", res_idxs) 
    if (!suppress_warnings) {
      warning(paste0(
        "Only ", floor(sum(samp_ok) / num_samples * 1000) / 10, "% of the upper samples ",
        "are in the support of the bottom-up distribution; ",
        "the others are discarded and the remaining ones are resampled with replacement."
      ))
    }
  }

  # Get bottom samples via the prob top-down
  B <- matrix(nrow = n_b, ncol = num_samples)
  for (j in 1:n_u_low) {
    mask_j <- as.logical(A[lowest_rows[j], ]) # mask for the position of the bottom referring to lowest upper j
    B[mask_j, ] <- .TD_sampling(U_js[[j]], L_pmf_js[[j]])
  }
  U <- A %*% B # dim: n_upper x num_samples

  # Prepare output: include the marginal pmfs and/or the samples (depending on "return" inputs)
  out <- list()
  if (return_type %in% c("pmf", "all")) {
    out$bottom_rec_pmf <- lapply(1:n_b, function(i) PMF_from_samples(B[i, ]))
    if (return_upper) {
      out$upper_rec_pmf <- lapply(1:n_u, function(i) PMF_from_samples(U[i, ]))
    }
  }
  if (return_type %in% c("samples", "all")) {
    out$bottom_rec_samples <- B
    if (return_upper) {
      out$upper_rec_samples <- U
    }
  }
  return(out)
}

#' @title Schäfer Strimmer covariance shrinkage
#'
#' @description
#'
#' Computes the Schäfer Strimmer shrinkage estimator for a covariance matrix
#' from a matrix of samples.
#'
#' @details
#'
#' This function computes the shrinkage to a diagonal covariance with unequal variances.
#' Note that here we use the estimators \eqn{S = X X^T/n} and \eqn{T = diag(S)} and we internally
#' use the correlation matrix in place of the covariance to compute the optimal shrinkage factor.
#'
#' @param x matrix of samples with dimensions nxp (n samples, p dimensions).
#'
#' @return A list containing the shrinkage estimator and the optimal lambda. The list has the following named elements:
#'
#' * `shrink_cov`: the shrinked covariance matrix (`p` x `p`);
#' * `lambda_star`: the optimal lambda for the shrinkage;
#'
#' @examples
#'
#' # Generate some multivariate normal samples
#' # Parameters
#' nSamples <- 200
#' pTrue <- 2
#'
#' # True moments
#' true_cov <- matrix(c(3, 2, 2, 2), nrow = 2)
#' chol_true_cov <- chol(true_cov)
#' true_mean <- c(0, 0)
#'
#' # Generate samples
#' set.seed(42)
#' x <- replicate(nSamples, true_mean) +
#'   t(chol_true_cov) %*% matrix(stats::rnorm(pTrue * nSamples),
#'     nrow = pTrue, ncol = nSamples
#'   )
#' x <- t(x)
#' res_shrinkage <- schaferStrimmer_cov(x)
#' res_shrinkage$lambda_star # should be 0.01287923
#'
#' @references
#' Schäfer, Juliane, and Korbinian Strimmer. (2005). *A Shrinkage Approach to Large-Scale Covariance Matrix Estimation and Implications for Functional Genomics.* Statistical Applications in Genetics and Molecular Biology 4: Article32. \doi{10.2202/1544-6115.1175}.
#'
#'
#'
#' @export
schaferStrimmer_cov <- function(x) {
  n <- nrow(x)
  p <- ncol(x)

  if (p == 1) {
    warning("you passed x with one column, returning E[x^2]")
    return(list(shrink_cov = crossprod(x) / n, lambda_star = 0))
  }

  # Compute the two estimators
  # Here we used the biased estimator
  sMat <- crossprod(x) / (n)

  # Alternative: unbiased estimator
  # sMat <- crossprod(x)/(n-1)

  tMat <- diag(diag(sMat))


  # Scale x by the standard deviation
  xscale <- scale(x, center = FALSE, scale = sqrt(diag(sMat)))
  # Compute the correlation equivalents
  rSmat <- stats::cov2cor(sMat)
  rTmat <- stats::cov2cor(tMat)


  # Compute numerator
  # we divide by (1 / (n*(n - 1))) because of biased estimator
  varSij <- (1 / (n * (n - 1))) * (crossprod(xscale^2) - 1 / n * (crossprod(xscale))^2)

  # Alternative for unbiased estimator: divide by (n / ((n - 1)^3))
  # varSij <- (n / ((n - 1)^3)) * (crossprod(xscale ^ 2) - 1 / n * (crossprod(xscale)) ^ 2)


  diag(varSij) <- 0

  # Compute denominator
  sqSij <- (rSmat - rTmat)^2

  # Compute Lambda
  lambda_star <- sum(varSij) / sum(sqSij)
  lambda_star <- max(min(lambda_star, 1), 0)

  # Compute shrinkage
  shrink_cov <- lambda_star * tMat + (1 - lambda_star) * sMat

  return(list(shrink_cov = shrink_cov, lambda_star = lambda_star))
}

test_that("Test shrinkage estimator", {
  # Parameters
  nSamples <- 500
  pTrue <- 2

  # True moments
  trueMean <- c(0, 0)
  trueSigma <- matrix(c(3, 2, 2, 2), nrow = 2)
  chol_trueSigma <- chol(trueSigma)

  # we run 100 shrinkage estimators
  lambdas <- rep(0, 100)
  for (i in seq(100)) {
    rr <- replicate(nSamples, trueMean) +
      t(chol_trueSigma) %*% matrix(stats::rnorm(pTrue * nSamples), nrow = pTrue, ncol = nSamples)
    # Estimate mean and covariance from samples
    mean_est <- rowMeans(rr)
    Sigma_est <- cov(t(rr))

    mean_est
    trueMean

    Sigma_est

    rr_centered <- rr - replicate(nSamples, mean_est)
    rowMeans(rr_centered)

    x <- t(rr)
    # x <- t(rr_centered)

    lambdas[i] <- schaferStrimmer_cov(x)$lambda_star
  }
  # The average over 100 runs must be within a certain range
  mean_lambdas <- mean(lambdas)

  # We expect this to be lower than the 99.99% quantile over 200 simulations of this scenario
  expect_lte(mean_lambdas, 0.005191634)

  # We expect this to be greater than the 0.01% quantile over 200 simulations of this scenario
  expect_gte(mean_lambdas, 0.004843328)
})

test_that("Test behavior for p=1", {
  # Parameters
  nSamples <- 500
  pTrue <- 1

  samples <- matrix(stats::rnorm(n = nSamples, sd = 2), ncol = pTrue)

  exp_sample_squared <- matrix(mean(samples^2), 1, 1)

  # The function should return E[x^2] and a warning
  expect_warning(expect_equal(schaferStrimmer_cov(samples)$shrink_cov, exp_sample_squared))
})

# Generate the weekly base forecasts parameters
# CHANGE THE WORKING DIRECTORY BEFORE RUNNING
rm(list = ls())
library(bayesRecon)

# Generate A matrix for weekly hierarchy
A <- .gen_weekly()
# Save matrix to file
write.table(A, file = "./Weekly-Gaussian_A.csv", row.names = FALSE, sep = ",", col.names = FALSE, quote = FALSE)

# Generate randomly bottom means
set.seed(1)
bottom_means <- runif(52, min = 5, max = 10)

# The upper means are generated by
# aggregating the bottom means (according to A) with an epsilon of incoherence
eps <- 0.1
upper_means <- (1 + eps) * A %*% bottom_means
upper_means <- as.vector(upper_means)

# Create the matrix to save to csv
base_forecasts_out <- cbind(c(upper_means, bottom_means), c(rep(3, length(upper_means)), rep(2, length(bottom_means))))
write.table(base_forecasts_out, file = "./Weekly-Gaussian_basef.csv", row.names = FALSE, sep = ",", col.names = FALSE, quote = FALSE)

#' Example of a time series from carparts
#'
#' A monthly time series from the `carparts` dataset, 51 observations, Jan 1998 - Mar 2002.
#'
#' @format
#' Univariate time series of class \link[stats]{ts}.
#'
#' @references
#' Hyndman, R.J., Koehler, A.B., Ord, J.K., and Snyder, R.D., (2008).
#' Forecasting with exponential smoothing: the state space approach.
#' Springer Science & Business Media.
#'
#' Godahewa, R., Bergmeir, C., Webb, G., Hyndman, R., & Montero-Manso, P. (2020).
#' Car Parts Dataset (without Missing Values) (Version 2)
#' \doi{10.5281/zenodo.4656021}
#'
#' @source
#' Godahewa, R., Bergmeir, C., Webb, G., Hyndman, R.J., & Montero-Manso, P. (2020).
#' Car Parts Dataset (without Missing Values) (Version 2)
#' \doi{10.5281/zenodo.4656021}
"carparts_example"


#' Infant Mortality grouped time series dataset
#'
#' A yearly grouped time series dataset, from 1901 to 2003, of infant mortality counts (deaths) in Australia;
#' disaggregated by state (see below), and sex (male and female).
#'
#' States: New South Wales (NSW), Victoria (VIC), Queensland (QLD), South Australia (SA), Western Australia
#' (WA), Northern Territory (NT), Australian Capital Territory (ACT), and Tasmania (TAS).
#'
#' @format
#' List of time series of class \link[stats]{ts}.
#'
#' @source
#' hts package [CRAN](https://cran.r-project.org/package=hts)
#'
#' @references
#' Hyndman, R.J., Ahmed, R.A., Athanasopoulos, G., Shang, H.L. (2011).
#' Optimal combination forecasts for hierarchical time series.
#' Computational Statistics and Data Analysis, 55(9), 2579-2589.
"infantMortality"


#' Example of a time series from the M3 forecasting competition
#'
#' A monthly time series, from the M3 forecasting competition ("N1485").
#'
#' @format
#' List of time series of class \link[stats]{ts}.
#'
#'
#' @source [https://forecasters.org/resources/time-series-data/m3-competition/](https://forecasters.org/resources/time-series-data/m3-competition/)
"M3_example"


#' Extreme market events dataset
#'
#' Count time series of extreme market events in five economic sectors.
#' The data refer to the trading days between 2004/12/31 and 2018/12/19 (3508 trading days in total).
#'
#' The counts are computed by considering 29 companies included in the Euro Stoxx
#' 50 index and observing if the value of the CDS spread on a given day exceeds
#' the 90-th percentile of its distribution in the last trading year.
#' The companies are divided in the following  sectors: Financial (FIN), Information
#' and Communication Technology (ICT), Manufacturing (MFG), Energy (ENG), and Trade (TRD).
#'
#' There are 6 time series:
#' - 5 bottom time series, corresponding to the daily counts for each sector
#' - 1 upper time series, which is the sum of all the bottom (ALL)
#'
#' @format
#' A multivariate time series of class \link[stats]{ts}.
#'
#' @source
#' Zambon, L., Agosto, A., Giudici, P., Corani, G. (2024).
#' *Properties of the reconciled distributions for Gaussian and count forecasts*.
#' International Journal of Forecasting (in press).
#' \doi{10.1016/j.ijforecast.2023.12.004}.
#'
#' @references
#' Zambon, L., Agosto, A., Giudici, P., Corani, G. (2024).
#' *Properties of the reconciled distributions for Gaussian and count forecasts*.
#' International Journal of Forecasting (in press).
#' \doi{10.1016/j.ijforecast.2023.12.004}.
#'
#' Agosto, A. (2022).
#' *Multivariate Score-Driven Models for Count Time Series to Assess Financial Contagion*.
#' \doi{10.2139/ssrn.4119895}
"extr_mkt_events"


#' Base forecasts for the extreme market events dataset
#'
#' Base forecasts for the `extr_mkt_events` dataset, computed using the model by
#' Agosto, A. (2022).
#' *Multivariate Score-Driven Models for Count Time Series to Assess Financial Contagion*.
#' \doi{10.2139/ssrn.4119895}.
#'
#' The predictive distribution for the bottom time series is a multivariate negative
#' binomial with a static vector of dispersion parameters and a time-varying vector
#' of location parameters following a score-driven dynamics.
#' The base forecasts for the upper time series are computed using a univariate version of this model.
#' They are in-sample forecasts: for each training instant, they are computed for
#' time t+1 by conditioning on the counts observed up to time t.
#'
#' @format
#' A list `extr_mkt_events_basefc` containing
#' \describe{
#'    \item{`extr_mkt_events_basefc$mu`}{data frame of the base forecast means, for each day}
#'    \item{`extr_mkt_events_basefc$size`}{data frame of the static base forecast size parameters}
#' }
#'
#' @source
#' Agosto, A. (2022).
#' *Multivariate Score-Driven Models for Count Time Series to Assess Financial Contagion*.
#' \doi{10.2139/ssrn.4119895}
#'
#' @references
#' Agosto, A. (2022).
#' *Multivariate Score-Driven Models for Count Time Series to Assess Financial Contagion*.
#' \doi{10.2139/ssrn.4119895}
#'
#' Zambon, L., Agosto, A., Giudici, P., Corani, G. (2024).
#' *Properties of the reconciled distributions for Gaussian and count forecasts*.
#' International Journal of Forecasting (in press).
#' \doi{10.1016/j.ijforecast.2023.12.004}.
"extr_mkt_events_basefc"


#' Example of hierarchical forecasts for a store from the M5 competition
#'
#' This dataset contains forecasts for the hierarchy of time series related to the store `CA_1` from the M5 competition.
#'
#' The store `CA_1` contains 3049 item level time series and 11 aggregate time series:
#' * Store level aggregation (`CA_1`)
#' * Category level aggregations  (`HOBBIES`, `HOUSEHOLD`, `FOODS`)
#' * Department level aggregations (`HOBBIES_1`, `HOBBIES_2`, `HOUSEHOLD_1`, `HOUSEHOLD_2`, `FOODS_1`, `FOODS_2`, `FOODS_3`)
#'
#' Forecasts are generated with the function `forecast` and the model `adam` from the package `smooth`.
#' * The models for the bottom time series are selected with multiplicative Gamma error term (`MNN`);
#' * The models for the upper time series (`AXZ`) is selected with Gaussian additive error term, seasonality selected based on information criterion.
#'
#' The raw data was downloaded with the package `m5`.
#'
#' @format
#' A list containing:
#' * `upper`: a list of 11 elements each representing an aggregation level. Each element contains: `mu`, `sigma` the mean and standard deviation of the Gaussian forecast, `actual` the actual value, `residuals` the residuals of the model used to estimate forecasts covariance.
#' * `lower`: a list of 3049 elements each representing a forecast for each item. Each element contains `pmf` the probability mass function of the item level forecast, `actual` the actual value.
#' * `A`: the aggregation matrix for A.
#' * `S`: the S matrix for the hierarchy.
#' * `Q_u`: scaling factors for computing MASE on the upper forecasts.
#' * `Q_b`: scaling factors for computing MASE on the bottom forecasts.
#'
#' @references
#' Joachimiak K (2022). *m5: 'M5 Forecasting' Challenges Data*. R package version 0.1.1, <https://CRAN.R-project.org/package=m5>.
#'
#' Makridakis, Spyros & Spiliotis, Evangelos & Assimakopoulos, Vassilis. (2020). *The M5 Accuracy competition: Results, findings and conclusions.* International Journal of Forecasting 38(4) 1346-1364. \doi{10.1016/j.ijforecast.2021.10.009}
#'
#' Svetunkov I (2023). *smooth: Forecasting Using State Space Models*. R package version 4.0.0, <https://CRAN.R-project.org/package=smooth>.
#'
#' @source
#' Makridakis, Spyros & Spiliotis, Evangelos & Assimakopoulos, Vassilis. (2020). *The M5 Accuracy competition: Results, findings and conclusions.* International Journal of Forecasting 38(4) 1346-1364. \doi{10.1016/j.ijforecast.2021.10.009}
"M5_CA1_basefc"


#' Swiss tourism: monthly Swiss overnight stay data
#'
#' A dataset of 27 time series with monthly observations of the number of 
#' overnight stays in Switzerland, from Jan 2005 to Jan 2025.
#' 
#' This is a hierarchy of time series that contains:
#' - 26 bottom-level series corresponding to the 26 Swiss cantons 
#' - 1 top-level series corresponding to the Swiss total.
#' Each time series has 241 observations. 
#' 
#'
#' @format
#' A list that contains: 
#' - \code{ts}: a multivariate time series of class \link[stats]{ts};
#' - \code{n_bottom}: integer with the number of bottom time series (26);
#' - \code{n_upper}: integer with the number of upper time series (1);
#' - \code{agg_mat}: the aggregation matrix for the hierarchy. (1x26)
#' 
#' @source 
#' Federal Statistical Office of Switzerland (2025).
#' *Hotel sector: arrivals and overnight stays of open establishments by year, month, canton and visitors' country of residence*.
#' \url{https://www.bfs.admin.ch/asset/en/px-x-1003020000_102}
#' 
#' @references 
#' 
#' Carrara, C., Azzimonti, D., Corani, G., Zambon, L. (2025).
#' *Modeling the uncertainty on the covariance matrix for probabilistic forecast reconciliation*.
#' arXiv:2506.19554
#' \doi{10.48550/arXiv.2506.19554}.
#' 
"swiss_tourism"

# Function to find the best hierarchy contained in A;
# "best" means that minimizes the number of IS steps required by BUIS.
# It is not the maximal hierarchy contained in A!
# Returns a vector with length equal to the number of rows of A
# Each entry is 1 if the corresponding row has to be picked, and 0 otherwise
.get_hier_rows <- function(A, scale = 196) {
  k <- nrow(A)
  m <- ncol(A)

  # matrix C of the coefficients of the non-linear problem
  # (k variables, 1 constraint)

  C <- A %*% t(A)

  for (i in 1:k) {
    for (j in 1:k) {
      C[i, j] <- C[i, j] * sum((A[i, ] - A[j, ]) * (A[i, ] - A[j, ] - 1)) *
        sum((A[j, ] - A[i, ]) * (A[j, ] - A[i, ] - 1))
    }
  }

  #-------------------
  # LINEARIZED PROBLEM
  # Number of variables: k + k^2 + 1
  # Number of constraints: 1 + k^2 + k^2 + k^2 + m

  # Set coefficients of the objective function
  f.obj <- c(rep(-1, k), rep(0, k^2), 1 - 1 / (2 * k))

  # Set matrix corresponding to coefficients of constraints by rows

  coeff <-
    c(rep(0, k), as.vector(C), 0) # first constraint

  M1 <- matrix(0, k^2, k + k^2 + 1) # z_{ij} <= x_i
  for (i in 1:k) {
    temp <- matrix(0, k, k)
    temp[i, ] <- rep(-1, k)
    M1[, i] <- as.vector(temp)
  }
  M1[, (k + 1):(k + k^2)] <- diag(1, k^2)

  M2 <- matrix(0, k^2, k + k^2 + 1) # z_{ij} <= x_j
  for (i in 1:k) {
    temp <- matrix(0, k, k)
    temp[, i] <- rep(-1, k)
    M2[, i] <- as.vector(temp)
  }
  M2[, (k + 1):(k + k^2)] <- diag(1, k^2)

  M3 <- matrix(0, k^2, k + k^2 + 1) # z_{ij} >= x_i + x_j - 1
  M3[, 1:k] <- M1[, 1:k] + M2[, 1:k]
  M3[, (k + 1):(k + k^2)] <- diag(1, k^2)

  M4 <- matrix(0, m, k + k^2 + 1) # \sum_i x_i A_{ij} <= y
  M4[, 1:k] <- t(A)
  M4[, k + k^2 + 1] <- rep(-1, m)

  f.con <- rbind(coeff, M1, M2, M3, M4)

  # Set inequality/equality signs
  f.dir <- c(
    "=",
    rep("<=", k^2),
    rep("<=", k^2),
    rep(">=", k^2),
    rep("<=", m)
  )

  # Set right hand side coefficients
  f.rhs <- c(0, rep(0, k^2), rep(0, k^2), rep(-1, k^2), rep(0, m))

  #---------------------
  # Solve the LP problem

  # Variables final values
  indices_sol <-
    lpSolve::lp(
      "min",
      f.obj,
      f.con,
      f.dir,
      f.rhs,
      all.int = TRUE,
      binary.vec = 1:(k + k^2),
      scale = 196
    )$solution[1:k]

  return(indices_sol)
}

#' Extract and sort hierarchy rows
#'
#' Function that extracts the "best hierarchy rows" from A (see .get_hier_rows),
#' and sorts them in the correct order (i.e. bottom-up).
#' Also sorts accordingly the vectors v, d, it (e.g. of parameters).
#'
#' @param A Aggregation matrix
#' @param v Vector of values
#' @param d Distribution parameters
#' @param it Input type parameters
#'
#' @return List containing sorted hierarchy rows and remaining rows
#'
#' @keywords internal
#' @export
.get_HG <- function(A, v, d, it) {
  # get the indices of the "hierarchy rows" of A
  indices_sol <- .get_hier_rows(A)

  # extract rows from A
  ind_h <- as.logical(indices_sol)
  H <- matrix(A[ind_h, ], ncol = ncol(A))
  v_h <- v[ind_h]
  d_h <- d[ind_h]
  it_h <- it[ind_h]

  # sort bottom-up
  ord <- order(rowSums(H))
  H <- matrix(H[ord, ], ncol = ncol(A))
  v_h <- v_h[ord]
  d_h <- d_h[ord]
  it_h <- it_h[ord]

  # collect remaining rows in matrix G
  ind_g <- as.logical(1 - indices_sol)
  if (sum(ind_g) == 0) {
    G <- NULL
    v_g <- NULL
    d_g <- NULL
    it_g <- NULL
  } else {
    G <- matrix(A[ind_g, ], ncol = ncol(A))
    v_g <- v[ind_g]
    d_g <- d[ind_g]
    it_g <- it[ind_g]
  }

  out <- list(
    H = H,
    G = G,
    Hv = v_h,
    Gv = v_g,
    Hdistr = d_h,
    Gdistr = d_g,
    Hin_type = it_h,
    Gin_type = it_g
  )
  return(out)
}

# Functions to generate the monthly and weekly A matrices
.gen_monthly <- function() {
  H <- matrix(0, nrow = 10, ncol = 12)
  for (j in 1:6) {
    H[j, (2 * (j - 1) + 1):(2 * j)] <- 1
  }
  for (j in 1:3) {
    H[6 + j, (4 * (j - 1) + 1):(4 * j)] <- 1
  }
  H[6 + 3 + 1, ] <- 1

  G <- matrix(0, nrow = 6, ncol = 12)
  for (j in 1:4) {
    G[j, (3 * (j - 1) + 1):(3 * j)] <- 1
  }
  for (j in 1:2) {
    G[4 + j, (6 * (j - 1) + 1):(6 * j)] <- 1
  }

  return(rbind(H, G))
}
.gen_weekly <- function() {
  H <- matrix(0, nrow = 40, ncol = 52)
  for (j in 1:26) {
    H[j, (2 * (j - 1) + 1):(2 * j)] <- 1
  }
  for (j in 1:13) {
    H[26 + j, (4 * (j - 1) + 1):(4 * j)] <- 1
  }
  H[26 + 13 + 1, ] <- 1

  G <- matrix(0, nrow = 6, ncol = 52)
  for (j in 1:4) {
    G[j, (13 * (j - 1) + 1):(13 * j)] <- 1
  }
  for (j in 1:2) {
    G[4 + j, (26 * (j - 1) + 1):(26 * j)] <- 1
  }

  return(rbind(H, G))
}

#' @title Temporal aggregation of a time series
#'
#' @description
#' Creates a list of aggregated time series from a time series of class \link[stats]{ts}.
#'
#' @param y univariate time series of class \link[stats]{ts}.
#' @param agg_levels user-selected list of aggregation levels.
#'
#' @details
#' If `agg_levels=NULL` then `agg_levels` is automatically generated by taking all the factors of the time series frequency.
#'
#' @return A list of \link[stats]{ts} objects each containing the aggregates time series in the order defined by `agg_levels`.
#'
#' @seealso [get_reconc_matrices()]
#'
#' @examples
#'
#' # Create a monthly count time series with 100 observations
#' y <- ts(data = stats::rpois(100, lambda = 2), frequency = 12)
#'
#' # Create the aggregate time series according to agg_levels
#' y_agg <- temporal_aggregation(y, agg_levels = c(2, 3, 4, 6, 12))
#'
#' # Show annual aggregate time series
#' print(y_agg$`f=1`)
#'
#' @export
temporal_aggregation <- function(y, agg_levels = NULL) {
  f <- stats::frequency(y)
  L <- length(y)
  s <- stats::time(y)[1]
  if (is.null(agg_levels)) {
    agg_levels <- c()
    for (i in 1:f) {
      if (f %% i == 0 && L >= i) {
        agg_levels <- c(agg_levels, i)
      }
    }
  } else {
    agg_levels <- sort(agg_levels[agg_levels <= L])
    if (!(1 %in% agg_levels)) {
      agg_levels <- c(1, agg_levels)
    }
  }
  out <- list()
  for (i in 1:length(agg_levels)) {
    k <- agg_levels[i]
    num_aggs <- floor(L / k)
    y_trunc <- y[(L - num_aggs * k + 1):L]
    y_matrix <- matrix(y_trunc, nrow = k, ncol = num_aggs)
    y_start <- s + (L - num_aggs * k) / f
    y_f <- f / k
    y_agg <- stats::ts(data = apply(y_matrix, 2, sum), frequency = y_f, start = y_start)
    out[[i]] <- y_agg
  }
  names(out) <- paste0("f=", f / agg_levels)
  out <- rev(out)
  return(out)
}

#' @title Build hierarchy matrices
#'
#' @description
#' Creates the aggregation and summing matrices for a temporal hierarchy of time series
#' from a user-selected list of aggregation levels.
#'
#' @param agg_levels user-selected list of aggregation levels.
#' @param h number of steps ahead for the bottom level forecasts.
#'
#' @return A list containing the named elements:
#'
#' * `A` the aggregation matrix;
#' * `S` the summing matrix.
#'
#' @examples
#'
#' library(bayesRecon)
#'
#' # Create monthly hierarchy
#' agg_levels <- c(1, 2, 3, 4, 6, 12)
#' h <- 12
#' rec_mat <- get_reconc_matrices(agg_levels, h)
#' S <- rec_mat$S
#' A <- rec_mat$A
#'
#' @seealso [temporal_aggregation()]
#'
#' @export
get_reconc_matrices <- function(agg_levels, h) {
  A <- list()
  for (i in 1:length(agg_levels)) {
    k <- agg_levels[i]
    if (k == 1) {
      next
    }
    k.r <- h / k
    k.A <- matrix(data = 0, nrow = k.r, ncol = h)
    coli <- 1
    for (r in 1:k.r) {
      k.A[r, coli:(coli + k - 1)] <- 1
      coli <- coli + k
    }
    A[[i]] <- k.A
  }
  A <- do.call(rbind, rev(A))
  S <- rbind(A, diag(h))
  out <- list(A = A, S = S)
  return(out)
}

# Get A from S
.get_A_from_S <- function(S) {
  # Bottom rows are those with a single 1; if there are replicated bottom rows,
  # only one is treated as bottom, the copies will be upper
  bottom_idxs <- which(rowSums(S) == 1 & !duplicated(S))
  if (length(bottom_idxs) < ncol(S)) {
    stop("Check S: some bottom rows are missing")
  }
  upper_idxs <- setdiff(1:nrow(S), bottom_idxs)
  A <- matrix(S[upper_idxs, ], ncol = ncol(S))
  out <- list(
    A = A,
    upper_idxs = upper_idxs,
    bottom_idxs = bottom_idxs
  )
  return(out)
}

# Split bottoms, uppers
.split_hierarchy <- function(S, Y) {
  if (nrow(S) != length(Y)) {
    stop(sprintf("Error: summing matrix rows (%d) != length base forecasts (%d)", nrow(S), length(Y)))
  }

  getAfromS.res <- .get_A_from_S(S)
  upper <- Y[getAfromS.res$upper_idxs]
  bottom <- Y[getAfromS.res$bottom_idxs]
  out <- list(
    A = matrix(getAfromS.res$A, ncol = ncol(S)),
    upper = upper,
    bottom = bottom,
    upper_idxs = getAfromS.res$upper_idxs,
    bottom_idxs = getAfromS.res$bottom_idxs
  )
  return(out)
}

#' Check if matrix is hierarchical
#'
#' Returns TRUE if A is a hierarchy matrix.
#'
#' @param A Aggregation matrix
#'
#' @return Logical value indicating if matrix is hierarchical
#'
#' @keywords internal
#' @export
.check_hierarchical <- function(A) {
  
  # Number of bottom series shared by upper i and upper j:
  overlaps <- tcrossprod(A)
  # Number of bottom series for each upper:
  node_sizes <- diag(overlaps)
  
  # Find indices that are non-zero and in the upper triangle (i < j)
  nz_idx <- which(overlaps != 0 & upper.tri(overlaps, diag = FALSE), arr.ind = TRUE)
  # If no off-diagonal overlaps exist, it's a hierarchy
  if (nrow(nz_idx) == 0) return(TRUE)
  
  # Extract intersection values and sizes of the two nodes for overlapping pairs
  intersections <- overlaps[nz_idx]
  size_i <- node_sizes[nz_idx[, 1]]
  size_j <- node_sizes[nz_idx[, 2]]
  
  # Hierarchy Condition: if two nodes overlap, one must be a subset of the other
  # If smaller than both sizes, then neither is a subset of the other, which violates hierarchy
  violations <- (intersections < size_i) & (intersections < size_j)
  
  return(!any(violations))
}

# Check if A is a hierarchical matrix and if the rows are in bottom-up order,
# i.e., parents appear after children
.check_BU_matr <- function(A) {
  
  # Number of bottom series shared by upper i and upper j:
  overlaps <- tcrossprod(A)
  # Number of bottom series for each upper:
  node_sizes <- diag(overlaps)
  
  # Find indices that are non-zero and in the upper triangle (i < j)
  nz_idx <- which(overlaps != 0 & upper.tri(overlaps, diag = FALSE), arr.ind = TRUE)
  # If there are no overlaps in the upper triangle (disjoint series), order is fine.
  if (nrow(nz_idx) == 0) return(TRUE)
  
  # Extract intersection values and size of the first node of each overlapping pair
  intersections <- overlaps[nz_idx]
  size_i <- node_sizes[nz_idx[, 1]]
  
  # i is not a subset of j IFF intersection(i, j) < size(i)
  violations <- intersections < size_i
  
  return(!any(violations))
}

#' Find rows corresponding to the lowest level
#'
#' Identifies the rows of aggregation matrix A that correspond to the lowest
#' hierarchical level (i.e., the most disaggregated upper variables that
#' collectively sum to all bottom variables).
#'
#' @param A Aggregation matrix (must be hierarchical)
#'
#' @return Integer vector of row indices corresponding to the lowest level
#'
#' @keywords internal
#' @export
.lowest_lev <- function(A) {
  
  if (nrow(A) == 1) return(1)
  
  if (!.check_hierarchical(A)) stop("Matrix A is not hierarchical")

  # First, only keep unique rows of A
  A_uni <- unique(A)
  
  # Number of bottom series shared by upper i and upper j:
  overlaps <- tcrossprod(A_uni)
  # Number of bottom series for each upper:
  sizes <- diag(overlaps)
  
  # Logic: row i is a parent if there exists another row j such that:
  #   a) j is strictly smaller than i (sizes[j] < sizes[i])
  #   b) j is fully contained in i (overlaps[i,j] == sizes[j])
  
  k <- nrow(A_uni)
  is_parent <- logical(k)
  
  for (i in 1:k) {
    # Check against all other rows 'j'
    # Condition: Intersection count equals size of j (j is subset) AND size j < size i
    children_mask = (overlaps[i, ] == sizes) & (sizes < sizes[i])
    if (any(children_mask)) {
      is_parent[i] <- TRUE
    }
  }

  # Now, change the indices of the lowest rows to match with A (instead of A_uni)
  # If there are duplicated rows for some lowest rows, only take one copy
  low_rows_A <- (1:nrow(A))[!duplicated(A)][!is_parent]

  # The sum of the rows corresponding to the lowest level should be a vector of 1
  if (any(colSums(A[low_rows_A, , drop = FALSE]) != 1)) {
    unbal_bott <- which(colSums(A[low_rows_A, , drop = FALSE]) != 1)
    err_mess <- "It is impossible to find the lowest upper level. Probably the hierarchy is unbalanced, the following bottom should be duplicated (see example): "
    err_mess <- paste0(c(err_mess, unbal_bott), collapse = " ")
    stop(err_mess)
  }

  return(low_rows_A)
}

#' Get aggregation matrix of upper sub-hierarchy
#'
#' Constructs the aggregation matrix Au for the sub-hierarchy composed only
#' of the upper variables. This matrix relates the "upper uppers" to the
#' "lowest uppers" (the most disaggregated level of the upper hierarchy).
#'
#' @param A Aggregation matrix (must be hierarchical)
#' @param lowest_rows Optional. Integer vector of row indices corresponding to
#'   the lowest upper level. If NULL, computed using `.lowest_lev(A)`.
#'
#' @return Aggregation matrix for the upper sub-hierarchy, or NULL if all
#'   uppers are at the lowest level.
#'
#' @keywords internal
#' @export
.get_Au <- function(A, lowest_rows = NULL) {
  
  if (is.null(lowest_rows)) lowest_rows <- .lowest_lev(A)

  if (length(lowest_rows) == nrow(A)) {
    warning("All the upper are lowest-upper. Return NULL")
    return(NULL)
  }

  A_ <- A[-lowest_rows, , drop = FALSE]
  n_upp_u <- nrow(A_)
  n_bott_u <- length(lowest_rows)
  A_u <- matrix(nrow = n_upp_u, ncol = n_bott_u)
  for (j in 1:n_bott_u) {
    l <- lowest_rows[[j]]
    for (i in 1:n_upp_u) {
      A_u[i, j] <- all(A[l, ] <= A_[i, ]) # check that "lower upper" j is a descendant of "upper upper" i
    }
  }

  return(1 * A_u) # to get numerical values instead of TRUE / FALSE
}

# Check if the rows of A are ordered
.check_ordered_A <- function(A) {
  aggregates_sum <- rowSums(A)
  ordered_aggreg <- order(aggregates_sum, decreasing = TRUE)
  if (all(aggregates_sum == aggregates_sum[ordered_aggreg])) {
    return(list(value = TRUE))
  } else {
    return(list(value = FALSE, order = ordered_aggreg))
  }
}

test_that("reconc_MixCond simple example", {
  # Simple example with
  # - 12 bottom
  # - 10 upper: year, 6 bi-monthly, 3 quarterly
  A <- matrix(
    data = c(
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
      1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1,
      1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1
    ),
    nrow = 10, byrow = TRUE
  )

  # Define means and vars for the forecasts
  means <- c(90, 62, 63, 64, 31, 32, 31, 33, 31, 32, rep(15, 12))
  vars <- c(20, 8, 8, 8, 4, 4, 4, 4, 4, 4, rep(2, 12))^2

  # create the lists for reconciliation
  ## upper
  fc_upper <- list(
    mean = means[1:10],
    cov = diag(vars[1:10])
  )

  ## bottom
  fc_bottom <- list()
  for (i in seq(ncol(A))) {
    fc_bottom[[i]] <- as.integer(.distr_sample(list(mean = means[i + 10], sd = vars[i + 10]), "gaussian", 2e4))
    fc_bottom[[i]][which(fc_bottom[[i]] < 0)] <- 0 # set-negative-to-zero
  }


  res.MixCond <- reconc_MixCond(A, fc_bottom, fc_upper, bottom_in_type = "samples", seed = 42)

  bott_rec_means <- unlist(lapply(res.MixCond$bottom_rec_pmf, PMF_get_mean))
  bott_rec_vars <- unlist(lapply(res.MixCond$bottom_rec_pmf, PMF_get_var))


  # Create PMF from samples
  fc_bottom_pmf <- list()
  for (i in seq(ncol(A))) {
    fc_bottom_pmf[[i]] <- PMF_from_samples(fc_bottom[[i]])
  }

  # Reconcile from bottom PMF
  res.MixCond_pmf <- reconc_MixCond(A, fc_bottom_pmf, fc_upper, seed = 42)

  bott_rec_means_pmf <- unlist(lapply(res.MixCond_pmf$bottom_rec_pmf, PMF_get_mean))
  bott_rec_vars_pmf <- unlist(lapply(res.MixCond_pmf$bottom_rec_pmf, PMF_get_var))

  expect_equal(bott_rec_means, bott_rec_means_pmf, tolerance = 0.01)
  expect_equal(bott_rec_vars, bott_rec_vars_pmf, tolerance = 0.1)
})

test_that("reconc_MixCond and reconc_TDcond with temporal hier and params", {
  # Read samples from dataForTests (reproducibility)
  vals <- read.csv(file = "dataForTests/Monthly-Count_ts.csv", header = FALSE)

  # Create a count time series with monthly observations for 10 years
  y <- ts(data = vals, frequency = 12)

  # Create the aggregated yearly time series
  y_agg <- temporal_aggregation(y, agg_levels = c(1, 12))

  # We use a marginal forecast that computes for each month
  # the empirical mean and forecasts a Poisson with that value
  fc_bottom <- list()
  for (i in seq(12)) {
    fc_bottom[[i]] <- list(lambda = mean(y_agg$`f=12`[seq(i, 120, 12)]))
  }

  # We compute the empirical mean and variance of the yearly ts
  # we forecast with a Gaussian with those parameters
  fc_upper <- list(mean = mean(y_agg$`f=1`), cov = matrix(var(y_agg$`f=1`)))

  # Obtain the aggregation matrix for this hierarchy
  rec_mat <- get_reconc_matrices(c(1, 12), 12)

  # Do a couple of checks on S and A
  expect_no_error(.check_S(rec_mat$S))
  expect_error(.check_S(rec_mat$A))
  expect_true(.check_BU_matr(rec_mat$A))
  expect_false(.check_BU_matr(rec_mat$S))

  # We can reconcile with reconc_MixCond
  res.mixCond <- reconc_MixCond(rec_mat$A, fc_bottom, fc_upper, bottom_in_type = "params", distr = "poisson")

  # We can reconcile with reconc_TDcond
  res.TDcond <- reconc_TDcond(rec_mat$A, fc_bottom, fc_upper, bottom_in_type = "params", distr = "poisson")

  # Summary of the upper reconciled with TDcond
  pmfSum <- PMF_summary(res.TDcond$upper_rec_pmf[[1]])
  # We expect that the reconciled mean is very similar to the initial mean (should be equal)
  expect_equal(pmfSum$Mean, fc_upper$mean, tolerance = 0.01)

  # Check that all bottom and upper reconciled PMF sum to 1
  check_pmf_bott_mixCond <- sum(unlist(lapply(res.mixCond$bottom_rec_pmf, function(x) {
    sum(x)
  })))
  check_pmf_upp_mixCond <- sum(unlist(lapply(res.mixCond$upper_rec_pmf, function(x) {
    sum(x)
  })))
  expect_equal(check_pmf_bott_mixCond, 12)
  expect_equal(check_pmf_upp_mixCond, 1)

  # Check that all bottom and upper reconciled PMF sum to 1
  check_pmf_bott_TDcond <- sum(unlist(lapply(res.TDcond$bottom_rec_pmf, function(x) {
    sum(x)
  })))
  check_pmf_upp_TDcond <- sum(unlist(lapply(res.TDcond$upper_rec_pmf, function(x) {
    sum(x)
  })))
  expect_equal(check_pmf_bott_TDcond, 12)
  expect_equal(check_pmf_upp_TDcond, 1)
})

test_that(".get_hier_rows works on a monthly hierarchy", {
  month_hier <- .gen_monthly()
  ind_best <- .get_hier_rows(month_hier)

  best_obj_fun <- max(colSums(month_hier[as.logical(ind_best), ])) - sum(ind_best)

  expect_equal(best_obj_fun, -7)
})


test_that(".get_hier_rows works on a weekly hierarchy", {
  week_hier <- .gen_weekly()
  ind_best <- .get_hier_rows(week_hier)

  best_obj_fun <- max(colSums(week_hier[as.logical(ind_best), ])) - sum(ind_best)

  expect_equal(best_obj_fun, -37)
})

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(bayesRecon)

## ----finTShier, fig.cap="**Figure 1**: financial time series hierarchy.", out.width = '80%', echo = FALSE----
knitr::include_graphics("img/finTS_hier.jpg")

## -----------------------------------------------------------------------------
# Hierarchy composed by 6 time series: 5 bottom and 1 upper
n_b <- 5
n_u <- 1
n <- n_b + n_u

A <- matrix(1, ncol = n_b, nrow = n_u) # aggregation matrix

# Actual values:
actuals <- data.frame(extr_mkt_events) # convert to data frame
# Base forecasts:
base_fc <- extr_mkt_events_basefc

N <- nrow(actuals) # number of days (3508)

# # If you want to run only N reconciliations (instead of 3508):
# N <- 200
# actuals <- actuals[1:N,]
# base_fc$mu <- base_fc$mu[1:N,]

## -----------------------------------------------------------------------------
# We need to save the mean and median of the reconciled distribution
# in order to compute the squared error and the absolute error:
rec_means <- matrix(NA, ncol = n, nrow = N)
rec_medians <- matrix(NA, ncol = n, nrow = N)

# We need to save the lower and upper quantiles of the reconciled distribution
# in order to compute the interval score:
rec_L <- matrix(NA, ncol = n, nrow = N)
rec_U <- matrix(NA, ncol = n, nrow = N)
int_cov <- 0.9 # use 90% interval
q1 <- (1 - int_cov) / 2
q2 <- (1 + int_cov) / 2

# Set the number of samples to draw from the reconciled distribution:
N_samples <- 1e4

start <- Sys.time()
for (j in 1:N) {
  # Base forecasts:
  base_fc_j <- c()
  for (i in 1:n) {
    base_fc_j[[i]] <- list(size = base_fc$size[[i]], mu = base_fc$mu[[j, i]])
  }

  # Reconcile via importance sampling:
  buis <- reconc_BUIS(A, base_fc_j, "params", "nbinom",
    num_samples = N_samples, seed = 42
  )
  samples_y <- rbind(buis$upper_rec_samples, buis$bottom_rec_samples)

  # Save mean, median, and lower and upper quantiles:
  rec_means[j, ] <- rowMeans(samples_y)
  rec_medians[j, ] <- apply(samples_y, 1, median)
  rec_L[j, ] <- apply(samples_y, 1, quantile, q1)
  rec_U[j, ] <- apply(samples_y, 1, quantile, q2)
}
stop <- Sys.time()
cat(
  "Computational time for ", N, " reconciliations: ",
  round(difftime(stop, start, units = "secs"), 2), "s"
)

## -----------------------------------------------------------------------------
base_means <- base_fc$mu
base_medians <- matrix(NA, ncol = n, nrow = N)
base_L <- matrix(NA, ncol = n, nrow = N)
base_U <- matrix(NA, ncol = n, nrow = N)

for (i in 1:n) {
  base_medians[, i] <- sapply(
    base_means[, i],
    function(mu) qnbinom(p = 0.5, size = base_fc$size[[i]], mu = mu)
  )
  base_L[, i] <- sapply(
    base_means[, i],
    function(mu) qnbinom(p = q1, size = base_fc$size[[i]], mu = mu)
  )
  base_U[, i] <- sapply(
    base_means[, i],
    function(mu) qnbinom(p = q2, size = base_fc$size[[i]], mu = mu)
  )
}

## -----------------------------------------------------------------------------
# Compute the squared errors
SE_base <- (base_means - actuals)^2
SE_rec <- (rec_means - actuals)^2

# Compute the absolute errors
AE_base <- abs(base_medians - actuals)
AE_rec <- abs(rec_medians - actuals)

# Define the function for the interval score
int_score <- function(l, u, actual, int_cov = 0.9) {
  is <- (u - l) +
    2 / (1 - int_cov) * (actual - u) * (actual > u) +
    2 / (1 - int_cov) * (l - actual) * (l > actual)
  return(is)
}

# Compute the interval scores
IS_base <- mapply(int_score, base_L, base_U, data.matrix(actuals))
IS_base <- matrix(IS_base, nrow = N)
IS_rec <- mapply(int_score, rec_L, rec_U, data.matrix(actuals))
IS_rec <- matrix(IS_rec, nrow = N)

## -----------------------------------------------------------------------------
SS_AE <- (AE_base - AE_rec) / (AE_base + AE_rec) * 2
SS_SE <- (SE_base - SE_rec) / (SE_base + SE_rec) * 2
SS_IS <- (IS_base - IS_rec) / (IS_base + IS_rec) * 2
SS_AE[is.na(SS_AE)] <- 0
SS_SE[is.na(SS_SE)] <- 0
SS_IS[is.na(SS_IS)] <- 0

mean_skill_scores <- c(
  round(colMeans(SS_IS), 2),
  round(colMeans(SS_SE), 2),
  round(colMeans(SS_AE), 2)
)
mean_skill_scores <- data.frame(t(matrix(mean_skill_scores, nrow = n)))
colnames(mean_skill_scores) <- names(actuals)
rownames(mean_skill_scores) <- c("Interval score", "Squared error", "Absolute error")
knitr::kable(mean_skill_scores)

## -----------------------------------------------------------------------------
# Now we draw a larger number of samples:
N_samples <- 1e5

### Example of concordant-shift effect
j <- 124
base_fc_j <- c()
for (i in 1:n) {
  base_fc_j[[i]] <- list(
    size = extr_mkt_events_basefc$size[[i]],
    mu = extr_mkt_events_basefc$mu[[j, i]]
  )
}
# Reconcile
buis <- reconc_BUIS(A, base_fc_j, "params", "nbinom",
  num_samples = N_samples, seed = 42
)
samples_y <- rbind(buis$upper_rec_samples, buis$bottom_rec_samples)

# The reconciled mean is lower than both the base and bottom-up means:
means <- c(
  round(extr_mkt_events_basefc$mu[[j, 1]], 2),
  round(sum(extr_mkt_events_basefc$mu[j, 2:n]), 2),
  round(mean(samples_y[1, ]), 2)
)
col_names <- c("Base upper mean", "Bottom-up upper mean", "Reconciled upper mean")
knitr::kable(matrix(means, nrow = 1), col.names = col_names)

### Example of combination effect
j <- 1700
base_fc_j <- c()
for (i in 1:n) {
  base_fc_j[[i]] <- list(
    size = extr_mkt_events_basefc$size[[i]],
    mu = extr_mkt_events_basefc$mu[[j, i]]
  )
}
# Reconcile
buis <- reconc_BUIS(A, base_fc_j, "params", "nbinom",
  num_samples = N_samples, seed = 42
)
samples_y <- rbind(buis$upper_rec_samples, buis$bottom_rec_samples)

# The reconciled mean is between the base and the bottom-up mean:
means <- c(
  round(extr_mkt_events_basefc$mu[[j, 1]], 2),
  round(sum(extr_mkt_events_basefc$mu[j, 2:n]), 2),
  round(mean(samples_y[1, ]), 2)
)
col_names <- c("Base upper mean", "Bottom-up upper mean", "Reconciled upper mean")
knitr::kable(matrix(means, nrow = 1), col.names = col_names)

## -----------------------------------------------------------------------------
j <- 2308
base_fc_j <- c()
for (i in 1:n) {
  base_fc_j[[i]] <- list(
    size = extr_mkt_events_basefc$size[[i]],
    mu = extr_mkt_events_basefc$mu[[j, i]]
  )
}
# Reconcile
buis <- reconc_BUIS(A, base_fc_j, "params", "nbinom", num_samples = N_samples, seed = 42)
samples_y <- rbind(buis$upper_rec_samples, buis$bottom_rec_samples)

# Compute the variance of the base and reconciled bottom forecasts
base_bottom_var <- mapply(
  function(mu, size) var(rnbinom(n = 1e5, size = size, mu = mu)),
  extr_mkt_events_basefc$mu[j, 2:n],
  extr_mkt_events_basefc$size[2:n]
)
rec_bottom_var <- apply(samples_y[2:n, ], MARGIN = 1, var)

# The reconciled variance is greater than the base variance:
bottom_var <- rbind(base_bottom_var, rec_bottom_var)
rownames(bottom_var) <- c("var base", "var reconc")
knitr::kable(round(bottom_var, 2))

#' Estimate covariance from naive or seasonal-naive residuals
#'
#' Estimates via shrinkage the covariance matrix of the residuals of the naive 
#' or seasonal naive forecasts.
#' If the frequency of the time series is > 1, the function chooses between the
#' two methods series-by-series according to a selection criterion.
#' If the frequency is 1 or is not provided, only the naive residuals are used.
#'
#' @param y_train Multivariate time series object or numeric matrix of historical 
#'   observations, with dimensions `T x n` (rows are time points, columns are series).
#' @param freq Positive integer seasonal frequency (optional).
#'   If not provided and `y_train` is a multivariate time series, the frequency 
#'   of the data is used. 
#' @param criterion Character string used when `freq > 1` to choose residuals.
#'   Supported values are `"RSS"` (default) and `"seas-test"`.
#'   `"RSS"` chooses the method with the lower residual sum of squares (RSS), 
#'   while `"seas-test"` uses a statistical test for seasonality (requires `forecast` package).
#'
#' @return A numeric `n x n` shrinkage covariance matrix estimated with
#'   [schaferStrimmer_cov()].
#'
#' @seealso [schaferStrimmer_cov()], [reconc_t()]
#'
#' @keywords internal
#' @export
.compute_naive_cov = function(y_train, freq = NULL, criterion = "RSS") {
  
  n <- ncol(y_train)
  L <- nrow(y_train)
  
  # Compute residuals of naive
  res_n <- y_train[2:L, ] - y_train[1:(L - 1), ]
  
  # If freq is not provided and the data is a multivariate time series, 
  # set freq to the frequency of the data
  if (stats::is.mts(y_train)) {
    if (is.null(freq)) {
      freq <- stats::frequency(y_train)
    }
  }

  if (is.null(freq) || freq == 1) {
    residuals <- res_n
  } else {
    # compute residuals of seasonal naive
    res_seas <- y_train[(freq + 1):L, ] - y_train[1:(L - freq), ]

    if (is.null(criterion)) {
      criterion <- "RSS"
    }

    # Choose which residuals to use based on the criterion
    if (criterion == "seas-test") {
      # check if forecast package is installed
      if (!requireNamespace("forecast", quietly = TRUE)) {
        stop("Package 'forecast' is required for criterion = 'seas-test'. 
              Please install it using install.packages('forecast') or use criterion = 'RSS'."
        )
      }
      # Seasonality test for each time series
      is_seas = as.logical(apply(y_train, 2, 
                                 function(col) forecast::nsdiffs(stats::ts(col,frequency = freq))))
      
    } else if (criterion == "RSS") {
      # Compute RSS for both methods, for each series
      RSS_n <- colSums(res_n^2, na.rm = TRUE)
      RSS_seas <- colSums(res_seas^2, na.rm = TRUE)
      is_seas <- RSS_seas < RSS_n
      
    } else {
      stop("Input error: criterion must be either 'RSS' or 'seas-test'")
    }

    # No series is seasonal
    if (sum(is_seas) == 0) {
      residuals <- res_n

      # All series are seasonal
    } else if (sum(is_seas) == n) {
      residuals <- res_seas

      # Some series are seasonal, some are not (use criterion)
    } else {
      residuals <- matrix(NA, nrow = L - freq, ncol = n)
      residuals[, is_seas] <- res_seas[, is_seas]
      residuals[, !is_seas] <- res_n[freq:(L - 1), !is_seas]
    }
  }

  Sigma_naive <- schaferStrimmer_cov(residuals)$shrink_cov

  return(Sigma_naive)
}


#' Optimize degrees of freedom (nu) via LOO Cross-Validation
#'
#' @param res Matrix of residuals (n_obs x n_var).
#' @param prior_mean The prior mean covariance matrix (n_var x n_var).
#' @param trim Fraction of observations to trim (0 to 1). Default is 0.1 (10%).
#'
#' @details
#' TODO: check these details
#'
#' \strong{Leave-One-Out (LOO) Cross-Validation:}
#' This function estimates the optimal degrees of freedom \eqn{\nu} by maximizing the
#' out-of-sample predictive performance. This is achieved by computing the
#' log-density of each held-out observation \eqn{\mathbf{r}_i} given the remaining
#' data \eqn{\mathbf{R}_{-i}}. The total objective function is the sum of these
#' predictive log-densities:
#' \deqn{\mathcal{L}(\nu) = \sum_{i=1}^T \log f(\mathbf{r}_i | \mathbf{R}_{-i}, \nu)}
#'
#' \strong{The Log-Density Function:}
#' For each LOO step, the residuals are assumed to follow a
#' multivariate t-distribution. The density is expressed directly as a function of the
#' posterior sum-of-squares matrix \eqn{\Psi}, where \eqn{\Psi} scales implicitly with \eqn{\nu}:
#' \deqn{f(\mathbf{r}_i | \Psi, \nu) = \frac{\Gamma(\frac{\nu + T}{2})}{\Gamma(\frac{\nu + T - p}{2}) \pi^{p/2}} |\Psi|^{-1/2} \left( 1 + \mathbf{r}_i^\top \Psi^{-1} \mathbf{r}_i \right)^{-\frac{\nu + T}{2}}}
#' In the code, \eqn{\Psi} is constructed as:
#' \deqn{\Psi = (\nu - p - 1)\bar{\Sigma}_{prior} + \mathbf{R}^\top\mathbf{R}}
#' By using this formulation, the standard scaling factors \eqn{1/\nu} and \eqn{\nu^{-p/2}}
#' are absorbed into the matrix inverse and determinant, respectively.
#'
#' \strong{Efficient Computation via Sherman-Morrison:}
#' Rather than recomputing \eqn{\Psi_{-i}} and its inverse \eqn{T} times, the function uses
#' the full-sample matrix \eqn{\Psi} and adjusts it using the leverage
#' \eqn{h_i = \mathbf{r}_i^\top \Psi^{-1} \mathbf{r}_i}.
#'
#' Through the Matrix Determinant Lemma and the Sherman-Morrison formula, the
#' internal term \eqn{(1 + \mathbf{r}_i^\top \Psi_{-i}^{-1} \mathbf{r}_i)} simplifies to \eqn{(1 - h_i)^{-1}}.
#' The final log-density contribution used in the code is:
#' \deqn{\log f_i \propto \textrm{const} - \frac{1}{2}\log|\Psi| + \frac{\nu + T - 1}{2} \log(1 - h_i)}
#'
#' @return A list containing the optimization results:
#' * `optimal_nu`: The optimal degrees of freedom found.
#' * `min_neg_log_score`: The minimum negative log score achieved.
#' * `convergence`: The convergence status of the optimizer.
#' * `time_elapsed`: The time taken for the optimization.
#'
#' @import nloptr
#'
#' @export
multi_log_score_optimization <- function(res, prior_mean, trim = 0.1) {
  n_obs <- nrow(res)
  n_var <- ncol(res)

  # Pre-compute cross-product of the residuals (to avoid recomputing inside the loop)
  RRt <- crossprod(res)

  # Define the Objective Function (Negative Log Score to minimize)
  objective_function <- function(nu) {
    # Compute posterior Psi
    Psi <- (prior_mean * (nu - n_var - 1)) + RRt

    # Compute Cholesky decomposition
    Psi_chol <- tryCatch(chol(Psi), error = function(e) {
      return(NULL)
    })
    if (is.null(Psi_chol)) {
      return(Inf)
    } # Return Inf if Psi is not positive definite

    # Invert (using Cholesky)
    inv_Psi <- chol2inv(Psi_chol)

    # Calculate log-determinant
    log_det_Psi <- 2 * sum(log(diag(Psi_chol)))

    # Log Gamma terms and determinant
    const_term <- lgamma((nu + n_obs) / 2) - lgamma((nu + n_obs - n_var) / 2) - log_det_Psi / 2

    # Compute LOO Terms (using Sherman-Morrison)
    h_i <- rowSums((res %*% inv_Psi) * res)
    if (any(h_i >= 1)) {
      return(Inf)
    } # If h_i >= 1, the log is undefined
    log_adjustments <- log(1 - h_i)

    # Combine to get vector of LOO Log-Likelihoods per observation
    loo_log_liks <- const_term + ((nu + n_obs - 1) / 2) * log_adjustments

    # Calculate how many observations to remove (based on 'trim')
    n_trim <- round(trim * n_obs)

    # Sort likelihoods (ascending) and remove the lowest 'n_trim' values
    # Return negative sum because optimizer minimizes
    return(-sum(sort(loo_log_liks)[(n_trim + 1):n_obs]))
  }

  # Optimization Setup
  initial_guess <- n_var + 3
  lower_bound <- n_var + 2
  upper_bound <- max(5 * n_var, n_obs)

  opts <- list(
    "algorithm" = "NLOPT_LN_BOBYQA",
    "xtol_rel" = 1e-5,
    "maxeval" = 1000
  )

  # Run Optimization
  start_time <- Sys.time()
  results <- nloptr::nloptr(
    x0     = initial_guess,
    eval_f = objective_function,
    lb     = lower_bound,
    ub     = upper_bound,
    opts   = opts
  )
  end_time <- Sys.time()

  return(list(
    optimal_nu = results$solution,
    min_neg_log_score = results$objective,
    convergence = results$status,
    time_elapsed = end_time - start_time
  ))
}


#' t-Rec: reconciliation via conditioning with uncertain covariance via multivariate t-distribution
#'
#' Reconciles base forecasts in a hierarchy by conditioning on the hierarchical
#' constraints, specified by the aggregation matrix A.
#' The base forecasts are assumed to be jointly Gaussian, conditionally on the
#' covariance matrix of the forecast errors. 
#' A Bayesian approach is adopted to account for the uncertainty of the covariance matrix. 
#' An Inverse-Wishart prior is specified on the covariance matrix, 
#' leading to a multivariate t-distribution for the base forecasts.
#' The reconciliation via conditioning is in closed-form, yielding a multivariate t 
#' reconciled distribution.
#'
#' @param A Matrix (n_upp x n_bott) defining the hierarchy (u = Ab).
#' @param base_fc_mean Vector of base forecasts (length n = n_upp + n_bott).
#' @param y_train mts (or matrix) of historical training data (T x n) used for setting prior parameters.
#' @param residuals Matrix (T x n) of base forecast residuals.
#' @param ... Additional arguments for advanced usage: see details.
#' @param return_upper Logical; if TRUE, also returns parameters for the upper-level reconciled distribution.
#' @param return_parameters Logical; if TRUE, also returns prior and posterior parameters.
#'
#' @details
#' \strong{Standard usage.}
#' 
#' The standard workflow for this function is to provide the in-sample \code{residuals}
#' and the historical training data \code{y_train}.
#' The parameters of the Inverse-Wishart prior distribution of the covariance matrix 
#' are set as follows: 
#' \itemize{
#'   \item Prior scale matrix (Psi): set as the covariance of the residuals of naive (or seasonal naive,
#'         a criterion is used to choose between the two) forecasts computed on \code{y_train}.
#'   \item Prior degrees of freedom (nu): estimated via Bayesian Leave-One-Out
#'         Cross-Validation (LOOCV) to maximize out-of-sample performance.
#' }
#' The posterior distribution of the covariance matrix is still Inverse-Wishart.
#' The parameters of the posterior are computed in closed-form using the sample 
#' covariance of the provided \code{residuals}.
#'
#' \strong{Advanced Options.}
#' 
#' Users can bypass the automated estimation by specifying:
#' \enumerate{
#'   \item \code{prior}: a list with entries 'nu' and 'Psi'.
#'         This skips the LOOCV step for \eqn{\nu_0} and the covariance estimation from \code{y_train}.
#'         It requires \code{residuals} to compute the posterior.
#'   \item \code{posterior}: a list with entries 'nu' and 'Psi'.
#'         This skips all internal estimation and updating logic.
#' }
#' Moreover, users can specify:
#' \itemize{
#'   \item \code{freq}: positive integer, used as frequency of data for the seasonal naive forecast in the specification of the prior scale matrix.
#'         By default, if \code{y_train} is a multivariate time series, the frequency of the data is used; otherwise, it is set to 1 (no seasonality).
#'   \item \code{criterion}: either 'RSS' (default) or 'seas-test', specifying which criterior is used to choose between 
#'                           the naive and seasonal naive forecasts for the specification of the prior scale matrix. 
#'                           'RSS' computes the residual sum of squares for both methods and chooses the one with lower RSS, 
#'                           while 'seas-test' uses a statistical test for seasonality 
#'                           (currently implemented using the number of seasonal differences suggested by the `forecast` package, 
#'                           which must be installed). 
#' }
#' 
#' \strong{Reconciled distribution.}
#' 
#' The reconciled distribution is a multivariate t-distribution, 
#' specified by a vector of means, a scale matrix, and a number of degrees of freedom.
#' These parameters are computed in closed-form.
#' By default, only the parameters of the reconciled distribution for the bottom-level 
#' series are returned. See examples.
#'
#'
#' @references
#' Carrara, C., Corani, G., Azzimonti, D., & Zambon, L. (2025). Modeling the uncertainty on the covariance
#' matrix for probabilistic forecast reconciliation. arXiv preprint arXiv:2506.19554.
#' \url{https://arxiv.org/abs/2506.19554}
#'
#' @return A list containing:
#' \itemize{
#'   \item \code{bottom_rec_mean}: reconciled bottom-level mean.
#'   \item \code{bottom_rec_scale_matrix}: reconciled bottom-level scale matrix.
#'   \item \code{bottom_rec_df}: reconciled degrees of freedom.
#' }
#' If \code{return_upper} is TRUE, also returns:
#' \itemize{
#'  \item \code{upper_rec_mean}: reconciled upper-level mean.
#'  \item \code{upper_rec_scale_matrix}: reconciled upper-level scale matrix.
#'  \item \code{upper_rec_df}: reconciled upper-level degrees of freedom.
#'  }
#' If \code{return_parameters} is TRUE, also returns:
#' \itemize{
#'   \item \code{prior_nu}: prior degrees of freedom.
#'   \item \code{prior_Psi}: prior scale matrix.
#'   \item \code{posterior_nu}: posterior degrees of freedom.
#'   \item \code{posterior_Psi}: posterior scale matrix.
#' }
#'
#' @examples
#' 
#' \donttest{
#' library(bayesRecon)
#'
#' if (requireNamespace("forecast", quietly = TRUE)) {
#' 
#'   set.seed(1234)
#'   n_obs <- 100
#' 
#'   # Simulate 2 bottom series from AR(1) processes
#'   y1 <- arima.sim(model = list(ar = 0.8), n = n_obs)
#'   y2 <- arima.sim(model = list(ar = 0.5), n = n_obs)
#'
#'   y_upper <- y1 + y2   # upper series is the sum of the two bottoms
#'   A <- matrix(c(1, 1), nrow = 1)  # Aggregation matrix
#'
#'   # Fit additive ETS models
#'   fit1 <- forecast::ets(y1, additive.only = TRUE)
#'   fit2 <- forecast::ets(y2, additive.only = TRUE)
#'   fit_upper <- forecast::ets(y_upper, additive.only = TRUE)
#'
#'   # Point forecasts (h = 1)
#'   fc_upper <- as.numeric(forecast::forecast(fit_upper, h = 1)$mean)
#'   fc1 <- as.numeric(forecast::forecast(fit1, h = 1)$mean)
#'   fc2 <- as.numeric(forecast::forecast(fit2, h = 1)$mean)
#'   base_fc_mean <- c(fc_upper, fc1, fc2)
#'
#'   # Residuals and training data (n_obs x n matrices, columns in same order as base_fc_mean)
#'   res <- cbind(residuals(fit_upper), residuals(fit1), residuals(fit2))
#'   y_train <- cbind(y_upper, y1, y2)
#'
#'   # --- 1) Generate joint reconciled samples ---
#'   result <- reconc_t(A, base_fc_mean, y_train = y_train, residuals = res)
#'
#'   # Sample from the reconciled bottom-level t-distribution
#'   n_samples <- 2000
#'   L_chol <- t(chol(result$bottom_rec_scale_matrix))
#'   z <- matrix(rt(ncol(A) * n_samples, df = result$bottom_rec_df), nrow = ncol(A))
#'   bottom_samples <- result$bottom_rec_mean + L_chol %*% z  # 2 x n_samples
#'
#'   # Aggregate bottom samples to get upper samples
#'   upper_samples <- A %*% bottom_samples              
#'   joint_samples <- rbind(upper_samples, bottom_samples)
#'   rownames(joint_samples) <- c("upper", "bottom_1", "bottom_2")
#'
#'   cat("Reconciled means (from samples):\n")
#'   print(round(rowMeans(joint_samples), 3))
#' 
#'   cat("Reconciled standard deviations (from samples):\n")
#'   print(round(apply(joint_samples, 1, sd), 3))
#'
#'   # --- 2) 95% prediction intervals via t-distribution quantiles ---
#'   result2 <- reconc_t(A, base_fc_mean, y_train = y_train,
#'                       residuals = res, return_upper = TRUE)
#'
#'   alpha <- 0.05
#'   # Bottom series intervals
#'   for (i in seq_len(ncol(A))) {
#'     s_i <- sqrt(result2$bottom_rec_scale_matrix[i, i])
#'     lo  <- result2$bottom_rec_mean[i] + s_i * qt(alpha / 2,     df = result2$bottom_rec_df)
#'     hi  <- result2$bottom_rec_mean[i] + s_i * qt(1 - alpha / 2, df = result2$bottom_rec_df)
#'     cat(sprintf("Bottom %d: 95%% PI = [%.3f, %.3f]\n", i, lo, hi))
#'   }
#'   # Upper series interval
#'   s_u <- sqrt(result2$upper_rec_scale_matrix[1, 1])
#'   lo  <- result2$upper_rec_mean[1] + s_u * qt(alpha / 2,     df = result2$upper_rec_df)
#'   hi  <- result2$upper_rec_mean[1] + s_u * qt(1 - alpha / 2, df = result2$upper_rec_df)
#'   cat(sprintf("Upper:    95%% PI = [%.3f, %.3f]\n", lo, hi))
#' }
#' }
#' 
#' @seealso 
#' [reconc_gaussian()]
#'
#' @export
reconc_t <- function(A,
                     base_fc_mean,
                     y_train = NULL,
                     residuals = NULL,
                     ...,
                     return_upper = FALSE,
                     return_parameters = FALSE) {
  
  add_args <- list(...)
  unused_names <- setdiff(names(add_args), c("prior", "posterior", "freq", "criterion"))
  if (length(unused_names) > 0) {
    warning(paste("The following additional arguments are not used:", paste(unused_names, collapse = ", ")))
  }
  
  prior <- add_args$prior
  posterior <- add_args$posterior
  freq <- add_args$freq
  criterion <- add_args$criterion
  .check_input_t(A, base_fc_mean, y_train, residuals, ...)  

  ##############################################################################
  ### CASE 1 ###
  # If posterior is provided, check if is a list with entries nu and Psi and extract values
  if (!is.null(posterior)) {
    nu_post <- posterior$nu
    Psi_post <- posterior$Psi
    
    ### CASE 2 ###
    # If posterior not provided, first check that residuals are provided
  } else {
    L <- nrow(residuals) # number of residual samples (i.e., training length)
    n <- length(base_fc_mean) # number of series
    
    # Compute sample covariance of the residuals
    Samp_cov <- crossprod(residuals) / nrow(residuals) 
    # Shrink to diagonal for numerical reasons
    Samp_cov <- (1 - .L_SHRINK_RECONC_T) * Samp_cov + .L_SHRINK_RECONC_T * diag(diag(Samp_cov)) 
    
    ### CASE 2a ###
    # If prior is provided, check if is a list with entries nu and Psi and extract values
    if (!is.null(prior)) {
      nu_prior <- prior$nu
      Psi_prior <- prior$Psi
      
      ### CASE 2b ###
      # If prior not provided:
      # - compute Psi using the (shrinked) covariance matrix of the residuals of the naive
      #   or seasonal naive forecasts
      # - set nu using LOOCV
    } else {
      # Compute the covariance residuals of the naive or seasonal naive forecasts on training data
      cov_naive <- .compute_naive_cov(y_train, freq = freq, criterion = criterion)
      # Use it to set the prior
      bayesian_LOO <- multi_log_score_optimization(residuals, cov_naive)
      nu_prior <- bayesian_LOO$optimal_nu
      Psi_prior <- (nu_prior - n - 1) * cov_naive
    }
    # Compute posterior parameters
    Psi_post <- Psi_prior + L * Samp_cov
    nu_post <- nu_prior + nrow(residuals)
  }

  ##############################################################################
  # Reconcile via conditioning the t-distribution in closed form

  out <- .core_reconc_t(
    A = A, base_fc_mean = base_fc_mean, Psi_post = Psi_post, nu_post = nu_post,
    return_upper = return_upper)

  if (return_parameters) {
    if(!is.null(posterior)) {
      warning("Prior parameters are not returned when 'posterior' is provided, 
              as the prior is not used in this case.")
    } else {
      out$prior_nu <- nu_prior
      out$prior_Psi <- Psi_prior
    }
    out$posterior_nu <- nu_post
    out$posterior_Psi <- Psi_post
  }

  return(out)
}

#' Core reconciliation via multivariate t-distribution.
#'
#' Internal function that performs the core reconciliation logic for the t-distribution based
#' reconciliation method. This function assumes an uncertain covariance matrix with an Inverse-Wishart prior.
#'
#' @param A Matrix (n_upper x n_bottom) defining the hierarchy where upper = A %*% bottom.
#' @param base_fc_mean Vector of length (n_upper + n_bottom) containing the base forecast means
#'   for both upper and bottom levels (upper first, then bottom).
#' @param Psi_post Scale matrix (n_upper + n_bottom x n_upper + n_bottom) of the posterior
#'   multivariate t-distribution.
#' @param nu_post Degrees of freedom of the posterior multivariate t-distribution.
#' @param return_upper Logical, whether to return the reconciled parameters for the upper variables (default is FALSE).
#'
#' @return A list containing:
#' \itemize{
#'   \item \code{bottom_rec_mean}: reconciled bottom-level mean.
#'   \item \code{bottom_rec_scale_matrix}: reconciled bottom-level scale matrix.
#'   \item \code{bottom_rec_df}: reconciled degrees of freedom.
#' }
#' If \code{return_upper} is TRUE, also returns:
#' \itemize{
#'  \item \code{upper_rec_mean}: reconciled upper-level mean.
#'  \item \code{upper_rec_scale_matrix}: reconciled upper-level scale matrix.
#'  \item \code{upper_rec_df}: reconciled upper-level degrees of freedom.
#'  }
#'
#' @keywords internal
#' @export
.core_reconc_t <- function(A, base_fc_mean, Psi_post, nu_post, return_upper = FALSE,
                           return_parameters = FALSE) {
  # Indices for Upper and Bottom
  k <- nrow(A)
  m <- ncol(A)
  idx_u <- 1:k
  idx_b <- (k + 1):(k + m)

  # Extract Psi blocks
  Psi_U <- Psi_post[idx_u, idx_u, drop = FALSE]
  Psi_B <- Psi_post[idx_b, idx_b, drop = FALSE]
  Psi_UB <- Psi_post[idx_u, idx_b, drop = FALSE]

  # Extract means
  u_hat <- base_fc_mean[idx_u]
  b_hat <- base_fc_mean[idx_b]

  # Compute Q = Psi_U - (Psi_UB %*% t(A)) - (A %*% t(Psi_UB)) + (A %*% Psi_B %*% t(A))
  Psi_UB_At <- tcrossprod(Psi_UB, A)
  A_Psi_B <- A %*% Psi_B
  Q <- Psi_U - Psi_UB_At - t(Psi_UB_At) + tcrossprod(A_Psi_B, A)

  # Invert Q using Cholesky (should be p.d.)
  Q_chol <- tryCatch(chol(Q), error = function(e) NULL)
  # If Cholesky fails, stop
  if (is.null(Q_chol)) {
    stop("A numerical error occured during a Cholesky decomposition. 
         This may be due to redundant aggregated series.")
  }
  inv_Q <- chol2inv(Q_chol)

  # Incoherence
  delta <- (A %*% b_hat) - u_hat

  # Lambda = Psi_UB^T - Psi_B A^T
  Lambda <- t(Psi_UB) - tcrossprod(Psi_B, A)
  Lambda_invQ <- Lambda %*% inv_Q

  # Compute b_tilde = b_hat + Lambda * Q^{-1} * delta
  b_tilde <- b_hat + (Lambda_invQ %*% delta)

  # Compute nu_tilde = nu' - n_b + 1
  nu_tilde <- nu_post - m + 1

  # Compute C = (1 + delta^T Q^{-1} delta) / nu_tilde
  mahalanobis_term <- sum(delta * (inv_Q %*% delta)) # efficient x^T A x
  C <- (1 + mahalanobis_term) / nu_tilde

  # Compute Sigma_B_tilde = C * [ Psi_B - Lambda * Q^{-1} * Lambda^T ]
  Sigma_tilde_B <- as.numeric(C) * (Psi_B - tcrossprod(Lambda_invQ, Lambda))

  # Prepare output
  out <- list(
    bottom_rec_mean = as.vector(b_tilde),
    bottom_rec_scale_matrix = Sigma_tilde_B,
    bottom_rec_df = nu_tilde
  )
  if (return_upper) {
    # Compute the parameters of the uppers using closure property
    u_tilde <- A %*% b_tilde
    Sigma_tilde_U <- A %*% Sigma_tilde_B %*% t(A)
    out$upper_rec_mean <- as.vector(u_tilde)
    out$upper_rec_scale_matrix <- Sigma_tilde_U
    out$upper_rec_df <- nu_tilde
  }

  return(out)
}

# Shared fixtures --------------------------------------------------------------

# Simple 1-upper, 2-bottom hierarchy: u = b1 + b2
.A_simple <- matrix(c(1, 1), nrow = 1)

# Coherent base forecasts: upper = sum of bottoms
.mu_coh <- c(10, 4, 6) # u=10, b1=4, b2=6  =>  delta = A%*%b - u = 0

# Incoherent base forecasts: upper != sum of bottoms
.mu_inc <- c(11, 4, 6) # u=11, b1=4, b2=6  =>  delta = -1

# Posterior parameters
.nu_post <- 50
.Psi_post <- diag(c(3, 1, 1.5)) * .nu_post  # scale ~ nu * sigma^2 per series


# 1. Output structure ----------------------------------------------------------

test_that("reconc_t output contains expected fields (bottom only)", {
  res <- reconc_t(.A_simple, .mu_coh,
    posterior = list(nu = .nu_post, Psi = .Psi_post)
  )

  expect_named(res, c("bottom_rec_mean", "bottom_rec_scale_matrix", "bottom_rec_df"),
    ignore.order = TRUE
  )
  expect_length(res$bottom_rec_mean, ncol(.A_simple))
  expect_equal(dim(res$bottom_rec_scale_matrix), c(ncol(.A_simple), ncol(.A_simple)))
  expect_length(res$bottom_rec_df, 1)
})


# 2. Degrees of freedom formula ------------------------------------------------

test_that("bottom_rec_df equals nu_post - n_bottom + 1", {
  res <- reconc_t(.A_simple, .mu_coh,
    posterior = list(nu = .nu_post, Psi = .Psi_post)
  )
  expect_equal(res$bottom_rec_df, .nu_post - ncol(.A_simple) + 1)
})


# 3. Coherent forecasts: reconciled mean unchanged -----------------------------

test_that("reconciled mean is unchanged when base forecasts are already coherent", {
  res <- reconc_t(.A_simple, .mu_coh,
    posterior = list(nu = .nu_post, Psi = .Psi_post)
  )
  expect_equal(res$bottom_rec_mean, .mu_coh[(nrow(.A_simple) + 1):length(.mu_coh)],
    tolerance = 1e-10
  )
})


# 4. return_upper: upper_rec_mean == A %*% bottom_rec_mean ----------------------

test_that("return_upper=TRUE returns correct upper fields consistent with bottom", {
  res <- reconc_t(.A_simple, .mu_inc,
    posterior = list(nu = .nu_post, Psi = .Psi_post),
    return_upper = TRUE
  )

  expect_true(all(c("upper_rec_mean", "upper_rec_scale_matrix", "upper_rec_df") %in% names(res)))

  # upper_rec_mean must equal A %*% bottom_rec_mean (closure property)
  expect_equal(
    as.vector(res$upper_rec_mean),
    as.vector(.A_simple %*% res$bottom_rec_mean),
    tolerance = 1e-10
  )

  # upper_rec_scale_matrix must equal A %*% bottom_rec_scale_matrix %*% t(A)
  expect_equal(
    res$upper_rec_scale_matrix,
    .A_simple %*% res$bottom_rec_scale_matrix %*% t(.A_simple),
    tolerance = 1e-10
  )

  # degrees of freedom are the same for upper and bottom
  expect_equal(res$upper_rec_df, res$bottom_rec_df)
})

test_that("return_upper=FALSE (default) does not return upper fields", {
  res <- reconc_t(.A_simple, .mu_inc,
    posterior = list(nu = .nu_post, Psi = .Psi_post)
  )
  expect_false("upper_rec_mean" %in% names(res))
  expect_false("upper_rec_scale_matrix" %in% names(res))
})


# 5. Scale matrix is symmetric and positive-definite ---------------------------

test_that("bottom_rec_scale_matrix is symmetric and positive definite", {
  res <- reconc_t(.A_simple, .mu_inc,
    posterior = list(nu = .nu_post, Psi = .Psi_post)
  )
  S <- res$bottom_rec_scale_matrix

  # Symmetry
  expect_equal(S, t(S), tolerance = 1e-12)

  # Positive definiteness: all eigenvalues > 0
  eigs <- eigen(S, symmetric = TRUE, only.values = TRUE)$values
  expect_true(all(eigs > 0))
})


# 6. Incoherence: reconciled mean is coherent ----------------------------------
# After reconciliation, A %*% bottom_rec_mean always equals upper_rec_mean

test_that("reconciled forecasts satisfy the hierarchical constraint", {
  res <- reconc_t(.A_simple, .mu_inc,
    posterior = list(nu = .nu_post, Psi = .Psi_post),
    return_upper = TRUE
  )
  expect_equal(
    as.vector(.A_simple %*% res$bottom_rec_mean),
    as.vector(res$upper_rec_mean),
    tolerance = 1e-10
  )
})


# 7. Input via residuals (no error) --------------------------------------------

test_that("reconc_t runs without error when residuals and prior are provided", {
  set.seed(42)
  n <- 3
  L <- 30  # training length

  A <- .A_simple
  mu <- .mu_inc

  # Simulate residuals
  res_mat <- matrix(rnorm(L * n), nrow = L, ncol = n)

  # Simple diagonal prior
  nu_prior <- 10
  Psi_prior <- diag(n) * (nu_prior - n - 1)

  expect_no_error(
    reconc_t(A, mu, prior = list(nu = nu_prior, Psi = Psi_prior), residuals = res_mat)
  )
})


# 8. Posterior input reproduces prior+residuals path -------------------------

test_that("direct posterior input gives same result as prior+residuals path", {
  set.seed(7)
  n <- 3
  L <- 20

  A <- .A_simple
  mu <- .mu_inc
  res_mat <- matrix(rnorm(L * n), nrow = L, ncol = n)

  nu_prior <- 15
  Psi_prior <- diag(n) * (nu_prior - n - 1)

  # Path A: via prior + residuals
  res_via_prior <- reconc_t(A, mu,
    prior = list(nu = nu_prior, Psi = Psi_prior),
    residuals = res_mat
  )

  # Manual posterior computation (mirrors internal logic)
  Samp_cov <- crossprod(res_mat) / L
  lam <- .L_SHRINK_RECONC_T
  Samp_cov_shr <- (1 - lam) * Samp_cov + lam * diag(diag(Samp_cov))
  nu_post_man <- nu_prior + L
  Psi_post_man <- Psi_prior + L * Samp_cov_shr

  # Path B: directly provide posterior
  res_via_post <- reconc_t(A, mu,
    posterior = list(nu = nu_post_man, Psi = Psi_post_man)
  )

  expect_equal(res_via_prior$bottom_rec_mean, res_via_post$bottom_rec_mean, tolerance = 1e-10)
  expect_equal(res_via_prior$bottom_rec_scale_matrix, res_via_post$bottom_rec_scale_matrix,
    tolerance = 1e-10
  )
  expect_equal(res_via_prior$bottom_rec_df, res_via_post$bottom_rec_df)
})


# 9. Unknown argument triggers a warning -------------------------------------

test_that("unknown arguments in ... trigger a warning", {
  expect_warning(
    reconc_t(.A_simple, .mu_coh,
      posterior = list(nu = .nu_post, Psi = .Psi_post),
      unknown_arg = 99
    ),
    regexp = "unknown_arg"
  )
})


# 10. Standard usage: y_train + residuals (CASE 2b) ---------------------------
# This path runs .compute_naive_cov + multi_log_score_optimization to estimate
# nu_prior and Psi_prior, then updates to the posterior and reconciles.

# Shared fixture for CASE 2b tests (mts y_train with frequency=1, no seasonality)
set.seed(123)
.n_series <- 3L   # 1 upper + 2 bottom
.L_train  <- 50L
.A_2b     <- matrix(c(1, 1), nrow = 1)
.y_mat    <- ts(matrix(rnorm(.L_train * .n_series, mean = 10), nrow = .L_train, ncol = .n_series),
                frequency = 4)  # mts object with frequency=4, but no actual seasonality
.res_mat  <- matrix(rnorm(.L_train * .n_series), nrow = .L_train, ncol = .n_series)
.mu_2b    <- c(11, 4, 6)  # incoherent: 11 != 4 + 6 = 10

test_that("reconc_t runs without error with y_train + residuals (standard usage)", {
  expect_no_error(
    reconc_t(.A_2b, .mu_2b, y_train = .y_mat, residuals = .res_mat)
  )
})

test_that("reconc_t output structure is correct with y_train + residuals", {
  result <- reconc_t(.A_2b, .mu_2b, y_train = .y_mat, residuals = .res_mat)

  expect_named(result, c("bottom_rec_mean", "bottom_rec_scale_matrix", "bottom_rec_df"),
    ignore.order = TRUE
  )
  expect_length(result$bottom_rec_mean, ncol(.A_2b))
  expect_equal(dim(result$bottom_rec_scale_matrix), c(ncol(.A_2b), ncol(.A_2b)))
  expect_length(result$bottom_rec_df, 1)
  expect_true(result$bottom_rec_df > 0)
})

test_that("reconciled forecasts satisfy hierarchical constraint with y_train + residuals", {
  result <- reconc_t(.A_2b, .mu_2b, y_train = .y_mat, residuals = .res_mat,
    return_upper = TRUE
  )

  expect_equal(
    as.vector(.A_2b %*% result$bottom_rec_mean),
    as.vector(result$upper_rec_mean),
    tolerance = 1e-10
  )
})

test_that("bottom_rec_df equals posterior_nu - n_bottom + 1 with y_train + residuals", {
  result <- reconc_t(.A_2b, .mu_2b, y_train = .y_mat, residuals = .res_mat,
    return_parameters = TRUE
  )

  expect_equal(result$bottom_rec_df, result$posterior_nu - ncol(.A_2b) + 1)
})

test_that("posterior_nu equals optimized nu_prior + L with y_train + residuals", {
  result <- reconc_t(.A_2b, .mu_2b, y_train = .y_mat, residuals = .res_mat,
    return_parameters = TRUE
  )

  # posterior_nu = nu_prior + L; since nu_prior >= n + 2, posterior_nu > L
  expect_true(result$posterior_nu > .L_train)
  # Also must be valid for the t-distribution (> n_series - 1)
  expect_true(result$posterior_nu > .n_series - 1)
})

test_that("prior and posterior parameters are returned with y_train + residuals (return_parameters=TRUE)", {
  result <- reconc_t(.A_2b, .mu_2b, y_train = .y_mat, residuals = .res_mat,
    return_parameters = TRUE
  )

  expect_true(all(c("prior_nu", "prior_Psi", "posterior_nu", "posterior_Psi") %in% names(result)))
  expect_true(is.numeric(result$prior_nu))
  expect_true(is.matrix(result$prior_Psi))
  expect_equal(dim(result$prior_Psi), c(.n_series, .n_series))
  expect_true(is.numeric(result$posterior_nu))
  expect_true(is.matrix(result$posterior_Psi))
  expect_equal(dim(result$posterior_Psi), c(.n_series, .n_series))
})

test_that("bottom_rec_scale_matrix is symmetric and positive definite with y_train + residuals", {
  result <- reconc_t(.A_2b, .mu_2b, y_train = .y_mat, residuals = .res_mat)

  S <- result$bottom_rec_scale_matrix
  expect_equal(S, t(S), tolerance = 1e-12)
  eigs <- eigen(S, symmetric = TRUE, only.values = TRUE)$values
  expect_true(all(eigs > 0))
})

test_that("return_upper=TRUE works correctly with y_train + residuals", {
  result <- reconc_t(.A_2b, .mu_2b, y_train = .y_mat, residuals = .res_mat,
    return_upper = TRUE
  )

  expect_true(all(c("upper_rec_mean", "upper_rec_scale_matrix", "upper_rec_df") %in% names(result)))
  expect_equal(
    result$upper_rec_scale_matrix,
    .A_2b %*% result$bottom_rec_scale_matrix %*% t(.A_2b),
    tolerance = 1e-10
  )
  expect_equal(result$upper_rec_df, result$bottom_rec_df)
})

test_that("reconc_t standard usage works with explicit freq argument", {
  set.seed(42)
  freq <- 4L
  L <- 40L
  y_seas <- matrix(rnorm(L * .n_series, mean = 10), nrow = L, ncol = .n_series)
  res_seas <- matrix(rnorm(L * .n_series), nrow = L, ncol = .n_series)

  expect_no_error(
    reconc_t(.A_2b, .mu_2b, y_train = y_seas, residuals = res_seas, freq = freq)
  )
})

test_that("reconc_t standard usage works when y_train is an mts object", {
  set.seed(42)
  freq <- 4L
  L <- 40L
  y_mts <- ts(matrix(rnorm(L * .n_series, mean = 10), nrow = L, ncol = .n_series),
    frequency = freq
  )
  res_seas <- matrix(rnorm(L * .n_series), nrow = L, ncol = .n_series)

  expect_no_error(
    reconc_t(.A_2b, .mu_2b, y_train = y_mts, residuals = res_seas)
  )
})

test_that("reconc_t standard usage gives deterministic results for fixed inputs", {
  result1 <- reconc_t(.A_2b, .mu_2b, y_train = .y_mat, residuals = .res_mat)
  result2 <- reconc_t(.A_2b, .mu_2b, y_train = .y_mat, residuals = .res_mat)

  expect_equal(result1$bottom_rec_mean, result2$bottom_rec_mean)
  expect_equal(result1$bottom_rec_scale_matrix, result2$bottom_rec_scale_matrix)
  expect_equal(result1$bottom_rec_df, result2$bottom_rec_df)
})


# 12. Regression test: hard-coded reference values ----------------------------
# Reference computed with:
#   set.seed(123)
#   n_series <- 3L; L_train <- 50L
#   A_2b <- matrix(c(1,1), nrow=1)
#   y_mat <- matrix(rnorm(L_train * n_series, mean=10), nrow=L_train, ncol=n_series)
#   res_mat <- matrix(rnorm(L_train * n_series), nrow=L_train, ncol=n_series)
#   mu_2b <- c(11, 4, 6)
#   reconc_t(A_2b, mu_2b, y_train=y_mat, residuals=res_mat, return_parameters=TRUE)

test_that("reconc_t standard usage matches hard-coded reference values", {
  set.seed(123)
  n_series <- 3L
  L_train  <- 50L
  A        <- matrix(c(1, 1), nrow = 1)
  y_mat    <- matrix(rnorm(L_train * n_series, mean = 10), nrow = L_train, ncol = n_series)
  res_mat  <- matrix(rnorm(L_train * n_series),            nrow = L_train, ncol = n_series)
  mu       <- c(11, 4, 6)

  result <- suppressWarnings(
    reconc_t(A, mu, y_train = y_mat, residuals = res_mat, return_parameters = TRUE)
  )

  # bottom_rec_mean
  expect_equal(result$bottom_rec_mean,
    c(4.314540, 6.391355),
    tolerance = 1e-5
  )

  # bottom_rec_scale_matrix
  expect_equal(result$bottom_rec_scale_matrix,
    matrix(c(0.6503992, -0.3002090, -0.3002090, 0.5975068), nrow = 2),
    tolerance = 1e-5
  )

  # bottom_rec_df
  expect_equal(result$bottom_rec_df, 54)

  # posterior_nu
  expect_equal(result$posterior_nu, 55)
})

#'
#' @title Analytical reconciliation of Gaussian base forecasts
#'
#' @description
#' Closed form computation of the reconciled forecasts in case of Gaussian base forecasts.
#'
#' @param A aggregation matrix (n_upper x n_bottom).
#' @param base_fc_mean a vector containing the means of the base forecasts.
#' @param base_fc_cov a matrix containing the covariance matrix of the base forecasts.
#' @param residuals a matrix with the residuals of the base forecasts, with n_upper + n_bottom columns.
#' The covariance matrix of the base forecasts is computed from the residuals using the Schäfer Strimmer shrinkage estimator.
#' If base_fc_cov is provided, residuals are ignored.
#' @param return_upper logical, whether to return the reconciled parameters for the upper variables (default is FALSE).
#'
#' @details
#' In the vector of the means of the base forecasts the order must be: first the upper,
#' then the bottom; the order within the uppers is given by the rows of A,
#' the order within the bottoms by the columns of A.
#' The order of the rows of the covariance matrix of the base forecasts is the same.
#'
#' Unless `return_upper = TRUE`, the function returns only the reconciled parameters of the bottom variables.
#'
#' @return A list containing the reconciled forecasts. The list has the following named elements:
#'
#' * `bottom_rec_mean`: reconciled mean for the bottom forecasts;
#' * `bottom_rec_cov`: reconciled covariance for the bottom forecasts;
#' * `upper_rec_mean`: (only if `return_upper = TRUE`) reconciled mean for the upper forecasts;
#' * `upper_rec_cov`: (only if `return_upper = TRUE`) reconciled covariance for the upper forecasts.
#'
#'
#' @examples
#'
#' library(bayesRecon)
#'
#' #' # ---- Example 1: base forecasts are given ----
#' 
#' # Create a minimal hierarchy with 2 bottom and 1 upper variable
#' A <- get_reconc_matrices(agg_levels = c(1, 2), h = 2)$A
#'
#' # Set the parameters of the Gaussian base forecast distributions
#' mu1 <- 2
#' mu2 <- 4
#' muY <- 9
#' base_fc_mean <- c(muY, mu1, mu2)  # vector of means
#'
#' sigma1 <- 2
#' sigma2 <- 2
#' sigmaY <- 3
#' sigmas <- c(sigmaY, sigma1, sigma2)
#' base_fc_cov <- diag(sigmas^2)  # covariance matrix
#' 
#' analytic_rec <- reconc_gaussian(A,
#'   base_fc_mean = base_fc_mean,
#'   base_fc_cov = base_fc_cov
#' )
#'
#' bottom_mean_rec <- analytic_rec$bottom_rec_mean
#' bottom_cov_rec <- analytic_rec$bottom_rec_cov
#'
#' # To obtain reconciled samples for the entire hierarchy, sample from the reconciled 
#' # bottom distribution and then aggregate using A. 
#' 
#' # Sample from the reconciled bottom-level Gaussian distribution
#' # First, compute the Cholesky decomposition of the reconciled covariance matrix:
#' chol_decomp <- chol(bottom_cov_rec)
#' # Then, sample from the standard normal distribution and apply the transformation:
#' Z <- matrix(stats::rnorm(n = 2000), nrow = 2) 
#' B <- t(chol_decomp) %*% Z + matrix(rep(bottom_mean_rec, 1000), nrow = 2) 
#'
#' # Aggregate bottom samples to get upper samples, then stack
#' U <- A %*% B
#' Y_reconc <- rbind(U, B)
#' 
#' cat("Dimensions of reconciled samples (upper + bottom):", dim(Y_reconc), "\n")
#'
#' 
#' # ---- Example 2: using residuals from fitted ETS models ----
#' \donttest{
#' if (requireNamespace("forecast", quietly = TRUE)) {
#'
#'   # Simulate 2 bottom series from AR(1) processes
#'   set.seed(1234)
#'   n_obs <- 200
#'   y1 <- arima.sim(model = list(ar = 0.8), n = n_obs)
#'   y2 <- arima.sim(model = list(ar = 0.5), n = n_obs)
#'
#'   # Upper series is the sum of the two bottom series
#'   y_upper <- y1 + y2
#'
#'   # Aggregation matrix A:
#'   A <- matrix(c(1, 1), nrow = 1)
#'
#'   # Fit additive ETS models
#'   fit1 <- forecast::ets(y1, additive.only = TRUE)
#'   fit2 <- forecast::ets(y2, additive.only = TRUE)
#'   fit_upper <- forecast::ets(y_upper, additive.only = TRUE)
#'
#'   # Point forecasts (h = 1):
#'   fc1 <- forecast::forecast(fit1, h = 1)$mean
#'   fc2 <- forecast::forecast(fit2, h = 1)$mean
#'   fc_upper <- forecast::forecast(fit_upper, h = 1)$mean
#'   base_fc_mean <- c(fc_upper, fc1, fc2)
#'
#'   # Residuals matrix (T x n, columns in same order as base_fc_mean)
#'   res <- cbind(residuals(fit_upper),
#'                residuals(fit1),
#'                residuals(fit2))
#'
#'   # Reconcile (covariance estimated internally via Schafer-Strimmer)
#'   result <- reconc_gaussian(A, base_fc_mean = base_fc_mean, residuals = res, return_upper = TRUE)
#'
#'   bottom_mean <- result$bottom_rec_mean
#'   bottom_cov <- result$bottom_rec_cov
#'   upper_mean <- result$upper_rec_mean
#'   upper_cov <- result$upper_rec_cov
#'
#'   # Print reconciled means
#'   cat("Reconciled bottom means:", round(bottom_mean, 3), "\n")
#'   cat("Reconciled upper mean:", round(upper_mean, 3), "\n")
#' 
#'   # Print 95% predictions intervals
#'   cat("Reconciled bottom 95% prediction intervals:\n")
#'   for (i in 1:length(bottom_mean)) {
#'     lower <- bottom_mean[i] - 1.96 * sqrt(bottom_cov[i, i])
#'     upper <- bottom_mean[i] + 1.96 * sqrt(bottom_cov[i, i])
#'     cat(paste0("Bottom ", i, ": [", round(lower, 3), ", ", round(upper, 3), "]\n"))
#'   }
#'   cat("Reconciled upper 95% prediction interval:\n")
#'   lower <- upper_mean - 1.96 * sqrt(upper_cov[1, 1])
#'   upper <- upper_mean + 1.96 * sqrt(upper_cov[1, 1])
#'   cat(paste0("Upper: [", round(lower, 3), ", ", round(upper, 3), "]\n"))
#' 
#' }
#' }
#'
#' @references
#' Corani, G., Azzimonti, D., Augusto, J.P.S.C., Zaffalon, M. (2021).
#' *Probabilistic Reconciliation of Hierarchical Forecast via Bayes' Rule*.
#' ECML PKDD 2020. Lecture Notes in Computer Science, vol 12459.
#' \doi{10.1007/978-3-030-67664-3_13}.
#'
#' Zambon, L., Agosto, A., Giudici, P., Corani, G. (2024).
#' *Properties of the reconciled distributions for Gaussian and count forecasts*.
#' International Journal of Forecasting (in press).
#' \doi{10.1016/j.ijforecast.2023.12.004}.
#'
#' @seealso [reconc_t()], [reconc_BUIS()]
#'
#' @export
reconc_gaussian <- function(A, base_fc_mean,
                            base_fc_cov = NULL,
                            residuals = NULL,
                            return_upper = FALSE) {
  # Check matrix A
  .check_A(A)
  k <- nrow(A) # number of upper TS
  m <- ncol(A) # number of bottom TS
  n <- length(base_fc_mean) # total number of TS
  if (!(k + m == n)) {
    stop("Input error: the shape of A is not correct")
  }

  # If residuals are not provided, base_fc_cov must be provided
  if (is.null(residuals)) {
    if (is.null(base_fc_cov)) {
      stop("Input error: either residuals or base_fc_cov must be provided")
    }
    if (!(nrow(base_fc_cov) == n)) {
      stop("Input error: nrow(base_fc_cov) != length(base_fc_mean)")
    }
    .check_cov(base_fc_cov, "base_fc_cov", pd_check = FALSE, symm_check = TRUE)
  } else {
    if (!is.null(base_fc_cov)) {
      warning("Input warning: both residuals and base_fc_cov are provided, ignoring residuals")
      .check_cov(base_fc_cov, "base_fc_cov", pd_check = FALSE, symm_check = TRUE)
    } else if (ncol(residuals) != n) {
      stop("Input error: ncol(residuals) != length(base_fc_mean)")
    } else {
      # Compute the covariance matrix of the base forecasts from the residuals
      base_fc_cov <- schaferStrimmer_cov(residuals)$shrink_cov
    }
  }

  Sigma_u <- base_fc_cov[1:k, 1:k]
  Sigma_b <- base_fc_cov[(k + 1):n, (k + 1):n]
  Sigma_ub <- base_fc_cov[1:k, (k + 1):n, drop = FALSE]
  mu_u <- base_fc_mean[1:k]
  mu_b <- base_fc_mean[(k + 1):n]

  # Formulation from:
  # Zambon, L., et al. "Properties of the reconciled distributions for
  # Gaussian and count forecasts." (2023)
  Sigma_ub_At <- tcrossprod(Sigma_ub, A)
  Sigma_b_At <- tcrossprod(Sigma_b, A)
  Q <- Sigma_u - Sigma_ub_At - t(Sigma_ub_At) + A %*% Sigma_b_At
  # we only need to check if Q is p.d.
  .check_cov(Q, "Q", pd_check = TRUE, symm_check = FALSE)
  temp_diff <- t(Sigma_ub) - Sigma_b_At
  K <- t(solve(Q, t(temp_diff))) # equal to temp_diff %*% Q^-1

  mu_b_tilde <- mu_b + K %*% (A %*% mu_b - mu_u)
  Sigma_b_tilde <- Sigma_b - K %*% t(temp_diff)

  out <- list(
    bottom_rec_mean = as.vector(mu_b_tilde),
    bottom_rec_cov = Sigma_b_tilde
  )

  if (return_upper) {
    out$upper_rec_mean <- as.vector(A %*% mu_b_tilde)
    out$upper_rec_cov <- A %*% Sigma_b_tilde %*% t(A)
  }

  return(out)
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = TRUE ### !!!! set to FALSE here to render only the text !!!!
)
set.seed(42)

## ----klippy, echo=FALSE, include=TRUE, eval=FALSE-----------------------------
# klippy::klippy(position = c("top", "right"), tooltip_message = "Copy", tooltip_success = "Done", color = "black")

## ----install, eval=FALSE------------------------------------------------------
# install.packages("bayesRecon", dependencies = TRUE)

## ----load---------------------------------------------------------------------
library(bayesRecon)

## ----carpart-plot, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 1**: Carpart - monthly car part sales.", fig.dim = c(6, 3)----
layout(mat = matrix(c(1, 2), nrow = 1, ncol = 2), widths = c(2, 1))
plot(carparts_example, xlab = "Time", ylab = "Car part sales", main = NULL)
hist(carparts_example, xlab = "Car part sales", main = NULL)

## ----train-test---------------------------------------------------------------
train <- window(carparts_example, end = c(2001, 3))
test <- window(carparts_example, start = c(2001, 4))

## ----temp-agg-----------------------------------------------------------------
train.agg <- bayesRecon::temporal_aggregation(train, agg_levels = c(2, 3, 4, 6, 12))
levels <- c("Annual", "Biannual", "4-Monthly", "Quarterly", "2-Monthly", "Monthly")
names(train.agg) <- levels

## ----temp-agg-plot, dpi=300, fig.show="hold", out.width="100%", out.heigth="100%", fig.align='center', fig.cap="**Figure 2**: The aggregated time series of the temporal hierarchy.", fig.dim=c(6,3.5)----
par(mfrow = c(2, 3), mai = c(0.6, 0.6, 0.5, 0.5))
for (l in levels) {
  plot(train.agg[[l]], xlab = "Time", ylab = "Car part sales", main = l)
}

## ----hier-fore, cache=TRUE----------------------------------------------------
# install.packages("glarma", dependencies = TRUE)
# library(glarma)

fc.samples <- list()
D <- 20000
fc.count <- 1

# iterating over the temporal aggregation levels
for (l in seq_along(train.agg)) {
  f.level <- frequency(train.agg[[l]])
  print(paste("Forecasting at ", levels[l], "...", sep = ""))
  # fit an independent model for each aggregation level
  model <- glarma::glarma(
    train.agg[[l]],
    phiLags = if (f.level == 1) 1 else 1:(min(6, f.level - 1)),
    thetaLags = if (f.level == 1) NULL else f.level,
    X = cbind(intercept = rep(1, length(train.agg[[l]]))),
    offset = cbind(intercept = rep(0, length(train.agg[[l]]))),
    type = "Poi"
  )
  # forecast 1 year ahead
  h <- f.level
  tmp <- matrix(data = NA, nrow = h, ncol = D)
  for (s in (1:D)) {
    # each call to 'forecast.glarma' returns a simulation path
    tmp[, s] <- glarma::forecast(
      model,
      n.ahead = h,
      newdata = cbind(intercept = rep(1, h)),
      newoffset = rep(0, h)
    )$Y
  }
  # collect the forecasted samples
  for (i in 1:h) {
    fc.samples[[fc.count]] <- tmp[i, ]
    fc.count <- fc.count + 1
  }
}

## ----aggregationMatrix--------------------------------------------------------
recon.matrices <- bayesRecon::get_reconc_matrices(agg_levels = c(2, 3, 4, 6, 12), h = 12)
# Aggregation matrix
A <- recon.matrices$A

## ----reconc-------------------------------------------------------------------
recon.res <- bayesRecon::reconc_BUIS(
  A,
  base_fc = fc.samples,
  in_type = "samples",
  distr = "discrete",
  seed = 42
)

## ----res----------------------------------------------------------------------
reconciled_samples <- recon.res$bottom_rec_samples
dim(reconciled_samples)

## ----metrics------------------------------------------------------------------
# install.packages("scoringRules", dependencies = TRUE)
library(scoringRules)

ae.fc <- list()
ae.reconc <- list()
crps.fc <- list()
crps.reconc <- list()
for (h in 1:length(test)) {
  y.hat_ <- median(fc.samples[[nrow(A) + h]])
  y.reconc_ <- median(recon.res$bottom_rec_samples[h, ])
  # Compute Absolute Errors
  ae.fc[[h]] <- abs(test[h] - y.hat_)
  ae.reconc[[h]] <- abs(test[h] - y.reconc_)
  # Compute Continuous Ranked Probability Score (CRPS)
  crps.fc[[h]] <-
    scoringRules::crps_sample(y = test[h], dat = fc.samples[[nrow(A) + h]])
  crps.reconc[[h]] <-
    scoringRules::crps_sample(y = test[h], dat = recon.res$bottom_rec_samples[h, ])
}

mae.fc <- mean(unlist(ae.fc))
mae.reconc <- mean(unlist(ae.reconc))
crps.fc <- mean(unlist(crps.fc))
crps.reconc <- mean(unlist(crps.reconc))
metrics <- data.frame(
  row.names = c("MAE", "CRPS"),
  base.forecasts = c(mae.fc, crps.fc),
  reconciled.forecasts = c(mae.reconc, crps.reconc)
)
metrics

## ----m3-plot, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 3**: M3 - N1485 time series.", fig.dim = c(6, 3)----
plot(M3_example$train, xlab = "Time", ylab = "y", main = "N1485")

## ----m3-agg-------------------------------------------------------------------
train.agg <- bayesRecon::temporal_aggregation(M3_example$train)
levels <- c("Annual", "Biannual", "4-Monthly", "Quarterly", "2-Monthly", "Monthly")
names(train.agg) <- levels

## ----m3-fore------------------------------------------------------------------
# install.packages("forecast", dependencies = TRUE)
library(forecast)

H <- length(M3_example$test)
H

fc <- list()
level.idx <- 1
fc.idx <- 1
for (level in train.agg) {
  level.name <- names(train.agg)[level.idx]
  # fit an ETS model for each temporal level
  model <- ets(level)
  # generate forecasts for each level within 18 months
  h <- floor(H / (12 / frequency(level)))
  print(paste("Forecasting at ", level.name, ", h=", h, "...", sep = ""))
  level.fc <- forecast(model, h = h)
  # save mean and sd of the gaussian predictive distribution
  for (i in 1:h) {
    fc[[fc.idx]] <- list(
      mean = level.fc$mean[[i]],
      sd = (level.fc$upper[, "95%"][[i]] - level.fc$mean[[i]]) / qnorm(0.975)
    )
    fc.idx <- fc.idx + 1
  }
  level.idx <- level.idx + 1
}

## ----m3-rmat, dpi=300, out.width = '70%', fig.align='center', fig.cap="**Figure 4**: M3 - The aggregation matrix A (red=1, yellow=0).", fig.dim = c(8, 8)----
rmat <- get_reconc_matrices(agg_levels = c(2, 3, 4, 6, 12), h = 18)

par(mai = c(1, 1, 0.5, 0.5))
image(1:ncol(rmat$A), 1:nrow(rmat$A),
  t(apply(t(rmat$A), 1, rev)),
  xaxt = "n", yaxt = "n", ylab = "", xlab = levels[6]
)
axis(1, at = 1:ncol(rmat$A), label = 1:ncol(rmat$A), las = 1)
axis(2, at = c(23, 22, 19, 15, 9), label = levels[1:5], las = 2)

## ----m3-reco------------------------------------------------------------------
recon.gauss <- bayesRecon::reconc_gaussian(
  A = rmat$A,
  base_fc_mean = sapply(fc, "[[", 1),
  base_fc_cov = diag(sapply(fc, "[[", 2))^2
)

reconc.buis <- bayesRecon::reconc_BUIS(
  A = rmat$A,
  base_fc = fc,
  in_type = "params",
  distr = "gaussian",
  num_samples = 20000,
  seed = 42
)

# check that the algorithms return consistent results
round(rbind(
  c(rmat$S %*% recon.gauss$bottom_rec_mean),
  rowMeans(rbind(reconc.buis$upper_rec_samples, reconc.buis$bottom_rec_samples))
))

## ----m3-plotfore, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 5**: M3 - Base and reconciled forecasts. The black line shows the actual data (dashed in the test). The orange line is the  mean of the base forecasts, the blu line is the reconciled mean. We also show the 95% prediction intervals.", fig.dim = c(6, 4)----
yhat.mu <- tail(sapply(fc, "[[", 1), 18)
yhat.sigma <- tail(sapply(fc, "[[", 2), 18)
yhat.hi95 <- qnorm(0.975, mean = yhat.mu, sd = yhat.sigma)
yhat.lo95 <- qnorm(0.025, mean = yhat.mu, sd = yhat.sigma)
yreconc.mu <- rowMeans(reconc.buis$bottom_rec_samples)
yreconc.hi95 <- apply(
  reconc.buis$bottom_rec_samples, 1,
  function(x) quantile(x, 0.975)
)
yreconc.lo95 <- apply(
  reconc.buis$bottom_rec_samples, 1,
  function(x) quantile(x, 0.025)
)

ylim_ <- c(
  min(M3_example$train, M3_example$test, yhat.lo95, yreconc.lo95) - 1,
  max(M3_example$train, M3_example$test, yhat.hi95, yreconc.hi95) + 1
)

plot(M3_example$train,
  xlim = c(1990, 1995.6), ylim = ylim_,
  ylab = "y", main = "N1485 Forecasts"
)
lines(M3_example$test, lty = "dashed")
lines(ts(yhat.mu, start = start(M3_example$test), frequency = 12),
  col = "coral", lwd = 2
)
lines(ts(yreconc.mu, start = start(M3_example$test), frequency = 12),
  col = "blue2", lwd = 2
)
xtest <- time(M3_example$test)
polygon(c(xtest, rev(xtest)), c(yhat.mu, rev(yhat.hi95)),
  col = "#FF7F5066", border = "#FF7F5066"
)
polygon(c(xtest, rev(xtest)), c(yhat.mu, rev(yhat.lo95)),
  col = "#FF7F5066", border = "#FF7F5066"
)
polygon(c(xtest, rev(xtest)), c(yreconc.mu, rev(yreconc.hi95)),
  col = "#0000EE4D", border = "#0000EE4D"
)
polygon(c(xtest, rev(xtest)), c(yreconc.mu, rev(yreconc.lo95)),
  col = "#0000EE4D", border = "#0000EE4D"
)

## ----infants-forecasts--------------------------------------------------------
# install.packages("forecast", dependencies = TRUE)
library(forecast)

fc <- list()
residuals <- matrix(NA,
  nrow = length(infantMortality$total),
  ncol = length(infantMortality)
)
fc.idx <- 1
for (s in infantMortality) {
  s.name <- names(infantMortality)[fc.idx]
  print(paste("Forecasting at ", s.name, "...", sep = ""))
  # fit an auto.arima model and forecast with h=1
  model <- auto.arima(s)
  s.fc <- forecast(model, h = 1)
  # save mean and sd of the gaussian predictive distribution
  fc[[s.name]] <- c(
    s.fc$mean,
    (s.fc$upper[, "95%"][[1]] - s.fc$mean) / qnorm(0.975)
  )
  residuals[, fc.idx] <- s.fc$residuals
  fc.idx <- fc.idx + 1
}

## ----infants-s, dpi=300, out.width = '70%', fig.align='center', fig.cap="**Figure 6**: Infants mortality - The aggregation matrix A (red=1, yellow=0).", fig.dim = c(8, 8)----
# we have 16 bottom time series, and 11 upper time series
A <- matrix(data = c(
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
  1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
  0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1
), byrow = TRUE, ncol = 16)

# plot of A
par(mai = c(1.5, 1, 0.5, 0.5))
image(1:ncol(A), 1:nrow(A),
  t(apply(t(A), 1, rev)),
  xaxt = "n", yaxt = "n", ann = FALSE
)
axis(1, at = 1:ncol(A), label = names(infantMortality)[12:27], las = 2)
axis(2, at = c(1:11), label = rev(names(infantMortality)[1:11]), las = 2)

## ----infants reconc-----------------------------------------------------------
# means
mu <- sapply(fc, "[[", 1)
# Shrinkage covariance
shrink.res <- bayesRecon::schaferStrimmer_cov(residuals)
print(paste("The estimated shrinkage intensity is", round(shrink.res$lambda_star, 3)))
Sigma <- shrink.res$shrink_cov

## ----infants-recon------------------------------------------------------------
recon.gauss <- bayesRecon::reconc_gaussian(A,
  base_fc_mean = mu,
  base_fc_cov = Sigma
)

bottom_mu_reconc <- recon.gauss$bottom_rec_mean
bottom_Sigma_reconc <- recon.gauss$bottom_rec_cov

# Obtain reconciled mu and Sigma for the upper variable
upper_mu_reconc <- A %*% bottom_mu_reconc
upper_Sigma_reconc <- A %*% bottom_Sigma_reconc %*% t(A)

upper_mu_reconc

###############################################################################
# Reconciliation with mixed-conditioning (Mix-Cond)
###############################################################################


#' @title Probabilistic forecast reconciliation of mixed hierarchies 
#'
#' @description
#'
#' `reconc_MixCond()` uses importance sampling to draw samples from the reconciled
#' forecast distribution, obtained via conditioning, in the case of a mixed hierarchy.
#'
#' `reconc_TDcond()` uses a top-down conditioning algorithm: first, upper base forecasts are
#' reconciled via conditioning using only the hierarchical constraints between the
#' upper; then, the bottom distributions are updated via a probabilistic top-down procedure.
#'
#' @details
#'
#' The base bottom forecasts `base_fc_bottom` must be a list of length n_bottom, where each element is either
#' * a PMF object (see details below), if `bottom_in_type='pmf'`;
#' * a vector of samples, if `bottom_in_type='samples'`;
#' * a list of parameters, if `bottom_in_type='params'`:
#'    * lambda for the Poisson base forecast if `distr`='poisson', see \link[stats]{Poisson};
#'    * size and prob (or mu) for the negative binomial base forecast if `distr`='nbinom',
#'      see \link[stats]{NegBinomial}.
#'
#' The base upper forecasts `base_fc_upper` must be a list containing the parameters of
#' the multivariate Gaussian distribution of the upper forecasts.
#' The list must contain only the named elements `mean` (vector of length n_upper)
#' and `cov` (n_upper x n_upper matrix).
#'
#' The order of the upper and bottom base forecasts must match the order of (respectively) the rows and the columns of A.
#'
#' A PMF object is a numerical vector containing the probability mass function of a discrete distribution.
#' Each element corresponds to the probability of the integers from 0 to the last value of the support.
#' See also [PMF] for functions that handle PMF objects.
#'
#' \strong{Warnings and errors.}
#'
#' In `reconc_MixCond`, warnings are triggered from the importance sampling step if:
#' * weights are all zeros, then the upper forecast is ignored during reconciliation;
#' * the effective sample size is < 200;
#' * the effective sample size is < 1% of the sample size.
#' 
#' These warnings are an indication that the base forecasts might have issues.
#' Please check the base forecasts in case of warnings.
#' 
#' In `reconc_TDcond`, if some of the reconciled upper samples lie outside the support of the bottom-up
#' distribution, those samples are discarded; the remaining ones are resampled with
#' replacement, so that the number of output samples is equal to `num_samples`.
#' In this case, a warning is issued if `suppress_warnings=FALSE` (default is `TRUE`).
#' If the fraction of discarded samples is above 50%, the function returns an error.
#'
#' @param A Aggregation matrix (n_upper x n_bottom).
#' @param base_fc_bottom A list containing the bottom base forecasts, see details.
#' @param base_fc_upper A list containing the upper base forecasts, see details.
#' @param bottom_in_type A string with three possible values:
#'
#' * 'pmf' if the bottom base forecasts are in the form of pmf, see details;
#' * 'samples' if the bottom base forecasts are in the form of samples;
#' * 'params'  if the bottom base forecasts are in the form of estimated parameters.
#'
#' @param distr A string describing the type of bottom base forecasts ('poisson' or 'nbinom').
#'
#' This is only used if `bottom_in_type='params'`.
#'
#' @param num_samples Number of samples drawn from the reconciled distribution.
#'        This is ignored if `bottom_in_type='samples'`; in this case, the number of
#'        reconciled samples is equal to the number of samples of the base forecasts.
#'
#' @param return_type The return type of the reconciled distributions.
#'        A string with three possible values:
#'
#' * 'pmf' returns a list containing the reconciled marginal pmf objects;
#' * 'samples' returns a list containing the reconciled multivariate samples;
#' * 'all' returns a list with both pmf objects and samples.
#'
#' @param return_upper Logical, whether to return the reconciled parameters for the upper variables (default is TRUE).
#'
#' @param suppress_warnings Logical. If \code{TRUE}, no warnings about samples are triggered;
#'        if \code{FALSE}, warnings are generated. Default is \code{FALSE} for `reconc_MixCond`
#'        and \code{TRUE} for `reconc_TDcond`. See the respective sections above.
#' 
#' @param seed Seed for reproducibility.
#'
#' @return A list containing the reconciled forecasts. The list has the following named elements:
#'
#' * `bottom_rec_pmf`: a list of PMF objects for each bottom series
#'    (only if `return_type` is `'pmf'` or `'all'`);
#' * `bottom_rec_samples`: a matrix (n_bottom x `num_samples`) of reconciled bottom samples
#'    (only if `return_type` is `'samples'` or `'all'`);
#' * `upper_rec_pmf`: a list of PMF objects for each upper series
#'    (only if `return_type` is `'pmf'` or `'all'`, and `return_upper = TRUE`);
#' * `upper_rec_samples`: a matrix (n_upper x `num_samples`) of reconciled upper samples
#'    (only if `return_type` is `'samples'` or `'all'`, and `return_upper = TRUE`).
#'
#' @references
#' Zambon, L., Azzimonti, D., Rubattu, N., Corani, G. (2024).
#' *Probabilistic reconciliation of mixed-type hierarchical time series*.
#' Proceedings of the Fortieth Conference on Uncertainty in Artificial Intelligence,
#' PMLR 244:4078-4095. <https://proceedings.mlr.press/v244/zambon24a.html>.
#'
#' @seealso [reconc_BUIS()], [reconc_gaussian()], [PMF]
#'
#' @name Mixed_reconciliation
NULL

#' @rdname Mixed_reconciliation
#' @examples
#'
#' library(bayesRecon)
#'
#' # Consider a simple hierarchy with two bottom and one upper
#' A <- matrix(c(1, 1), nrow = 1)
#' # The bottom forecasts are Poisson with lambda=15
#' lambda <- 15
#' n_tot <- 60
#' base_fc_bottom <- list()
#' base_fc_bottom[[1]] <- apply(matrix(seq(0, n_tot)), MARGIN = 1,
#'                              FUN = \(x) dpois(x, lambda = lambda))
#' base_fc_bottom[[2]] <- apply(matrix(seq(0, n_tot)), MARGIN = 1,
#'                              FUN = \(x) dpois(x, lambda = lambda))
#'
#' # The upper forecast is a Normal with mean 40 and std 5
#' base_fc_upper <- list(mean = 40, cov = matrix(5^2))
#'
#' # Reconcile with reconc_MixCond
#' res.mixCond <- reconc_MixCond(A, base_fc_bottom, base_fc_upper)
#'
#' # Note that the bottom distributions are slightly shifted to the right
#' PMF_summary(res.mixCond$bottom_rec_pmf[[1]])
#' PMF_summary(base_fc_bottom[[1]])
#'
#' PMF_summary(res.mixCond$bottom_rec_pmf[[2]])
#' PMF_summary(base_fc_bottom[[2]])
#'
#' # The upper distribution is slightly shifted to the left
#' PMF_summary(res.mixCond$upper_rec_pmf[[1]])
#' PMF_get_var(res.mixCond$upper_rec_pmf[[1]])
#'
#' @export
reconc_MixCond <- function(A, base_fc_bottom, base_fc_upper,
                           bottom_in_type = "pmf", distr = NULL,
                           num_samples = 2e4, return_type = "pmf",
                           return_upper = TRUE, suppress_warnings = FALSE, 
                           seed = NULL) {
  if (!is.null(seed)) set.seed(seed)

  # Check inputs
  .check_input_TD(
    A, base_fc_bottom, base_fc_upper,
    bottom_in_type, distr,
    return_type
  )

  # Prepare samples from the base bottom distribution
  if (bottom_in_type == "pmf") {
    B <- lapply(base_fc_bottom, PMF_sample, N_samples = num_samples)
    B <- do.call("cbind", B) # matrix of bottom samples (N_samples x n_bottom)
  } else if (bottom_in_type == "samples") {
    B <- do.call("cbind", base_fc_bottom)
    num_samples <- nrow(B)
  } else if (bottom_in_type == "params") {
    L_pmf <- lapply(base_fc_bottom, PMF_from_params, distr = distr)
    B <- lapply(L_pmf, PMF_sample, N_samples = num_samples)
    B <- do.call("cbind", B) # matrix of bottom samples (N_samples x n_bottom)
  }

  # Get mean and covariance matrix of the MVN upper base forecasts
  mean_upper <- base_fc_upper$mean
  cov_upper <- as.matrix(base_fc_upper$cov)

  out <- .core_reconc_MixCond(
    A, B, mean_upper, cov_upper, num_samples,
    return_type = return_type, 
    return_ESS = FALSE,
    return_upper = return_upper,
    suppress_warnings = suppress_warnings
  )

  return(out)
}


#' Core Reconciliation via Importance Sampling for Mixed Hierarchies
#'
#' Internal function that performs the core reconciliation logic using importance sampling
#' (IS) to reconcile mixed-type hierarchies. The base bottom forecasts (provided as samples)
#' are reweighted according to their fit to the upper multivariate Gaussian forecasts.
#'
#' @param A Matrix (n_upper x n_bottom) defining the hierarchy where upper = A %*% bottom.
#' @param B Matrix (n_samples x n_bottom) of bottom base forecast samples to be reconciled.
#' @param mean_upper Vector of upper level means.
#' @param cov_upper Covariance matrix of upper level.
#' @param num_samples Number of samples to draw/resample from.
#' @param return_type Character string specifying return format: 'pmf', 'samples', or 'all'.
#' @param return_ESS Logical, whether to return the Effective Sample Size (ESS) from importance sampling weights (default TRUE).
#' @param return_upper Logical, whether to return the reconciled parameters for the upper variables (default TRUE).
#' @param suppress_warnings Logical. If TRUE, suppresses warnings about sample quality (default FALSE).

#' @return A list containing:
#'   \itemize{
#'     \item `bottom_rec`: List with reconciled bottom forecasts (pmf and/or samples).
#'     \item `bottom_rec_pmf`: list of PMF objects for each bottom series (only if `return_type` is 'pmf' or 'all').
#'     \item `bottom_rec_samples`: matrix (n_bottom x num_samples) of reconciled bottom samples (only if `return_type` is 'samples' or 'all').
#'     \item `upper_rec_pmf`: list of PMF objects for each upper series (only if `return_type` is 'pmf' or 'all', and `return_upper = TRUE`).
#'     \item `upper_rec_samples`: matrix (n_upper x num_samples) of reconciled upper samples (only if `return_type` is 'samples' or 'all', and `return_upper = TRUE`).
#'     \item `ESS`: Effective Sample Size resulting from importance sampling reweighting (only if `return_ESS = TRUE`).
#'   }
#'
#' @keywords internal
#' @export
.core_reconc_MixCond <- function(A, B, mean_upper, cov_upper, num_samples, return_type, 
                                 return_ESS = TRUE, return_upper = TRUE,
                                 suppress_warnings = FALSE) {
  # Get dimensions
  n_u <- nrow(A)
  n_b <- ncol(A)

  # IS using MVN
  U <- B %*% t(A)
  weights <- .MVN_density(x = U, mu = mean_upper, Sigma = cov_upper)


  check_weights_res <- .check_weights(weights)
  if (check_weights_res$warning & !suppress_warnings) {
    warning_msg <- check_weights_res$warning_msg
    warning(warning_msg)
  }
  if (!(check_weights_res$warning & (1 %in% check_weights_res$warning_code))) {
    B <- .resample(B, weights, num_samples)
  }

  B <- t(B)
  U <- A %*% B

  # Prepare output: include the marginal pmfs and/or the samples (depending on "return" inputs)
  out <- list()
  if (return_type %in% c("pmf", "all")) {
    out$bottom_rec_pmf <- lapply(1:n_b, function(i) PMF_from_samples(B[i, ]))
    if (return_upper) {
      out$upper_rec_pmf <- lapply(1:n_u, function(i) PMF_from_samples(U[i, ]))
    }
  }
  if (return_type %in% c("samples", "all")) {
    out$bottom_rec_samples <- B
    if (return_upper) {
      out$upper_rec_samples <- U
    }
  }
  
  if (return_ESS) {
    out$ESS <- sum(weights)**2 / sum(weights**2)
  }

  return(out)
}

---
title: "Reconciliation of M5 hierarchy with mixed-type forecasts"
author: "Dario Azzimonti, Lorenzo Zambon, Nicolò Rubattu, Giorgio Corani"
date: "2024-05-29"
lang: "en"
bibliography: references.bib
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Reconciliation of M5 hierarchy with mixed-type forecasts}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(bayesRecon)
```

# Introduction

This vignette partially reproduces the results of *Probabilistic reconciliation of mixed-type hierarchical time series* [@zambon2024mixed], 
published at UAI 2024 (the 40th Conference on Uncertainty in Artificial Intelligence).

In particular, we replicate the reconciliation of the one-step ahead (h=1) forecasts of one store of the  M5 competition [@MAKRIDAKIS20221325]. 
Sect. 5 of the paper presents the results for 10  stores, each reconciled
14 times using rolling one-step ahead forecasts. 


# Data and base forecasts

The M5 competition [@MAKRIDAKIS20221325] is about daily time series of sales data referring to 10 different stores. 
Each store has the same hierarchy: 3049 bottom time series (single items) and 11 upper time series,  obtained by aggregating the items by department, product category, and store; see the figure below.

```{r M5hier, fig.cap="**Figure 1**: graph of the M5 hierarchy.", out.width = '100%', echo = FALSE}
knitr::include_graphics("img/M5store_hier.png")
```

We reproduce  the results of the store "CA_1". The  base forecasts (for h=1)  of the bottom and upper time series are stored in `M5_CA1_basefc`, available as data in the package.
The base forecast are computed  using ADAM [@svetunkov2023iets], implemented in the R package smooth [@smooth_pkg]. 


```{r InitializeHierarchy}
# Hierarchy composed by 3060 time series: 3049 bottom and 11 upper
n_b <- 3049
n_u <- 11
n <- n_b + n_u

# Load matrix A
A <- M5_CA1_basefc$A

# Load base forecasts:
base_fc_upper <- M5_CA1_basefc$upper
base_fc_bottom <- M5_CA1_basefc$bottom

# We will save all the results in the list rec_fc
rec_fc <- list(
  Gauss      = list(),
  Mixed_cond = list(),
  TD_cond    = list()
)
```

# Gaussian reconciliation

We first perform Gaussian reconciliation (`Gauss`, @corani2021probabilistic). 
It assumes  all forecasts to be Gaussian, even though the bottom base forecasts are not Gaussian. 

We assume the upper base forecasts to be a multivariate Gaussian and we estimate their  covariance matrix from the in-sample residuals. We assume also the bottom base forecasts to be independent Gaussians.


```{r readBaseForecasts}
# Parameters of the upper base forecast distributions
mu_u <- unlist(lapply(base_fc_upper, "[[", "mu")) # upper means
# Compute the (shrinked) covariance matrix of the residuals
residuals.upper <- lapply(base_fc_upper, "[[", "residuals")
residuals.upper <- t(do.call("rbind", residuals.upper))
Sigma_u <- schaferStrimmer_cov(residuals.upper)$shrink_cov

# Parameters of the bottom base forecast distributions
mu_b <- c()
sd_b <- c()
for (fc_b in base_fc_bottom) {
  pmf <- fc_b$pmf
  mu_b <- c(mu_b, PMF_get_mean(pmf))
  sd_b <- c(sd_b, PMF_get_var(pmf)**0.5)
}
Sigma_b <- diag(sd_b**2)

# Mean and covariance matrix of the base forecasts
base_fc_mean <- c(mu_u, mu_b)
base_fc_cov <- matrix(0, nrow = n, ncol = n)
base_fc_cov[1:n_u, 1:n_u] <- Sigma_u
base_fc_cov[(n_u + 1):n, (n_u + 1):n] <- Sigma_b
```

We reconcile using the function `reconc_gaussian()`, which takes as input: 

* the summing matrix `A`;
* the means of the  base forecast,  `base_fc_mean`; 
* the covariance of the base forecast,  `base_fc_cov`. 

The function returns the reconciled mean and covariance for the bottom time series. 

```{r GaussianReconciliation}    
# Gaussian reconciliation
start <- Sys.time()
gauss <- reconc_gaussian(A, base_fc_mean, base_fc_cov)
stop <- Sys.time()

rec_fc$Gauss <- list(
  mu_b = gauss$bottom_rec_mean,
  Sigma_b = gauss$bottom_rec_cov,
  mu_u = A %*% gauss$bottom_rec_mean,
  Sigma_u = A %*% gauss$bottom_rec_cov %*% t(A)
)

Gauss_time <- as.double(round(difftime(stop, start, units = "secs"), 2))
cat("Time taken by Gaussian reconciliation: ", Gauss_time, "s")
```


# Reconciliation with mixed-conditioning 

We now reconcile the forecasts using the mixed-conditioning 
approach of @zambon2024mixed, Sect. 3. 
The algorithm is implemented in the function `reconc_MixCond()`. The function takes as input: 

* the aggregation matrix `A`;
* the probability mass functions of the bottom base forecasts, stored in the list `fc_bottom_4rec`;
* the parameters of the multivariate Gaussian distribution for the upper variables, `fc_upper_4rec` (a list with `mean` and `cov` entries);
* additional function parameters; among those note that `num_samples` specifies the number of samples used in the internal importance sampling (IS) algorithm. 

The function returns the reconciled forecasts in the form of probability mass functions for both the upper and bottom time series. The function parameter `return_type` can be changed to `samples` or `all` to obtain the IS samples. 

```{r MixedCondReconciliation}
seed <- 1
N_samples_IS <- 5e4

# Base forecasts
fc_upper_4rec <- list(mean = mu_u, cov = Sigma_u)
fc_bottom_4rec <- lapply(base_fc_bottom, "[[", "pmf") # list of PMFs

# MixCond reconciliation
start <- Sys.time()
mix_cond <- reconc_MixCond(A, fc_bottom_4rec, fc_upper_4rec,
  bottom_in_type = "pmf",
  num_samples = N_samples_IS, return_type = "pmf", seed = seed
)
stop <- Sys.time()

rec_fc$Mixed_cond <- list(
  bottom = mix_cond$bottom_rec_pmf,
  upper  = mix_cond$upper_rec_pmf,
  ESS    = mix_cond$ESS
)

MixCond_time <- as.double(round(difftime(stop, start, units = "secs"), 2))
cat("Computational time for Mix-cond reconciliation: ", MixCond_time, "s")
```

As discussed in @zambon2024mixed, Sect. 3, conditioning with mixed variables performs poorly in high dimensions.
This is because the bottom-up distribution, built by assuming the bottom forecasts to be independent, is untenable
in high dimensions.
Moreover, forecasts for count time series are usually biased and their sum tends to be strongly biased; see @zambon2024mixed, Fig. 3, for a graphical example. 

# Top down conditioning
Top down conditioning (TD-cond; see @zambon2024mixed, Sect. 4) is a more reliable approach for reconciling  mixed variables in high dimensions.
The algorithm is implemented in the function `reconc_TDcond()`; it takes the same arguments as `reconc_MixCond()` and returns reconciled forecasts in the same format. 

```{r TDcondReconciliation}
N_samples_TD <- 1e4

# TDcond reconciliation
start <- Sys.time()
td <- reconc_TDcond(A, fc_bottom_4rec, fc_upper_4rec,
  bottom_in_type = "pmf", num_samples = N_samples_TD,
  return_type = "pmf", seed = seed
)
stop <- Sys.time()
```
The algorithm TD-cond raises a warning regarding the incoherence between the joint bottom-up and the upper base forecasts. 
We will see that this warning does not impact the performances of TD-cond.

```{r print computational time TD cond}
rec_fc$TD_cond <- list(
  bottom = td$bottom_rec_pmf,
  upper  = td$upper_rec_pmf
)

TDCond_time <- as.double(round(difftime(stop, start, units = "secs"), 2))
cat("Computational time for TD-cond reconciliation: ", TDCond_time, "s")
```


 
# Comparison

The computational time required for the Gaussian reconciliation is `r Gauss_time` seconds, Mix-cond requires `r MixCond_time` seconds and TD-cond requires `r TDCond_time` seconds. 

For each time series in the hierarchy, we compute the following scores for each method:

- MASE: Mean Absolute Scaled Error

- MIS: Mean Interval Score

- RPS: Ranked Probability Score

```{r InitializeMetrics}
# Parameters for computing the scores
alpha <- 0.1 # MIS uses 90% coverage intervals
jitt <- 1e-9 # jitter for numerical stability

# Save actual values
actuals_u <- unlist(lapply(base_fc_upper, "[[", "actual"))
actuals_b <- unlist(lapply(base_fc_bottom, "[[", "actual"))
actuals <- c(actuals_u, actuals_b)

# Scaling factor for computing MASE
Q_u <- M5_CA1_basefc$Q_u
Q_b <- M5_CA1_basefc$Q_b
Q <- c(Q_u, Q_b)

# Initialize lists to save the results
mase <- list()
mis <- list()
rps <- list()
```

The following functions are used for computing the scores:

* `AE_pmf`: compute the absolute error for a PMF;
* `MIS_pmf`: compute interval score for a PMF;
* `RPS_pmf`: compute RPS for a PMF;
* `MIS_gauss`: compute MIS for a Gaussian distribution. 

The implementation of these functions is available in the source code of the vignette but not shown here. 

```{r metricFunctions,include=FALSE}
# Functions for computing the scores of a PMF
AE_pmf <- function(pmf, actual) {
  return(abs(PMF_get_quantile(pmf, p = 0.5) - actual))
}
MIS_pmf <- function(pmf, actual, alpha) {
  u <- PMF_get_quantile(pmf, p = 1 - (alpha / 2))
  l <- PMF_get_quantile(pmf, p = alpha / 2)
  return(u - l + (2 / alpha) * (l - actual) * (actual < l) +
    (2 / alpha) * (actual - u) * (actual > u))
}
RPS_pmf <- function(pmf, actual) {
  cdf <- cumsum(pmf) / sum(pmf)
  M <- length(cdf)
  # if actual is outside the supp of pmf, add ones to the end of the cdf:
  if (actual >= M) {
    cdf <- c(cdf, rep(1, (actual - M + 1)))
    M <- length(cdf)
  }
  cdf_act <- (0:(M - 1)) >= actual # unit step function in actual
  crps_ <- sum((cdf - cdf_act)**2)
  return(crps_)
}

# Function for computing the MIS of (possibly truncated) Gaussian forecasts
MIS_gauss <- function(mus, sds, actuals, alpha, trunc = FALSE) {
  z <- qnorm(1 - (alpha / 2))
  u <- mus + z * sds
  l <- mus - z * sds
  # If it is truncated, set negative quantiles to zero
  if (trunc) {
    l[l < 0] <- 0
    u[u < 0] <- 0
  }
  return(u - l + (2 / alpha) * (l - actuals) * (actuals < l) +
    (2 / alpha) * (actuals - u) * (actuals > u))
}
```




```{r computeScores}
# Compute scores for the base forecasts
# Upper
mu_u <- unlist(lapply(base_fc_upper, "[[", "mu"))
sd_u <- unlist(lapply(base_fc_upper, "[[", "sigma"))
mase$base[1:n_u] <- abs(mu_u - actuals_u) / Q_u
mis$base[1:n_u] <- MIS_gauss(mu_u, sd_u, actuals_u, alpha)
rps$base[1:n_u] <- scoringRules::crps(actuals_u, "norm", mean = mu_u, sd = sd_u)
# Bottom
pmfs <- lapply(base_fc_bottom, "[[", "pmf")
mase$base[(n_u + 1):n] <- mapply(AE_pmf, pmfs, actuals_b) / Q_b
mis$base[(n_u + 1):n] <- mapply(MIS_pmf, pmfs, actuals_b, MoreArgs = list(alpha = alpha))
rps$base[(n_u + 1):n] <- mapply(RPS_pmf, pmfs, actuals_b)

# Compute scores for Gauss reconciliation
mu <- c(rec_fc$Gauss$mu_u, rec_fc$Gauss$mu_b)
sd <- c(diag(rec_fc$Gauss$Sigma_u), diag(rec_fc$Gauss$Sigma_b))**0.5
sd <- sd + jitt
mase$Gauss <- abs(mu - actuals) / Q
mis$Gauss <- MIS_gauss(mu, sd, actuals, alpha)
rps$Gauss <- scoringRules::crps(actuals, "norm", mean = mu, sd = sd)

# Compute scores for Mix-cond reconciliation
pmfs <- c(rec_fc$Mixed_cond$upper, rec_fc$Mixed_cond$bottom)
mase$MixCond <- mapply(AE_pmf, pmfs, actuals) / Q
mis$MixCond <- mapply(MIS_pmf, pmfs, actuals, MoreArgs = list(alpha = alpha))
rps$MixCond <- mapply(RPS_pmf, pmfs, actuals)

# Compute scores for TD-cond reconciliation
pmfs <- c(rec_fc$TD_cond$upper, rec_fc$TD_cond$bottom)
mase$TDcond <- mapply(AE_pmf, pmfs, actuals) / Q
mis$TDcond <- mapply(MIS_pmf, pmfs, actuals, MoreArgs = list(alpha = alpha))
rps$TDcond <- mapply(RPS_pmf, pmfs, actuals)
```


## Skill scores
We report the improvement over the base forecasts using the skill score values and averaging them across experiments.
For instance, the skill score of Gauss on RPS is:

$$ \text{Skill}_{\%}\,(\text{RPS, }Gauss) = 100 \cdot
\frac{\text{RPS}(base) - \text{RPS}(Gauss)}
{(\text{RPS}(base) + \text{RPS}(Gauss))/2}$$

This formula is implemented in the function `skill.score`, available in the source code of the vignette but not shown here.

```{r skillScoreFunction,include=FALSE}
# Function for computing the skill score
skill.score <- function(ref, met) {
  s <- (2 * (ref - met) / (ref + met)) * 100
  s[is.na(s)] <- 0 # if both numerator and denominator are 0, set skill score to 0
  return(s)
}
```


```{r ComputeScores}
scores <- list(
  mase = mase,
  mis = mis,
  rps = rps
)
scores_ <- names(scores)

ref_met <- "base"
methods_ <- c("Gauss", "MixCond", "TDcond")

# For each score and method we compute the skill score with respect to the base forecasts
skill_scores <- list()
for (s in scores_) {
  skill_scores[[s]] <- list()
  for (met in methods_) {
    skill_scores[[s]][["upper"]][[met]] <- skill.score(
      scores[[s]][[ref_met]][1:n_u],
      scores[[s]][[met]][1:n_u]
    )
    skill_scores[[s]][["bottom"]][[met]] <- skill.score(
      scores[[s]][[ref_met]][(n_u + 1):n],
      scores[[s]][[met]][(n_u + 1):n]
    )
  }
}
```

We report in the tables below the mean values for each skill score. 


```{r AverageScores}
mean_skill_scores <- list()

for (s in scores_) {
  mean_skill_scores[[s]] <- rbind(
    data.frame(lapply(skill_scores[[s]][["upper"]], mean)),
    data.frame(lapply(skill_scores[[s]][["bottom"]], mean))
  )
  rownames(mean_skill_scores[[s]]) <- c("upper", "bottom")
}
```

```{r printMASETable}
knitr::kable(mean_skill_scores$mase, digits = 2, caption = "Mean skill score on MASE.", align = "lccc")
```
The mean MASE skill score is positive only for the TD-cond reconciliation. Both Mix-cond and Gauss achieve scores lower than the base forecasts, even if Mix-cond degrades less the base forecasts compared to Gauss. 

```{r PrintMIStable}
knitr::kable(mean_skill_scores$mis, digits = 2, caption = "Mean skill score on MIS.")
```
The mean MIS score of TD-cond is slightly above that of the base forecasts. Mix-cond achieves slightly higher scores than the base forecasts only on the bottom variables. Gauss strongly degrades the base forecasts according to this metric.

```{r printRPStable}
knitr::kable(mean_skill_scores$rps, digits = 2, caption = "Mean skill score on RPS.")
```
The mean RPS skill score for TD-cond is positive for both upper and bottom time series. Mix-cond slightly improves the base forecasts on the bottom variables, however it degrades the upper base forecasts. Gauss strongly degrades both upper and bottom base forecasts. 

## Boxplots

Finally, we show the boxplots of the skill scores for each method divided in upper and bottom levels. 

```{r MASEboxplots, fig.cap="**Figure 2**: boxplot of MASE skill scores for upper and bottom time series.", fig.width=7,fig.height=8}
custom_colors <- c(
  "#a8a8e4",
  "#a9c7e4",
  "#aae4df"
)

# Boxplots of MASE skill scores
par(mfrow = c(2, 1))
boxplot(skill_scores$mase$upper,
  main = "MASE upper time series",
  col = custom_colors, ylim = c(-80, 80)
)
abline(h = 0, lty = 3)
boxplot(skill_scores$mase$bottom,
  main = "MASE bottom time series",
  col = custom_colors, ylim = c(-200, 200)
)
abline(h = 0, lty = 3)
```
```{r setupParams,eval=TRUE,include=FALSE}
par(mfrow = c(1, 1))
```

Both Mix-cond and TD-cond do not improve the bottom MASE over the base forecasts (boxplot flattened on the value zero), however TD-cond provides a slight improvement over the upper base forecasts (boxplot over the zero line).

```{r MISboxplots, fig.cap="**Figure 3**: boxplot of MIS skill scores for upper and bottom time series.", fig.width=7,fig.height=8}
# Boxplots of MIS skill scores
par(mfrow = c(2, 1))
boxplot(skill_scores$mis$upper,
  main = "MIS upper time series",
  col = custom_colors, ylim = c(-150, 150)
)
abline(h = 0, lty = 3)
boxplot(skill_scores$mis$bottom,
  main = "MIS bottom time series",
  col = custom_colors, ylim = c(-200, 200)
)
abline(h = 0, lty = 3)
```
```{r,eval=TRUE,include=FALSE}
par(mfrow = c(1, 1))
```

Both Mix-cond and TD-cond do not improve nor degrade the bottom base forecasts in MIS score as shown by the small boxplots centered around zero. On the upper variables instead only TD-cond does not degrade the MIS score of the base forecasts. 


```{r RPSboxplots,fig.cap="**Figure 4**: boxplot of RPS skill scores for upper and bottom time series.", fig.width=7,fig.height=8}
# Boxplots of RPS skill scores
par(mfrow = c(2, 1))
boxplot(skill_scores$rps$upper,
  main = "RPS upper time series",
  col = custom_colors, ylim = c(-80, 80)
)
abline(h = 0, lty = 3)
boxplot(skill_scores$rps$bottom,
  main = "RPS bottom time series",
  col = custom_colors, ylim = c(-200, 200)
)
abline(h = 0, lty = 3)
```
```{r,eval=TRUE,include=FALSE}
par(mfrow = c(1, 1))
```

According to RPS, TD-cond does not degrade the bottom base forecasts and improves the upper base forecasts. On the other hand both Gauss and Mix-cond strongly degrade the upper base forecasts. 


# Full reproducibility

The full experiment described in [@zambon2024mixed] can be reproduced by using the code available [here](https://github.com/LorenzoZambon/M5_MixedReconc).

# References
<div id="refs"></div>

---
title: "Properties of the reconciled distribution via conditioning"
author: "Lorenzo Zambon, Giorgio Corani"
date: "2023-12-19"
lang: "en"
output: rmarkdown::html_vignette
bibliography: references.bib
vignette: >
  %\VignetteIndexEntry{Properties of the reconciled distribution via conditioning}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```


# Introduction

This vignette reproduces the results of the paper *Properties of the reconciled distributions for Gaussian and count forecasts* [@zambon2024properties], 
accepted for publication in the International Journal of Forecasting.
We  replicate here the main experiment of the paper (see sec. 3, @zambon2024properties)  where we  reconcile  base forecasts constituted by  negative binomial distributions. 

We use the R package `BayesRecon`.

```{r setup}
library(bayesRecon)
```

# Data and base forecasts

We release a new data set, containing time series of counts of extreme market events 
in five economic sectors in the period 2005-2018 (3508 trading days). 
The counts are computed by considering 29 companies included in the Euro Stoxx 50 index 
and observing if the value of the CDS spread on a given day exceeds the 90-th percentile of its distribution in the last trading year.
The companies are divided into the following  sectors: Financial (FIN), Information and Communication Technology (ICT), 
Manufacturing (MFG), Energy (ENG), and Trade (TRD). 

The hierarchy is composed of 5 bottom time series, the daily number of extreme market  
events in each sector, and 1 upper time series (the sum of the different sectors). 
Data are stored in `extr_mkt_events`.

```{r finTShier, fig.cap="**Figure 1**: financial time series hierarchy.", out.width = '80%', echo = FALSE}
knitr::include_graphics("img/finTS_hier.jpg")
```

We also make available as `extr_mkt_events_basefc` the base forecasts to be reconciled. 
As discussed in the paper, they are  produced
using the model by [@agosto2022multivariate];
the predictive distributions
are  negative binomial. 

```{r}
# Hierarchy composed by 6 time series: 5 bottom and 1 upper
n_b <- 5
n_u <- 1
n <- n_b + n_u

A <- matrix(1, ncol = n_b, nrow = n_u) # aggregation matrix

# Actual values:
actuals <- data.frame(extr_mkt_events) # convert to data frame
# Base forecasts:
base_fc <- extr_mkt_events_basefc

N <- nrow(actuals) # number of days (3508)

# # If you want to run only N reconciliations (instead of 3508):
# N <- 200
# actuals <- actuals[1:N,]
# base_fc$mu <- base_fc$mu[1:N,]
```

# Reconciliation via conditioning

We reconcile the base forecasts via conditioning, using importance sampling.
We use the `reconc_BUIS` function, which implements the BUIS algorithm [@zambon2022efficient];
since there is only one upper time series in the hierarchy, the BUIS algorithm is
equivalent to importance sampling.
We perform 3508 reconciliations, one for each day, drawing each time 10,000 samples from the reconciled distribution.
We use 10,000 samples instead of 100,000 (as in the paper) to speed up the computation.

For each day, we save the empirical mean, median, and quantiles of the reconciled distribution.

```{r}
# We need to save the mean and median of the reconciled distribution
# in order to compute the squared error and the absolute error:
rec_means <- matrix(NA, ncol = n, nrow = N)
rec_medians <- matrix(NA, ncol = n, nrow = N)

# We need to save the lower and upper quantiles of the reconciled distribution
# in order to compute the interval score:
rec_L <- matrix(NA, ncol = n, nrow = N)
rec_U <- matrix(NA, ncol = n, nrow = N)
int_cov <- 0.9 # use 90% interval
q1 <- (1 - int_cov) / 2
q2 <- (1 + int_cov) / 2

# Set the number of samples to draw from the reconciled distribution:
N_samples <- 1e4

start <- Sys.time()
for (j in 1:N) {
  # Base forecasts:
  base_fc_j <- c()
  for (i in 1:n) {
    base_fc_j[[i]] <- list(size = base_fc$size[[i]], mu = base_fc$mu[[j, i]])
  }

  # Reconcile via importance sampling:
  buis <- reconc_BUIS(A, base_fc_j, "params", "nbinom",
    num_samples = N_samples, seed = 42
  )
  samples_y <- rbind(buis$upper_rec_samples, buis$bottom_rec_samples)

  # Save mean, median, and lower and upper quantiles:
  rec_means[j, ] <- rowMeans(samples_y)
  rec_medians[j, ] <- apply(samples_y, 1, median)
  rec_L[j, ] <- apply(samples_y, 1, quantile, q1)
  rec_U[j, ] <- apply(samples_y, 1, quantile, q2)
}
stop <- Sys.time()
cat(
  "Computational time for ", N, " reconciliations: ",
  round(difftime(stop, start, units = "secs"), 2), "s"
)
```

We compute the median and the quantiles of the negative binomial base forecasts.

```{r}
base_means <- base_fc$mu
base_medians <- matrix(NA, ncol = n, nrow = N)
base_L <- matrix(NA, ncol = n, nrow = N)
base_U <- matrix(NA, ncol = n, nrow = N)

for (i in 1:n) {
  base_medians[, i] <- sapply(
    base_means[, i],
    function(mu) qnbinom(p = 0.5, size = base_fc$size[[i]], mu = mu)
  )
  base_L[, i] <- sapply(
    base_means[, i],
    function(mu) qnbinom(p = q1, size = base_fc$size[[i]], mu = mu)
  )
  base_U[, i] <- sapply(
    base_means[, i],
    function(mu) qnbinom(p = q2, size = base_fc$size[[i]], mu = mu)
  )
}
```

For each day and for each time series, we compute the absolute error, the squared error, 
and the interval score for the base and reconciled forecasts.

```{r}
# Compute the squared errors
SE_base <- (base_means - actuals)^2
SE_rec <- (rec_means - actuals)^2

# Compute the absolute errors
AE_base <- abs(base_medians - actuals)
AE_rec <- abs(rec_medians - actuals)

# Define the function for the interval score
int_score <- function(l, u, actual, int_cov = 0.9) {
  is <- (u - l) +
    2 / (1 - int_cov) * (actual - u) * (actual > u) +
    2 / (1 - int_cov) * (l - actual) * (l > actual)
  return(is)
}

# Compute the interval scores
IS_base <- mapply(int_score, base_L, base_U, data.matrix(actuals))
IS_base <- matrix(IS_base, nrow = N)
IS_rec <- mapply(int_score, rec_L, rec_U, data.matrix(actuals))
IS_rec <- matrix(IS_rec, nrow = N)
```

We compute and show the skill scores, which measure the improvement of the 
reconciled forecasts over the base forecasts.
The skill score is symmetric and bounded between -2 and 2.

```{r}
SS_AE <- (AE_base - AE_rec) / (AE_base + AE_rec) * 2
SS_SE <- (SE_base - SE_rec) / (SE_base + SE_rec) * 2
SS_IS <- (IS_base - IS_rec) / (IS_base + IS_rec) * 2
SS_AE[is.na(SS_AE)] <- 0
SS_SE[is.na(SS_SE)] <- 0
SS_IS[is.na(SS_IS)] <- 0

mean_skill_scores <- c(
  round(colMeans(SS_IS), 2),
  round(colMeans(SS_SE), 2),
  round(colMeans(SS_AE), 2)
)
mean_skill_scores <- data.frame(t(matrix(mean_skill_scores, nrow = n)))
colnames(mean_skill_scores) <- names(actuals)
rownames(mean_skill_scores) <- c("Interval score", "Squared error", "Absolute error")
knitr::kable(mean_skill_scores)
```

The table closely matches Table 4 of the paper.
In order to exactly reproduce  the paper table, it is necessary increase the number of samples  drawn from the reconciled distribution to 100,000.

# Reconciled mean and variance

We now show the effects of the reconciliation on the mean and variance of the
forecast distribution. For more details, we refer to Section 3.2 of the paper.

We observe two different behaviors for the reconciled upper mean:
it can be between the base and the bottom-up mean (*combination* effect)
or it can be lower than both (*concordant-shift* effect).
We show them for two different days.

```{r}
# Now we draw a larger number of samples:
N_samples <- 1e5

### Example of concordant-shift effect
j <- 124
base_fc_j <- c()
for (i in 1:n) {
  base_fc_j[[i]] <- list(
    size = extr_mkt_events_basefc$size[[i]],
    mu = extr_mkt_events_basefc$mu[[j, i]]
  )
}
# Reconcile
buis <- reconc_BUIS(A, base_fc_j, "params", "nbinom",
  num_samples = N_samples, seed = 42
)
samples_y <- rbind(buis$upper_rec_samples, buis$bottom_rec_samples)

# The reconciled mean is lower than both the base and bottom-up means:
means <- c(
  round(extr_mkt_events_basefc$mu[[j, 1]], 2),
  round(sum(extr_mkt_events_basefc$mu[j, 2:n]), 2),
  round(mean(samples_y[1, ]), 2)
)
col_names <- c("Base upper mean", "Bottom-up upper mean", "Reconciled upper mean")
knitr::kable(matrix(means, nrow = 1), col.names = col_names)

### Example of combination effect
j <- 1700
base_fc_j <- c()
for (i in 1:n) {
  base_fc_j[[i]] <- list(
    size = extr_mkt_events_basefc$size[[i]],
    mu = extr_mkt_events_basefc$mu[[j, i]]
  )
}
# Reconcile
buis <- reconc_BUIS(A, base_fc_j, "params", "nbinom",
  num_samples = N_samples, seed = 42
)
samples_y <- rbind(buis$upper_rec_samples, buis$bottom_rec_samples)

# The reconciled mean is between the base and the bottom-up mean:
means <- c(
  round(extr_mkt_events_basefc$mu[[j, 1]], 2),
  round(sum(extr_mkt_events_basefc$mu[j, 2:n]), 2),
  round(mean(samples_y[1, ]), 2)
)
col_names <- c("Base upper mean", "Bottom-up upper mean", "Reconciled upper mean")
knitr::kable(matrix(means, nrow = 1), col.names = col_names)
```

Finally, we show an example in which the variance of the bottom time series increases
after reconciliation. 
This is a major difference with the Gaussian reconciliation, for which the reconciled 
variance is always smaller than the base variance.

```{r}
j <- 2308
base_fc_j <- c()
for (i in 1:n) {
  base_fc_j[[i]] <- list(
    size = extr_mkt_events_basefc$size[[i]],
    mu = extr_mkt_events_basefc$mu[[j, i]]
  )
}
# Reconcile
buis <- reconc_BUIS(A, base_fc_j, "params", "nbinom", num_samples = N_samples, seed = 42)
samples_y <- rbind(buis$upper_rec_samples, buis$bottom_rec_samples)

# Compute the variance of the base and reconciled bottom forecasts
base_bottom_var <- mapply(
  function(mu, size) var(rnbinom(n = 1e5, size = size, mu = mu)),
  extr_mkt_events_basefc$mu[j, 2:n],
  extr_mkt_events_basefc$size[2:n]
)
rec_bottom_var <- apply(samples_y[2:n, ], MARGIN = 1, var)

# The reconciled variance is greater than the base variance:
bottom_var <- rbind(base_bottom_var, rec_bottom_var)
rownames(bottom_var) <- c("var base", "var reconc")
knitr::kable(round(bottom_var, 2))
```

# References
<div id="refs"></div>

---
title: "Probabilistic Reconciliation via Conditioning with `bayesRecon`"
author: "Nicolò Rubattu, Giorgio Corani, Dario Azzimonti, Lorenzo Zambon"
date: "2023-11-26"
lang: "en"
output: html_vignette
bibliography: references.bib
cite:
- '@zambon2022efficient'
- '@zambon2023properties'
- '@corani2023probabilistic'
- '@corani2021probabilistic'
vignette: >
  %\VignetteIndexEntry{Probabilistic Reconciliation via Conditioning with `bayesRecon`}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval=TRUE ### !!!! set to FALSE here to render only the text !!!!
)
set.seed(42)
```

```{r klippy, echo=FALSE, include=TRUE, eval=FALSE}
klippy::klippy(position = c('top', 'right'), tooltip_message = 'Copy', tooltip_success = 'Done', color="black")
```

# Introduction

This vignette shows how to perform  *probabilistic reconciliation* with
the `bayesRecon` package. We provide three examples:

1. *Temporal hierarchy for a count time series*: we build a temporal hierarchy over a count time series,  produce the base forecasts using `glarma` and  reconcile them via Bottom-Up Importance Sampling (BUIS).

2.  *Temporal hierarchy for a smooth time series*: we build a temporal hierarchy over a smooth time series,  compute the base forecasts using `ets` and  we reconcile them in closed form using Gaussian reconciliation. The covariance matrix is diagonal.

3.   *Hierarchical of smooth time series*: this is an example of a cross-sectional hierarchy. We generate the base forecasts using `ets` and we reconcile them via Gaussian reconciliation.
The covariance matrix is full and estimated via shrinkage.



# Installation

The package, available at this 
[CRAN page](https://cran.r-project.org/package=bayesRecon), can be installed and loaded with the usual commands:
```{r install, eval=FALSE}
install.packages('bayesRecon', dependencies = TRUE)
```

Load the package:

```{r load}
library(bayesRecon)
```

# Temporal hierarchy over a count time series
We select a monthly time series of counts from  the *carparts* dataset, available from
the expsmooth package [@expsmooth_pkg].
The data set contains time series of sales of cars part from Jan. 1998 to Mar. 2002. 
For this example we select time series #2655, which we make  available  as `bayesRecon::carparts_example`.

This time series  has a skewed distribution of  values. 
```{r carpart-plot, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 1**: Carpart - monthly car part sales.", fig.dim = c(6, 3)}
layout(mat = matrix(c(1, 2), nrow = 1, ncol = 2), widths = c(2, 1))
plot(carparts_example, xlab = "Time", ylab = "Car part sales", main = NULL)
hist(carparts_example, xlab = "Car part sales", main = NULL)
```
<br><br>
We divide the time series into train and test; the test set contains  the last 12 months.
```{r train-test}
train <- window(carparts_example, end = c(2001, 3))
test  <- window(carparts_example, start = c(2001, 4))
```

<!-- In order to do hierarchical forecasting: -->

<!-- 1.    we build a temporal hierarchy; -->
<!-- 2.    we compute the *base forecasts*) at each temporal scale. -->

We build the temporal hierarchy using the `temporal aggregation` function.
We specify the aggregation levels using the 
`agg_levels` argument; in this case they are
*2-Monthly*, *Quarterly*, *4-Monthly*, *Biannual*, and *Annual*.

``` {r temp-agg}
train.agg <- bayesRecon::temporal_aggregation(train, agg_levels = c(2, 3, 4, 6, 12))
levels <- c("Annual", "Biannual", "4-Monthly", "Quarterly", "2-Monthly", "Monthly")
names(train.agg) <- levels
```

The function returns a list of aggregated time series, ordered from the most aggregated (top of the hierarchy) to the most disagreggated (bottom of the hierarchy). We  plot them below.
``` {r temp-agg-plot, dpi=300, fig.show="hold", out.width="100%", out.heigth="100%", fig.align='center', fig.cap="**Figure 2**: The aggregated time series of the temporal hierarchy.", fig.dim=c(6,3.5)}
par(mfrow = c(2, 3), mai = c(0.6, 0.6, 0.5, 0.5))
for (l in levels) {
  plot(train.agg[[l]], xlab = "Time", ylab = "Car part sales", main = l)
}
```
<br><br>
We  compute the  *base forecasts* using the package  [`glarma`](https://cran.r-project.org/package=glarma),
a package specific for forecasting count time series.
We forecast 12 steps ahead at the monthly level,  4 steps ahead  at the quarterly level, etc. by iterating over the levels of the  hierarchy,
At each level, we fit a `glarma` model
with Poisson predictive distribution and we forecast;
each forecast is constituted by 20000  integer samples.


Eventually we collect the samples of the  28 predictive distributions (one at the  *Annual* level, two at the *Biannual* level, etc.) in a list.
The  code below takes about 30 seconds on a standard computer.

``` {r hier-fore, cache=TRUE}
# install.packages("glarma", dependencies = TRUE)
#library(glarma)

fc.samples <- list()
D <- 20000
fc.count <- 1

# iterating over the temporal aggregation levels
for (l in seq_along(train.agg)) {
  f.level <- frequency(train.agg[[l]])
  print(paste("Forecasting at ", levels[l], "...", sep = ""))
  # fit an independent model for each aggregation level
  model <- glarma::glarma(
    train.agg[[l]],
    phiLags = if (f.level == 1) 1 else 1:(min(6, f.level - 1)),
    thetaLags = if (f.level == 1) NULL else f.level,
    X = cbind(intercept = rep(1, length(train.agg[[l]]))),
    offset = cbind(intercept = rep(0, length(train.agg[[l]]))),
    type = "Poi"
  )
  # forecast 1 year ahead
  h <- f.level
  tmp <- matrix(data = NA, nrow = h, ncol = D)
  for (s in (1:D)) {
    # each call to 'forecast.glarma' returns a simulation path
    tmp[, s] <- glarma::forecast(
      model,
      n.ahead = h,
      newdata = cbind(intercept = rep(1, h)),
      newoffset = rep(0, h)
    )$Y
  }
  # collect the forecasted samples
  for (i in 1:h) {
    fc.samples[[fc.count]] <- tmp[i, ]
    fc.count <- fc.count + 1
  }
}

```


Reconciliation requires the aggregation matrix $\mathbf{A}$, which we obtain using the function `get_reconc_matrices`. 
It requires: 

*   the aggregation factors of the hierarchy, which in this example are $\{2, 3, 4, 6, 12\}$;
*   the length of the forecasting horizon at the  bottom level, which is 12 in this example.

``` {r aggregationMatrix}
recon.matrices <- bayesRecon::get_reconc_matrices(agg_levels = c(2, 3, 4, 6, 12), h = 12)
# Aggregation matrix
A <- recon.matrices$A 
```

To reconcile using Bottom-Up Important Sampling (BUIS) we
we use the function `reconc_BUIS`, passing to it
the $\mathbf{A}$ matrix, the *base forecasts*, the  type  of the base forecasts (`in_type`="samples") and whether the samples are discrete or integer (`distr` = "discrete").


``` {r reconc}
recon.res <- bayesRecon::reconc_BUIS(
  A,
  base_forecasts = fc.samples,
  in_type = "samples",
  distr = "discrete",
  seed = 42
)
```

Here we obtain samples from the reconciled forecast distribution.
```{r res}
reconciled_samples <- recon.res$reconciled_samples
dim(reconciled_samples)

```

We now compute the Mean Absolute Error (MAE) and the Continuous Ranked Probability Score (CRPS) for the bottom (i.e., *monthly*) time series.  For computing CRPS, we use the package [`scoringRules`](https://cran.r-project.org/package=scoringRules).

```{r metrics}
# install.packages("scoringRules", dependencies = TRUE)
library(scoringRules)

ae.fc <- list()
ae.reconc <- list()
crps.fc <- list()
crps.reconc <- list()
for (h in 1:length(test)) {
  y.hat_ <- median(fc.samples[[nrow(A) + h]])
  y.reconc_ <- median(recon.res$bottom_reconciled_samples[, h])
  # Compute Absolute Errors
  ae.fc[[h]] <- abs(test[h] - y.hat_)
  ae.reconc[[h]] <- abs(test[h] - y.reconc_)
  # Compute Continuous Ranked Probability Score (CRPS)
  crps.fc[[h]] <-
    scoringRules::crps_sample(y = test[h], dat = fc.samples[[nrow(A) + h]])
  crps.reconc[[h]] <-
    scoringRules::crps_sample(y = test[h], dat = recon.res$bottom_reconciled_samples[, h])
}

mae.fc <- mean(unlist(ae.fc))
mae.reconc <- mean(unlist(ae.reconc))
crps.fc <- mean(unlist(crps.fc))
crps.reconc <- mean(unlist(crps.reconc))
metrics <- data.frame(
  row.names = c("MAE", "CRPS"),
  base.forecasts = c(mae.fc, crps.fc),
  reconciled.forecasts = c(mae.reconc, crps.reconc)
)
metrics

```


# Temporal hierarchy over a smooth time series

In this second example, we select a smooth monthly time series (N1485) from the M3 forecasting competition [@makridakis2000m3]. The data set is available in
the Mcomp package [@mcomp_pkg] and we make available this time series as `bayesRecon::m3_example`.


```{r m3-plot, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 3**: M3 - N1485 time series.", fig.dim = c(6, 3)}
plot(M3_example$train, xlab = "Time", ylab = "y", main = "N1485")
```
<br>
We build the temporal hierarchy using the `temporal_aggregation` function. 

Without specifying  `agg_levels`, the function  generates by default all the feasible aggregation: 2-Monthly, Quarterly, 4-Monthly, Biannual, and Annual.

```{r m3-agg} 
train.agg <- bayesRecon::temporal_aggregation(M3_example$train)
levels <- c("Annual", "Biannual", "4-Monthly", "Quarterly", "2-Monthly", "Monthly")
names(train.agg) <- levels
```


We generate the base forecasts using `ets` from the   [forecast](https://cran.r-project.org/package=forecast) package [@pkg_forecast].
At every level we predict 18 months ahead.
All the  predictive distributions are Gaussian.
```{r m3-fore}
# install.packages("forecast", dependencies = TRUE)
library(forecast)

H <- length(M3_example$test)
H

fc <- list()
level.idx <- 1
fc.idx <- 1
for (level in train.agg) {
  level.name <- names(train.agg)[level.idx]
  # fit an ETS model for each temporal level
  model <- ets(level)
  # generate forecasts for each level within 18 months
  h <- floor(H / (12 / frequency(level)))
  print(paste("Forecasting at ", level.name, ", h=", h, "...", sep = ""))
  level.fc <- forecast(model, h = h)
  # save mean and sd of the gaussian predictive distribution
  for (i in 1:h) {
    fc[[fc.idx]] <- list(mean = level.fc$mean[[i]],
                         sd = (level.fc$upper[, "95%"][[i]] - level.fc$mean[[i]]) / qnorm(0.975))
    fc.idx <- fc.idx + 1
  }
  level.idx <- level.idx + 1
}
```

Using the function `get_reconc_matrices`,  we get  matrix $\mathbf{A}$.

```{r m3-rmat, dpi=300, out.width = '70%', fig.align='center', fig.cap="**Figure 4**: M3 - The aggregation matrix A (red=1, yellow=0).", fig.dim = c(8, 8)}
rmat <- get_reconc_matrices(agg_levels = c(2, 3, 4, 6, 12), h = 18)

par(mai = c(1,1,0.5,0.5))
image(1:ncol(rmat$A), 1:nrow(rmat$A), 
      t(apply(t(rmat$A),1,rev)), 
      xaxt='n', yaxt='n', ylab = "", xlab = levels[6])
axis(1, at=1:ncol(rmat$A), label=1:ncol(rmat$A), las=1)
axis(2, at=c(23,22,19,15,9), label=levels[1:5], las=2)
```
<br>
The function `reconc_gaussian`  implements the exact Gaussian reconciliation.
We also run `reconc_BUIS`, to check the consistency 
between the two approaches.


```{r m3-reco} 
recon.gauss <- bayesRecon::reconc_gaussian(
  A = rmat$A,
  base_forecasts.mu = sapply(fc, "[[", 1),
  base_forecasts.Sigma = diag(sapply(fc, "[[", 2)) ^ 2
)

reconc.buis <- bayesRecon::reconc_BUIS(
  A = rmat$A,
  base_forecasts = fc,
  in_type = "params",
  distr = "gaussian",
  num_samples = 20000,
  seed = 42
)

# check that the algorithms return consistent results
round(rbind(
  c(rmat$S %*% recon.gauss$bottom_reconciled_mean),
  rowMeans(reconc.buis$reconciled_samples)
))
```

We now compare  *base forecasts* and  *reconciled forecasts*:

```{r m3-plotfore, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 5**: M3 - Base and reconciled forecasts. The black line shows the actual data (dashed in the test). The orange line is the  mean of the base forecasts, the blu line is the reconciled mean. We also show the 95% prediction intervals.", fig.dim = c(6, 4)}
yhat.mu <- tail(sapply(fc, "[[", 1), 18)
yhat.sigma <- tail(sapply(fc, "[[", 2), 18)
yhat.hi95 <- qnorm(0.975, mean = yhat.mu, sd = yhat.sigma)
yhat.lo95 <- qnorm(0.025, mean = yhat.mu, sd = yhat.sigma)
yreconc.mu <- rowMeans(reconc.buis$bottom_reconciled_samples)
yreconc.hi95 <- apply(reconc.buis$bottom_reconciled_samples, 1, 
                      function(x) quantile(x, 0.975))
yreconc.lo95 <- apply(reconc.buis$bottom_reconciled_samples, 1, 
                      function(x) quantile(x, 0.025))

ylim_ <- c(min(M3_example$train, M3_example$test, yhat.lo95, yreconc.lo95) - 1, 
           max(M3_example$train, M3_example$test, yhat.hi95, yreconc.hi95) + 1)

plot(M3_example$train, xlim = c(1990, 1995.6), ylim = ylim_, 
     ylab = "y", main = "N1485 Forecasts")
lines(M3_example$test, lty = "dashed")
lines(ts(yhat.mu, start = start(M3_example$test), frequency = 12), 
      col = "coral", lwd = 2)
lines(ts(yreconc.mu, start = start(M3_example$test), frequency = 12), 
      col = "blue2", lwd = 2)
xtest <- time(M3_example$test)
polygon(c(xtest, rev(xtest)), c(yhat.mu, rev(yhat.hi95)), 
        col = "#FF7F5066", border = "#FF7F5066")
polygon(c(xtest, rev(xtest)), c(yhat.mu, rev(yhat.lo95)), 
        col = "#FF7F5066", border = "#FF7F5066")
polygon(c(xtest, rev(xtest)), c(yreconc.mu, rev(yreconc.hi95)), 
        col = "#0000EE4D", border = "#0000EE4D")
polygon(c(xtest, rev(xtest)), c(yreconc.mu, rev(yreconc.lo95)), 
        col = "#0000EE4D", border = "#0000EE4D")
```



#   Gaussian reconciliation of a cross-sectional hierarchy

In this example, we consider the hierarchical time series *infantgts*, which is available 
from the `hts` package [@hts_pkg]
We make it available also in our package as `bayesRecon::infantMortality`.

It contains counts of infant mortality (deaths) in Australia, disaggregated by state and sex (male and female). 

<!-- State codes refer to: New South Wales (NSW), Victoria (VIC), Queensland (QLD), South Australia (SA), Western Australia (WA), Northern Territory (NT), Australian Capital Territory (ACT), and Tasmania (TAS). -->

We forecast one year ahead using `auto.arima`  from the [`forecast` ](https://cran.r-project.org/package=forecast) package. 
We collect the residuals, which we will later use to compute the covariance matrix. 

```{r infants-forecasts}
# install.packages("forecast", dependencies = TRUE)
library(forecast)

fc <- list()
residuals <- matrix(NA,
                    nrow = length(infantMortality$total),
                    ncol = length(infantMortality))
fc.idx <- 1
for (s in infantMortality) {
  s.name <- names(infantMortality)[fc.idx]
  print(paste("Forecasting at ", s.name, "...", sep = ""))
  # fit an auto.arima model and forecast with h=1
  model <- auto.arima(s)
  s.fc <- forecast(model, h = 1)
  # save mean and sd of the gaussian predictive distribution
  fc[[s.name]] <- c(s.fc$mean,
                    (s.fc$upper[, "95%"][[1]] - s.fc$mean) / qnorm(0.975))
  residuals[, fc.idx] <- s.fc$residuals
  fc.idx <- fc.idx + 1
}
```

Now we build the $\mathbf{A}$ matrix.

```{r infants-s, dpi=300, out.width = '70%', fig.align='center', fig.cap="**Figure 6**: Infants mortality - The aggregation matrix A (red=1, yellow=0).", fig.dim = c(8, 8)}
# we have 16 bottom time series, and 11 upper time series
A <- matrix(data = c(1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
                     1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                     0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,
                     0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,
                     0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,
                     0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,
                     0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,
                     0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,
                     0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,
                     1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,
                     0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1), byrow=TRUE, ncol = 16)

# plot of A
par(mai = c(1.5,1,0.5,0.5))
image(1:ncol(A), 1:nrow(A), 
      t(apply(t(A),1,rev)), 
      xaxt='n', yaxt='n', ann=FALSE)
axis(1, at=1:ncol(A), label=names(infantMortality)[12:27], las=2)
axis(2, at=c(1:11), label=rev(names(infantMortality)[1:11]), las=2)
```
<br>
We use  `bayesRecon::schaferStrimmer_cov` to estimate the covariance matrix of the residuals with shrinkage [@schafer2005shrinkage].

```{r infants reconc}
# means
mu <- sapply(fc, "[[", 1)
# Shrinkage covariance
shrink.res <- bayesRecon::schaferStrimmer_cov(residuals)
print(paste("The estimated shrinkage intensity is", round(shrink.res$lambda_star, 3)))
Sigma <- shrink.res$shrink_cov
```

We now perform Gaussian reconciliation:

```{r infants-recon}
recon.gauss <- bayesRecon::reconc_gaussian(A,
                                           base_forecasts.mu = mu,
                                           base_forecasts.Sigma = Sigma)

bottom_mu_reconc <- recon.gauss$bottom_reconciled_mean
bottom_Sigma_reconc <- recon.gauss$bottom_reconciled_covariance

# Obtain reconciled mu and Sigma for the upper variable
upper_mu_reconc <- A %*% bottom_mu_reconc
upper_Sigma_reconc <- A %*% bottom_Sigma_reconc %*% t(A)

upper_mu_reconc
```

# References
<div id="refs"></div>

---
title: "t-Rec: Reconciliation of Gaussian forecasts by modelling the uncertainty on the covariance matrix"
author: "Dario Azzimonti, Chiara Carrara, Lorenzo Zambon, Giorgio Corani"
date: ""
lang: "en"
output: rmarkdown::html_vignette
bibliography: references.bib
cite:
- '@zambon2024properties'
- '@carrara2025'
vignette: >
  %\VignetteIndexEntry{t-Rec: Reconciliation of Gaussian forecasts by modelling the uncertainty on the covariance matrix}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{css, echo=FALSE}
p.caption {
  text-align: left;
}
caption {
  text-align: left;
}
```

# Introduction

This vignette showcases the method *t-Rec*, a Bayesian approach to forecast reconciliation that explicitly accounts for uncertainty in the covariance matrix of residuals. 

By exploiting the flexibility of Bayesian models to incorporate parameter uncertainty, t-Rec assumes an Inverse-Wishart prior on the covariance matrix. This choice allows the reconciliation to be derived in closed form, resulting in a reconciled predictive distribution that follows a multivariate t. See @carrara2025 for a detailed explanation.


```{r load, message=FALSE, warning=FALSE}
# load the packages
library(bayesRecon)
library(forecast) # base forecasts
library(ggplot2)  # plots
```

# Loading the data

We consider the monthly Swiss overnight tourist stays data divided by Canton and aggregated over the whole country. This is a cross-sectional hierarchical structure, consisting of two levels: the national total and the division by canton. 

The data is available in this package and can be loaded as `bayesRecon::swiss_tourism`. This is a list that contains the `mts` object, the aggregation matrix and the number of upper and bottom time series. The raw data is also publicly available on the official website of the Swiss Confederation at this [link](https://www.bfs.admin.ch/asset/en/px-x-1003020000_102).

The data set spans the period from January 2005 to January 2025, comprising 241 monthly observations.

The time series exhibit strong seasonality. See, in **Figure 1**, the top-level (aggregate) time series.

```{r swiss-tourism-plot, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 1**: Swiss tourism - monthly Swiss overnight stays.", fig.dim = c(8, 4)}
# save all time series 
Y = swiss_tourism$ts

# plot the first (top) time series
autoplot(Y[,1], ylab = "Overnight stays in Switzerland",linewidth=0.9)+
  scale_y_continuous(labels = function(x) paste0(formatC(x / 1e6, format = "g"), "M"))
```

We load the aggregation matrix and select the length of the training set. The length of the training set can be selected within the range [26, 240] observations.

```{r Swiss tourism}
# Save aggregation matrix
A = swiss_tourism$agg_mat

# Number of bottom and upper time series
n_b = ncol(A)
n_u = nrow(A)
n = n_b + n_u

# Frequency is monthly:
print(frequency(Y))

# Select the length of the training set and the forecast horizon
L = 60    
h = 1    

# Select the training set and the actuals for the forecast horizon
train = window(Y, end = time(Y)[L])
actuals = window(Y, start= time(Y)[L + 1], end = time(Y)[L + h])
```

# Forecasts

The base forecasts are generated using the `ets()` function from the `forecast` package, which fits individual exponential smoothing models to each time series.

For each time series forecast, we save the base forecast mean (`base_fc`) and the residuals from the fitted model (`res`), which are used to estimate the covariance matrix of the forecast errors.


```{r Base-forecasts and residuals computation}
# Compute base forecasts and residuals for each time series
base_fc = rep(NA, n)
res = matrix(NA, ncol = n, nrow = L)
for (i in 1:n){
    models = forecast::ets(train[,i], model = "AZZ")
    f = forecast(models, h = h)
    base_fc[i] = f$mean
    res[, i] = models$residuals
}
```

# Reconciliation 

We can now reconcile the forecasts with the function `reconc_t()`, which implements the t-Rec method. 

This function takes as input the aggregation matrix `A`, the base forecast means `base_fc_mean`, the training data `y_train` and the model residuals `residuals` and returns the parameters of the reconciled forecasts, which are distributed as a multivariate-t. The flag `return_upper =  TRUE` forces the function to return also the parameters of the upper-level reconciled forecasts. The flag `return_parameters = TRUE` forces the function to return also the parameters of the posterior distribution of the covariance matrix, which we will use to compare the covariance estimates obtained with t-Rec and MinT.

```{r t-reconciliation}
t_rec_results = reconc_t(A, base_fc_mean = base_fc, 
                         y_train = train, 
                         residuals = res,
                         return_parameters = TRUE,
                         return_upper = TRUE)
```

For the selected data window, we save the base forecasts (Base).

```{r Base}
# Base forecasts
Base_mean = base_fc
Base_cov_mat = crossprod(res)/nrow(res) # covariance of the residuals
```

We further compute the reconciled forecasts with the standard Gaussian reconciliation (MinT). Note the flag `return_upper = TRUE`, which forces the function to return the parameters of the upper-level reconciled forecasts.

```{r Gaussian}
# Gaussian/MinT: compute reconciliation with bayesRecon
gauss_results = reconc_gaussian(A, base_fc, residuals = res, return_upper = TRUE)
# Reconciled mean for the whole hierarchy:
MinT_reconciled_mean = c(gauss_results$upper_rec_mean, 
                         gauss_results$bottom_rec_mean)
```

# Comparison of results

We now compare the predictive densities for the 1-step ahead forecasts of the upper-level time series (Switzerland) obtained by the three methods above: the base forecasts (Base), Minimum Trace reconciliation (MinT), and t-Rec.

```{r compute uppers, echo=FALSE, eval=TRUE}
t_Rec_reconciled_mean <- c(t_rec_results$upper_rec_mean,t_rec_results$bottom_rec_mean)
t_Rec_corr_with_upper = A %*% t_rec_results$bottom_rec_scale_matrix
t_Rec_scale_par <- rbind(cbind(t_rec_results$upper_rec_scale_matrix, t_Rec_corr_with_upper),
                         cbind(t(t_Rec_corr_with_upper), t_rec_results$bottom_rec_scale_matrix))

# Index of the upper variable to plot
i_upper <- 1  # change this if a different index is needed

# Extract distribution's parameters for each method

# MinT
# Build the full reconciled covariance matrix
MinT_corr_with_upper = A %*% gauss_results$bottom_rec_cov
MinT_cov_mat = rbind(cbind(gauss_results$upper_rec_cov, MinT_corr_with_upper),
                        cbind(t(MinT_corr_with_upper),gauss_results$bottom_rec_cov))
mu_MinT <- as.numeric(MinT_reconciled_mean[i_upper])
sd_MinT <- sqrt(MinT_cov_mat[i_upper, i_upper])

# t-Rec
mu_tRec <- as.numeric(t_Rec_reconciled_mean[i_upper])
scale_tRec <- sqrt(t_Rec_scale_par[i_upper, i_upper])
df_tRec <- t_rec_results$bottom_rec_df

# Base
mu_Base <- as.numeric(Base_mean[i_upper])
sd_Base <- sqrt(Base_cov_mat[i_upper, i_upper])

# Create a grid of x values
x_vals <- seq(min(mu_MinT, mu_tRec, mu_Base) - 4*max(sd_MinT, scale_tRec, sd_Base),
              max(mu_MinT, mu_tRec, mu_Base) + 4*max(sd_MinT, scale_tRec, sd_Base),
              length.out = 1000)

# Compute densities
dens_MinT <- dnorm(x_vals, mean = mu_MinT, sd = sd_MinT)
dens_tRec <- dt((x_vals - mu_tRec) / scale_tRec, df = df_tRec) /scale_tRec
dens_Base <- dnorm(x_vals, mean = mu_Base, sd = sd_Base)

```

```{r comparison-plot, echo=FALSE, eval=TRUE, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 2**: Predictive densities of the upper time series obtained with MinT (purple), t-Rec (green) and Base (blue). The black triangle indicates the actual value.", fig.dim = c(8, 4), warning=FALSE}

dens_df <- data.frame(
  x       = rep(x_vals, 3),
  Method  = rep(c("MinT", "t-Rec", "Base"), each = length(x_vals)),
  Density = c(dens_MinT, dens_tRec, dens_Base)
)

method_colors <- c(
  "MinT" = "#440154",
  "t-Rec" = "#29AF7F",
  "Base" = "#3B528B"
)

ggplot(dens_df, aes(x = x, y = Density, color = Method)) +
  geom_area(
    data = dens_df[dens_df$Method %in% c("MinT", "t-Rec"),],
    aes(fill = Method),
    position = "identity",
    alpha = 0.3,
    color = NA
  ) +
  geom_line(linewidth = 1.5) +
  scale_color_manual(values = method_colors) +
  scale_fill_manual(values = method_colors) +
  geom_point(aes(x = actuals[i_upper], y = 0, shape = "Actual value"), color = "black", size = 3) +
  scale_shape_manual(values = c("Actual value" = 17)) +
  scale_x_continuous(labels = function(x) paste0(formatC(x / 1e3, format = "g"), "k")) +
  guides(fill = "none") +
  labs(
    title = "Predictive densities of upper time series",
    x = "",
    y = ""
  ) +
  theme_minimal(base_size = 12) +
  theme(
    legend.position = "top",
    legend.title = element_blank(),
    plot.title       = element_text(size = 18)
  )
```

The base forecast and the MinT reconciled forecast are normally distributed, while t-Rec has a Student's t distribution. 
Both reconciliation methods improve the forecast: the mean of the densities is closer to the actual value. However, the t-Rec method returns a wider density, which is more likely to contain the actual value. This is because t-Rec accounts for the uncertainty in the covariance matrix of the forecast errors, which leads to a more realistic representation of the forecast uncertainty.

# Comparison of the covariance matrix estimates

In this section we showcase the main strength of t-Rec: it provides an estimate of the covariance which accounts for the uncertainty in the estimation. To illustrate this point, we compare the covariance matrix estimates obtained with t-Rec and with the standard Gaussian reconciliation method (MinT).

We focus on the covariance matrix between the upper-level series, denoted as CH, and the bottom-level time series with the largest average values "Graubünden", denoted as GR. Analogous considerations apply also for other series. The code in the vignette is parametric so that a user could manually visualize different comparisons. 

While MinT only returns a point estimate for the covariance matrix, t-Rec provides a distribution for the covariance matrix. This is obtained in a Bayesian way: starting from a prior on the covariance, we include the data and obtain a posterior distribution which is an Inverse Wishart with known parameters $\nu$ and $\Psi$. Those parameters are stored in `t_rec_results$posterior_nu` and `t_rec_results$posterior_Psi`. 

Since we know analytically the distribution of the covariance, we can also compute the marginal distribution of the variances in closed-form. This is a scaled inverse Gamma distribution with the following parameters:
$$
\Sigma_{ii} \sim \text{Inv-Gamma}\left(\frac{\nu - n + 1}{2}, \frac{\Psi_{ii}}{2}\right).
$$

Instead of visualizing the variance, we show the standard deviation which produces a more intuitive interpretation. The standard deviation is obtained by applying the square root transformation to the variance, which results in a non-standard distribution. The density of the standard deviation can be computed using the change of variable formula, which leads to the following expression:
```{r compute density}
# Select which series to plot
i = 1     # CH
j = 19    # GR

# density of the inverse gamma
dinvgamma <- function(x, shape, rate) {
  dgamma(1/x, shape = shape, rate = rate) / x^2
}

# density of the standard deviation (square root transform of variance)
d_std_dev <- function(x, shape,rate){
  dinvgamma(x^2, shape = shape, rate = rate)*2*x
}

# compute the density of the variance parameters for CH
mean_ch <- t_rec_results$posterior_Psi[i, i]/(t_rec_results$posterior_nu - n - 1)
x_ch <- sqrt(seq(mean_ch*0.5, mean_ch*1.7, length.out = 1000))
shape_ch <- (t_rec_results$posterior_nu - n + 1) /2
rate_ch  <- t_rec_results$posterior_Psi[i, i] / 2
dens_ch <- d_std_dev(x_ch, shape = shape_ch, rate = rate_ch)

# compute the density of the variance parameters for GR
mean_gr <- t_rec_results$posterior_Psi[j, j]/(t_rec_results$posterior_nu - n - 1)
x_gr <- sqrt(seq(mean_gr*0.5, mean_gr*1.7, length.out = 1000))
shape_gr <- (t_rec_results$posterior_nu - n + 1) /2
rate_gr  <- t_rec_results$posterior_Psi[j, j] / 2
dens_gr <- d_std_dev(x_gr, shape = shape_gr, rate = rate_gr)
```

**Figure 3** shows the density of the posterior standard deviation of the forecasts for the upper time series (CH) and the bottom time series (GR).

```{r density plot, echo=FALSE, eval=TRUE, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 3**: Density of the posterior standard deviation of the forecasts.", fig.dim = c(8, 4), warning=FALSE}

df_dens <- data.frame(
  x = c(x_ch, x_gr), 
  y = c(dens_ch, dens_gr), 
  panel = rep(c("CH", "GR"), each = length(dens_ch))
)

# Plot the density of the standard deviation parameters for CH and GR
ggplot(df_dens, aes(x = x, y = y)) +
  geom_line(linewidth = 1.5, color="darkolivegreen", alpha=0.7) +
  scale_x_continuous(labels = function(x) paste0(formatC(x / 1e3, format = "g"), "k")) +
  guides(fill = "none") +
  labs(
    title = "Posterior standard deviation of the forecasts",
    x = "",
    y = ""
  ) +
  facet_wrap(~ panel, nrow = 1, scales = "free") +
  theme_minimal(base_size = 12) +
  theme(
    legend.position = "top",
    legend.title = element_blank(),
    plot.title       = element_text(size = 18)
  )
```

The posterior distribution for the covariance and for the correlation values is not available in closed form but it can be obtained via sampling. Since the posterior distribution is an Inverse-Wishart distribution, we can sample from this distribution using the custom function `rinvwishart()`, defined below.

```{r generate IW samples}
# generate k samples from an IW(Psi, nu) distribution
rinvwishart <- function(k, nu, Psi, seed=42) {
  p <- nrow(Psi)
  Sigma <- solve(Psi)
  
  set.seed(seed)
  all_W <- rWishart(k, df = nu, Sigma = Sigma)
  
  W <- array(NA, dim = c(p, p, k))
  for (i in 1:k) {
    W[,,i] <- solve(all_W[,,i])
  }
  return(W)
}

IW_post_samples <- rinvwishart(k=1000, nu = t_rec_results$posterior_nu,
                            Psi = t_rec_results$posterior_Psi)

```


```{r compute correlations, echo=FALSE, eval=TRUE}
corr_mint <- cov2cor(MinT_cov_mat)
corr_base <- cov2cor(Base_cov_mat)
corr_tRec <- cov2cor(t_rec_results$posterior_Psi/(t_rec_results$posterior_nu - n - 1))
```

**Figure 4** shows the posterior density of the correlation between CH and GR. The value estimated by MinT, plotted as a vertical dashed line, differs from the posterior mode estimated by t-Rec, illustrating how t-Rec captures a different view of the dependence structure. 

```{r plot densities, echo=FALSE, eval=TRUE, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 4**: Density of the posterior correlation between CH and GR obtained with t-Rec.", fig.dim = c(8, 4), warning=FALSE}
# Compute correlation samples for all sample matrices generated above
corr_post <- array(apply(IW_post_samples, FUN = function(M) cov2cor(M), MARGIN = c(3)),
                   dim = dim(IW_post_samples))

df_corr <- data.frame(x = corr_post[i, j, ])

ggplot(df_corr, aes(x = x)) +
  geom_density(fill = "#00BA38", alpha = 0.4) +
  geom_vline(xintercept = corr_mint[i, j], linetype = "dashed", color = "black") +
  annotate("text", x = corr_mint[i, j], y = max(density(corr_post[i, j, ])$y) * 0.95, label = "MinT", 
           hjust = -0.2, vjust = 0, size = 4.5) +
  labs(x = "", y = "Density",
       title = expression(paste("Posterior correlation ", rho["CH,GR"]))) +
  theme_minimal() +
  theme(
    legend.position  = "none",
    plot.title       = element_text(size = 18)
  )
```

Finally, we show in **Table 1** the standard deviation and correlation estimates for the upper-level series (CH) and the bottom-level series (GR) obtained with Base, MinT and t-Rec methods. While Base and MinT provide only point estimates for the standard deviation and correlation, t-Rec provides a distribution for these parameters. In the table, we report the mean of the posterior distribution for the standard deviation and the correlation.

```{r,echo=FALSE,eval=TRUE}
# Print table with values for Std and correlation
knitr::kable(data.frame(
  Method = c("Base", "MinT", "t-Rec"),
  sigma2_CH   = sqrt(c(Base_cov_mat[i, i], MinT_cov_mat[i, i], t_rec_results$posterior_Psi[i, i]/(t_rec_results$posterior_nu - n - 1))),
  sigma2_GR   = sqrt(c(Base_cov_mat[j, j], MinT_cov_mat[j, j], t_rec_results$posterior_Psi[j, j]/(t_rec_results$posterior_nu - n - 1))),
  rho_CH_GR   = c(corr_base[i, j], corr_mint[i, j], corr_tRec[i, j])
), digits = c(0, 0, 0, 2),
   format = "html",
   col.names = c("Method",
                 "$\\hat{\\sigma}_{\\text{CH}}$",
                 "$\\hat{\\sigma}_{\\text{GR}}$",
                 "$\\hat{\\rho}_{\\text{CH,GR}}$"),
   escape = FALSE,
   table.attr = "style='width:auto;'",
   format.args = list(big.mark = ",", scientific = FALSE),
   caption = "**Table 1**: Standard deviation and correlation estimates for CH and GR.")
```

Note that the standard deviation estimated by t-Rec is higher than both the base and the MinT estimates. This is because the posterior Inverse-Wishart distribution depends on the prior, which is initialized using the residuals of the naive forecasts. Since these residuals capture the full uncertainty of the non-reconciled forecasts, the prior inflates the posterior variance relative to the point estimate provided by MinT.
Moreover, the correlation between CH and GR is also estimated differently by t-Rec compared to MinT.




# References
<div id="refs"></div>

---
title: "Probabilistic Reconciliation via Conditioning with `bayesRecon`"
author: "Nicolò Rubattu, Giorgio Corani, Dario Azzimonti, Lorenzo Zambon"
date: "2023-11-26"
lang: "en"
output: html_vignette
bibliography: references.bib
cite:
- '@zambon2022efficient'
- '@zambon2024properties'
- '@corani2023probabilistic'
- '@corani2021probabilistic'
vignette: >
  %\VignetteIndexEntry{Probabilistic Reconciliation via Conditioning with `bayesRecon`}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = TRUE ### !!!! set to FALSE here to render only the text !!!!
)
set.seed(42)
```

```{r klippy, echo=FALSE, include=TRUE, eval=FALSE}
klippy::klippy(position = c("top", "right"), tooltip_message = "Copy", tooltip_success = "Done", color = "black")
```

# Introduction

This vignette shows how to perform  *probabilistic reconciliation* with
the `bayesRecon` package. We provide three examples:

1. *Temporal hierarchy for a count time series*: we build a temporal hierarchy over a count time series,  produce the base forecasts using `glarma` and  reconcile them via Bottom-Up Importance Sampling (BUIS).

2.  *Temporal hierarchy for a smooth time series*: we build a temporal hierarchy over a smooth time series,  compute the base forecasts using `ets` and  we reconcile them in closed form using Gaussian reconciliation. The covariance matrix is diagonal.

3.   *Hierarchical of smooth time series*: this is an example of a cross-sectional hierarchy. We generate the base forecasts using `ets` and we reconcile them via Gaussian reconciliation.
The covariance matrix is full and estimated via shrinkage.



# Installation

The package, available at this 
[CRAN page](https://cran.r-project.org/package=bayesRecon), can be installed and loaded with the usual commands:
```{r install, eval=FALSE}
install.packages("bayesRecon", dependencies = TRUE)
```

Load the package:

```{r load}
library(bayesRecon)
```

# Temporal hierarchy over a count time series
We select a monthly time series of counts from  the *carparts* dataset, available from
the expsmooth package [@expsmooth_pkg].
The data set contains time series of sales of cars part from Jan. 1998 to Mar. 2002. 
For this example we select time series #2655, which we make  available  as `bayesRecon::carparts_example`.

This time series  has a skewed distribution of  values. 
```{r carpart-plot, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 1**: Carpart - monthly car part sales.", fig.dim = c(6, 3)}
layout(mat = matrix(c(1, 2), nrow = 1, ncol = 2), widths = c(2, 1))
plot(carparts_example, xlab = "Time", ylab = "Car part sales", main = NULL)
hist(carparts_example, xlab = "Car part sales", main = NULL)
```
<br><br>
We divide the time series into train and test; the test set contains  the last 12 months.
```{r train-test}
train <- window(carparts_example, end = c(2001, 3))
test <- window(carparts_example, start = c(2001, 4))
```

<!-- In order to do hierarchical forecasting: -->

<!-- 1.    we build a temporal hierarchy; -->
<!-- 2.    we compute the *base forecasts*) at each temporal scale. -->

We build the temporal hierarchy using the `temporal aggregation` function.
We specify the aggregation levels using the 
`agg_levels` argument; in this case they are
*2-Monthly*, *Quarterly*, *4-Monthly*, *Biannual*, and *Annual*.

``` {r temp-agg}
train.agg <- bayesRecon::temporal_aggregation(train, agg_levels = c(2, 3, 4, 6, 12))
levels <- c("Annual", "Biannual", "4-Monthly", "Quarterly", "2-Monthly", "Monthly")
names(train.agg) <- levels
```

The function returns a list of aggregated time series, ordered from the most aggregated (top of the hierarchy) to the most disagreggated (bottom of the hierarchy). We  plot them below.
``` {r temp-agg-plot, dpi=300, fig.show="hold", out.width="100%", out.heigth="100%", fig.align='center', fig.cap="**Figure 2**: The aggregated time series of the temporal hierarchy.", fig.dim=c(6,3.5)}
par(mfrow = c(2, 3), mai = c(0.6, 0.6, 0.5, 0.5))
for (l in levels) {
  plot(train.agg[[l]], xlab = "Time", ylab = "Car part sales", main = l)
}
```
<br><br>
We  compute the  *base forecasts* using the package  [`glarma`](https://cran.r-project.org/package=glarma),
a package specific for forecasting count time series.
We forecast 12 steps ahead at the monthly level,  4 steps ahead  at the quarterly level, etc. by iterating over the levels of the  hierarchy,
At each level, we fit a `glarma` model
with Poisson predictive distribution and we forecast;
each forecast is constituted by 20000  integer samples.


Eventually we collect the samples of the  28 predictive distributions (one at the  *Annual* level, two at the *Biannual* level, etc.) in a list.
The  code below takes about 30 seconds on a standard computer.

``` {r hier-fore, cache=TRUE}
# install.packages("glarma", dependencies = TRUE)
# library(glarma)

fc.samples <- list()
D <- 20000
fc.count <- 1

# iterating over the temporal aggregation levels
for (l in seq_along(train.agg)) {
  f.level <- frequency(train.agg[[l]])
  print(paste("Forecasting at ", levels[l], "...", sep = ""))
  # fit an independent model for each aggregation level
  model <- glarma::glarma(
    train.agg[[l]],
    phiLags = if (f.level == 1) 1 else 1:(min(6, f.level - 1)),
    thetaLags = if (f.level == 1) NULL else f.level,
    X = cbind(intercept = rep(1, length(train.agg[[l]]))),
    offset = cbind(intercept = rep(0, length(train.agg[[l]]))),
    type = "Poi"
  )
  # forecast 1 year ahead
  h <- f.level
  tmp <- matrix(data = NA, nrow = h, ncol = D)
  for (s in (1:D)) {
    # each call to 'forecast.glarma' returns a simulation path
    tmp[, s] <- glarma::forecast(
      model,
      n.ahead = h,
      newdata = cbind(intercept = rep(1, h)),
      newoffset = rep(0, h)
    )$Y
  }
  # collect the forecasted samples
  for (i in 1:h) {
    fc.samples[[fc.count]] <- tmp[i, ]
    fc.count <- fc.count + 1
  }
}
```


Reconciliation requires the aggregation matrix $\mathbf{A}$, which we obtain using the function `get_reconc_matrices`. 
It requires: 

*   the aggregation factors of the hierarchy, which in this example are $\{2, 3, 4, 6, 12\}$;
*   the length of the forecasting horizon at the  bottom level, which is 12 in this example.

``` {r aggregationMatrix}
recon.matrices <- bayesRecon::get_reconc_matrices(agg_levels = c(2, 3, 4, 6, 12), h = 12)
# Aggregation matrix
A <- recon.matrices$A
```

To reconcile using Bottom-Up Important Sampling (BUIS) we
we use the function `reconc_BUIS`, passing to it
the $\mathbf{A}$ matrix, the *base forecasts*, the  type  of the base forecasts (`in_type`="samples") and whether the samples are continuous or discrete (`distr` = "discrete").


``` {r reconc}
recon.res <- bayesRecon::reconc_BUIS(
  A,
  base_fc = fc.samples,
  in_type = "samples",
  distr = "discrete",
  seed = 42
)
```

Here we obtain samples from the reconciled forecast distribution.
```{r res}
reconciled_samples <- recon.res$bottom_rec_samples
dim(reconciled_samples)
```

We now compute the Mean Absolute Error (MAE) and the Continuous Ranked Probability Score (CRPS) for the bottom (i.e., *monthly*) time series.  For computing CRPS, we use the package [`scoringRules`](https://cran.r-project.org/package=scoringRules).

```{r metrics}
# install.packages("scoringRules", dependencies = TRUE)
library(scoringRules)

ae.fc <- list()
ae.reconc <- list()
crps.fc <- list()
crps.reconc <- list()
for (h in 1:length(test)) {
  y.hat_ <- median(fc.samples[[nrow(A) + h]])
  y.reconc_ <- median(recon.res$bottom_rec_samples[h, ])
  # Compute Absolute Errors
  ae.fc[[h]] <- abs(test[h] - y.hat_)
  ae.reconc[[h]] <- abs(test[h] - y.reconc_)
  # Compute Continuous Ranked Probability Score (CRPS)
  crps.fc[[h]] <-
    scoringRules::crps_sample(y = test[h], dat = fc.samples[[nrow(A) + h]])
  crps.reconc[[h]] <-
    scoringRules::crps_sample(y = test[h], dat = recon.res$bottom_rec_samples[h, ])
}

mae.fc <- mean(unlist(ae.fc))
mae.reconc <- mean(unlist(ae.reconc))
crps.fc <- mean(unlist(crps.fc))
crps.reconc <- mean(unlist(crps.reconc))
metrics <- data.frame(
  row.names = c("MAE", "CRPS"),
  base.forecasts = c(mae.fc, crps.fc),
  reconciled.forecasts = c(mae.reconc, crps.reconc)
)
metrics
```


# Temporal hierarchy over a smooth time series

In this second example, we select a smooth monthly time series (N1485) from the M3 forecasting competition [@makridakis2000m3]. The data set is available in
the Mcomp package [@mcomp_pkg] and we make available this time series as `bayesRecon::M3_example`.


```{r m3-plot, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 3**: M3 - N1485 time series.", fig.dim = c(6, 3)}
plot(M3_example$train, xlab = "Time", ylab = "y", main = "N1485")
```
<br>
We build the temporal hierarchy using the `temporal_aggregation` function. 

Without specifying  `agg_levels`, the function  generates by default all the feasible aggregation: 2-Monthly, Quarterly, 4-Monthly, Biannual, and Annual.

```{r m3-agg} 
train.agg <- bayesRecon::temporal_aggregation(M3_example$train)
levels <- c("Annual", "Biannual", "4-Monthly", "Quarterly", "2-Monthly", "Monthly")
names(train.agg) <- levels
```


We generate the base forecasts using `ets` from the   [forecast](https://cran.r-project.org/package=forecast) package [@pkg_forecast].
At every level we predict 18 months ahead.
All the  predictive distributions are Gaussian.
```{r m3-fore}
# install.packages("forecast", dependencies = TRUE)
library(forecast)

H <- length(M3_example$test)
H

fc <- list()
level.idx <- 1
fc.idx <- 1
for (level in train.agg) {
  level.name <- names(train.agg)[level.idx]
  # fit an ETS model for each temporal level
  model <- ets(level)
  # generate forecasts for each level within 18 months
  h <- floor(H / (12 / frequency(level)))
  print(paste("Forecasting at ", level.name, ", h=", h, "...", sep = ""))
  level.fc <- forecast(model, h = h)
  # save mean and sd of the gaussian predictive distribution
  for (i in 1:h) {
    fc[[fc.idx]] <- list(
      mean = level.fc$mean[[i]],
      sd = (level.fc$upper[, "95%"][[i]] - level.fc$mean[[i]]) / qnorm(0.975)
    )
    fc.idx <- fc.idx + 1
  }
  level.idx <- level.idx + 1
}
```

Using the function `get_reconc_matrices`,  we get  matrix $\mathbf{A}$.

```{r m3-rmat, dpi=300, out.width = '70%', fig.align='center', fig.cap="**Figure 4**: M3 - The aggregation matrix A (red=1, yellow=0).", fig.dim = c(8, 8)}
rmat <- get_reconc_matrices(agg_levels = c(2, 3, 4, 6, 12), h = 18)

par(mai = c(1, 1, 0.5, 0.5))
image(1:ncol(rmat$A), 1:nrow(rmat$A),
  t(apply(t(rmat$A), 1, rev)),
  xaxt = "n", yaxt = "n", ylab = "", xlab = levels[6]
)
axis(1, at = 1:ncol(rmat$A), label = 1:ncol(rmat$A), las = 1)
axis(2, at = c(23, 22, 19, 15, 9), label = levels[1:5], las = 2)
```
<br>
The function `reconc_gaussian`  implements the exact Gaussian reconciliation.
We also run `reconc_BUIS`, to check the consistency 
between the two approaches.


```{r m3-reco} 
recon.gauss <- bayesRecon::reconc_gaussian(
  A = rmat$A,
  base_fc_mean = sapply(fc, "[[", 1),
  base_fc_cov = diag(sapply(fc, "[[", 2))^2
)

reconc.buis <- bayesRecon::reconc_BUIS(
  A = rmat$A,
  base_fc = fc,
  in_type = "params",
  distr = "gaussian",
  num_samples = 20000,
  seed = 42
)

# check that the algorithms return consistent results
round(rbind(
  c(rmat$S %*% recon.gauss$bottom_rec_mean),
  rowMeans(rbind(reconc.buis$upper_rec_samples, reconc.buis$bottom_rec_samples))
))
```

We now compare  *base forecasts* and  *reconciled forecasts*:

```{r m3-plotfore, dpi=300, out.width = "100%", fig.align='center', fig.cap="**Figure 5**: M3 - Base and reconciled forecasts. The black line shows the actual data (dashed in the test). The orange line is the  mean of the base forecasts, the blu line is the reconciled mean. We also show the 95% prediction intervals.", fig.dim = c(6, 4)}
yhat.mu <- tail(sapply(fc, "[[", 1), 18)
yhat.sigma <- tail(sapply(fc, "[[", 2), 18)
yhat.hi95 <- qnorm(0.975, mean = yhat.mu, sd = yhat.sigma)
yhat.lo95 <- qnorm(0.025, mean = yhat.mu, sd = yhat.sigma)
yreconc.mu <- rowMeans(reconc.buis$bottom_rec_samples)
yreconc.hi95 <- apply(
  reconc.buis$bottom_rec_samples, 1,
  function(x) quantile(x, 0.975)
)
yreconc.lo95 <- apply(
  reconc.buis$bottom_rec_samples, 1,
  function(x) quantile(x, 0.025)
)

ylim_ <- c(
  min(M3_example$train, M3_example$test, yhat.lo95, yreconc.lo95) - 1,
  max(M3_example$train, M3_example$test, yhat.hi95, yreconc.hi95) + 1
)

plot(M3_example$train,
  xlim = c(1990, 1995.6), ylim = ylim_,
  ylab = "y", main = "N1485 Forecasts"
)
lines(M3_example$test, lty = "dashed")
lines(ts(yhat.mu, start = start(M3_example$test), frequency = 12),
  col = "coral", lwd = 2
)
lines(ts(yreconc.mu, start = start(M3_example$test), frequency = 12),
  col = "blue2", lwd = 2
)
xtest <- time(M3_example$test)
polygon(c(xtest, rev(xtest)), c(yhat.mu, rev(yhat.hi95)),
  col = "#FF7F5066", border = "#FF7F5066"
)
polygon(c(xtest, rev(xtest)), c(yhat.mu, rev(yhat.lo95)),
  col = "#FF7F5066", border = "#FF7F5066"
)
polygon(c(xtest, rev(xtest)), c(yreconc.mu, rev(yreconc.hi95)),
  col = "#0000EE4D", border = "#0000EE4D"
)
polygon(c(xtest, rev(xtest)), c(yreconc.mu, rev(yreconc.lo95)),
  col = "#0000EE4D", border = "#0000EE4D"
)
```



#   Gaussian reconciliation of a cross-sectional hierarchy

In this example, we consider the hierarchical time series *infantgts*, which is available 
from the `hts` package [@hts_pkg]
We make it available also in our package as `bayesRecon::infantMortality`.

It contains counts of infant mortality (deaths) in Australia, disaggregated by state and sex (male and female). 

<!-- State codes refer to: New South Wales (NSW), Victoria (VIC), Queensland (QLD), South Australia (SA), Western Australia (WA), Northern Territory (NT), Australian Capital Territory (ACT), and Tasmania (TAS). -->

We forecast one year ahead using `auto.arima`  from the [`forecast` ](https://cran.r-project.org/package=forecast) package. 
We collect the residuals, which we will later use to compute the covariance matrix. 

```{r infants-forecasts}
# install.packages("forecast", dependencies = TRUE)
library(forecast)

fc <- list()
residuals <- matrix(NA,
  nrow = length(infantMortality$total),
  ncol = length(infantMortality)
)
fc.idx <- 1
for (s in infantMortality) {
  s.name <- names(infantMortality)[fc.idx]
  print(paste("Forecasting at ", s.name, "...", sep = ""))
  # fit an auto.arima model and forecast with h=1
  model <- auto.arima(s)
  s.fc <- forecast(model, h = 1)
  # save mean and sd of the gaussian predictive distribution
  fc[[s.name]] <- c(
    s.fc$mean,
    (s.fc$upper[, "95%"][[1]] - s.fc$mean) / qnorm(0.975)
  )
  residuals[, fc.idx] <- s.fc$residuals
  fc.idx <- fc.idx + 1
}
```

Now we build the $\mathbf{A}$ matrix.

```{r infants-s, dpi=300, out.width = '70%', fig.align='center', fig.cap="**Figure 6**: Infants mortality - The aggregation matrix A (red=1, yellow=0).", fig.dim = c(8, 8)}
# we have 16 bottom time series, and 11 upper time series
A <- matrix(data = c(
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
  1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
  0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1
), byrow = TRUE, ncol = 16)

# plot of A
par(mai = c(1.5, 1, 0.5, 0.5))
image(1:ncol(A), 1:nrow(A),
  t(apply(t(A), 1, rev)),
  xaxt = "n", yaxt = "n", ann = FALSE
)
axis(1, at = 1:ncol(A), label = names(infantMortality)[12:27], las = 2)
axis(2, at = c(1:11), label = rev(names(infantMortality)[1:11]), las = 2)
```
<br>
We use  `bayesRecon::schaferStrimmer_cov` to estimate the covariance matrix of the residuals with shrinkage [@schafer2005shrinkage].

```{r infants reconc}
# means
mu <- sapply(fc, "[[", 1)
# Shrinkage covariance
shrink.res <- bayesRecon::schaferStrimmer_cov(residuals)
print(paste("The estimated shrinkage intensity is", round(shrink.res$lambda_star, 3)))
Sigma <- shrink.res$shrink_cov
```

We now perform Gaussian reconciliation:

```{r infants-recon}
recon.gauss <- bayesRecon::reconc_gaussian(A,
  base_fc_mean = mu,
  base_fc_cov = Sigma
)

bottom_mu_reconc <- recon.gauss$bottom_rec_mean
bottom_Sigma_reconc <- recon.gauss$bottom_rec_cov

# Obtain reconciled mu and Sigma for the upper variable
upper_mu_reconc <- A %*% bottom_mu_reconc
upper_Sigma_reconc <- A %*% bottom_Sigma_reconc %*% t(A)

upper_mu_reconc
```

# References
<div id="refs"></div>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shrink_cov.R
\name{schaferStrimmer_cov}
\alias{schaferStrimmer_cov}
\title{Schäfer Strimmer covariance shrinkage}
\usage{
schaferStrimmer_cov(x)
}
\arguments{
\item{x}{matrix of samples with dimensions nxp (n samples, p dimensions).}
}
\value{
A list containing the shrinkage estimator and the optimal lambda. The list has the following named elements:
\itemize{
\item \code{shrink_cov}: the shrinked covariance matrix (\code{p} x \code{p});
\item \code{lambda_star}: the optimal lambda for the shrinkage;
}
}
\description{
Computes the Schäfer Strimmer shrinkage estimator for a covariance matrix
from a matrix of samples.
}
\details{
This function computes the shrinkage to a diagonal covariance with unequal variances.
Note that here we use the estimators \eqn{S = X X^T/n} and \eqn{T = diag(S)} and we internally
use the correlation matrix in place of the covariance to compute the optimal shrinkage factor.
}
\examples{

# Generate some multivariate normal samples
# Parameters
nSamples <- 200
pTrue <- 2

# True moments
true_cov <- matrix(c(3, 2, 2, 2), nrow = 2)
chol_true_cov <- chol(true_cov)
true_mean <- c(0, 0)

# Generate samples
set.seed(42)
x <- replicate(nSamples, true_mean) +
  t(chol_true_cov) \%*\% matrix(stats::rnorm(pTrue * nSamples),
    nrow = pTrue, ncol = nSamples
  )
x <- t(x)
res_shrinkage <- schaferStrimmer_cov(x)
res_shrinkage$lambda_star # should be 0.01287923

}
\references{
Schäfer, Juliane, and Korbinian Strimmer. (2005). \emph{A Shrinkage Approach to Large-Scale Covariance Matrix Estimation and Implications for Functional Genomics.} Statistical Applications in Genetics and Molecular Biology 4: Article32. \doi{10.2202/1544-6115.1175}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reconc_t.R
\name{.core_reconc_t}
\alias{.core_reconc_t}
\title{Core reconciliation via multivariate t-distribution.}
\usage{
.core_reconc_t(
  A,
  base_fc_mean,
  Psi_post,
  nu_post,
  return_upper = FALSE,
  return_parameters = FALSE
)
}
\arguments{
\item{A}{Matrix (n_upper x n_bottom) defining the hierarchy where upper = A \%*\% bottom.}

\item{base_fc_mean}{Vector of length (n_upper + n_bottom) containing the base forecast means
for both upper and bottom levels (upper first, then bottom).}

\item{Psi_post}{Scale matrix (n_upper + n_bottom x n_upper + n_bottom) of the posterior
multivariate t-distribution.}

\item{nu_post}{Degrees of freedom of the posterior multivariate t-distribution.}

\item{return_upper}{Logical, whether to return the reconciled parameters for the upper variables (default is FALSE).}
}
\value{
A list containing:
\itemize{
\item \code{bottom_rec_mean}: reconciled bottom-level mean.
\item \code{bottom_rec_scale_matrix}: reconciled bottom-level scale matrix.
\item \code{bottom_rec_df}: reconciled degrees of freedom.
}
If \code{return_upper} is TRUE, also returns:
\itemize{
\item \code{upper_rec_mean}: reconciled upper-level mean.
\item \code{upper_rec_scale_matrix}: reconciled upper-level scale matrix.
\item \code{upper_rec_df}: reconciled upper-level degrees of freedom.
}
}
\description{
Internal function that performs the core reconciliation logic for the t-distribution based
reconciliation method. This function assumes an uncertain covariance matrix with an Inverse-Wishart prior.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hierarchy.R
\name{.get_Au}
\alias{.get_Au}
\title{Get aggregation matrix of upper sub-hierarchy}
\usage{
.get_Au(A, lowest_rows = NULL)
}
\arguments{
\item{A}{Aggregation matrix (must be hierarchical)}

\item{lowest_rows}{Optional. Integer vector of row indices corresponding to
the lowest upper level. If NULL, computed using \code{.lowest_lev(A)}.}
}
\value{
Aggregation matrix for the upper sub-hierarchy, or NULL if all
uppers are at the lowest level.
}
\description{
Constructs the aggregation matrix Au for the sub-hierarchy composed only
of the upper variables. This matrix relates the "upper uppers" to the
"lowest uppers" (the most disaggregated level of the upper hierarchy).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{extr_mkt_events_basefc}
\alias{extr_mkt_events_basefc}
\title{Base forecasts for the extreme market events dataset}
\format{
A list \code{extr_mkt_events_basefc} containing
\describe{
\item{\code{extr_mkt_events_basefc$mu}}{data frame of the base forecast means, for each day}
\item{\code{extr_mkt_events_basefc$size}}{data frame of the static base forecast size parameters}
}
}
\source{
Agosto, A. (2022).
\emph{Multivariate Score-Driven Models for Count Time Series to Assess Financial Contagion}.
\doi{10.2139/ssrn.4119895}
}
\usage{
extr_mkt_events_basefc
}
\description{
Base forecasts for the \code{extr_mkt_events} dataset, computed using the model by
Agosto, A. (2022).
\emph{Multivariate Score-Driven Models for Count Time Series to Assess Financial Contagion}.
\doi{10.2139/ssrn.4119895}.
}
\details{
The predictive distribution for the bottom time series is a multivariate negative
binomial with a static vector of dispersion parameters and a time-varying vector
of location parameters following a score-driven dynamics.
The base forecasts for the upper time series are computed using a univariate version of this model.
They are in-sample forecasts: for each training instant, they are computed for
time t+1 by conditioning on the counts observed up to time t.
}
\references{
Agosto, A. (2022).
\emph{Multivariate Score-Driven Models for Count Time Series to Assess Financial Contagion}.
\doi{10.2139/ssrn.4119895}

Zambon, L., Agosto, A., Giudici, P., Corani, G. (2024).
\emph{Properties of the reconciled distributions for Gaussian and count forecasts}.
International Journal of Forecasting (in press).
\doi{10.1016/j.ijforecast.2023.12.004}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{infantMortality}
\alias{infantMortality}
\title{Infant Mortality grouped time series dataset}
\format{
List of time series of class \link[stats]{ts}.
}
\source{
hts package \href{https://cran.r-project.org/package=hts}{CRAN}
}
\usage{
infantMortality
}
\description{
A yearly grouped time series dataset, from 1901 to 2003, of infant mortality counts (deaths) in Australia;
disaggregated by state (see below), and sex (male and female).
}
\details{
States: New South Wales (NSW), Victoria (VIC), Queensland (QLD), South Australia (SA), Western Australia
(WA), Northern Territory (NT), Australian Capital Territory (ACT), and Tasmania (TAS).
}
\references{
Hyndman, R.J., Ahmed, R.A., Athanasopoulos, G., Shang, H.L. (2011).
Optimal combination forecasts for hierarchical time series.
Computational Statistics and Data Analysis, 55(9), 2579-2589.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reconc_MCMC.R
\name{reconc_MCMC}
\alias{reconc_MCMC}
\title{MCMC for Probabilistic Reconciliation of forecasts via conditioning}
\usage{
reconc_MCMC(
  A,
  base_fc,
  distr,
  num_samples = 10000,
  tuning_int = 100,
  init_scale = 1,
  burn_in = 1000,
  return_upper = TRUE,
  seed = NULL
)
}
\arguments{
\item{A}{aggregation matrix (n_upper x n_bottom).}

\item{base_fc}{list of the parameters of the base forecast distributions, see details.}

\item{distr}{a string describing the type of predictive distribution.}

\item{num_samples}{number of samples to draw using MCMC.}

\item{tuning_int}{number of iterations between scale updates of the proposal.}

\item{init_scale}{initial scale of the proposal.}

\item{burn_in}{number of initial samples to be discarded.}

\item{return_upper}{Logical. If \code{TRUE} (default), also returns the reconciled samples for the
upper time series. Default is \code{TRUE}.}

\item{seed}{seed for reproducibility.}
}
\value{
A list containing the reconciled forecasts. The list has the following named elements:
\itemize{
\item \code{bottom_rec_samples}: a matrix (n_bottom x \code{num_samples}) containing reconciled samples for the bottom time series;
\item \code{upper_rec_samples}: a matrix (n_upper x \code{num_samples}) containing reconciled samples for the upper time series (only if \code{return_upper = TRUE}).
}
}
\description{
Uses Markov Chain Monte Carlo algorithm to draw samples from the reconciled
forecast distribution, which is obtained via conditioning.

This is a bare-bones implementation of the Metropolis-Hastings algorithm,
we suggest the usage of tools to check the convergence.
The function only works with Poisson or Negative Binomial base forecasts.

The function \code{\link[=reconc_BUIS]{reconc_BUIS()}} is generally faster on most hierarchies.
}
\details{
The parameter \code{base_fc} is a list containing n = n_upper + n_bottom elements.
Each element is a list containing the estimated:
\itemize{
\item mean and sd for the Gaussian base forecast, see \link[stats]{Normal}, if \code{distr}='gaussian';
\item lambda for the Poisson base forecast, see \link[stats]{Poisson}, if \code{distr}='poisson';
\item size and prob (or mu) for the negative binomial base forecast, see \link[stats]{NegBinomial}, if \code{distr}='nbinom'.
}

The first n_upper elements of the list are the upper base forecasts, in the order given by the rows of A.
The elements from n_upper+1 until the end of the list are the bottom base forecasts, in the order given by the columns of A.
}
\examples{

library(bayesRecon)

# Create a minimal hierarchy with 2 bottom and 1 upper variable
rec_mat <- get_reconc_matrices(agg_levels = c(1, 2), h = 2)
A <- rec_mat$A

# Set the parameters of the Poisson base forecast distributions
lambda1 <- 2
lambda2 <- 4
lambdaY <- 9
lambdas <- c(lambdaY, lambda1, lambda2)

base_fc <- list()
for (i in 1:length(lambdas)) {
  base_fc[[i]] <- list(lambda = lambdas[i])
}

# Sample from the reconciled forecast distribution using MCMC
mcmc <- reconc_MCMC(A, base_fc,
  distr = "poisson",
  num_samples = 30000, seed = 42
)
samples_mcmc <- rbind(mcmc$upper_rec_samples, mcmc$bottom_rec_samples)

# Compare the reconciled means with those obtained via BUIS
buis <- reconc_BUIS(A, base_fc,
  in_type = "params",
  distr = "poisson", num_samples = 100000, seed = 42
)
samples_buis <- rbind(buis$upper_rec_samples, buis$bottom_rec_samples)

print(rowMeans(samples_mcmc))
print(rowMeans(samples_buis))

}
\references{
Corani, G., Azzimonti, D., Rubattu, N. (2024).
\emph{Probabilistic reconciliation of count time series}.
International Journal of Forecasting 40 (2), 457-469.
\doi{10.1016/j.ijforecast.2023.04.003}.
}
\seealso{
\code{\link[=reconc_BUIS]{reconc_BUIS()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reconc_MixCond.R, R/reconc_TDcond.R
\name{Mixed_reconciliation}
\alias{Mixed_reconciliation}
\alias{reconc_MixCond}
\alias{reconc_TDcond}
\title{Probabilistic forecast reconciliation of mixed hierarchies}
\usage{
reconc_MixCond(
  A,
  base_fc_bottom,
  base_fc_upper,
  bottom_in_type = "pmf",
  distr = NULL,
  num_samples = 20000,
  return_type = "pmf",
  return_upper = TRUE,
  suppress_warnings = FALSE,
  seed = NULL
)

reconc_TDcond(
  A,
  base_fc_bottom,
  base_fc_upper,
  bottom_in_type = "pmf",
  distr = NULL,
  num_samples = 20000,
  return_type = "pmf",
  return_upper = TRUE,
  suppress_warnings = TRUE,
  seed = NULL
)
}
\arguments{
\item{A}{Aggregation matrix (n_upper x n_bottom).}

\item{base_fc_bottom}{A list containing the bottom base forecasts, see details.}

\item{base_fc_upper}{A list containing the upper base forecasts, see details.}

\item{bottom_in_type}{A string with three possible values:
\itemize{
\item 'pmf' if the bottom base forecasts are in the form of pmf, see details;
\item 'samples' if the bottom base forecasts are in the form of samples;
\item 'params'  if the bottom base forecasts are in the form of estimated parameters.
}}

\item{distr}{A string describing the type of bottom base forecasts ('poisson' or 'nbinom').

This is only used if \code{bottom_in_type='params'}.}

\item{num_samples}{Number of samples drawn from the reconciled distribution.
This is ignored if \code{bottom_in_type='samples'}; in this case, the number of
reconciled samples is equal to the number of samples of the base forecasts.}

\item{return_type}{The return type of the reconciled distributions.
A string with three possible values:
\itemize{
\item 'pmf' returns a list containing the reconciled marginal pmf objects;
\item 'samples' returns a list containing the reconciled multivariate samples;
\item 'all' returns a list with both pmf objects and samples.
}}

\item{return_upper}{Logical, whether to return the reconciled parameters for the upper variables (default is TRUE).}

\item{suppress_warnings}{Logical. If \code{TRUE}, no warnings about samples are triggered;
if \code{FALSE}, warnings are generated. Default is \code{FALSE} for \code{reconc_MixCond}
and \code{TRUE} for \code{reconc_TDcond}. See the respective sections above.}

\item{seed}{Seed for reproducibility.}
}
\value{
A list containing the reconciled forecasts. The list has the following named elements:
\itemize{
\item \code{bottom_rec_pmf}: a list of PMF objects for each bottom series
(only if \code{return_type} is \code{'pmf'} or \code{'all'});
\item \code{bottom_rec_samples}: a matrix (n_bottom x \code{num_samples}) of reconciled bottom samples
(only if \code{return_type} is \code{'samples'} or \code{'all'});
\item \code{upper_rec_pmf}: a list of PMF objects for each upper series
(only if \code{return_type} is \code{'pmf'} or \code{'all'}, and \code{return_upper = TRUE});
\item \code{upper_rec_samples}: a matrix (n_upper x \code{num_samples}) of reconciled upper samples
(only if \code{return_type} is \code{'samples'} or \code{'all'}, and \code{return_upper = TRUE}).
}
}
\description{
\code{reconc_MixCond()} uses importance sampling to draw samples from the reconciled
forecast distribution, obtained via conditioning, in the case of a mixed hierarchy.

\code{reconc_TDcond()} uses a top-down conditioning algorithm: first, upper base forecasts are
reconciled via conditioning using only the hierarchical constraints between the
upper; then, the bottom distributions are updated via a probabilistic top-down procedure.
}
\details{
The base bottom forecasts \code{base_fc_bottom} must be a list of length n_bottom, where each element is either
\itemize{
\item a PMF object (see details below), if \code{bottom_in_type='pmf'};
\item a vector of samples, if \code{bottom_in_type='samples'};
\item a list of parameters, if \code{bottom_in_type='params'}:
\itemize{
\item lambda for the Poisson base forecast if \code{distr}='poisson', see \link[stats]{Poisson};
\item size and prob (or mu) for the negative binomial base forecast if \code{distr}='nbinom',
see \link[stats]{NegBinomial}.
}
}

The base upper forecasts \code{base_fc_upper} must be a list containing the parameters of
the multivariate Gaussian distribution of the upper forecasts.
The list must contain only the named elements \code{mean} (vector of length n_upper)
and \code{cov} (n_upper x n_upper matrix).

The order of the upper and bottom base forecasts must match the order of (respectively) the rows and the columns of A.

A PMF object is a numerical vector containing the probability mass function of a discrete distribution.
Each element corresponds to the probability of the integers from 0 to the last value of the support.
See also \link{PMF} for functions that handle PMF objects.

\strong{Warnings and errors.}

In \code{reconc_MixCond}, warnings are triggered from the importance sampling step if:
\itemize{
\item weights are all zeros, then the upper forecast is ignored during reconciliation;
\item the effective sample size is < 200;
\item the effective sample size is < 1\% of the sample size.
}

These warnings are an indication that the base forecasts might have issues.
Please check the base forecasts in case of warnings.

In \code{reconc_TDcond}, if some of the reconciled upper samples lie outside the support of the bottom-up
distribution, those samples are discarded; the remaining ones are resampled with
replacement, so that the number of output samples is equal to \code{num_samples}.
In this case, a warning is issued if \code{suppress_warnings=FALSE} (default is \code{TRUE}).
If the fraction of discarded samples is above 50\%, the function returns an error.
}
\examples{

library(bayesRecon)

# Consider a simple hierarchy with two bottom and one upper
A <- matrix(c(1, 1), nrow = 1)
# The bottom forecasts are Poisson with lambda=15
lambda <- 15
n_tot <- 60
base_fc_bottom <- list()
base_fc_bottom[[1]] <- apply(matrix(seq(0, n_tot)), MARGIN = 1,
                             FUN = \(x) dpois(x, lambda = lambda))
base_fc_bottom[[2]] <- apply(matrix(seq(0, n_tot)), MARGIN = 1,
                             FUN = \(x) dpois(x, lambda = lambda))

# The upper forecast is a Normal with mean 40 and std 5
base_fc_upper <- list(mean = 40, cov = matrix(5^2))

# Reconcile with reconc_MixCond
res.mixCond <- reconc_MixCond(A, base_fc_bottom, base_fc_upper)

# Note that the bottom distributions are slightly shifted to the right
PMF_summary(res.mixCond$bottom_rec_pmf[[1]])
PMF_summary(base_fc_bottom[[1]])

PMF_summary(res.mixCond$bottom_rec_pmf[[2]])
PMF_summary(base_fc_bottom[[2]])

# The upper distribution is slightly shifted to the left
PMF_summary(res.mixCond$upper_rec_pmf[[1]])
PMF_get_var(res.mixCond$upper_rec_pmf[[1]])


library(bayesRecon)

# Consider a simple hierarchy with two bottom and one upper
A <- matrix(c(1, 1), nrow = 1)
# The bottom forecasts are Poisson with lambda=15
lambda <- 15
n_tot <- 60
base_fc_bottom <- list()
base_fc_bottom[[1]] <- apply(matrix(seq(0, n_tot)), MARGIN = 1,
                             FUN = \(x) dpois(x, lambda = lambda))
base_fc_bottom[[2]] <- apply(matrix(seq(0, n_tot)), MARGIN = 1,
                             FUN = \(x) dpois(x, lambda = lambda))

# The upper forecast is a Normal with mean 40 and std 5
base_fc_upper <- list(mean = 40, cov = matrix(c(5^2)))

# Reconcile with reconc_TDcond
res.TDcond <- reconc_TDcond(A, base_fc_bottom, base_fc_upper)

# Note that the bottom distributions are shifted to the right
PMF_summary(res.TDcond$bottom_rec_pmf[[1]])
PMF_summary(base_fc_bottom[[1]])

PMF_summary(res.TDcond$bottom_rec_pmf[[2]])
PMF_summary(base_fc_bottom[[2]])

# The upper distribution remains similar
PMF_summary(res.TDcond$upper_rec_pmf[[1]])
PMF_get_var(res.TDcond$upper_rec_pmf[[1]])

## Example 2: reconciliation with unbalanced hierarchy
# We consider the example in Fig. 9 of Zambon et al. (2024).

# The hierarchy has 5 bottoms and 3 uppers
A <- matrix(c(
  1, 1, 1, 1, 1,
  1, 1, 0, 0, 0,
  0, 0, 1, 1, 0
), nrow = 3, byrow = TRUE)
# Note that the 5th bottom only appears in the highest level, this is an unbalanced hierarchy.
n_upper <- nrow(A)
n_bottom <- ncol(A)

# The bottom forecasts are Poisson with lambda=15
lambda <- 15
n_tot <- 60
base_fc_bottom <- list()
for (i in seq(n_bottom)) {
  base_fc_bottom[[i]] <- apply(matrix(seq(0, n_tot)), MARGIN = 1,
                               FUN = \(x) dpois(x, lambda = lambda))
}

# The upper forecasts are a multivariate Gaussian
mean <- c(75, 30, 30)
cov <- matrix(c(
  5^2, 5, 5,
  5, 10, 0,
  5, 0, 10
), nrow = 3, byrow = TRUE)

base_fc_upper <- list(mean = mean, cov = cov)
\dontrun{
# If we reconcile with reconc_TDcond it won't work (unbalanced hierarchy)
res.TDcond <- reconc_TDcond(A, base_fc_bottom, base_fc_upper)
}

# We can balance the hierarchy by duplicating the node b5:
# i) consider the time series observations for b5 as the upper u4,
# ii) fit the multivariate ts model for u1, u2, u3, u4.

# In this example we simply assume that the forecast for u1-u4 is
# Gaussian with the mean and variance of u4 given by the parameters in b5.
mean_b5 <- lambda
var_b5 <- lambda
mean <- c(75, 30, 30, mean_b5)
cov <- matrix(c(
  5^2, 5, 5, 5,
  5, 10, 0, 0,
  5, 0, 10, 0,
  5, 0, 0, var_b5
), nrow = 4, byrow = TRUE)
base_fc_upper <- list(mean = mean, cov = cov)

# We also need to update the aggregation matrix
A <- matrix(c(
  1, 1, 1, 1, 1,
  1, 1, 0, 0, 0,
  0, 0, 1, 1, 0,
  0, 0, 0, 0, 1
), nrow = 4, byrow = TRUE)

# We can now reconcile with TDcond
res.TDcond <- reconc_TDcond(A, base_fc_bottom, base_fc_upper)

# Note that the reconciled distribution of b5 and u4 are identical,
# keep this in mind when using the results of your reconciliation!
max(abs(res.TDcond$bottom_rec_pmf[[5]] - res.TDcond$upper_rec_pmf[[4]]))

}
\references{
Zambon, L., Azzimonti, D., Rubattu, N., Corani, G. (2024).
\emph{Probabilistic reconciliation of mixed-type hierarchical time series}.
Proceedings of the Fortieth Conference on Uncertainty in Artificial Intelligence,
PMLR 244:4078-4095. \url{https://proceedings.mlr.press/v244/zambon24a.html}.
}
\seealso{
\code{\link[=reconc_BUIS]{reconc_BUIS()}}, \code{\link[=reconc_gaussian]{reconc_gaussian()}}, \link{PMF}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bayesRecon-package.R
\docType{package}
\name{bayesRecon-package}
\alias{bayesRecon}
\alias{bayesRecon-package}
\title{bayesRecon: Probabilistic Reconciliation via Conditioning}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Provides methods for probabilistic reconciliation of hierarchical forecasts of time series. The available methods include analytical Gaussian reconciliation (Corani et al., 2021) \doi{10.1007/978-3-030-67664-3_13}, MCMC reconciliation of count time series (Corani et al., 2024) \doi{10.1016/j.ijforecast.2023.04.003}, Bottom-Up Importance Sampling (Zambon et al., 2024) \doi{10.1007/s11222-023-10343-y}, methods for the reconciliation of mixed hierarchies (Mix-Cond and TD-cond) (Zambon et al., 2024) \url{https://proceedings.mlr.press/v244/zambon24a.html}, analytical reconciliation with Bayesian treatment of the covariance matrix (Carrara et al., 2025) \doi{ 10.48550/arXiv.2506.19554}.
}
\section{Learn more}{


To learn more about \code{bayesRecon}, start with the vignettes: \code{browseVignettes(package = "bayesRecon")}
}

\section{Reconciliation functions}{


The package implements reconciliation via conditioning for probabilistic forecasts
of hierarchical time series. The reconciliation functions are:
\itemize{
\item \code{\link[=reconc_gaussian]{reconc_gaussian()}}: reconciliation via conditioning assuming multivariate Gaussian
base forecasts; this is done analytically;
\item \code{\link[=reconc_t]{reconc_t()}}: reconciliation via conditioning assuming multivariate Gaussian
base forecasts with uncertain covariance matrix;
the reconciled forecasts are multivariate Student-t; this is done analytically;
\item \code{\link[=reconc_BUIS]{reconc_BUIS()}}: reconciliation via conditioning of any probabilistic forecast
via bottom-up importance sampling; an alternative method for discrete forecasts
is implemented in \code{\link[=reconc_MCMC]{reconc_MCMC()}}, but we recommend using \code{reconc_BUIS};
\item \code{\link[=reconc_MixCond]{reconc_MixCond()}} and \code{\link[=reconc_TDcond]{reconc_TDcond()}}: reconciliation of mixed hierarchies, where
the upper forecasts are multivariate Gaussian and the bottom forecasts are discrete distributions;
\code{reconc_MixCond} implements conditioning via importance sampling,
while \code{reconc_TDcond} implements top-down conditioning.
}
}

\section{Utility functions}{

\itemize{
\item \code{\link[=temporal_aggregation]{temporal_aggregation()}}: temporal aggregation of a given time series object of class \link[stats]{ts};
\item \code{\link[=get_reconc_matrices]{get_reconc_matrices()}}: aggregation and summing matrices for a temporal hierarchy
of time series from user-selected list of aggregation levels;
\item \code{\link[=schaferStrimmer_cov]{schaferStrimmer_cov()}}: computes the Schäfer-Strimmer shrinkage estimator for the covariance matrix;
\item \code{\link[=multi_log_score_optimization]{multi_log_score_optimization()}}: estimates the optimal degrees of freedom for \code{\link[=reconc_t]{reconc_t()}}
by maximizing the leave-one-out (LOO) multivariate log density;
\item \link{PMF}: functions for handling PMF objects (sampling, computing statistics like mean, variance, quantiles, and summaries).
}
}

\references{
Carrara, C., Corani, G., Azzimonti, D., & Zambon, L. (2025).
\emph{Modeling the uncertainty on the covariance matrix for probabilistic forecast reconciliation}.
arXiv preprint arXiv:2506.19554. \url{https://arxiv.org/abs/2506.19554}.

Corani, G., Azzimonti, D., Augusto, J.P.S.C., Zaffalon, M. (2021).
\emph{Probabilistic Reconciliation of Hierarchical Forecast via Bayes' Rule}.
ECML PKDD 2020. Lecture Notes in Computer Science, vol 12459.
\doi{10.1007/978-3-030-67664-3_13}.

Corani, G., Azzimonti, D., Rubattu, N. (2024).
\emph{Probabilistic reconciliation of count time series}.
International Journal of Forecasting 40 (2), 457-469.
\doi{10.1016/j.ijforecast.2023.04.003}.

Zambon, L., Azzimonti, D. & Corani, G. (2024).
\emph{Efficient probabilistic reconciliation of forecasts for real-valued and count time series}.
Statistics and Computing 34 (1), 21.
\doi{10.1007/s11222-023-10343-y}.

Zambon, L., Agosto, A., Giudici, P., Corani, G. (2024).
\emph{Properties of the reconciled distributions for Gaussian and count forecasts}.
International Journal of Forecasting (in press).
\doi{10.1016/j.ijforecast.2023.12.004}.

Zambon, L., Azzimonti, D., Rubattu, N., Corani, G. (2024).
\emph{Probabilistic reconciliation of mixed-type hierarchical time series}.
Proceedings of the Fortieth Conference on Uncertainty in Artificial Intelligence,
PMLR 244:4078-4095. \url{https://proceedings.mlr.press/v244/zambon24a.html}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/IDSIA/bayesRecon}
  \item \url{https://idsia.github.io/bayesRecon/}
  \item Report bugs at \url{https://github.com/IDSIA/bayesRecon/issues}
}

}
\author{
\strong{Maintainer}: Dario Azzimonti \email{dario.azzimonti@gmail.com} (\href{https://orcid.org/0000-0001-5080-3061}{ORCID})

Authors:
\itemize{
  \item Lorenzo Zambon \email{lorenzo.zambon@idsia.ch} (\href{https://orcid.org/0000-0002-8939-993X}{ORCID})
  \item Stefano Damato \email{stefano.damato@idsia.ch} (\href{https://orcid.org/0009-0001-0225-3736}{ORCID})
  \item Nicolò Rubattu \email{nicolo.rubattu@idsia.ch} (\href{https://orcid.org/0000-0002-2703-1005}{ORCID})
  \item Giorgio Corani \email{giorgio.corani@idsia.ch} (\href{https://orcid.org/0000-0002-1541-8384}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reconc_gaussian.R
\name{reconc_gaussian}
\alias{reconc_gaussian}
\title{Analytical reconciliation of Gaussian base forecasts}
\usage{
reconc_gaussian(
  A,
  base_fc_mean,
  base_fc_cov = NULL,
  residuals = NULL,
  return_upper = FALSE
)
}
\arguments{
\item{A}{aggregation matrix (n_upper x n_bottom).}

\item{base_fc_mean}{a vector containing the means of the base forecasts.}

\item{base_fc_cov}{a matrix containing the covariance matrix of the base forecasts.}

\item{residuals}{a matrix with the residuals of the base forecasts, with n_upper + n_bottom columns.
The covariance matrix of the base forecasts is computed from the residuals using the Schäfer Strimmer shrinkage estimator.
If base_fc_cov is provided, residuals are ignored.}

\item{return_upper}{logical, whether to return the reconciled parameters for the upper variables (default is FALSE).}
}
\value{
A list containing the reconciled forecasts. The list has the following named elements:
\itemize{
\item \code{bottom_rec_mean}: reconciled mean for the bottom forecasts;
\item \code{bottom_rec_cov}: reconciled covariance for the bottom forecasts;
\item \code{upper_rec_mean}: (only if \code{return_upper = TRUE}) reconciled mean for the upper forecasts;
\item \code{upper_rec_cov}: (only if \code{return_upper = TRUE}) reconciled covariance for the upper forecasts.
}
}
\description{
Closed form computation of the reconciled forecasts in case of Gaussian base forecasts.
}
\details{
In the vector of the means of the base forecasts the order must be: first the upper,
then the bottom; the order within the uppers is given by the rows of A,
the order within the bottoms by the columns of A.
The order of the rows of the covariance matrix of the base forecasts is the same.

Unless \code{return_upper = TRUE}, the function returns only the reconciled parameters of the bottom variables.
}
\examples{

library(bayesRecon)

#' # ---- Example 1: base forecasts are given ----

# Create a minimal hierarchy with 2 bottom and 1 upper variable
A <- get_reconc_matrices(agg_levels = c(1, 2), h = 2)$A

# Set the parameters of the Gaussian base forecast distributions
mu1 <- 2
mu2 <- 4
muY <- 9
base_fc_mean <- c(muY, mu1, mu2)  # vector of means

sigma1 <- 2
sigma2 <- 2
sigmaY <- 3
sigmas <- c(sigmaY, sigma1, sigma2)
base_fc_cov <- diag(sigmas^2)  # covariance matrix

analytic_rec <- reconc_gaussian(A,
  base_fc_mean = base_fc_mean,
  base_fc_cov = base_fc_cov
)

bottom_mean_rec <- analytic_rec$bottom_rec_mean
bottom_cov_rec <- analytic_rec$bottom_rec_cov

# To obtain reconciled samples for the entire hierarchy, sample from the reconciled 
# bottom distribution and then aggregate using A. 

# Sample from the reconciled bottom-level Gaussian distribution
# First, compute the Cholesky decomposition of the reconciled covariance matrix:
chol_decomp <- chol(bottom_cov_rec)
# Then, sample from the standard normal distribution and apply the transformation:
Z <- matrix(stats::rnorm(n = 2000), nrow = 2) 
B <- t(chol_decomp) \%*\% Z + matrix(rep(bottom_mean_rec, 1000), nrow = 2) 

# Aggregate bottom samples to get upper samples, then stack
U <- A \%*\% B
Y_reconc <- rbind(U, B)

cat("Dimensions of reconciled samples (upper + bottom):", dim(Y_reconc), "\n")


# ---- Example 2: using residuals from fitted ETS models ----
\donttest{
if (requireNamespace("forecast", quietly = TRUE)) {

  # Simulate 2 bottom series from AR(1) processes
  set.seed(1234)
  n_obs <- 200
  y1 <- arima.sim(model = list(ar = 0.8), n = n_obs)
  y2 <- arima.sim(model = list(ar = 0.5), n = n_obs)

  # Upper series is the sum of the two bottom series
  y_upper <- y1 + y2

  # Aggregation matrix A:
  A <- matrix(c(1, 1), nrow = 1)

  # Fit additive ETS models
  fit1 <- forecast::ets(y1, additive.only = TRUE)
  fit2 <- forecast::ets(y2, additive.only = TRUE)
  fit_upper <- forecast::ets(y_upper, additive.only = TRUE)

  # Point forecasts (h = 1):
  fc1 <- forecast::forecast(fit1, h = 1)$mean
  fc2 <- forecast::forecast(fit2, h = 1)$mean
  fc_upper <- forecast::forecast(fit_upper, h = 1)$mean
  base_fc_mean <- c(fc_upper, fc1, fc2)

  # Residuals matrix (T x n, columns in same order as base_fc_mean)
  res <- cbind(residuals(fit_upper),
               residuals(fit1),
               residuals(fit2))

  # Reconcile (covariance estimated internally via Schafer-Strimmer)
  result <- reconc_gaussian(A, base_fc_mean = base_fc_mean, residuals = res, return_upper = TRUE)

  bottom_mean <- result$bottom_rec_mean
  bottom_cov <- result$bottom_rec_cov
  upper_mean <- result$upper_rec_mean
  upper_cov <- result$upper_rec_cov

  # Print reconciled means
  cat("Reconciled bottom means:", round(bottom_mean, 3), "\n")
  cat("Reconciled upper mean:", round(upper_mean, 3), "\n")

  # Print 95\% predictions intervals
  cat("Reconciled bottom 95\% prediction intervals:\n")
  for (i in 1:length(bottom_mean)) {
    lower <- bottom_mean[i] - 1.96 * sqrt(bottom_cov[i, i])
    upper <- bottom_mean[i] + 1.96 * sqrt(bottom_cov[i, i])
    cat(paste0("Bottom ", i, ": [", round(lower, 3), ", ", round(upper, 3), "]\n"))
  }
  cat("Reconciled upper 95\% prediction interval:\n")
  lower <- upper_mean - 1.96 * sqrt(upper_cov[1, 1])
  upper <- upper_mean + 1.96 * sqrt(upper_cov[1, 1])
  cat(paste0("Upper: [", round(lower, 3), ", ", round(upper, 3), "]\n"))

}
}

}
\references{
Corani, G., Azzimonti, D., Augusto, J.P.S.C., Zaffalon, M. (2021).
\emph{Probabilistic Reconciliation of Hierarchical Forecast via Bayes' Rule}.
ECML PKDD 2020. Lecture Notes in Computer Science, vol 12459.
\doi{10.1007/978-3-030-67664-3_13}.

Zambon, L., Agosto, A., Giudici, P., Corani, G. (2024).
\emph{Properties of the reconciled distributions for Gaussian and count forecasts}.
International Journal of Forecasting (in press).
\doi{10.1016/j.ijforecast.2023.12.004}.
}
\seealso{
\code{\link[=reconc_t]{reconc_t()}}, \code{\link[=reconc_BUIS]{reconc_BUIS()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{swiss_tourism}
\alias{swiss_tourism}
\title{Swiss tourism: monthly Swiss overnight stay data}
\format{
A list that contains:
\itemize{
\item \code{ts}: a multivariate time series of class \link[stats]{ts};
\item \code{n_bottom}: integer with the number of bottom time series (26);
\item \code{n_upper}: integer with the number of upper time series (1);
\item \code{agg_mat}: the aggregation matrix for the hierarchy. (1x26)
}
}
\source{
Federal Statistical Office of Switzerland (2025).
\emph{Hotel sector: arrivals and overnight stays of open establishments by year, month, canton and visitors' country of residence}.
\url{https://www.bfs.admin.ch/asset/en/px-x-1003020000_102}
}
\usage{
swiss_tourism
}
\description{
A dataset of 27 time series with monthly observations of the number of
overnight stays in Switzerland, from Jan 2005 to Jan 2025.
}
\details{
This is a hierarchy of time series that contains:
\itemize{
\item 26 bottom-level series corresponding to the 26 Swiss cantons
\item 1 top-level series corresponding to the Swiss total.
Each time series has 241 observations.
}
}
\references{
Carrara, C., Azzimonti, D., Corani, G., Zambon, L. (2025).
\emph{Modeling the uncertainty on the covariance matrix for probabilistic forecast reconciliation}.
arXiv:2506.19554
\doi{10.48550/arXiv.2506.19554}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reconc_t.R
\name{reconc_t}
\alias{reconc_t}
\title{t-Rec: reconciliation via conditioning with uncertain covariance via multivariate t-distribution}
\usage{
reconc_t(
  A,
  base_fc_mean,
  y_train = NULL,
  residuals = NULL,
  ...,
  return_upper = FALSE,
  return_parameters = FALSE
)
}
\arguments{
\item{A}{Matrix (n_upp x n_bott) defining the hierarchy (u = Ab).}

\item{base_fc_mean}{Vector of base forecasts (length n = n_upp + n_bott).}

\item{y_train}{mts (or matrix) of historical training data (T x n) used for setting prior parameters.}

\item{residuals}{Matrix (T x n) of base forecast residuals.}

\item{...}{Additional arguments for advanced usage: see details.}

\item{return_upper}{Logical; if TRUE, also returns parameters for the upper-level reconciled distribution.}

\item{return_parameters}{Logical; if TRUE, also returns prior and posterior parameters.}
}
\value{
A list containing:
\itemize{
\item \code{bottom_rec_mean}: reconciled bottom-level mean.
\item \code{bottom_rec_scale_matrix}: reconciled bottom-level scale matrix.
\item \code{bottom_rec_df}: reconciled degrees of freedom.
}
If \code{return_upper} is TRUE, also returns:
\itemize{
\item \code{upper_rec_mean}: reconciled upper-level mean.
\item \code{upper_rec_scale_matrix}: reconciled upper-level scale matrix.
\item \code{upper_rec_df}: reconciled upper-level degrees of freedom.
}
If \code{return_parameters} is TRUE, also returns:
\itemize{
\item \code{prior_nu}: prior degrees of freedom.
\item \code{prior_Psi}: prior scale matrix.
\item \code{posterior_nu}: posterior degrees of freedom.
\item \code{posterior_Psi}: posterior scale matrix.
}
}
\description{
Reconciles base forecasts in a hierarchy by conditioning on the hierarchical
constraints, specified by the aggregation matrix A.
The base forecasts are assumed to be jointly Gaussian, conditionally on the
covariance matrix of the forecast errors.
A Bayesian approach is adopted to account for the uncertainty of the covariance matrix.
An Inverse-Wishart prior is specified on the covariance matrix,
leading to a multivariate t-distribution for the base forecasts.
The reconciliation via conditioning is in closed-form, yielding a multivariate t
reconciled distribution.
}
\details{
\strong{Standard usage.}

The standard workflow for this function is to provide the in-sample \code{residuals}
and the historical training data \code{y_train}.
The parameters of the Inverse-Wishart prior distribution of the covariance matrix
are set as follows:
\itemize{
\item Prior scale matrix (Psi): set as the covariance of the residuals of naive (or seasonal naive,
a criterion is used to choose between the two) forecasts computed on \code{y_train}.
\item Prior degrees of freedom (nu): estimated via Bayesian Leave-One-Out
Cross-Validation (LOOCV) to maximize out-of-sample performance.
}
The posterior distribution of the covariance matrix is still Inverse-Wishart.
The parameters of the posterior are computed in closed-form using the sample
covariance of the provided \code{residuals}.

\strong{Advanced Options.}

Users can bypass the automated estimation by specifying:
\enumerate{
\item \code{prior}: a list with entries 'nu' and 'Psi'.
This skips the LOOCV step for \eqn{\nu_0} and the covariance estimation from \code{y_train}.
It requires \code{residuals} to compute the posterior.
\item \code{posterior}: a list with entries 'nu' and 'Psi'.
This skips all internal estimation and updating logic.
}
Moreover, users can specify:
\itemize{
\item \code{freq}: positive integer, used as frequency of data for the seasonal naive forecast in the specification of the prior scale matrix.
By default, if \code{y_train} is a multivariate time series, the frequency of the data is used; otherwise, it is set to 1 (no seasonality).
\item \code{criterion}: either 'RSS' (default) or 'seas-test', specifying which criterior is used to choose between
the naive and seasonal naive forecasts for the specification of the prior scale matrix.
'RSS' computes the residual sum of squares for both methods and chooses the one with lower RSS,
while 'seas-test' uses a statistical test for seasonality
(currently implemented using the number of seasonal differences suggested by the \code{forecast} package,
which must be installed).
}

\strong{Reconciled distribution.}

The reconciled distribution is a multivariate t-distribution,
specified by a vector of means, a scale matrix, and a number of degrees of freedom.
These parameters are computed in closed-form.
By default, only the parameters of the reconciled distribution for the bottom-level
series are returned. See examples.
}
\examples{

\donttest{
library(bayesRecon)

if (requireNamespace("forecast", quietly = TRUE)) {

  set.seed(1234)
  n_obs <- 100

  # Simulate 2 bottom series from AR(1) processes
  y1 <- arima.sim(model = list(ar = 0.8), n = n_obs)
  y2 <- arima.sim(model = list(ar = 0.5), n = n_obs)

  y_upper <- y1 + y2   # upper series is the sum of the two bottoms
  A <- matrix(c(1, 1), nrow = 1)  # Aggregation matrix

  # Fit additive ETS models
  fit1 <- forecast::ets(y1, additive.only = TRUE)
  fit2 <- forecast::ets(y2, additive.only = TRUE)
  fit_upper <- forecast::ets(y_upper, additive.only = TRUE)

  # Point forecasts (h = 1)
  fc_upper <- as.numeric(forecast::forecast(fit_upper, h = 1)$mean)
  fc1 <- as.numeric(forecast::forecast(fit1, h = 1)$mean)
  fc2 <- as.numeric(forecast::forecast(fit2, h = 1)$mean)
  base_fc_mean <- c(fc_upper, fc1, fc2)

  # Residuals and training data (n_obs x n matrices, columns in same order as base_fc_mean)
  res <- cbind(residuals(fit_upper), residuals(fit1), residuals(fit2))
  y_train <- cbind(y_upper, y1, y2)

  # --- 1) Generate joint reconciled samples ---
  result <- reconc_t(A, base_fc_mean, y_train = y_train, residuals = res)

  # Sample from the reconciled bottom-level t-distribution
  n_samples <- 2000
  L_chol <- t(chol(result$bottom_rec_scale_matrix))
  z <- matrix(rt(ncol(A) * n_samples, df = result$bottom_rec_df), nrow = ncol(A))
  bottom_samples <- result$bottom_rec_mean + L_chol \%*\% z  # 2 x n_samples

  # Aggregate bottom samples to get upper samples
  upper_samples <- A \%*\% bottom_samples              
  joint_samples <- rbind(upper_samples, bottom_samples)
  rownames(joint_samples) <- c("upper", "bottom_1", "bottom_2")

  cat("Reconciled means (from samples):\n")
  print(round(rowMeans(joint_samples), 3))

  cat("Reconciled standard deviations (from samples):\n")
  print(round(apply(joint_samples, 1, sd), 3))

  # --- 2) 95\% prediction intervals via t-distribution quantiles ---
  result2 <- reconc_t(A, base_fc_mean, y_train = y_train,
                      residuals = res, return_upper = TRUE)

  alpha <- 0.05
  # Bottom series intervals
  for (i in seq_len(ncol(A))) {
    s_i <- sqrt(result2$bottom_rec_scale_matrix[i, i])
    lo  <- result2$bottom_rec_mean[i] + s_i * qt(alpha / 2,     df = result2$bottom_rec_df)
    hi  <- result2$bottom_rec_mean[i] + s_i * qt(1 - alpha / 2, df = result2$bottom_rec_df)
    cat(sprintf("Bottom \%d: 95\%\% PI = [\%.3f, \%.3f]\n", i, lo, hi))
  }
  # Upper series interval
  s_u <- sqrt(result2$upper_rec_scale_matrix[1, 1])
  lo  <- result2$upper_rec_mean[1] + s_u * qt(alpha / 2,     df = result2$upper_rec_df)
  hi  <- result2$upper_rec_mean[1] + s_u * qt(1 - alpha / 2, df = result2$upper_rec_df)
  cat(sprintf("Upper:    95\%\% PI = [\%.3f, \%.3f]\n", lo, hi))
}
}

}
\references{
Carrara, C., Corani, G., Azzimonti, D., & Zambon, L. (2025). Modeling the uncertainty on the covariance
matrix for probabilistic forecast reconciliation. arXiv preprint arXiv:2506.19554.
\url{https://arxiv.org/abs/2506.19554}
}
\seealso{
\code{\link[=reconc_gaussian]{reconc_gaussian()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hierarchy.R
\name{.get_HG}
\alias{.get_HG}
\title{Extract and sort hierarchy rows}
\usage{
.get_HG(A, v, d, it)
}
\arguments{
\item{A}{Aggregation matrix}

\item{v}{Vector of values}

\item{d}{Distribution parameters}

\item{it}{Input type parameters}
}
\value{
List containing sorted hierarchy rows and remaining rows
}
\description{
Function that extracts the "best hierarchy rows" from A (see .get_hier_rows),
and sorts them in the correct order (i.e. bottom-up).
Also sorts accordingly the vectors v, d, it (e.g. of parameters).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reconc_t.R
\name{.compute_naive_cov}
\alias{.compute_naive_cov}
\title{Estimate covariance from naive or seasonal-naive residuals}
\usage{
.compute_naive_cov(y_train, freq = NULL, criterion = "RSS")
}
\arguments{
\item{y_train}{Multivariate time series object or numeric matrix of historical
observations, with dimensions \verb{T x n} (rows are time points, columns are series).}

\item{freq}{Positive integer seasonal frequency (optional).
If not provided and \code{y_train} is a multivariate time series, the frequency
of the data is used.}

\item{criterion}{Character string used when \code{freq > 1} to choose residuals.
Supported values are \code{"RSS"} (default) and \code{"seas-test"}.
\code{"RSS"} chooses the method with the lower residual sum of squares (RSS),
while \code{"seas-test"} uses a statistical test for seasonality (requires \code{forecast} package).}
}
\value{
A numeric \verb{n x n} shrinkage covariance matrix estimated with
\code{\link[=schaferStrimmer_cov]{schaferStrimmer_cov()}}.
}
\description{
Estimates via shrinkage the covariance matrix of the residuals of the naive
or seasonal naive forecasts.
If the frequency of the time series is > 1, the function chooses between the
two methods series-by-series according to a selection criterion.
If the frequency is 1 or is not provided, only the naive residuals are used.
}
\seealso{
\code{\link[=schaferStrimmer_cov]{schaferStrimmer_cov()}}, \code{\link[=reconc_t]{reconc_t()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{M3_example}
\alias{M3_example}
\title{Example of a time series from the M3 forecasting competition}
\format{
List of time series of class \link[stats]{ts}.
}
\source{
\url{https://forecasters.org/resources/time-series-data/m3-competition/}
}
\usage{
M3_example
}
\description{
A monthly time series, from the M3 forecasting competition ("N1485").
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{M5_CA1_basefc}
\alias{M5_CA1_basefc}
\title{Example of hierarchical forecasts for a store from the M5 competition}
\format{
A list containing:
\itemize{
\item \code{upper}: a list of 11 elements each representing an aggregation level. Each element contains: \code{mu}, \code{sigma} the mean and standard deviation of the Gaussian forecast, \code{actual} the actual value, \code{residuals} the residuals of the model used to estimate forecasts covariance.
\item \code{lower}: a list of 3049 elements each representing a forecast for each item. Each element contains \code{pmf} the probability mass function of the item level forecast, \code{actual} the actual value.
\item \code{A}: the aggregation matrix for A.
\item \code{S}: the S matrix for the hierarchy.
\item \code{Q_u}: scaling factors for computing MASE on the upper forecasts.
\item \code{Q_b}: scaling factors for computing MASE on the bottom forecasts.
}
}
\source{
Makridakis, Spyros & Spiliotis, Evangelos & Assimakopoulos, Vassilis. (2020). \emph{The M5 Accuracy competition: Results, findings and conclusions.} International Journal of Forecasting 38(4) 1346-1364. \doi{10.1016/j.ijforecast.2021.10.009}
}
\usage{
M5_CA1_basefc
}
\description{
This dataset contains forecasts for the hierarchy of time series related to the store \code{CA_1} from the M5 competition.
}
\details{
The store \code{CA_1} contains 3049 item level time series and 11 aggregate time series:
\itemize{
\item Store level aggregation (\code{CA_1})
\item Category level aggregations  (\code{HOBBIES}, \code{HOUSEHOLD}, \code{FOODS})
\item Department level aggregations (\code{HOBBIES_1}, \code{HOBBIES_2}, \code{HOUSEHOLD_1}, \code{HOUSEHOLD_2}, \code{FOODS_1}, \code{FOODS_2}, \code{FOODS_3})
}

Forecasts are generated with the function \code{forecast} and the model \code{adam} from the package \code{smooth}.
\itemize{
\item The models for the bottom time series are selected with multiplicative Gamma error term (\code{MNN});
\item The models for the upper time series (\code{AXZ}) is selected with Gaussian additive error term, seasonality selected based on information criterion.
}

The raw data was downloaded with the package \code{m5}.
}
\references{
Joachimiak K (2022). \emph{m5: 'M5 Forecasting' Challenges Data}. R package version 0.1.1, \url{https://CRAN.R-project.org/package=m5}.

Makridakis, Spyros & Spiliotis, Evangelos & Assimakopoulos, Vassilis. (2020). \emph{The M5 Accuracy competition: Results, findings and conclusions.} International Journal of Forecasting 38(4) 1346-1364. \doi{10.1016/j.ijforecast.2021.10.009}

Svetunkov I (2023). \emph{smooth: Forecasting Using State Space Models}. R package version 4.0.0, \url{https://CRAN.R-project.org/package=smooth}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hierarchy.R
\name{.lowest_lev}
\alias{.lowest_lev}
\title{Find rows corresponding to the lowest level}
\usage{
.lowest_lev(A)
}
\arguments{
\item{A}{Aggregation matrix (must be hierarchical)}
}
\value{
Integer vector of row indices corresponding to the lowest level
}
\description{
Identifies the rows of aggregation matrix A that correspond to the lowest
hierarchical level (i.e., the most disaggregated upper variables that
collectively sum to all bottom variables).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reconc_MixCond.R
\name{.core_reconc_MixCond}
\alias{.core_reconc_MixCond}
\title{Core Reconciliation via Importance Sampling for Mixed Hierarchies}
\usage{
.core_reconc_MixCond(
  A,
  B,
  mean_upper,
  cov_upper,
  num_samples,
  return_type,
  return_ESS = TRUE,
  return_upper = TRUE,
  suppress_warnings = FALSE
)
}
\arguments{
\item{A}{Matrix (n_upper x n_bottom) defining the hierarchy where upper = A \%*\% bottom.}

\item{B}{Matrix (n_samples x n_bottom) of bottom base forecast samples to be reconciled.}

\item{mean_upper}{Vector of upper level means.}

\item{cov_upper}{Covariance matrix of upper level.}

\item{num_samples}{Number of samples to draw/resample from.}

\item{return_type}{Character string specifying return format: 'pmf', 'samples', or 'all'.}

\item{return_ESS}{Logical, whether to return the Effective Sample Size (ESS) from importance sampling weights (default TRUE).}

\item{return_upper}{Logical, whether to return the reconciled parameters for the upper variables (default TRUE).}

\item{suppress_warnings}{Logical. If TRUE, suppresses warnings about sample quality (default FALSE).}
}
\value{
A list containing:
\itemize{
\item \code{bottom_rec}: List with reconciled bottom forecasts (pmf and/or samples).
\item \code{bottom_rec_pmf}: list of PMF objects for each bottom series (only if \code{return_type} is 'pmf' or 'all').
\item \code{bottom_rec_samples}: matrix (n_bottom x num_samples) of reconciled bottom samples (only if \code{return_type} is 'samples' or 'all').
\item \code{upper_rec_pmf}: list of PMF objects for each upper series (only if \code{return_type} is 'pmf' or 'all', and \code{return_upper = TRUE}).
\item \code{upper_rec_samples}: matrix (n_upper x num_samples) of reconciled upper samples (only if \code{return_type} is 'samples' or 'all', and \code{return_upper = TRUE}).
\item \code{ESS}: Effective Sample Size resulting from importance sampling reweighting (only if \code{return_ESS = TRUE}).
}
}
\description{
Internal function that performs the core reconciliation logic using importance sampling
(IS) to reconcile mixed-type hierarchies. The base bottom forecasts (provided as samples)
are reweighted according to their fit to the upper multivariate Gaussian forecasts.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hierarchy.R
\name{temporal_aggregation}
\alias{temporal_aggregation}
\title{Temporal aggregation of a time series}
\usage{
temporal_aggregation(y, agg_levels = NULL)
}
\arguments{
\item{y}{univariate time series of class \link[stats]{ts}.}

\item{agg_levels}{user-selected list of aggregation levels.}
}
\value{
A list of \link[stats]{ts} objects each containing the aggregates time series in the order defined by \code{agg_levels}.
}
\description{
Creates a list of aggregated time series from a time series of class \link[stats]{ts}.
}
\details{
If \code{agg_levels=NULL} then \code{agg_levels} is automatically generated by taking all the factors of the time series frequency.
}
\examples{

# Create a monthly count time series with 100 observations
y <- ts(data = stats::rpois(100, lambda = 2), frequency = 12)

# Create the aggregate time series according to agg_levels
y_agg <- temporal_aggregation(y, agg_levels = c(2, 3, 4, 6, 12))

# Show annual aggregate time series
print(y_agg$`f=1`)

}
\seealso{
\code{\link[=get_reconc_matrices]{get_reconc_matrices()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reconc_t.R
\name{multi_log_score_optimization}
\alias{multi_log_score_optimization}
\title{Optimize degrees of freedom (nu) via LOO Cross-Validation}
\usage{
multi_log_score_optimization(res, prior_mean, trim = 0.1)
}
\arguments{
\item{res}{Matrix of residuals (n_obs x n_var).}

\item{prior_mean}{The prior mean covariance matrix (n_var x n_var).}

\item{trim}{Fraction of observations to trim (0 to 1). Default is 0.1 (10\%).}
}
\value{
A list containing the optimization results:
\itemize{
\item \code{optimal_nu}: The optimal degrees of freedom found.
\item \code{min_neg_log_score}: The minimum negative log score achieved.
\item \code{convergence}: The convergence status of the optimizer.
\item \code{time_elapsed}: The time taken for the optimization.
}
}
\description{
Optimize degrees of freedom (nu) via LOO Cross-Validation
}
\details{
TODO: check these details

\strong{Leave-One-Out (LOO) Cross-Validation:}
This function estimates the optimal degrees of freedom \eqn{\nu} by maximizing the
out-of-sample predictive performance. This is achieved by computing the
log-density of each held-out observation \eqn{\mathbf{r}_i} given the remaining
data \eqn{\mathbf{R}_{-i}}. The total objective function is the sum of these
predictive log-densities:
\deqn{\mathcal{L}(\nu) = \sum_{i=1}^T \log f(\mathbf{r}_i | \mathbf{R}_{-i}, \nu)}

\strong{The Log-Density Function:}
For each LOO step, the residuals are assumed to follow a
multivariate t-distribution. The density is expressed directly as a function of the
posterior sum-of-squares matrix \eqn{\Psi}, where \eqn{\Psi} scales implicitly with \eqn{\nu}:
\deqn{f(\mathbf{r}_i | \Psi, \nu) = \frac{\Gamma(\frac{\nu + T}{2})}{\Gamma(\frac{\nu + T - p}{2}) \pi^{p/2}} |\Psi|^{-1/2} \left( 1 + \mathbf{r}_i^\top \Psi^{-1} \mathbf{r}_i \right)^{-\frac{\nu + T}{2}}}
In the code, \eqn{\Psi} is constructed as:
\deqn{\Psi = (\nu - p - 1)\bar{\Sigma}_{prior} + \mathbf{R}^\top\mathbf{R}}
By using this formulation, the standard scaling factors \eqn{1/\nu} and \eqn{\nu^{-p/2}}
are absorbed into the matrix inverse and determinant, respectively.

\strong{Efficient Computation via Sherman-Morrison:}
Rather than recomputing \eqn{\Psi_{-i}} and its inverse \eqn{T} times, the function uses
the full-sample matrix \eqn{\Psi} and adjusts it using the leverage
\eqn{h_i = \mathbf{r}_i^\top \Psi^{-1} \mathbf{r}_i}.

Through the Matrix Determinant Lemma and the Sherman-Morrison formula, the
internal term \eqn{(1 + \mathbf{r}_i^\top \Psi_{-i}^{-1} \mathbf{r}_i)} simplifies to \eqn{(1 - h_i)^{-1}}.
The final log-density contribution used in the code is:
\deqn{\log f_i \propto \textrm{const} - \frac{1}{2}\log|\Psi| + \frac{\nu + T - 1}{2} \log(1 - h_i)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hierarchy.R
\name{get_reconc_matrices}
\alias{get_reconc_matrices}
\title{Build hierarchy matrices}
\usage{
get_reconc_matrices(agg_levels, h)
}
\arguments{
\item{agg_levels}{user-selected list of aggregation levels.}

\item{h}{number of steps ahead for the bottom level forecasts.}
}
\value{
A list containing the named elements:
\itemize{
\item \code{A} the aggregation matrix;
\item \code{S} the summing matrix.
}
}
\description{
Creates the aggregation and summing matrices for a temporal hierarchy of time series
from a user-selected list of aggregation levels.
}
\examples{

library(bayesRecon)

# Create monthly hierarchy
agg_levels <- c(1, 2, 3, 4, 6, 12)
h <- 12
rec_mat <- get_reconc_matrices(agg_levels, h)
S <- rec_mat$S
A <- rec_mat$A

}
\seealso{
\code{\link[=temporal_aggregation]{temporal_aggregation()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{extr_mkt_events}
\alias{extr_mkt_events}
\title{Extreme market events dataset}
\format{
A multivariate time series of class \link[stats]{ts}.
}
\source{
Zambon, L., Agosto, A., Giudici, P., Corani, G. (2024).
\emph{Properties of the reconciled distributions for Gaussian and count forecasts}.
International Journal of Forecasting (in press).
\doi{10.1016/j.ijforecast.2023.12.004}.
}
\usage{
extr_mkt_events
}
\description{
Count time series of extreme market events in five economic sectors.
The data refer to the trading days between 2004/12/31 and 2018/12/19 (3508 trading days in total).
}
\details{
The counts are computed by considering 29 companies included in the Euro Stoxx
50 index and observing if the value of the CDS spread on a given day exceeds
the 90-th percentile of its distribution in the last trading year.
The companies are divided in the following  sectors: Financial (FIN), Information
and Communication Technology (ICT), Manufacturing (MFG), Energy (ENG), and Trade (TRD).

There are 6 time series:
\itemize{
\item 5 bottom time series, corresponding to the daily counts for each sector
\item 1 upper time series, which is the sum of all the bottom (ALL)
}
}
\references{
Zambon, L., Agosto, A., Giudici, P., Corani, G. (2024).
\emph{Properties of the reconciled distributions for Gaussian and count forecasts}.
International Journal of Forecasting (in press).
\doi{10.1016/j.ijforecast.2023.12.004}.

Agosto, A. (2022).
\emph{Multivariate Score-Driven Models for Count Time Series to Assess Financial Contagion}.
\doi{10.2139/ssrn.4119895}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{carparts_example}
\alias{carparts_example}
\title{Example of a time series from carparts}
\format{
Univariate time series of class \link[stats]{ts}.
}
\source{
Godahewa, R., Bergmeir, C., Webb, G., Hyndman, R.J., & Montero-Manso, P. (2020).
Car Parts Dataset (without Missing Values) (Version 2)
\doi{10.5281/zenodo.4656021}
}
\usage{
carparts_example
}
\description{
A monthly time series from the \code{carparts} dataset, 51 observations, Jan 1998 - Mar 2002.
}
\references{
Hyndman, R.J., Koehler, A.B., Ord, J.K., and Snyder, R.D., (2008).
Forecasting with exponential smoothing: the state space approach.
Springer Science & Business Media.

Godahewa, R., Bergmeir, C., Webb, G., Hyndman, R., & Montero-Manso, P. (2020).
Car Parts Dataset (without Missing Values) (Version 2)
\doi{10.5281/zenodo.4656021}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reconc_BUIS.R
\name{reconc_BUIS}
\alias{reconc_BUIS}
\title{BUIS for Probabilistic Reconciliation of forecasts via conditioning}
\usage{
reconc_BUIS(
  A,
  base_fc,
  in_type,
  distr,
  num_samples = 20000,
  suppress_warnings = FALSE,
  return_upper = TRUE,
  seed = NULL
)
}
\arguments{
\item{A}{aggregation matrix (n_upper x n_bottom).}

\item{base_fc}{A list containing the base_forecasts, see details.}

\item{in_type}{A string or a list of length n_upper + n_bottom. If it is a list the i-th element is a string with two possible values:
\itemize{
\item 'samples' if the i-th base forecasts are in the form of samples;
\item 'params'  if the i-th base forecasts are in the form of estimated parameters.
}

If it \code{in_type} is a string it is assumed that all base forecasts are of the same type.}

\item{distr}{A string or a list of length n_upper + n_bottom describing the type of base forecasts.
If it is a list the i-th element is a string with two possible values:
\itemize{
\item 'continuous' or 'discrete' if \code{in_type[[i]]}='samples';
\item 'gaussian', 'poisson' or 'nbinom' if \code{in_type[[i]]}='params'.
}

If \code{distr} is a string it is assumed that all distributions are of the same type.}

\item{num_samples}{Number of samples drawn from the reconciled distribution.
This is ignored if \code{bottom_in_type='samples'}; in this case, the number of reconciled samples is equal to
the number of samples of the base forecasts.}

\item{suppress_warnings}{Logical. If \code{TRUE}, no warnings about effective sample size
are triggered. If \code{FALSE}, warnings are generated. Default is \code{FALSE}. See Details.}

\item{return_upper}{Logical, whether to return the reconciled parameters for the upper variables (default is TRUE).}

\item{seed}{Seed for reproducibility.}
}
\value{
A list containing the reconciled forecasts. The list has the following named elements:
\itemize{
\item \code{bottom_rec_samples}: a matrix (n_bottom x \code{num_samples}) containing the reconciled samples for the bottom time series;
\item \code{upper_rec_samples}: (only if \code{return_upper = TRUE}) a matrix (n_upper x \code{num_samples}) containing the reconciled samples for the upper time series.
}
}
\description{
Uses the Bottom-Up Importance Sampling algorithm to draw samples from the reconciled
forecast distribution, obtained via conditioning.
}
\details{
The parameter \code{base_fc} is a list containing n = n_upper + n_bottom elements.
The first n_upper elements of the list are the upper base forecasts, in the order given by the rows of A.
The elements from n_upper+1 until the end of the list are the bottom base forecasts, in the order given by the columns of A.

The i-th element depends on the values of \code{in_type[[i]]} and \code{distr[[i]]}.

If \code{in_type[[i]]}='samples', then \code{base_fc[[i]]} is a vector containing samples from the base forecast distribution.

If \code{in_type[[i]]}='params', then \code{base_fc[[i]]} is a list containing the estimated:
\itemize{
\item mean and sd for the Gaussian base forecast if \code{distr[[i]]}='gaussian', see \link[stats]{Normal};
\item lambda for the Poisson base forecast if \code{distr[[i]]}='poisson', see \link[stats]{Poisson};
\item size and prob (or mu) for the negative binomial base forecast if \code{distr[[i]]}='nbinom', see \link[stats]{NegBinomial}.
}

See the description of the parameters \code{in_type} and \code{distr} for more details.

Warnings are triggered from the Importance Sampling step if:
\itemize{
\item weights are all zeros, then the upper is ignored during reconciliation;
\item the effective sample size is < 200;
\item the effective sample size is < 1\% of the sample size (\code{num_samples} if \code{in_type} is 'params' or the size of the base forecast if if \code{in_type} is 'samples').
}

Note that warnings are an indication that the base forecasts might have issues.
Please check the base forecasts in case of warnings.
}
\examples{

library(bayesRecon)

# Create a minimal hierarchy with 2 bottom and 1 upper variable
rec_mat <- get_reconc_matrices(agg_levels = c(1, 2), h = 2)
A <- rec_mat$A
S <- rec_mat$S


# 1) Gaussian base forecasts

# Set the parameters of the Gaussian base forecast distributions
mu1 <- 2
mu2 <- 4
muY <- 9
mus <- c(muY, mu1, mu2)

sigma1 <- 2
sigma2 <- 2
sigmaY <- 3
sigmas <- c(sigmaY, sigma1, sigma2)

base_fc <- list()
for (i in 1:length(mus)) {
  base_fc[[i]] <- list(mean = mus[[i]], sd = sigmas[[i]])
}


# Sample from the reconciled forecast distribution using the BUIS algorithm
buis <- reconc_BUIS(A, base_fc,
  in_type = "params",
  distr = "gaussian", num_samples = 100000, seed = 42
)

samples_buis <- rbind(buis$upper_rec_samples, buis$bottom_rec_samples)

# In the Gaussian case, the reconciled distribution is still Gaussian and can be
# computed in closed form
Sigma <- diag(sigmas^2) # transform into covariance matrix
analytic_rec <- reconc_gaussian(A,
  base_fc_mean = mus,
  base_fc_cov = Sigma
)

# Compare the reconciled means obtained analytically and via BUIS
print(c(S \%*\% analytic_rec$bottom_rec_mean))
print(rowMeans(samples_buis))


# 2) Poisson base forecasts

# Set the parameters of the Poisson base forecast distributions
lambda1 <- 2
lambda2 <- 4
lambdaY <- 9
lambdas <- c(lambdaY, lambda1, lambda2)

base_fc <- list()
for (i in 1:length(lambdas)) {
  base_fc[[i]] <- list(lambda = lambdas[i])
}

# Sample from the reconciled forecast distribution using the BUIS algorithm
buis <- reconc_BUIS(A, base_fc,
  in_type = "params",
  distr = "poisson", num_samples = 100000, seed = 42
)
samples_buis <- rbind(buis$upper_rec_samples, buis$bottom_rec_samples)

# Print the reconciled means
print(rowMeans(samples_buis))

}
\references{
Zambon, L., Azzimonti, D. & Corani, G. (2024).
\emph{Efficient probabilistic reconciliation of forecasts for real-valued and count time series}.
Statistics and Computing 34 (1), 21.
\doi{10.1007/s11222-023-10343-y}.
}
\seealso{
\code{\link[=reconc_gaussian]{reconc_gaussian()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reconc_TDcond.R
\name{.core_reconc_TDcond}
\alias{.core_reconc_TDcond}
\title{Core Reconciliation via Top-Down Conditioning for Mixed Hierarchies}
\usage{
.core_reconc_TDcond(
  A,
  mean_upper,
  cov_upper,
  L_pmf,
  num_samples,
  return_type,
  return_upper = TRUE,
  suppress_warnings = TRUE,
  min_fraction_samples_ok = .MIN_FRACTION_SAMPLES_OK
)
}
\arguments{
\item{A}{Matrix (n_upper x n_bottom) defining the hierarchy where upper = A \%*\% bottom.}

\item{mean_upper}{Vector of mean forecasts for upper level.}

\item{cov_upper}{Covariance matrix of upper level forecasts.}

\item{L_pmf}{List of PMF objects representing the bottom level base forecasts.}

\item{num_samples}{Number of samples to draw from the reconciled distribution.}

\item{return_type}{Character string specifying return format: 'pmf', 'samples', or 'all'.}

\item{return_upper}{Logical, whether to return reconciled upper forecasts (default TRUE).}

\item{suppress_warnings}{Logical, whether to suppress warnings about samples outside support (default TRUE).}

\item{min_fraction_samples_ok}{Numeric between 0 and 1, minimum fraction of reconciled
upper samples that must lie in the support of the bottom-up distribution (default 0.5).
If the fraction is below this threshold, the function returns an error.}
}
\value{
A list containing:
\itemize{
\item \code{bottom_rec_pmf}: list of PMF objects for each bottom series (only if \code{return_type} is 'pmf' or 'all').
\item \code{bottom_rec_samples}: matrix (n_bottom x num_samples) of reconciled bottom samples (only if \code{return_type} is 'samples' or 'all').
\item \code{upper_rec_pmf}: list of PMF objects for each upper series (only if \code{return_type} is 'pmf' or 'all', and \code{return_upper = TRUE}).
\item \code{upper_rec_samples}: matrix (n_upper x num_samples) of reconciled upper samples (only if \code{return_type} is 'samples' or 'all', and \code{return_upper = TRUE}).
}
}
\description{
Internal function that performs the core reconciliation logic using top-down conditioning
(TD) for mixed hierarchies. First, upper forecasts are reconciled analytically via conditioning
(if necessary), then bottom distributions are updated through probabilistic top-down procedure
by conditioning on the reconciled upper values.
}
\details{
The function internally:
\enumerate{
\item Identifies the "lowest upper" nodes in the hierarchy.
\item If all uppers are lowest-uppers, samples directly from the upper MVN. Otherwise,
analytically reconciles the upper hierarchy and samples from the lowest level.
\item Reconciles bottom distributions by conditioning on the sampled/reconciled upper values
using the probabilistic top-down algorithm.
\item Discards samples that fall outside the support of the bottom-up distribution.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reconc_BUIS.R
\name{.core_reconc_BUIS}
\alias{.core_reconc_BUIS}
\title{Core Reconciliation via Bayesian Universality Information Sharing}
\usage{
.core_reconc_BUIS(
  A,
  H,
  G,
  B,
  upper_base_fc_H,
  in_typeH,
  distr_H,
  upper_base_fc_G,
  in_typeG,
  distr_G,
  .comp_w = .compute_weights,
  suppress_warnings = FALSE,
  return_upper = TRUE
)
}
\arguments{
\item{A}{Matrix defining the overall hierarchy.}

\item{H}{Matrix defining hierarchical constraints.}

\item{G}{Matrix defining general linear constraints.}

\item{B}{Matrix of bottom level base forecast samples.}

\item{upper_base_fc_H}{List of upper base forecasts for hierarchical constraints.}

\item{in_typeH}{Character string specifying input type for H forecasts ('pmf', 'samples', or 'params').}

\item{distr_H}{Character string specifying distribution type for H forecasts ('poisson' or 'nbinom').}

\item{upper_base_fc_G}{List of upper base forecasts for general constraints.}

\item{in_typeG}{Character string specifying input type for G forecasts ('pmf', 'samples', or 'params').}

\item{distr_G}{Character string specifying distribution type for G forecasts ('poisson' or 'nbinom').}

\item{.comp_w}{Function to compute weights for importance sampling. Default is \code{.compute_weights}.}

\item{suppress_warnings}{Logical. If TRUE, suppresses warnings about sample quality. Default is FALSE.}
}
\value{
A list containing:
\itemize{
\item \code{bottom_rec}: List with reconciled bottom forecasts (pmf and/or samples).
\item \code{upper_rec_H}: List with reconciled upper forecasts for H constraints.
\item \code{upper_rec_G}: List with reconciled upper forecasts for G constraints.
}
}
\description{
Internal function that performs the core reconciliation logic for the BUIS method, which
reconciles forecasts using importance sampling based on both hierarchical and equality
constraints through a Bayesian framework.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hierarchy.R
\name{.check_hierarchical}
\alias{.check_hierarchical}
\title{Check if matrix is hierarchical}
\usage{
.check_hierarchical(A)
}
\arguments{
\item{A}{Aggregation matrix}
}
\value{
Logical value indicating if matrix is hierarchical
}
\description{
Returns TRUE if A is a hierarchy matrix.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PMF.R
\name{PMF}
\alias{PMF}
\alias{PMF_sample}
\alias{PMF_get_mean}
\alias{PMF_get_var}
\alias{PMF_get_quantile}
\alias{PMF_summary}
\title{PMF operations and utilities}
\usage{
PMF_sample(pmf, N_samples)
PMF_get_mean(pmf)
PMF_get_var(pmf)
PMF_get_quantile(pmf, p)
PMF_summary(pmf, Ltoll = .TOLL, Rtoll = .RTOLL)

PMF_get_mean(pmf)

PMF_get_var(pmf)

PMF_get_quantile(pmf, p)

PMF_summary(pmf, Ltoll = .TOLL, Rtoll = .RTOLL)
}
\arguments{
\item{pmf}{A PMF object (numeric vector where element j+1 is the probability of value j).}

\item{N_samples}{Number of samples to draw from the PMF.}

\item{p}{Probability level for quantile computation. Either a number or a vector with all elements between 0 and 1.}

\item{Ltoll}{Lower tolerance for computing the minimum of the PMF (default: 1e-15).}

\item{Rtoll}{Upper tolerance for computing the maximum of the PMF (default: 1e-9).}
}
\description{
A set of functions for working with Probability Mass Functions (PMFs) in bayesRecon.
A PMF is represented as a normalized numeric vector where element \code{v[j+1]} represents
the probability of value \code{j} (support starts from 0).

These functions provide utilities for:
\itemize{
\item Drawing samples from PMFs
\item Computing summary statistics (mean, variance, quantiles)
\item Summarizing PMF distributions
}
}
\section{Functions}{


\describe{
\item{\code{PMF_sample(pmf, N_samples)}}{
Draws random samples from the probability distribution specified by the PMF.
Uses sampling with replacement from the discrete support values, weighted by
their probabilities. This is useful for generating synthetic data or for
Monte Carlo simulations.
}

\item{\code{PMF_get_mean(pmf)}}{
Computes the expected value (mean) of the distribution represented by the PMF.
The mean is calculated as the sum of each value in the support weighted by
its probability: \eqn{\sum_{x} x \cdot P(X=x)}.
}

\item{\code{PMF_get_var(pmf)}}{
Computes the variance of the distribution represented by the PMF.
The variance measures the spread of the distribution and is calculated as
\eqn{E[X^2] - (E[X])^2}, where \eqn{E[X]} is the mean.
}

\item{\code{PMF_get_quantile(pmf, p)}}{
Computes the quantile of the distribution at probability level \code{p}.
Returns the smallest value \code{x} such that the cumulative probability up to \code{x}
is greater than or equal to \code{p}. For example, \code{p=0.5} gives the median.
The parameter \code{p} can be a single number or a vector of probabilities.
}

\item{\code{PMF_summary(pmf, Ltoll, Rtoll)}}{
Provides a comprehensive summary of the distribution including minimum, maximum,
quartiles, median, and mean. The minimum and maximum are determined based on
probability thresholds to handle the potentially infinite tails of discrete distributions.
}
}
}

\examples{
library(bayesRecon)

# Let's build the pmf of a Binomial distribution with parameters n and p
n <- 10
p <- 0.6
pmf_binomial <- apply(matrix(seq(0, n)), MARGIN = 1, FUN = \(x) dbinom(x, size = n, prob = p))

# Draw samples from the PMF object
set.seed(1)
samples <- PMF_sample(pmf = pmf_binomial, N_samples = 1e4)

# Compute statistics
PMF_get_mean(pmf_binomial) # Mean: should be close to n*p = 6
PMF_get_var(pmf_binomial) # Variance: should be close to n*p*(1-p) = 2.4
PMF_get_quantile(pmf_binomial, 0.5) # Median
PMF_summary(pmf_binomial) # Full summary

}
